define(function (require, exports, module) {
	'use strict';

	var AmountHelper = require('ft/taxonomy/helpers/amounthelper'),
		NotificationCenter = require('ft/system/notificationcenter').NotificationCenter,
		Extensions = require('ft/core/extensions').Extensions,
		NodeSet = require('ft/core/nodeset').NodeSet,
		DateUtils = require('ft/util/date'),
		timersToLastValidStartDates = {},
		dom = require('ft/util/dom'),
		timersToStartDates = {},
		timersToCurrentSteps = {},
		timersToFinishLines = {},
		currentDate = new Date(),
		timers = new NodeSet();

	function getStartDate(node) {
		var startDate;

		if (node.mode() === 'timer') {
			startDate = timersToStartDates[node.id];
			if (startDate) {
				return startDate;
			}

			var dateProperty = node.propertyNode('Start');
			if (dateProperty) {
				var timestamp = Date.parse(dateProperty.tag('value'));
				if (isNaN(timestamp) === false) {
					startDate = new Date(timestamp);
				}
			}

			if (!startDate) {
				startDate = timersToLastValidStartDates[node.id];
				if (!startDate) {
					startDate = currentDate;
				}
			}

			timersToStartDates[node.id] = startDate;
		} else if (node.modeContext() === 'timer') {
			if (node.parent.mode() === 'timer') {
				var startTime = 0,
					eachSibling = node.previousSibling;
				while (eachSibling) {
					var eachDuration = eachSibling.tag('duration');
					if (eachDuration) {
						startTime += eachDuration;
					}
					eachSibling = eachSibling.previousSibling;
				}
				startDate = new Date(getStartDate(node.parent).getTime() + (startTime * 1000));
			}
		}

		return startDate;
	}

	function getFinishDate(node) {
		var each = node,
			duration;
		while (each) {
			duration = each.tag('duration');
			if (duration) {
				var startDate = getStartDate(each);
				return new Date(startDate.getTime() + (duration * 1000));
			}
			each = each.previousSibling;
		}
		return getStartDate(node);
	}

	function isDateOverlappingStep(date, stepNode) {
		var stepStartDate = getStartDate(stepNode);
		if (stepStartDate) {
			var stepDuration = stepNode.tag('duration');
			if (stepDuration) {
				stepDuration *= 1000;
				var dateTime = date.getTime(),
					stepStartTime = stepStartDate.getTime();
				return dateTime >= stepStartTime && dateTime < (stepStartTime + stepDuration);
			}
		}
		return false;
	}

	function refreshCurrentStepForTimer(editor, timerNode, postNotificationOnChangeCurrentAction) {
		var currentStep = timersToCurrentSteps[timerNode.id];
		if (currentStep && currentStep.parent) {
			if (isDateOverlappingStep(currentDate, currentStep)) {
				return;
			} else {
				editor.setNeedsRender(currentStep);
			}
		}

		var eachStep = timerNode.firstChild,
			newCurrentStep;
		while (eachStep) {
			if (isDateOverlappingStep(currentDate, eachStep)) {
				newCurrentStep = eachStep;
				if (newCurrentStep) {
					editor.setNeedsRender(newCurrentStep);
				}
				break;
			}
			eachStep = eachStep.nextSibling;
		}

		if (postNotificationOnChangeCurrentAction && newCurrentStep !== currentStep) {
			if (newCurrentStep) {
				NotificationCenter.deliverNotification(newCurrentStep.text(), timerNode.line());
			} else {
				NotificationCenter.deliverNotification(timerNode.line() + ' ended');
			}
		}

		timersToCurrentSteps[timerNode.id] = newCurrentStep;
	}

	function refreshEndStepForTimer(editor, timerNode) {
		var endStep = timersToFinishLines[timerNode.id],
			newEndStep = timerNode.lastChild;

		if (endStep !== newEndStep) {
			if (endStep) {
				editor.setNeedsRender(endStep);
			}

			if (newEndStep) {
				editor.setNeedsRender(newEndStep);
			}

			timersToFinishLines[timerNode.id] = newEndStep;
		}
	}

	Extensions.addMode({
		name: 'timer'
	});

	Extensions.addInit(function (editor) {
		timers = NodeSet.setFromArray(editor.tree().evaluateNodePath('//@mode = timer'));
	});

	Extensions.add('com.foldingtext.taxonomy.classifier', {
		classify: function (text, state, previousState) {
			switch (state.type) {
			case 'empty':
			case 'codeblock':
			case 'fencedcode':
				return;
			}

			var modeStack = state.modeStack;
			if (modeStack) {
				var modeRecord = modeStack.lastObject();
				if (!modeRecord || modeRecord.mode !== 'timer') {
					return;
				}
			} else {
				return;
			}

			var duration = AmountHelper.parseDuration(text, state, previousState);
			if (duration) {
				state.tags.duration = duration;
			}
		},
		attributesToClear: [].concat(AmountHelper.attributesToClear)
	});

	Extensions.add('com.foldingtext.editor.clockTick', function (editor, date) {
		currentDate = date;
		timers.forEachNodeInSet(function (each) {
			if (each.mode() !== 'timer') {
				timers.removeNode(each);
			} else {
				refreshCurrentStepForTimer(editor, each, true);
			}
		});
	});

	Extensions.addTreeChanged(function (editor, e) {
		var deltas = e.deltas,
			effectedTimers = new NodeSet();

		for (var i = 0; i < deltas.length; i++) {
			var insertedNodes = deltas[i].insertedNodes,
				removedNodes = deltas[i].removedNodes,
				updatedNode = deltas[i].updatedNode,
				eachSibling,
				length,
				each,
				j;

			length = removedNodes.length;
			for (j = 0; j < length; j++) {
				each = removedNodes[j];
				if (timers.containsNode(each)) {
					timers.removeNode(each);
				} else {
					var previousParent = e.previousParent(each);
					if (previousParent) {
						if (previousParent.mode() === 'timer') {
							eachSibling = previousParent.firstChild;
							while (eachSibling) {
								editor.setNeedsRender(eachSibling);
								eachSibling = eachSibling.nextSibling;
							}
							effectedTimers.addNode(previousParent);
						}
					}
				}
			}

			length = insertedNodes.length;
			for (j = 0; j < length; j++) {
				each = insertedNodes[j];
				if (each.mode() === 'timer') {
					timers.addNode(each);
				} else if (each.modeContext() === 'timer') {
					eachSibling = each.nextSibling;
					while (eachSibling) {
						editor.setNeedsRender(eachSibling);
						eachSibling = eachSibling.nextSibling;
					}
					effectedTimers.addNode(each.parent);
				}
			}

			if (updatedNode) {
				if (updatedNode.mode() === 'timer') {
					timers.addNode(updatedNode);
					var eachChild = updatedNode.firstChild;
					while (eachChild) {
						editor.setNeedsRender(eachChild);
						eachChild = eachChild.nextSibling;
					}
				} else if (updatedNode.modeContext() === 'timer') {
					eachSibling = updatedNode.nextSibling;
					while (eachSibling) {
						editor.setNeedsRender(eachSibling);
						eachSibling = eachSibling.nextSibling;
					}
					effectedTimers.addNode(updatedNode.parent);
				}
			}
		}

		if (effectedTimers.count > 0) {
			effectedTimers.forEachNodeInSet(function (each) {
				if (each.mode() === 'timer') {
					// rest timer, mostly to check for case where 'Start' property gets modified
					timersToLastValidStartDates[each.id] = timersToStartDates[each.id];
					timersToStartDates[each.id] = undefined;
					refreshCurrentStepForTimer(editor, each, false);
					refreshEndStepForTimer(editor, each, false);
				}
			});
		}
	});

	function createWidget(identifier) {
		var widget = document.createElement('span');

		widget.className = 'cm-' + identifier + '-renderwidget';

		widget.addEventListener('mousedown', function (e) {
			e.preventDefault();
		});

		widget.addEventListener('mouseup', function (e) {
			e.preventDefault();
			var node = widget.ftNode,
				timerNode = node.parent,
				timerStartTime = getStartDate(timerNode).getTime(),
				stepStartTime = getStartDate(node).getTime(),
				durationToScheduleStart = stepStartTime - timerStartTime,
				newTimerStartDate = new Date(currentDate.getTime() - durationToScheduleStart);

			timerNode.setPropertyNodeValue('Start', newTimerStartDate.format('mmm d, yyyy h:MM:ss TT'));
		});

		return widget;
	}

	Extensions.addRenderNode(function (editor, node, nodeRenderer) {
		if (node.modeContext() === 'timer') {
			var duration = node.tag('duration'),
				widget;

			if (duration) {
				var startDate = getStartDate(node);
				if (startDate) {
					var anchoreDate = node.previousSibling ? getStartDate(node.previousSibling) : getStartDate(node.parent);
					widget = createWidget('timerstartdate');
					widget.textContent = startDate.formatDifference(anchoreDate);
					widget.ftNode = node;
					nodeRenderer.renderLineWidget(widget, {
						overlay : true,
						positionWidget : function (widgetWidth, widgetHeight) {
							var line = node.lineNumber(),
								leadingSpace = node.line().match(/\s*/)[0],
								coords = editor.cursorCoords({ line : line, ch : leadingSpace.length}, 'div');
							return {
								left: Math.round(coords.left - (widgetWidth + editor.defaultSpaceWidth())) + 'px'
							};
						}
					});
				}
			}

			if (timersToFinishLines[node.parent.id] === node) {
				var alignTo = node,
					alignToLeadingSpace;

				while (alignTo && !alignTo.tag('duration')) {
					alignTo = alignTo.previousSibling;
				}

				if (!alignTo) {
					alignTo = node.parent;
				}

				alignToLeadingSpace = alignTo.line().match(/\s*/)[0];
				var timerFinishDate = getFinishDate(node),
					timerStartDate = getStartDate(node.parent);

				if (currentDate > timerFinishDate) {
					widget = createWidget('timerendedfinishline');
					widget.textContent = 'Ended: ' + timerFinishDate.formatDifference(timerStartDate);
				} else {
					widget = createWidget('timerfinishline');
					widget.textContent = 'Ends: ' + timerFinishDate.formatDifference(timerStartDate);
				}

				nodeRenderer.renderLineWidget(widget, {
					showIfHidden : false,
					handleMouseEvents: true,
					positionWidget : function (widgetWidth, widgetHeight) {
						return {
							marginLeft: Math.round(editor.charCoords({ line: alignTo.lineNumber(), ch: alignToLeadingSpace.length }, 'local').left) + 'px'
						};
					}
				});
			}
		}
	});

	Extensions.addRenderNodeElementStyles(function (editor, node, elementRenderer) {
		if (timersToCurrentSteps[node.parent.id] === node) {
			elementRenderer.addWrapClass('cm-timercurrentstep');
		}
	});
});