//
// Copyright (c) 2014 Jesse Grosjean. All rights reserved.
//
// Permission is given for you to use this software ONLY while creating themes
// and plugins for the "FoldingText" software platform. Permission is NOT
// given for any other purpose.
//
define(function (require, exports, module) {
	'use strict';

	var AttributeRun = require('./attributerun').AttributeRun,
		Environment = require('../util/environment'),
		ObjectUtils = require('../util/objectutils'),
		assert = require('../util/assert');

	var AttributedString = function AttributedString(string) {
		if (string === undefined) {
			string = '';
		}
		this._clean = false;
		this._string = string;
		this._pendingAddAttributes = [];
		this._cmStyleArray = null;
	};

	(function () {

		this.copy = function copy() {
			this._ensureClean();

			var theCopy = new AttributedString(this._string),
				attributeRuns = this.attributeRuns();
			if (attributeRuns) {
				var attributeRunsCopy = [],
					length = attributeRuns.length;
				for (var i = 0; i < length; i++) {
					attributeRunsCopy.push(attributeRuns[i].copy());
				}
				theCopy._attributeRuns = attributeRunsCopy;
			}
			return theCopy;
		};

		this.string = function string() {
			return this._string;
		};

		this.replaceCharactersInRange = function replaceCharactersInRange(location, length, insertedString) {
			this._validateAttributeRuns();

			var attributeRuns = this.attributeRuns(),
				maxReplacedRange = location + length,
				string = this._string;

			if (attributeRuns) {
				var updateFromIndex,
					startingIndex,
					endingIndex;

				// If replace length is 0 then favor run at previous index to location, and send ending run to start run.
				if (length === 0) {
					if (location === 0) {
						startingIndex = 0;
					} else {
						startingIndex = this._indexOfAttributeRunWithCharacterIndex(location > 0 ? location - 1 : location);
					}
					endingIndex = startingIndex;
				// Else if replace length is not 0 get start run at location and end run at maxReplacedRange - 1.
				} else {
					startingIndex = this._indexOfAttributeRunWithCharacterIndex(location);
					endingIndex = this._indexOfAttributeRunWithCharacterIndex(Math.max(maxReplacedRange - 1, 0));
				}

				var startingAttributeRun = attributeRuns[startingIndex],
					endingAttributeRun = attributeRuns[endingIndex],
					additionalLength = insertedString.length - length;

				// If dealing with single run just chagne it's length.
				if (startingIndex === endingIndex) {
					startingAttributeRun.length += additionalLength;
					// If it's length changed to 0 and their are more runs, then remove starting run.
					if (startingAttributeRun.length === 0 && attributeRuns.length > 1) {
						attributeRuns.splice(startingIndex, 1);
						updateFromIndex = startingIndex;
					} else {
						updateFromIndex = startingIndex + 1;
					}
				// Else if dealing with different start and end attibute runs then extend start run to include inserted text
				// and shift end run to start after end of inserted text. Also remove all runs between start and end.
				} else {
					var maxInsertedRange = location + insertedString.length;
					startingAttributeRun.length = maxInsertedRange - startingAttributeRun.location;
					endingAttributeRun.length += endingAttributeRun.location - maxReplacedRange;
					endingAttributeRun.location = maxInsertedRange;

					// 1. Begin removing at start run + 1 and remove up to end run.
					var startingRemoveIndex = startingIndex + 1,
						removeLength = endingIndex - startingRemoveIndex;

					// 2. If starting run is left with length 0 then remove it too.
					if (startingAttributeRun.length === 0) {
						startingRemoveIndex--;
						removeLength++;
					}

					// 3. If ending run is left with length 0 then remove it and start updating from that index.
					if (endingAttributeRun.length === 0) {
						removeLength++;
						updateFromIndex = startingRemoveIndex;
					// 4. Else start updating from the next index since the end run is already valid.
					} else {
						updateFromIndex = startingRemoveIndex + 1;
					}

					if (removeLength > 0) {
						attributeRuns.splice(startingRemoveIndex, removeLength);

						// 5. If Removed all runs we need to add an empty one. We'll just
						// reset and reuse the startingAttributeRun in this case.
						if (attributeRuns.length === 0) {
							attributeRuns.push(startingAttributeRun);
							startingAttributeRun.attributes = {};
							updateFromIndex = 1; // nothing to update
						}
					}
				}

				// Update trailing runs start locations if length changed.
				if (additionalLength !== 0) {
					while (updateFromIndex < attributeRuns.length) {
						attributeRuns[updateFromIndex++].location += additionalLength;
					}
				}
			}

			this._string = string.substring(0, location) + insertedString + string.substring(maxReplacedRange);
			this._clean = false;
			this._cmStyleArray = null;
			this._validateAttributeRuns();
		};

		this.attributeRuns = function attributeRuns() {
			var runs = this._attributeRuns,
				pendingAddAttributes = this._pendingAddAttributes;

			if (!runs) {
				runs = [new AttributeRun(0, this._string.length, {})];
				this._attributeRuns = runs;
			}

			var length = pendingAddAttributes.length;
			if (length) {
				this._pendingAddAttributes = [];
				var eachPending;
				for (var i = 0; i < length; i++) {
					eachPending = pendingAddAttributes[i];
					this._addAttributeInRange(eachPending.attribute, eachPending.value, eachPending.location, eachPending.length);
				}
			}

			return runs;
		};

		this.attributesAtIndex = function attributesAtIndex(index, effectiveRange, longestEffectiveRange) {
			this._ensureClean();

			var runIndex = this._indexOfAttributeRunWithCharacterIndex(index);
			if (runIndex === -1) {
				return null;
			}
			var attributeRun = this.attributeRuns()[runIndex];
			if (effectiveRange) {
				effectiveRange.location = attributeRun.location;
				effectiveRange.length = attributeRun.length;
			}
			if (longestEffectiveRange) {
				var attributes = attributeRun.attributes;
				this._longestEffectiveRange(runIndex, attributeRun, longestEffectiveRange, function (candiateRun) {
					return ObjectUtils.shallowEquals(candiateRun.attributes, attributes);
				});
			}
			return attributeRun.attributes;
		};

		this.attributeAtIndex = function attributeAtIndex(attribute, index, effectiveRange, longestEffectiveRange) {
			this._ensureClean();

			var runIndex = this._indexOfAttributeRunWithCharacterIndex(index);
			if (runIndex === -1) {
				return null;
			}
			var attributeRun = this.attributeRuns()[runIndex];
			if (effectiveRange) {
				effectiveRange.location = attributeRun.location;
				effectiveRange.length = attributeRun.length;
			}

			if (longestEffectiveRange) {
				var comparisonAttribute = attributeRun.attributes[attribute];
				this._longestEffectiveRange(runIndex, attributeRun, longestEffectiveRange, function (candiateRun) {
					return candiateRun.attributes[attribute] === comparisonAttribute;
				});
			}

			return attributeRun.attributes[attribute];
		};

		this.findAttributeValue = function findAttributeValue(attribute, start) {
			var runs = this.attributeRuns(),
				runIndex = !start ? 0 : this._indexOfAttributeRunWithCharacterIndex(start),
				runCount = runs.length;

			for (; runIndex < runCount; runIndex++) {
				var runAttributes = runs[runIndex].attributes;
				if (runAttributes[attribute]) {
					return runAttributes[attribute];
				}
			}
		};

		this.attributeRangeCalculatedByFunction = function attributeRangeCalculatedByFunction(index, shouldExtendRangeToIncludeRun) {
			this._ensureClean();

			var runIndex = this._indexOfAttributeRunWithCharacterIndex(index);
			if (runIndex === -1) {
				return null;
			}

			var longestEffectiveRange = { location : index, length : 0 },
				attributeRun = this.attributeRuns()[runIndex];
			if (shouldExtendRangeToIncludeRun(attributeRun)) {
				this._longestEffectiveRange(runIndex, attributeRun, longestEffectiveRange, shouldExtendRangeToIncludeRun);
			}
			return longestEffectiveRange;
		};

		this._longestEffectiveRange = function _longestEffectiveRange(runIndex, attributeRun, longestEffectiveRange, shouldExtendRunToInclude) {
			var attributeRuns = this.attributeRuns(),
				length = attributeRuns.length,
				nextRun,
				nextIndex = runIndex - 1,
				currentRun = attributeRun;
			// scan backwards
			while (nextIndex >= 0) {
				nextRun = attributeRuns[nextIndex];
				if (shouldExtendRunToInclude(nextRun)) {
					currentRun = nextRun;
					nextIndex--;
				} else {
					break;
				}
			}

			longestEffectiveRange.location = currentRun.location;

			nextIndex = runIndex + 1;
			currentRun = attributeRun;
			// scan forwards
			while (nextIndex < length) {
				nextRun = attributeRuns[nextIndex];
				if (shouldExtendRunToInclude(nextRun)) {
					currentRun = nextRun;
					nextIndex++;
				} else {
					break;
				}
			}

			longestEffectiveRange.length = (currentRun.location + currentRun.length) - longestEffectiveRange.location;
		};

		this._addAttributeInRange = function _addAttributeInRange(attribute, value, location, length) {
			if (length === 0) {
				return;
			}

			var attributeRuns = this.attributeRuns(),
				startRunIndex = this._indexOfAttributeRunForCharacterIndex(location),
				endRunIndex = this._indexOfAttributeRunForCharacterIndex(location + length),
				i = startRunIndex;

			if (endRunIndex === -1) {
				endRunIndex = attributeRuns.length;
			}

			while (i < endRunIndex) {
				attributeRuns[i].attributes[attribute] = value;
				i++;
			}

			this._clean = false;
			this._cmStyleArray = null;
		};

		this.addAttributeInRange = function addAttributeInRange(attribute, value, location, length) {
			if (length === 0) {
				return;
			}

			// Add attribute in range is somewhat expensive. Particuarly 
			this._pendingAddAttributes.push({
				attribute: attribute,
				value: value,
				location: location,
				length: length
			});

			this._clean = false;
			this._cmStyleArray = null;
		};

		this.hasAttributes = function hasAttributes() {
			return this._attributeRuns || (this._pendingAddAttributes && this._pendingAddAttributes.length > 0);
		};

		this.removeAttributeInRange = function removeAttributeInRange(attribute, location, length) {
			if (length === 0 || !this.hasAttributes()) {
				return false;
			}
			return this.removeAttributesInRange([attribute], location, length);
		};

		this.removeAttributesInRange = function removeAttributesInRange(attributes, location, length) {
			if (length === 0 || !this.hasAttributes()) {
				return false;
			}

			var didRemove = false,
				attributeRuns = this.attributeRuns(),
				startRunIndex = location === undefined ? 0 : this._indexOfAttributeRunForCharacterIndex(location),
				endRunIndex = location === undefined ? -1 : this._indexOfAttributeRunForCharacterIndex(location + length),
				attributesLength = attributes.length,
				i = startRunIndex,
				j;

			if (endRunIndex === -1) {
				endRunIndex = attributeRuns.length;
			}

			while (i < endRunIndex) {
				var attributeRunAttributes = attributeRuns[i].attributes;
				for (j = 0; j < attributesLength; j++) {
					var attribute = attributes[j];
					if (attributeRunAttributes[attribute] !== undefined) {
						delete attributeRunAttributes[attribute];
						didRemove = true;
					}
				}
				i++;
			}

			if (didRemove) {
				this._clean = false;
				this._cmStyleArray = null;
			}

			return didRemove;
		};

		this.cacheKeyForRange = function cacheKeyForRange(location, length) {
			var key = '',
				string = this.string(),
				attributeRuns = this.attributeRuns(),
				start = this._indexOfAttributeRunForCharacterIndex(location),
				end = this._indexOfAttributeRunForCharacterIndex((location + length) - 1),
				eachRun,
				eachAttributes,
				eachName,
				i;

			for (i = start; i <= end; i++) {
				eachRun = attributeRuns[i];
				key += string.substr(eachRun.location, eachRun.length);
				eachAttributes = eachRun.attributes;
				for (eachName in eachAttributes) {
					if (eachAttributes.hasOwnProperty(eachName)) {
						key += eachName;
					}
				}
			}

			return key;
		};

		this._ensureClean = function cleanUp() {
			if (this._clean) {
				return;
			}

			var attributeRuns = this.attributeRuns(),
				length = attributeRuns.length,
				previousAttributeRun = attributeRuns[0],
				attributeRun;

			for (var i = 1; i < length; i++) {
				attributeRun = attributeRuns[i];
				if (previousAttributeRun._mergeWithNext(attributeRun)) {
					attributeRuns.splice(i, 1);
					length--;
				} else {
					previousAttributeRun = attributeRun;
				}
			}

			this._clean = true;
		};

		this.toCodeMirrorStyleArray = function toCodeMirrorStyleArray(generateStartAndEndMarkers) {
			this._ensureClean();

			var cmStyleArray = this._cmStyleArray;
			if (!cmStyleArray) {
				var attributeRuns = this.attributeRuns(),
					length = attributeRuns.length,
					prevRun,
					nextRun,
					attributeRun,
					attributes,
					attributeNames,
					eachAttributeName;

				cmStyleArray = [];

				for (var i = 0; i < length; i++) {
					attributeRun = attributeRuns[i];
					if (attributeRun.length > 0) {
						prevRun = attributeRuns[i - 1];
						nextRun = attributeRuns[i + 1];

						attributeNames = [];
						attributes = attributeRun.attributes;
						for (eachAttributeName in attributes) {
							if (attributes[eachAttributeName] !== undefined) {
								if (generateStartAndEndMarkers && (!prevRun || prevRun.attributes[eachAttributeName] === undefined)) {
									attributeNames.push(eachAttributeName + '-start');
								}

								attributeNames.push(eachAttributeName);

								if (generateStartAndEndMarkers && (!nextRun || nextRun.attributes[eachAttributeName] === undefined)) {
									attributeNames.push(eachAttributeName + '-end');
								}
							}
						}						
						cmStyleArray.push(attributeRun.location + attributeRun.length);
						cmStyleArray.push(attributeNames.join(' '));
					}
				}

				this._cmStyleArray = cmStyleArray;
			}

			return cmStyleArray;
		};

		// no idea what i'm doing here... The intention is to figured out if
		// attribute ranges changed durring syntax hiding. Assuptions are
		// underlying text didn't change, just checking to see if syntax
		// highlighting ranges changed.
		this._attributeRunsHash = function _attributeRunsHash() {
			var attributeRuns = this.attributeRuns(),
				length = attributeRuns.length,
				attributeRun,
				hash = 0;

			for (var i = 0; i < length; i++) {
				attributeRun = attributeRuns[i];
				hash = ((hash << 5) - hash) + attributeRun.location;
				hash = ((hash << 5) - hash) + attributeRun.length;
			}
		};

		this.toString = function toString(showAttributeValues) {
			this._ensureClean();

			var string = this._string,
				attributeRuns = this.attributeRuns(),
				length = attributeRuns.length,
				results = [];

			for (var i = 0; i < length; i++) {
				var eachRun = attributeRuns[i],
					eachRunAttributes = eachRun.attributes;

				var sortedNames = [];
				for (var each in eachRunAttributes) {
					if (eachRunAttributes[each] !== undefined) {
						sortedNames.push(each);
					}
				}

				sortedNames.sort();

				var nameValues = [],
					sortedNamesLength = sortedNames.length;
				for (var j = 0; j < sortedNamesLength; j++) {
					var name = sortedNames[j];
					if (showAttributeValues) {
						nameValues.push(name + '=\"' + eachRunAttributes[name] + '\"');
					} else {
						nameValues.push(name);
					}
				}

				results.push(string.substr(eachRun.location, eachRun.length) + '/' + nameValues.join(', '));
			}

			return '(' + results.join(')(') + ')';
		};

		this._indexOfAttributeRunWithCharacterIndex = function _indexOfAttributeRunWithCharacterIndex(characterIndex) {
			if (characterIndex < 0 || characterIndex >= this._string.length) {
				return -1;
			}

			var attributeRuns = this.attributeRuns(),
				low = 0,
				high = attributeRuns.length - 1,
				i, comparison,
				run, location, length, end,
				result = -1;

			while (low <= high) {
				/*jshint bitwise: false */
				i = (low + high) >> 1;
				/*jshint bitwise: true */
				run = attributeRuns[i];
				location = run.location;
				length = run.length;
				end = location + length;

				if (characterIndex >= location && characterIndex < end) {
					result = i;
					break;
				} else if (end <= characterIndex) {
					low = i + 1;
					continue;
				} else {
					high = i - 1;
					continue;
				}
			}

			return result;
		};

		this._indexOfAttributeRunForCharacterIndex = function _indexOfAttributeRunForCharacterIndex(characterIndex, lastOnMaxIndex) {
			var runIndex = this._indexOfAttributeRunWithCharacterIndex(characterIndex);

			if (runIndex === -1) {
				return -1;
			}

			var attributeRuns = this.attributeRuns(),
				attributeRun = attributeRuns[runIndex],
				location = attributeRun.location;

			if (location === characterIndex) {
				return runIndex;
			}

			var newAttributeRun = attributeRun.splitAtIndex(characterIndex);
			attributeRuns.splice(runIndex + 1, 0, newAttributeRun);
			return runIndex + 1;
		};

		this._validateAttributeRuns = function _validateAttributeRuns() {
			if (Environment.debug) {
				var attributeRuns = this.attributeRuns();

				if (attributeRuns) {
					var length = attributeRuns.length;

					if (length) {
						var offset = 0;
						for (var i = 0; i < length; i++) {
							var each = attributeRuns[i];
							if (length > 1) {
								assert.ok(each.length > 0, 'Attribute Run Empty');
							}
							assert.ok(each.location === offset, 'Attribute Run Invalid Location');
							offset += each.length;
						}
						assert.ok(offset === this._string.length, 'Attribute Run Invalid Location');
					} else {
						assert.ok(false, 'Empty Attribute Runs');
					}
				}
			}
		};

	}).call(AttributedString.prototype);

	exports.AttributedString = AttributedString;
});
