//
// Copyright (c) 2014 Jesse Grosjean. All rights reserved.
//
// Permission is given for you to use this software ONLY while creating themes
// and plugins for the "FoldingText" software platform. Permission is NOT
// given for any other purpose.
//
define(function (require, exports, module) {
	'use strict';

	require('ft/util/array');

	var ClassifierText = require('./classifiertext').ClassifierText,
		NodeSet = require('./nodeset').NodeSet,
		Delay = require('../util/delay').Delay,
		Delta = require('./delta').Delta,
		assert = require('../util/assert'),
		DEFAULT_WORK_TIME = 10;

	var Classifier = function Classifier(tree) {
		this.tree = tree;
		this.invalidLineStarts = [0];
		this.delay = new Delay();
		this.isThisUpdatingTree = false;
		this.startWorker(100);
		tree.addEventListener('treeChanged', this._treeChanged.bind(this));
	};

	var CLASSIFY_NEXT = 1,
		RECLASSIFY_PREVIOUS = 2;

	(function () {

		//
		// Cleanup
		//

		this.cleanupForCollection = function cleanupForCollection() {
			this.delay.clear();
		};

		//
		// Classification
		//

		this.classify = function classify(workTime) {
			this.classifyWorker(null, null, workTime);
		};

		this.ensureClassified = function ensureClassified(fromNode, toNode) {
			var tree = this.tree;

			if (!fromNode) {
				fromNode = tree.firstLineNode();
			}

			if (!toNode) {
				toNode = tree.lastLineNode();
			}

			this.classifyWorker(fromNode, toNode, Infinity);
		};

		this.startWorker = function startWorker(time) {
			if (time) {
				this.delay.clear();
				this.delay.set(time, this.classifyWorker.bind(this, null, null, DEFAULT_WORK_TIME));
			}
		};

		this.classifyWorker = function classifyWorker(fromNode, toNode, workTime) {
			workTime = workTime ? workTime : Infinity;

			var tree = this.tree,
				trackOriginalStates = true,
				invalidLineStarts = this.invalidLineStarts,
				classifierText = new ClassifierText(),
				classifiedNodesOriginalStates = {},
				endTime = Date.now() + workTime,
				previousAction = CLASSIFY_NEXT,
				taxonomy = tree.taxonomy,
				classifiedNodeRanges = [],
				classifiedNodesStartLine,
				classifiedNodes,
				outerThis = this,
				returnToLine,
				returnTo,
				eachLine,
				each,
				state,
				newState;

			function prepareStateToBeginClassification() {
				var previousNode = each ? each.previousLineNode() : null;
				if (previousNode && previousNode._taxonomyState) {
					state = taxonomy.copyState(previousNode._taxonomyState);
				} else {
					state = taxonomy.startState();
				}

				if (!classifiedNodes || classifiedNodes.length > 0) {
					classifiedNodes = [];
					classifiedNodeRanges.push(classifiedNodes);
				}

				classifiedNodesStartLine = eachLine;
			}

			function retreatToValidLineToBeginClassification() {
				if (!each) {
					invalidLineStarts.sortedRemove(eachLine);
					each = tree.lastLineNode();
					if (each) {
						eachLine = each.lineNumber();
					}
				}

				if (each) {
					var skippedBackwards = false;
					while (each && each._reclassifiedRegion) {
						skippedBackwards = true;
						each = each.previousLineNode();
						eachLine--;
					}
					if (!each) {
						each = tree.firstLineNode();
						eachLine = 0;
					} else if (skippedBackwards) {
						each = each.nextLineNode();
						eachLine++;
					}
				}
			}

			function advanceToNextInvalidStart() {
				if (invalidLineStarts.length) {
					eachLine = invalidLineStarts[0];
					each = tree.lineNumberToNode(eachLine);
					retreatToValidLineToBeginClassification();
				} else {
					eachLine = null;
					each = null;
				}
				prepareStateToBeginClassification();
			}

			function goToNext() {
				each = each.nextLineNode();
				if (each) {
					eachLine++;
				} else {
					eachLine = null;
				}
			}

			function goToPrevious() {
				each = each.previousLineNode();
				eachLine--;
			}

			function goToNextInvalid() {
				if (fromNode || toNode) {
					while (each && each !== toNode && each._taxonomyState !== null && !invalidLineStarts.sortedContains(eachLine)) {
						goToNext();
					}
					if (each === toNode) {
						each = null;
						eachLine = null;
					}
					prepareStateToBeginClassification();
				} else {
					advanceToNextInvalidStart();
				}
			}

			function recordClassifiedNodeAndState(node, taxonomyState, goingForward) {
				node._taxonomyState = taxonomyState;
				if (goingForward) {
					classifiedNodes.push(node);
				} else if (eachLine < classifiedNodesStartLine) {
					classifiedNodesStartLine = eachLine;
					classifiedNodes.unshift(node);
				}
			}

			function copyStateAndGoToPrevious() {
				state = taxonomy.copyState(each._taxonomyState);
				goToPrevious();
			}

			function recordMarkAndGoPrevious() {
				returnTo = each;
				returnToLine = eachLine;
				goToPrevious();
			}

			function returnToMark() {
				var iter = each,
					end = returnTo.nextLineNode(),
					iterPrevLine,// = iter ? iter.previousLineNode() : null,
					iterState;

				while (iter && iter !== end) {
					iterState = iter._taxonomyState;
					if (iterState) {
						classifierText._setNode(iter, iterState);
						taxonomy.reclassifyDown(classifierText, iterState, iterPrevLine ? iterPrevLine._taxonomyState : null);
						iter._taxonomyState = iterState;
					}
					iterPrevLine = iter;
					iter = iter.nextLineNode();
				}
				each = returnTo;
				eachLine = returnToLine;
				returnTo = null;
				returnToLine = null;
				state = taxonomy.copyState(each._taxonomyState);
			}

			function lastPossibleEffectedNodeInRange(startLine, endNode) {
				var endNodeParent = endNode.parent;
				while (endNodeParent.lineNumber() >= startLine) {
					endNode = endNodeParent;
					endNodeParent = endNode.parent;
				}
				return endNode.lastDescendantOrSelf();
			}

			function snapshotOriginalState(node) {
				if (!trackOriginalStates) {
					return;
				}

				if (!classifiedNodesOriginalStates[node.id]) {
					var snapshot = node._taxonomyState;
					if (snapshot) {
						snapshot = taxonomy.copyState(node._taxonomyState);
					} else {
						// if we encounter a node that's never been classified
						// before then disable all trackOriginalStates
						// tracking for this classification time. We only want
						// to do that extra checking when it will make a
						// difference such as when user is typing. If they are
						// loading a file or pasting in a big chunck of text
						// then we especially want to disalbe this since it's
						// expensive to calcualte on many nodes.
						trackOriginalStates = false;
						return;
					}
					snapshot.styleArray = node.lineAttributedString().toCodeMirrorStyleArray();
					classifiedNodesOriginalStates[node.id] = snapshot;
				}
			}

			function hasChangedFromOriginalStateOrDiffersFromNodeState(node) {
				if (!trackOriginalStates) {
					return true;
				}

				var originalState = classifiedNodesOriginalStates[node.id];
				if (originalState) {
					var currentState = node._taxonomyState;

					// Check if the nodes original state was even valid in the
					// first place. It can be that that state is invalid. For
					// example if the API user does setType,
					// setTypeIndentLevel, or other taxonomy calculated
					// properties it might be that the original state is out
					// of sync. In that case then make sure to reupdate the
					// node.
					if (node._hasChangedFromTaxonomyState(originalState)) {
						return true;
					}

					if (taxonomy.equalStates(originalState, currentState)) {
						var currentStyleArray = node.lineAttributedString().toCodeMirrorStyleArray();
						if (currentStyleArray.shallowEquals(originalState.styleArray)) {
							return false;
						}
					}
				}
				return true;
			}

			function saveStartAndApplyStateToClassifiedNodes(nextInvalidLine) {
				tree.beginUpdates();

				for (var j = 0; j < classifiedNodeRanges.length; j++) {
					var nodesInRange = classifiedNodeRanges[j],
						nodeStructureLevelChanged = false,
						length = nodesInRange.length,
						eachStructureLevel,
						each;

					each = nodesInRange[0];
					while (each && !hasChangedFromOriginalStateOrDiffersFromNodeState(each)) {
						nodesInRange.shift();
						each = nodesInRange[0];
						length--;
					}

					each = nodesInRange.lastObject();
					while (each && !hasChangedFromOriginalStateOrDiffersFromNodeState(each)) {
						nodesInRange.pop();
						each = nodesInRange.lastObject();
						length--;
					}

					if (length) {
						var rangeStartNode = nodesInRange[0],
							rangeStartLine = rangeStartNode.lineNumber(),
							rangeEndNode = nodesInRange[nodesInRange.length - 1],
							rangeMaxEffectedNode = rangeEndNode,
							rangeMaxEffectedNodeBeforeUpdate = lastPossibleEffectedNodeInRange(rangeStartLine, rangeEndNode);

						tree.disableDeltaRegistration();

						var i;
						for (i = 0; i < length; i++) {
							each = nodesInRange[i];
							eachStructureLevel = each._structureLevel;
							each._updateFromTaxonomyState(each._taxonomyState);
							if (eachStructureLevel !== each._structureLevel) {
								nodeStructureLevelChanged = true;
							}
						}

						if (nextInvalidLine !== null) {
							invalidLineStarts.sortedAdd(nextInvalidLine, true);
							outerThis.startWorker(100);
						}

						tree.enableDeltaRegistration();

						if (nodeStructureLevelChanged) {
							var rangeMaxEffectedNodeAfterUpdate = lastPossibleEffectedNodeInRange(rangeStartLine, rangeEndNode);

							if (rangeMaxEffectedNodeAfterUpdate.lineNumber() > rangeMaxEffectedNodeBeforeUpdate.lineNumber()) {
								rangeMaxEffectedNode = rangeMaxEffectedNodeAfterUpdate;
							} else {
								rangeMaxEffectedNode = rangeMaxEffectedNodeBeforeUpdate;
							}

							// grow the list of nodes in range to encompase all
							// possilbe nodes that could be effected by
							// adding/removing nodes in range. That entire set is
							// sent off in the delta as added then removed.

							while (rangeEndNode !== rangeMaxEffectedNode) {
								rangeEndNode = rangeEndNode.nextLineNode();
								nodesInRange.push(rangeEndNode);
							}
						}

						if (nodesInRange.length === 1) {
							if (tree.isDeltaRegistrationEnabled()) {
								tree.recordDelta(new Delta(
									rangeStartNode.lineTextStart(),
									undefined,
									undefined,
									{line: rangeStartLine, ch: 0},
									{line: rangeMaxEffectedNode.lineNumber(), ch: rangeMaxEffectedNode.line().length},
									rangeStartNode,
									[],
									[]
								));
							}
						} else {
							// When the classifier runs no text content is ever
							// changing only the tree structure is changing. Nodes
							// are added/removed to reflect this new structure.
							// The default implementation of those methods
							// generate text deltas. These are expensive, and even
							// just changed() events are expensive too. Becaues of
							// this when the classifier applies it's changes it
							// disables delta registration and then records a
							// single delta assuming that all effected nodes are
							// removed and then inserted. This is over agressive
							// reporting for cases where node structure level
							// didn't change but it's efficient and will make any
							// tree changed listeners fully aware of all changes
							// in the tree.
							if (tree.isDeltaRegistrationEnabled()) {
								tree.recordDelta(new Delta(
									rangeStartNode.lineTextStart(),
									undefined,
									undefined,
									{line: rangeStartLine, ch: 0},
									{line: rangeMaxEffectedNode.lineNumber(), ch: rangeMaxEffectedNode.line().length},
									null,
									nodesInRange,
									nodesInRange
								));
							}
						}
					}
				}

				outerThis.isThisUpdatingTree = true;
				tree.endUpdates();
				outerThis.isThisUpdatingTree = false;
			}

			if (toNode) {
				toNode = toNode.nextLineNode();
			}

			if (fromNode) {
				each = fromNode;
				eachLine = each.lineNumber();
				goToNextInvalid();
				if (!each) {
					return;
				}
				retreatToValidLineToBeginClassification();
				prepareStateToBeginClassification();
			} else {
				advanceToNextInvalidStart();
			}

			if (!each) {
				return;
			}

			while (each) {
				var originalState = each._taxonomyState,
					originalWasReclassifiedRegion = each._reclassifiedRegion,
					nextAction;

				snapshotOriginalState(each);

				//
				// Calculate state
				//

				if (previousAction === CLASSIFY_NEXT) {
					newState = taxonomy.startState();
					classifierText._setNode(each, null);
					classifierText.removeAttributesInRange(taxonomy.textAttributesToClearBeforeClassify());
					taxonomy.classify(classifierText, newState, state);
					state = newState;
					nextAction = state.nextAction;
					recordClassifiedNodeAndState(each, state, true);
					each._reclassifiedRegion = false;
					invalidLineStarts.sortedRemove(eachLine);
				} else if (previousAction === RECLASSIFY_PREVIOUS) {
					var reclassifiedState = taxonomy.copyState(each._taxonomyState);
					classifierText._setNode(each, reclassifiedState);
					taxonomy.reclassifyUp(classifierText, reclassifiedState, state);
					nextAction = reclassifiedState.nextAction;
					recordClassifiedNodeAndState(each, reclassifiedState, false);
					each._reclassifiedRegion = true; // not sure about this, but taking it out fixes a bug case and the rest of the specs still pass.
				} else {
					throw 'Invalid next action';
				}

				//
				// Figure out next step
				//

				if (nextAction === CLASSIFY_NEXT) {
					if (returnTo) {
						returnToMark();
						goToNext();
					} else {
						if (!originalWasReclassifiedRegion && taxonomy.equalStates(originalState || {}, state)) {
							// if node state didn't change, don't include node
							// in set of classified nodes since that's the set
							// used to generate change delta and we want those
							// change deltas to be as small as possible

							// NO, this is problematic. If node gets edited
							// such that it's original string is replaced with
							// exact same text then it will lose all syntax
							// highlight attributes in that edit (and get
							// rendered to screen with out them) but then
							// later when it's reclassified it won't be
							// included in code because of this check. Better
							// approach is to catch nodes that didn't really
							// change in render logic.

							//var popped = classifiedNodes.pop();
							//assert.ok(each === popped, 'State did not change, so pop off the current node');
							goToNextInvalid();
						} else {
							goToNext();
						}
					}

					if (each === toNode) {
						saveStartAndApplyStateToClassifiedNodes(eachLine);
						return;
					}

					if (workTime !== Infinity && endTime < Date.now()) {
						saveStartAndApplyStateToClassifiedNodes(eachLine);
						return;
					}
				} else {
					each._reclassifiedRegion = true;
					if (returnTo) {
						copyStateAndGoToPrevious();
						if (!each || !each._taxonomyState) {
							returnToMark();
							goToNext();							
							nextAction = CLASSIFY_NEXT;
						}
					} else {
						recordMarkAndGoPrevious();
					}
				}

				previousAction = nextAction;
			}

			saveStartAndApplyStateToClassifiedNodes(null);

			this.delay.clear();
		};

		//
		// Invalidation
		//

		this._treeChanged = function treeChanged(e) {
			if (this.isThisUpdatingTree) {
				return;
			}
			this.startWorker(100);
		};

		this._invalidateLinesForDelta = function _invalidateLinesForDelta(delta) {
			if (delta.textChanged()) {
				var invalidLineStarts = this.invalidLineStarts,
					hasInsertedOrDeletedNodes = delta.insertedNodes.length > 0 || delta.removedNodes.length > 0,
					changeInLinesLength = delta.changeInLinesLength(),
					length = invalidLineStarts.length,
					from = delta.fromLineCh.line,
					newInvalidLineStarts = [],
					to = delta.toLineCh.line,
					eachLine,
					i;

				for (i = length - 1; i >= 0; i--) {
					eachLine = invalidLineStarts[i];
					if (eachLine > from) {
						if (eachLine < to) {
							eachLine = from;
						} else if (changeInLinesLength !== 0) {
							eachLine += changeInLinesLength;
						}
					}
					newInvalidLineStarts.sortedAdd(eachLine, true);
				}

				invalidLineStarts = newInvalidLineStarts;
				this.invalidLineStarts = invalidLineStarts;

				if (delta.updatedNode) {
					invalidLineStarts.sortedAdd(from, true);
					if (hasInsertedOrDeletedNodes) {
						invalidLineStarts.sortedAdd(from + 1, true);
					}
				} else if (hasInsertedOrDeletedNodes) {
					if (delta.insertedText[0] === '\n') {
						// Handle case where node is appended at end of tree.
						// In that case the from is equal to the previous last
						// line in tree... but the new line is actually
						// inserted at from + 1, and thats the line that we
						// want to track.
						invalidLineStarts.sortedAdd(from, true);
						invalidLineStarts.sortedAdd(from + 1, true);
					} else {
						if (from > 0) {
							invalidLineStarts.sortedAdd(from - 1, true);
						}
						invalidLineStarts.sortedAdd(from, true);
					}
				}

				var tree = this.tree,
					maxLine = tree.nodeCount() - 1;

				while (invalidLineStarts.lastObject() > maxLine) {
					invalidLineStarts.pop();
				}				
			}
		};

	}).call(Classifier.prototype);

	Classifier.CLASSIFY_NEXT = CLASSIFY_NEXT;
	Classifier.RECLASSIFY_PREVIOUS = RECLASSIFY_PREVIOUS;

	exports.Classifier = Classifier;
});