//
// Copyright (c) 2014 Jesse Grosjean. All rights reserved.
//
// Permission is given for you to use this software ONLY while creating themes
// and plugins for the "FoldingText" software platform. Permission is NOT
// given for any other purpose.
//
define(function (require, exports, module) {
	'use strict';

	var ESCAPE_SYNTAX_CHAR = '\uE002';

	var ClassifierText = function ClassifierText() {
		this._node = null;
		this._attributedString = null;
		this._text = null;
		this._textStart = 0;
		this._escapedText = null;
		this._textStream = null;

		var codemirrorModuleID = 'codemirror';
		if (require.defined(codemirrorModuleID)) {
			var CodeMirror = require(codemirrorModuleID).CodeMirror;
			this._textStream = new CodeMirror.StringStream('', 4);
		}
	};

	(function () {

		this._setNode = function _setNode(node, state) {
			this._node = node;
			this._attributedString = node.lineAttributedString();
			this._text = state !== null ? state.text : this._attributedString.string();
			this._textStart = state !== null ? state.textStart : 0;
			this._escapedText = null;
		};

		this.textStart = function textStart() {
			return this._textStart;
		};

		this.text = function text() {
			return this._text;
		};

		this.textCharAt = function textCharAt(index) {
			return this._text.charAt(index);
		};

		this.textMatch = function textMatch(regex) {
			return this._text.match(regex);
		};

		this.textLength = function textLength() {
			return this._text.length;
		};

		this.textSubstr = function substr(start, length) {
			return this._text.substr(start, length);
		};

		this.textSubstring = function substring(from, to) {
			return this._text.substring(from, to);
		};

		this.sliceText = function sliceText(start, end) {
			this._textStart += start;
			this._text = this._text.slice(start, end);
			if (this._escapedText) {
				this._escapedText = this._escapedText.slice(start, end);
			}
		};

		this.sliceAllText = function sliceAllText(start, end) {
			var length = this._text.length;
			this.sliceText(length, length);
		};

		this.addAttributeInRange = function addAttributeInRange(attribute, value, location, length) {
			this._attributedString.addAttributeInRange(attribute, value, this._textStart + location, length);
		};

		this.removeAttributeInRange = function removeAttributeInRange(attribute, location, length) {
			this.removeAttributesInRange([attribute], location, length);
		};

		this.removeAttributesInRange = function removeAttributesInRange(attributes, location, length) {
			if (location !== undefined) {
				location += this._textStart;
			}
			this._attributedString.removeAttributesInRange(attributes, location, length);
		};

		/*this.attributesAtIndex = function attributesAtIndex(location, effectiveRange, longestEffectiveRange) {
			var attributes = this._attributedString.attributesAtIndex(this._textStart + location, effectiveRange, longestEffectiveRange);
			if (effectiveRange) {
				effectiveRange.location += this._textStart;
			}
			if (longestEffectiveRange) {
				longestEffectiveRange.location += this._textStart;
			}
			return attributes;
		};

		this.attributeAtIndex = function attributeAtIndex(attribute, location, effectiveRange, longestEffectiveRange) {
			var attributeValue = this._attributedString.attributeAtIndex(attribute, this._textStart + location, effectiveRange, longestEffectiveRange);
			if (effectiveRange) {
				effectiveRange.location += this._textStart;
			}
			if (longestEffectiveRange) {
				longestEffectiveRange.location += this._textStart;
			}
			return attributeValue;
		};*/

		this.textStream = function textStream() {
			var s = this._textStream;
			if (s) {
				s.start = 0;
				s.pos = 0;
				s.string = this._text;
				s.lastColumnPos = 0;
				s.lastColumnValue = 0;
			}
			return s;
		};

		this.escapedTextStream = function escapedTextStream() {
			var s = this._textStream;
			if (s) {
				s.start = 0;
				s.pos = 0;
				s.string = this.escapedText();
				s.lastColumnPos = 0;
				s.lastColumnValue = 0;
			}
			return s;
		};

		this.escapedText = function escapedText() {
			var result = this._escapedText;
			if (result === null) {
				result = this._text;
				if (result.indexOf('\\') !== -1) {
					/*jshint regexp: false */
					result = result.replace(/\\./g, ESCAPE_SYNTAX_CHAR + ESCAPE_SYNTAX_CHAR);
					/*jshint regexp: true */
				}
				this._escapedText = result;
			}
			return result;
		};

		this.escapedTextMatch = function escapedTextMatch(regex) {
			return this.escapedText().match(regex);
		};

		this.isFirstLine = function isFirstLine() {
			return this._node.previousLineNode() === null;
		};

		this.isLastLine = function isLastLine() {
			return this._node.nextLineNode() === null;
		};

	}).call(ClassifierText.prototype);

	exports.ClassifierText = ClassifierText;
});