//
// Copyright (c) 2013 Jesse Grosjean. All rights reserved.
//
// Permission is given for you to use this software ONLY while creating themes
// and plugins for the "FoldingText" software platform. Permission is NOT
// given for any other purpose.
//

/** @module ft/core/delta */
define(function (require, exports, module) {
	'use strict';

	var emptyArray = [];

	/**
	 * A contiguous region of change in a {@link
	 * module:ft/core/treechangedevent.TreeChangedEvent|TreeChangedEvent}.
	 *
	 * The delta may be reporting changes in the underlying text content, in
	 * which case the `replacedText` and `insertedText` members are defined
	 * and `textChanged()` returns true. Or it may only be reporting node
	 * structural changes (such as those generated by the {@tutorial
	 * Classification} process) in which case those members are not defined
	 * and `textChanged` returns false.
	 *
	 * @private
	 * @constructor
	 * @alias module:ft/core/delta.Delta*/
	var Delta = function Delta(location, replacedText, insertedText, fromLineCh, toLineCh, updatedNode, removedNodes, insertedNodes) {

		/**
		 * Text offset location where this delta starts.
		 * @type Number
		 */
		this.location = location;

		/**
		 * Text replaced by this delta. May be undefined if delta is only
		 * reporting node structural change.
		 * @type String
		 */
		this.replacedText = replacedText;

		/**
		 * Text inserted by this delta. May be undefined if delta is only
		 * reporting node structural change.
		 * @type String
		 */
		this.insertedText = insertedText;

		/**
		 * LineChar location where this delta starts.
		 * @type LineChar
		 */
		this.fromLineCh = fromLineCh;

		/**
		 * LineCh location where this delta ends.
		 * @type LineChar
		 */
		this.toLineCh = toLineCh;

		/**
		 * Node updated by this delta. May be undefined if no node was
		 * updated. For example it could be that nodes were only inserted or
		 * removed.
		 * @type module:ft/core/node.Node
		 */
		this.updatedNode = updatedNode;

		/**
		 * Nodes removed by this delta change. It may be that a single node is
		 * in both removed and inserted in the same delta. For example this
		 * could happen if a node is restructed by the {@tutorial
		 * Classification} process.
		 * @type Array.<module:ft/core/node.Node>
		 */
		this.removedNodes = removedNodes ? removedNodes : emptyArray;

		/**
		 * Nodes inserted by this delta change. It may be that a single node
		 * is in both removed and inserted in the same delta. For example this
		 * could happen if a node is restructed by the {@tutorial
		 * Classification} process.
		 * @type Array.<module:ft/core/node.Node>
		 */
		this.insertedNodes = insertedNodes ? insertedNodes : emptyArray;
	};

	/**
	 * Return true if this delta represents text change.
	 *
	 * @returns {Boolean}
	 */
	Delta.prototype.textChanged = function textChanged() {
		return this.insertedText !== undefined;
	};

	/**
	 * The change in text length represented by this delta.
	 *
	 * @returns {Number}
	 */
	Delta.prototype.changeInTextLength = function changeInTextLength() {
		if (this.insertedText !== undefined) {
			return this.insertedText.length - this.replacedText.length;
		} else {
			return 0;
		}
	};

	/**
	 * The change in lines length represented by this delta.
	 *
	 * @returns {Number}
	 */
	Delta.prototype.changeInLinesLength = function changeInLinesLength() {
		return this.insertedNodes.length - this.removedNodes.length;
	};

	Delta.prototype.toString = function toString() {
		return 'replace(' + this.location + ', ' + this.replacedText + ',' + this.insertedText + ')';
	};

	exports.Delta = Delta;
});
