//
// Copyright (c) 2014 Jesse Grosjean. All rights reserved.
//
// Permission is given for you to use this software ONLY while creating themes
// and plugins for the "FoldingText" software platform. Permission is NOT
// given for any other purpose.
//

/**
 * @module ft/core/extensions
 */
define(function (require, exports, module) {
	'use strict';

	/**
	 * Extension point process order.
	 *
	 * @typedef {(module:ft/core/extensions.PriorityNormal |
	 *			module:ft/core/extensions.PriorityFirst |
	 *			module:ft/core/extensions.PriorityLast |
	 *			Number)} ExtensionPriority
	 */

	var pointsToExtensions = {},
		PriorityNormal = 0,
		PriorityFirst = -1,
		PriorityLast = 1;

	/**
	 * Extensions add new behavior to FoldingText.
	 * 
	 * This class manages those extensions. Use it to add your extension with
	 * to a specific extension point. The extension logic is different for
	 * each extension point, see each extension points example code to get
	 * started.
	 *
	 * Each extension point may have multiple extensions registered with it.
	 * When you add an extension use the optional `priority` parameter to help
	 * determine the order that your extension is called in.
	 *
	 * @private
	 * @constructor
	 * @alias module:ft/core/extensions.Extensions*/
	var Extensions = function Extensions() {
	};


	Extensions.add = function add(extensionPoint, extension, priority) {
		var extensions = pointsToExtensions[extensionPoint];
		if (!extensions) {
			extensions = [];
			pointsToExtensions[extensionPoint] = extensions;
		}

		if (priority === undefined) {
			priority = PriorityNormal;
		}

		extensions.push({
			extension: extension,
			priority: priority
		});

		extensions.needsSort = true;
	};

	Extensions.remove = function remove(extensionPoint, filter) {
		var extensions = pointsToExtensions[extensionPoint];
		if (extensions) {
			var length = extensions.length,
				i;
			for (i = length - 1; i >= 0; i--) {
				if (filter(extensions[i].extension)) {
					extensions.splice(i, 1);
				}
			}
		}
	};

	Extensions.hasExtensions = function hasExtensions(extensionPoint) {
		var extensions = pointsToExtensions[extensionPoint];
		return extensions ? extensions.length > 0 : false;
	};

	Extensions.processExtensions = function processExtensions(extensionPoint, block, returnFirst) {
		var extensions = pointsToExtensions[extensionPoint],
			length = extensions ? extensions.length : 0,
			i;

		if (extensions && extensions.needsSort) {
			extensions.sort(function (a, b) {
				return a.priority - b.priority;
			});
			extensions.needsSort = false;
		}

		for (i = 0; i < length; i++) {
			var eachResult = block(extensions[i].extension);
			if (eachResult !== undefined && returnFirst) {
				return eachResult;
			}
		}
	};

	/**
	 * Default extension processing priority.
	 */
	Extensions.PriorityNormal = PriorityNormal;

	/**
	 * Extensions added with PriorityFirst priority are processed before other
	 * extensions.
	 */
	Extensions.PriorityFirst = PriorityFirst;

	/**
	 * Extensions added with PriorityLast priority are processed before other
	 * extensions.
	 */
	Extensions.PriorityLast = PriorityLast;


	/**
	 * Make a new command available to the editor. If you include a
	 * `description` then the command will be included in the Edit > Run
	 * Command... popup.
	 *
	 * @example
	 * var Extensions = require('ft/core/extensions').Extensions;
	 * Extensions.addCommand({
	 *	name: 'date',
	 *	description: 'Insert the current date',
	 *	performCommand: function (editor) {
	 *		editor.replaceSelection(new Date().format('mediumDate'), 'around');
	 *	}
	 * });
	 *
	 * @param {Object}
	 * @param {ExtensionPriority=} - Processing priority.
	 */
	Extensions.addCommand = function (extension, priority) {
		Extensions.add('com.foldingtext.editor.commands', extension, priority);
	};

	/**
	 * Get notified when the {@link module:ft/editor/editor.Editor|Editor}
	 * is first initialized.
	 *
	 * @example
	 * var Extensions = require('ft/core/extensions').Extensions;
	 * Extensions.addInit(function (editor) {
	 * });
	 *
	 * @param {Function}
	 * @param {ExtensionPriority=} - Processing priority.
	 */
	Extensions.addInit = function (extension, priority) {
		Extensions.add('com.foldingtext.editor.init', extension, priority);
	};

	/**
	 * Declare a new mode. Modes are whitelisted, so only modes declared
	 * using this extension point will get recognized by the classification
	 * process. See the source code for FoldingText's existing `todo` and
	 * `timer` modes for an examples of how to attach special styling and
	 * behavior to a mode.
	 *
	 * @example
	 * var Extensions = require('ft/core/extensions').Extensions;
	 * Extensions.addMode({
	 *	name: 'mymodeextension'
	 * });
	 *
	 * @param {Object}
	 * @param {ExtensionPriority=} - Processing priority.
	 */
	Extensions.addMode = function (extension, priority) {
		Extensions.add('com.foldingtext.editor.mode', extension, priority);
	};

	/**
	 * Use this extension point to store additional state with the
	 * saved document. If you want to save system level defaults then use
	 * {@link module:ft/system/userdefaults.UserDefaults|UserDefaults} instead.
	 *
	 * @example
	 * var Extensions = require('ft/core/extensions').Extensions;
	 * Extensions.addState({
	 *	readState: function (editor, state) {
	 *		var mystate = state.mystate;
	 *	},
	 *	writeState: function (editor, state) {
	 *		state.mystate = 'some special state I want to save';
	 *	}
	 * });
	 *
	 * @param {Object}
	 * @param {ExtensionPriority=} - Processing priority.
	 */
	Extensions.addState = function (extension, priority) {
		Extensions.add('com.foldingtext.editor.state', extension, priority);
	};

	/**
	 * Learn when the selection changes in the editor.
	 *
	 * @example
	 * var Extensions = require('ft/core/extensions').Extensions;
	 * Extensions.addSelectionDidChange(function (editor) {
	 *	console.log(editor.selectedText());
	 * });
	 *
	 * @param {Object}
	 * @param {ExtensionPriority=} - Processing priority.
	 */
	Extensions.addSelectionDidChange = function (extension, priority) {
		Extensions.add('com.foldingtext.editor.selectionDidChange', extension, priority);
	};

	/**
	 * Learn when the node path changes in the editor.
	 *
	 * @example
	 * var Extensions = require('ft/core/extensions').Extensions;
	 * Extensions.addNodePathDidChange(function (editor) {
	 *	console.log(editor.nodePathString());
	 * });
	 *
	 * @param {Object}
	 * @param {ExtensionPriority=} - Processing priority.
	 */
	Extensions.addNodePathDidChange = function (extension, priority) {
		Extensions.add('com.foldingtext.editor.nodePathDidChange', extension, priority);
	};

	/**
	 * Adjust DOM element attributes and styles for nodes rendered in the
	 * viewport. This is called in response to CodeMirror's
	 * [renderLine](http://codemirror.net/doc/manual.html#renderLine)
	 * event.
	 *
	 * @example
	 * var Extensions = require('ft/core/extensions').Extensions;
	 * Extensions.addRenderNodeElementStyles(function (editor, node, elementRenderer) {
	 *	if (node.mode() === 'mycustommode') {
	 *		elementRenderer.addWrapClass('cm-mymodewrapper');
	 *	}
	 * });
	 *
	 * @param {Object}
	 * @param {ExtensionPriority=} - Processing priority.
	 */
	Extensions.addRenderNodeElementStyles = function (extension, priority) {
		Extensions.add('com.foldingtext.editor.renderNodeElementStyles', extension, priority);
	};

	/**
	 * Get {@link module:ft/core/treechangedevent.TreeChangedEvent|TreeChangedEvents}.
	 *
	 * @example
	 * var Extensions = require('ft/core/extensions').Extensions;
	 * Extensions.addTreeChanged(function (editor, e) {
	 *		e.deltas.forEach(function (eachDelta) {
	 *			// process each delta
	 *		})
	 *	});
	 *
	 * @param {Object}
	 * @param {ExtensionPriority=} - Processing priority.
	 */
	Extensions.addTreeChanged = function (extension, priority) {
		Extensions.add('com.foldingtext.editor.treeChange', extension, priority);
	};

	/**
	 * Add DOM widgets to nodes when they are displayed. For example this
	 * code adds a "current time" overlay widget in the whitespace before
	 * lines that have a `clock` mode.
	 *
	 * @example
	 * var Extensions = require('ft/core/extensions').Extensions;
	 * Extensions.addRenderNode(function (editor, node, nodeRenderer) {
	 *	if (node.mode() === 'clock') {
	 *		var widget = document.createElement('span');
	 *		widget.textContent = formatDate(node);
	 *		nodeRenderer.renderLineWidget(widget, {
	 *			overlay: true,
	 *			positionWidget : function (widgetWidth, widgetHeight) {
	 *				var line = node.lineNumber(),
	 *				leadingSpace = node.line().match(/\s*\/)[0],
	 *				coords = editor.cursorCoords({ line : line, ch : leadingSpace.length}, 'div');
	 *				return {
	 *					left: Math.round(coords.left - (widgetWidth + editor.defaultSpaceWidth())) + 'px'
	 *				};
	 *			}
	 *		});
	 *	}
	 * });
	 *
	 * @param {Object}
	 * @param {ExtensionPriority=} - Processing priority.
	 */
	Extensions.addRenderNode = function (extension, priority) {
		Extensions.add('com.foldingtext.editor.renderNode', extension, priority);
	};

	/**
	 * Add a text snippet to the editor.
	 *
	 * @example
	 * var Extensions = require('ft/core/extensions').Extensions;
	 * Extensions.addSnippet({
	 *	'omw' : 'On my way!'
	 * });
	 *
	 * @param {Object}
	 * @param {ExtensionPriority=} - Processing priority.
	 */
	Extensions.addSnippet = function (extension, priority) {
		Extensions.add('com.foldingtext.editor.snippets', extension, priority);
	};

	exports.Extensions = Extensions;
});