//
// Copyright (c) 2014 Jesse Grosjean. All rights reserved.
//
// Permission is given for you to use this software ONLY while creating themes
// and plugins for the "FoldingText" software platform. Permission is NOT
// given for any other purpose.
//

/** @module ft/core/node */
define(function (require, exports, module) {
	'use strict';

	var NodeSet = require('./nodeset').NodeSet,
		Taxonomy = require('./taxonomy').Taxonomy,
		NodePath = require('./nodepath').NodePath,
		AttributeRun = require('./attributerun').AttributeRun,
		AttributedString = require('./attributedstring').AttributedString,

		assert = require('../util/assert'),
		Environment = require('../util/environment'),
		ObjectUtils = require('../util/objectutils');

	/**
	 * Represents a line of text in a {@link module:ft/core/tree.Tree|Tree}.
	 *
	 * For example if we have a Tree representing a Markdown document, then
	 * each line in the document is represented by a node. The nodes can have
	 * different types, such as 'heading', 'body', etc. Nodes form a tree
	 * structure. For example a node of type 'heading' contain "body" nodes as
	 * children.
	 *
	 * To create a new node use {@link
	 * module:ft/core/tree.Tree#createNode|Tree#createNode}.
	 *
	 * @private
	 * @constructor
	 * @tutorial Nodes
	 * @alias module:ft/core/node.Node
	 */
	var Node = function Node(line, tree) {
		if (!line) {
			line = '';
		}

		/**
		 * Owning Tree.
		 * @type module:ft/core/tree.Tree
		 */
		this.tree = tree;

		/**
		 * Unique ID within Tree.
		 * @type String
		 */
		this.id = tree._nextNodeID();

		this.isTracked = false;

		/**
		 * Parent node, or null.
		 * @type module:ft/core/node.Node
		 */
		this.parent = null;

		/**
		 * First child node, or null.
		 * @type module:ft/core/node.Node
		 */
		this.firstChild = null;

		/**
		 * Last child node, or null.
		 * @type module:ft/core/node.Node
		 */
		this.lastChild = null;

		/**
		 * Previous siblingnode, or null.
		 * @type module:ft/core/node.Node
		 */
		this.previousSibling = null;

		/**
		 * Next sibling node, or null.
		 * @type module:ft/core/node.Node
		 */
		this.nextSibling = null;

		this._RBParent = null;
		this._RBLeft = null;
		this._RBRight = null;
		this._RBColor = Node.BLACK;
		this._RBTotalLineCount = 1;
		this._RBTotalLength = line.length + 1;
		this._userData = null;
		this._type = undefined;
		this._typeLevel = 0;
		this._typeIndentLevel = 0;
		this._structureLevel = 0;
		this._text = line;
		this._textStart = 0;
		this._mode = undefined;
		this._modeContext = undefined;
		this._tags = {};
		this._trackedRanges = undefined;
		this._reclassifiedRegion = false;
		this._taxonomyState = null;
		this._lineAttributedString = new AttributedString(line);
		this._renderMarks = [];
	};

	//
	// Properties
	//

	/**
	 * Node "type" determined by classification process. For example in a
	 * Markdown document a nodes' type might be "heading" or "body".
	 *
	 * @returns {String}
	 * @methodgroup Type
	 */
	Node.prototype.type = function type() {
		return this._type;
	};

	/**
	 * Set new node type. Updates the node text content to match formatting of
	 * new type. For example if changing to "heading" in a Markdown document a
	 * "#" might be added to start of line. Preserves any trailing tags. Does
	 * not preserve type indent level.
	 *
	 * @param {String}
	 * @returns {String}
	 */
	Node.prototype.setType = function setType(type) {
		return this._setTaxonomyCalculatedProperty('_type', 'setType', type, function (node, newType) {
			node._typeLevel = node.tree.taxonomy.typeLevelForType(node._type);
			node._typeIndentLevel = node.validatedTypeIndentLevel(0);
			node._structureLevel = node._typeLevel + node._typeIndentLevel;
		});
	};

	Node.prototype.typeLevel = function typeLevel() {
		return this._typeLevel;
	};

	/**
	 * The type-specific indent level of this node. For example a "heading"
	 * node will have an indent level from 1 to 6.
	 *
	 * @returns {Number}
	 */
	Node.prototype.typeIndentLevel = function typeIndentLevel() {
		return this._typeIndentLevel;
	};

	Node.prototype.validatedTypeIndentLevel = function validatedTypeIndentLevel(typeIndentLevel) {
		return this.tree.taxonomy.validatedTypeIndentLevel(this, typeIndentLevel);
	};

	/**
	 * Set type indent level for this node. For example, in a Markdown
	 * document, if you set a "heading" node to have a level of 2 then the
	 * line text will be adjusted so that it start with "##".
	 *
	 * @param {Number} - New level requested.
	 * @returns {Number} - Actual type level set, may be different than asked
	 *     for level if asked for level isn't allowed. For example "heading"
	 *     type levels stop at level 6.
	 */
	Node.prototype.setTypeIndentLevel = function setTypeIndentLevel(typeIndentLevel) {
		return this._setTaxonomyCalculatedProperty('_typeIndentLevel', 'setTypeIndentLevel', typeIndentLevel, function (node, newLevel) {
			node._structureLevel = node._typeLevel + node._typeIndentLevel;
		});
	};

	/**
	 * Set type indent level of this node and also shift descendant node
	 * levels as needed. For example if this node is a level 1 "heading" node
	 * with a level 2 "heading" as a child, then setting this node to level 2
	 * will adjust the child to be at level 3.
	 *
	 * @param {Number} - New level requested.
	 */
	Node.prototype.setBranchTypeIndentLevel = function setBranchTypeIndentLevel(typeIndentLevel) {
		typeIndentLevel = this.tree.taxonomy.validatedTypeIndentLevel(this, typeIndentLevel);

		var oldLevel = this.typeIndentLevel(),
			delta = typeIndentLevel - oldLevel;

		if (delta !== 0) {
			var tree = this.tree,
				isTracked = this.isTracked;

			if (isTracked) {
				tree.beginUpdates();
			}

			var children = [],
				typeLevel = this._typeLevel,
				child = this.firstChild;
			while (child) {
				if (child._typeLevel === typeLevel) {
					children.push(child);
				}
				child = child.nextSibling;
			}

			typeIndentLevel = this.setTypeIndentLevel(typeIndentLevel);

			delta = typeIndentLevel - oldLevel;

			if (delta !== 0) {
				var length = children.length;
				for (var i = 0; i < length; i++) {
					child = children[i];
					child.setBranchTypeIndentLevel(child.typeIndentLevel() + delta);
				}
			}

			if (isTracked) {
				tree.endUpdates();
			}
		}
	};

	Node.prototype.validatedBranchTypeIndentLevel = function validatedBranchTypeIndentLevel(requestedTypeIndentLevel) {
		var possibleTypeIndentLevel = this.tree.taxonomy.validatedTypeIndentLevel(this, requestedTypeIndentLevel),
			oldLevel = this.typeIndentLevel(),
			delta = possibleTypeIndentLevel - oldLevel;

		if (delta !== 0) {
			var typeLevel = this.typeLevel(),
				child = this.firstChild;

			while (child) {
				if (child.typeLevel() === typeLevel) {
					var childRequestedTypeIndentLevel = child.typeIndentLevel() + delta,
						childPossibleTypeIndentLevel = child.validatedBranchTypeIndentLevel(childRequestedTypeIndentLevel);

					if (childRequestedTypeIndentLevel !== childPossibleTypeIndentLevel) {
						return oldLevel; // child can't move so branch can't move either.
					}
				}
				child = child.nextSibling;
			}
		}

		return possibleTypeIndentLevel;
	};

	Node.prototype.structureLevel = function structureLevel() {
		return this._structureLevel;
	};

	/**
	 * The mode declared by this node, may be undefined. Only useful in
	 * Taxonomies that support modes.
	 *
	 * @returns {String}
	 * @methodgroup Mode
	 */
	Node.prototype.mode = function mode() {
		return this._mode;
	};

	/**
	 * Set the mode associated with this node. Only useful in Taxonomies that
	 * support modes.
	 *
	 * @param {String} - New mode requested.
	 * @returns {String} - Actual mode applied.
	 */
	Node.prototype.setMode = function setMode(mode) {
		return this._setTaxonomyCalculatedProperty('_mode', 'setMode', mode, function (node, newMode) {
			node._modeContext = node._mode;
		});
	};

	/**
	 * The mode affecting this node. For example if a "heading" node declares
	 * a mode, then all nodes contained by that heading will have that mode as
	 * their mode context. Only useful in Taxonomies that support modes.
	 *
	 * @returns {String}
	 */
	Node.prototype.modeContext = function modeContext() {
		return this._modeContext;
	};

	Node.prototype.userData = function userData(key) {
		var record = this._userData;
		if (!record) {
			record = {};
			this._userData = record;
		}
		return record[key];
	};

	Node.prototype.setUserData = function setUserData(key, value) {
		var record = this._userData,
			isTracked = this.isTracked,
			tree = this.tree;

		if (!record) {
			record = {};
			this._userData = record;
		}

		if (isTracked) {
			tree.beginUpdates();
		}

		record[key] = value;

		if (isTracked) {
			this.changed();
			tree.endUpdates();
		}
	};

	Node.prototype._setTaxonomyCalculatedProperty = function _setTaxonomyCalculatedProperty(name, setter, value, onSet) {
		var tree = this.tree,
			oldValue = this[name],
			isTracked = this.isTracked;

		if (isTracked) {
			tree.beginUpdates();
		}

		var newValue = tree.taxonomy[setter](this, value);
		if (oldValue !== newValue) {
			var oldStructureLevel = this._structureLevel;

			this[name] = newValue;
			if (onSet) {
				onSet(this, newValue);
			}
			if (isTracked) {
				if (oldStructureLevel !== this._structureLevel) {
					var nextLineNode = this.nextLineNode();
					tree.removeNode(this);
					tree.insertNodeBefore(this, nextLineNode);
				}
			}
		}

		if (isTracked) {
			tree.endUpdates();
		}

		return newValue;
	};

	Node.prototype.changed = function changed() {
		if (this.isTracked) {
			var tree = this.tree;
			tree.beginUpdates();
			tree._nodeLineWasUpdated(this);
			tree.endUpdates();
		}
	};

	//
	// Tags
	//

	/**
	 * Readonly Object of tags associated with this node. Properties in
	 * returned object are tag names, property values are the tag values. Use
	 * `addTag` and `removeTag` to change tags.
	 *
	 * @returns {Object}
	 * @methodgroup Tags
	 */
	Node.prototype.tags = function tags() {
		return this._tags;
	};

	/**
	 * Returns value of tag. If value is the empty string then the tag exists
	 * in this node, but has no associated value. If value is `undefined`
	 * then tag doesn't exist in this node.
	 *
	 * @param {String}
	 * @returns {String}
	 */
	Node.prototype.tag = function tag(tagName) {
		return this.tags()[tagName];
	};

	/**
	 * Return true of this node has the given tag.
	 *
	 * @param {String}
	 * @returns {Boolean}
	 */
	Node.prototype.hasTag = function hasTag(tagName) {
		return this.tags()[tagName] !== undefined;
	};

	/**
	 * Add a tag and optional tag value to this node.
	 *
	 * @param {String}
	 * @param {String=}
	 */
	Node.prototype.addTag = function addTag(tagName, tagValue) {
		var tree = this.tree,
			isTracked = this.isTracked;

		if (isTracked) {
			tree.beginUpdates();
		}

		if (tree.taxonomy.addTag(this, tagName, tagValue)) {
			this._tags[tagName] = tagValue;
		}

		if (isTracked) {
			tree.endUpdates();
		}
	};

	/**
	 * Remove a tag from this node.
	 *
	 * @param {String}
	 */
	Node.prototype.removeTag = function removeTag(tagName) {
		var tree = this.tree,
			isTracked = this.isTracked;

		if (isTracked) {
			tree.beginUpdates();
		}

		if (tree.taxonomy.removeTag(this, tagName)) {
			delete this._tags[tagName];
		}

		if (isTracked) {
			tree.endUpdates();
		}
	};

	//
	// Property Nodes
	//

	Node.prototype.propertyNode = function propertyNode(propertyName, createIfNeeded) {
		var normalizedPropertyName = propertyName.toLowerCase(),
			each = this.firstChild;
		while (each && each.type() === 'property') {
			if (each.tag('name') === normalizedPropertyName) {
				return each;
			}
			each = each.nextSibling;
		}

		if (createIfNeeded) {
			return this.addPropertyNode(propertyName, '');
		} else {
			return null;
		}
	};

	Node.prototype.addPropertyNode = function addPropertyNode(propertyName, propertyValue) {
		var property = this.tree.createNode(propertyName + ' : ' + propertyValue),
			each = this.firstChild;
		while (each && each.type() === 'property') {
			each = each.nextSibling;
		}
		this.insertChildBefore(property, each);
		return property;
	};

	Node.prototype.removePropertyNode = function removePropertyNode(propertyName) {
		propertyName = propertyName.toLowerCase();
		var each = this.firstChild;
		while (each && each.type() === 'property') {
			if (each.tag('name') === propertyName) {
				each.removeFromParent();
				return;
			}
			each = each.nextSibling;
		}
	};

	Node.prototype.setPropertyNodeValue = function setPropertyNodeValue(propertyName, propertyValue) {
		this.tree.taxonomy.setPropertyNodeValue(this.propertyNode(propertyName, true), propertyValue);
	};

	//
	// Attribute Paths
	//

	Node.prototype._valueForPropertyPath = function _valueForPropertyPath(propertyPath, pathIndex) {
		var each = this.firstChild,
			end = propertyPath.length - 1,
			eachTags;

		while (each && each.type() === 'property') {
			eachTags = each.tags();
			if (eachTags.name === propertyPath[pathIndex]) {
				if (pathIndex === end) {
					return eachTags.value;
				} else {
					var result = each._valueForPropertyPath(propertyPath, pathIndex + 1);
					if (result !== undefined) {
						return result;
					}
				}
			}
			each = each.nextSibling;
		}
		return undefined;
	};

	Node.prototype._valueForLineAttributesPath = function _valueForLineAttributesPath(lineAttributesPath) {
		var attributePathLength = lineAttributesPath.length;
		if (attributePathLength === 2 && lineAttributesPath[1] === 'text') {
			// Optimize text case because it's common
			return this.text();
		} else {
			// Looking for attribute range that contains all attributePath elements
			// after 'line' and if found then return value of last attribute in path.
			var attributeRuns = this._lineAttributedString.attributeRuns(),
				length = attributeRuns.length,
				i;

			for (i = 0; i < length; i++) {
				var attributeRun = attributeRuns[i],
					attributes = attributeRun.attributes,
					matchedAll = true,
					j;

				for (j = 1; j < attributePathLength; j++) {
					if (attributes[lineAttributesPath[j]] === undefined) {
						matchedAll = false;
						break;
					}
				}

				if (matchedAll) {
					return attributes[lineAttributesPath.lastObject()];
				}
			}
			return undefined;
		}
	};

	Node.prototype.valueForAttributePath = function valueForAttributePath(attributePath) {
		var attributeName = attributePath[0];

		switch (attributeName) {
		case 'id':
			return this.id;
		case 'type':
			return this.type();
		case 'mode':
			return this.mode();
		case 'modeContext':
			return this.modeContext();
		case 'index':
			return this.indexToSelf();
		case 'line':
			if (attributePath.length === 1) {
				return this.line();
			} else {
				return this._valueForLineAttributesPath(attributePath);
			}
			break;
		case 'property':
			return this._valueForPropertyPath(attributePath, 1);
		default:
			return this.tag(attributeName);
		}
	};

	//
	// Text Content
	//

	/**
	 * Node's line of text minus leading and trailing syntax. Value depends on
	 * Taxonomy used to classify nodes. For example the `text` value of a
	 * heading in a markdown document will strip off the leading #'s.
	 *
	 * @returns {String}
	 * @methodgroup Text
	 */
	Node.prototype.text = function text() {
		return this._text;
	};

	Node.prototype.textStart = function textStart() {
		return this._textStart;
	};

	/**
	 * Replace node's text with new text. For example if you set the text of a
	 * heading in a markdown document the heading's leading # will be left in
	 * place.
	 *
	 * @param {String}
	 * @returns {String}
	 */
	Node.prototype.setText = function setText(text) {
		return this._setTaxonomyCalculatedProperty('_text', 'setText', text);
	};

	/**
	 * Entire line of text represented by node. Does not including line break.
	 *
	 * @returns {String}
	 * @methodgroup Line
	 */
	Node.prototype.line = function line() {
		return this._lineAttributedString.string();
	};

	/**
	 * Return substring of nodes line of text.
	 *
	 * @param {Number} - Relative to line start.
	 * @param {Number} - Length of text to extract.
	 * @returns {String}
	 */
	Node.prototype.lineInRange = function lineInRange(location, length) {
		var line = this.line();
		if (length === line.length + 1) {
			return line + '\n';
		} else {
			return line.substr(location, length);
		}
	};

	/**
	 * Replace represented line of text with a new line of text.
	 *
	 * @param {String}
	 */
	Node.prototype.setLine = function setLine(line) {
		this.replaceLineInRange(line, 0, this.line().length);
	};

	/**
	 * Replace a range of text in this node's line with new text.
	 *
	 * @param {String} - Text to insert, should not include line break.
	 * @param {Number} - Location, relative to start of line, to start
	 *     replacing.
	 * @param {Number} - Length of line region to replace.
	 */
	Node.prototype.replaceLineInRange = function replaceLineInRange(insertedText, location, length) {
		assert.ok(insertedText.indexOf('\n') === -1, 'Node line cannot contain newlines');

		var attributedString = this._lineAttributedString,
			taxonomy = this.tree.taxonomy,
			oldLine = this.line(),
			oldLineLength = oldLine.length,
			replacedText,
			rangeEnd,
			newLine;

		if (location !== null) {
			rangeEnd = location + length;
			replacedText = oldLine.substr(location, length);
			newLine = oldLine.substr(0, location) + insertedText + oldLine.substr(location + length);
		} else {
			location = 0;
			length = oldLineLength;
			rangeEnd = location + length;
			replacedText = oldLine;
			newLine = insertedText;
		}

		var tree = this.tree,
			isTracked = this.isTracked;

		if (isTracked) {
			tree.beginUpdates();
		}

		attributedString.replaceCharactersInRange(location, length, insertedText);

		var trackedRanges = this._trackedRanges;
		if (trackedRanges) {
			for (var i = 0; i < trackedRanges.length; i++) {
				var eachRange = trackedRanges[i];

				if (eachRange.startNode === this) {
					var startOffset = eachRange.startOffset;
					if (startOffset >= rangeEnd) {
						startOffset += insertedText.length - replacedText.length;
					} else if (startOffset > location) {
						startOffset = location;
					}
					eachRange.startOffset = startOffset;
				}

				if (eachRange.endNode === this) {
					var endOffset = eachRange.endOffset;
					if (endOffset >= rangeEnd) {
						endOffset += insertedText.length - replacedText.length;
					} else if (endOffset > location) {
						endOffset = location;
					}
					eachRange.endOffset = endOffset;
				}
			}
		}

		this._RBClearCache();

		if (isTracked) {
			tree._nodeLineWasUpdated(this, location, replacedText, insertedText);
			tree.endUpdates();
		}
	};

	Node.prototype._hasChangedFromTaxonomyState = function _hasChangedFromTaxonomyState(newState) {
		return this._type !== newState.type ||
			this._typeLevel !== newState.typeLevel ||
			this._typeIndentLevel !== newState.typeIndentLevel ||
			this._text !== newState.text ||
			this._textStart !== newState.textStart;
	};

	Node.prototype._updateFromTaxonomyState = function _updateFromTaxonomyState(newState, treeLocation, lineNumber) {
		var tree = this.tree,
			isTracked = this.isTracked,
			oldStructureLevel = this._structureLevel,
			oldType = this._type;

		this._taxonomyState = newState;
		this._type = newState.type;
		this._typeLevel = newState.typeLevel;
		this._typeIndentLevel = newState.typeIndentLevel;
		this._structureLevel = this._typeLevel + this._typeIndentLevel;
		this._text = newState.text;
		this._textStart = newState.textStart;
		this._tags = newState.tags;

		this._mode = undefined;
		this._modeContext = undefined;
		var modeStack = newState.modeStack;
		if (modeStack) {
			var modeRecord = modeStack.lastObject();
			if (modeRecord) {
				this._modeContext = modeRecord.mode;
				if (modeRecord.node === this) {
					this._mode = modeRecord.mode;
				}
			}
		}

		if (isTracked) {
			var structureLevelChanged = oldStructureLevel !== this._structureLevel;
			if (structureLevelChanged) {
				var nextLineNode = this.nextLineNode();
				tree.beginUpdates();
				tree.removeNode(this);
				tree.insertNodeBefore(this, nextLineNode);
				tree.endUpdates();
			} else {
				if (oldType === 'property' || this._type === 'property') {
					var nonPropertyAncestor = this.parent.nonPropertyAncestorOrSelf();
					if (nonPropertyAncestor && !nonPropertyAncestor.isRoot()) {
						nonPropertyAncestor.changed();
					}
				}
				this.changed();
			}
		}
	};

	//
	// Line Attributes
	//

	/**
	 * Zero-based line number of this node in containing `Tree`.
	 *
	 * @return {Number}
	 */
	Node.prototype.lineNumber = function lineNumber() {
		return this.tree.nodeToLineNumber(this);
	};

	/**
	 * Text offset position of the start of this node's line in containing
	 * `Tree`.
	 *
	 * @return {Number}
	 */
	Node.prototype.lineTextStart = function lineTextStart() {
		return this.tree.nodeOffsetToTextOffset(this, 0);
	};

	/**
	 * Return AttributedString for this node. The line of text represented by
	 * this line together with associated attribute ranges.
	 *
	 * @return {AttributedString}
	 */
	Node.prototype.lineAttributedString = function lineAttributedString() {
		return this._lineAttributedString;
	};

	Node.prototype.lineAttributeAtIndex = function lineAttributeAtIndex(attribute, index, effectiveRange, longestEffectiveRange) {
		return this._lineAttributedString.attributeAtIndex(attribute, index, effectiveRange, longestEffectiveRange);
	};

	Node.prototype.lineAttributesAtIndex = function lineAttributesAtIndex(index, effectiveRange, longestEffectiveRange) {
		return this._lineAttributedString.attributesAtIndex(index, effectiveRange, longestEffectiveRange);
	};

	Node.prototype.addLineAttributeInRange = function addLineAttributeInRange(attribute, value, location, length) {
		var isTracked = this.isTracked;
		this._lineAttributedString.addAttributeInRange(attribute, value, location, length);
		if (isTracked) {
			this.changed();
		}
	};

	Node.prototype.removeLineAttributeInRange = function removeLineAttributeInRange(attribute, location, length) {
		this.removeLineAttributesInRange([attribute], location, length);
	};

	Node.prototype.removeLineAttributesInRange = function removeLineAttributesInRange(attributes, location, length) {
		var isTracked = this.isTracked,
			didRemove = this._lineAttributedString.removeAttributesInRange(attributes, location, length);
		if (isTracked && didRemove) {
			this.changed();
		}
	};

	//
	// Tracked Ranges
	//

	Node.prototype.addTrackedRange = function addTrackedRange(range) {
		var trackedRanges = this._trackedRanges;
		if (!trackedRanges) {
			trackedRanges = [];
			this._trackedRanges = trackedRanges;
		}
		trackedRanges.push(range);
	};

	Node.prototype.removeTrackedRange = function removeTrackedRange(range) {
		var trackedRanges = this._trackedRanges;
		if (trackedRanges) {
			var index = trackedRanges.indexOf(range);
			if (index !== -1) {
				trackedRanges.splice(index, 1);
			}
		}
	};

	//
	// Children
	//

	/**
	 * Return true if this node has child nodes.
	 *
	 * @return {Boolean}
	 * @methodgroup Children
	 */
	Node.prototype.hasChildren = function hasChildren() {
		return this.firstChild !== null;
	};

	/**
	 * Return this node's children.
	 *
	 * @return {Array.<module:ft/core/node.Node>}
	 */
	Node.prototype.children = function children() {
		var results = [],
			each = this.firstChild;
		while (each) {
			results.push(each);
			each = each.nextSibling;
		}
		return results;
	};

	/**
	 * Return this node's child at the given index
	 *
	 * @param {Number}
	 * @return {module:ft/core/node.Node}
	 */
	Node.prototype.childAtIndex = function childAtIndex(index) {
		if (index === -1) { return this.lastChild; }

		var each = this.firstChild,
			runningIndex = 0;
		while (each) {
			if (runningIndex === index) {
				return each;
			}
			runningIndex++;
			each = each.nextSibling;
		}

		return undefined;
	};

	/**
	 * Return this node's position relative to its siblings.
	 *
	 * @param {Boolean=}
	 * @return {Number}
	 */
	Node.prototype.indexToSelf = function indexToSelf(onlyCountSiblingsWithSameType) {
		if (!this.parent) {
			return undefined;
		}

		var index = 0,
			each = this,
			type = this.type();
		while (each.previousSibling) {
			each = each.previousSibling;
			if (onlyCountSiblingsWithSameType) {
				if (each.type() === type) {
					index++;
				}
			} else {
				index++;
			}
		}
		return index;
	};

	Node.prototype.indexPathToSelf = function indexToSelf(onlyCountSiblingsWithSameType) {
		var indexPath = [],
			each = this,
			index = each.indexToSelf(onlyCountSiblingsWithSameType);
		while (index !== undefined) {
			indexPath.unshift(index);
			each = each.parent;
			index = each.indexToSelf(onlyCountSiblingsWithSameType);
		}
		return indexPath;
	};

	Node.prototype.validateInsertChildBefore = function insertChildBefore(aChild, referenceSibling) {
		var childMinIndent = aChild.validatedTypeIndentLevel(0),
			childTypeLevel = aChild.typeLevel(),
			thisTypeLevel = this.typeLevel();

		if (thisTypeLevel > childTypeLevel) {
			return false;
		}

		// validate against previous and next non empty lines. from those
		// figure out the min and max possible indent levels for inserted
		// child and then validate that childs level can be set given those
		// constraints.

		var previousConstraint = referenceSibling ? referenceSibling.previousSibling : this.lastChild;
		while (previousConstraint && previousConstraint.type() === 'empty') {
			previousConstraint = previousConstraint.previousSibling;
		}
		if (previousConstraint) {
			if (previousConstraint.typeLevel() < childTypeLevel) {
				return false;
			}
		}

		var nextConstraint = referenceSibling;
		while (nextConstraint && nextConstraint.type() === 'empty') {
			nextConstraint = nextConstraint.nextSibling;
		}
		if (nextConstraint) {
			if (nextConstraint.typeLevel() > childTypeLevel) {
				return false;
			}
		}		

		var minIndent = 0,
			maxIndent = Number.MAX_VALUE,
			siblingConstrained = false;

		if (previousConstraint && previousConstraint.typeLevel() === childTypeLevel) {
			maxIndent = previousConstraint.typeIndentLevel();
			siblingConstrained = true;
		}

		if (nextConstraint && nextConstraint.typeLevel() === childTypeLevel) {
			minIndent = nextConstraint.typeIndentLevel();
			siblingConstrained = true;
		}

		if (!siblingConstrained) {
			minIndent = this.typeIndentLevel() + 1;			
		}

		var childPossibleIndent = aChild.validatedTypeIndentLevel(minIndent);
		if (childPossibleIndent < minIndent || childPossibleIndent > maxIndent) {
			return false;
		}

		childPossibleIndent = aChild.validatedTypeIndentLevel(childPossibleIndent);
		if (childPossibleIndent < minIndent || childPossibleIndent > maxIndent) {
			return false;
		}

		return true;
	};

	Node.prototype._adjustNewChildBranchTypeIndentLevelToMatchSibling = function _adjustNewChildBranchTypeIndentLevelToMatchSibling(newChild, referenceSibling) {
		if (referenceSibling && (referenceSibling._typeLevel === newChild._typeLevel)) {
			newChild.setBranchTypeIndentLevel(referenceSibling._typeIndentLevel);
		} else if (this.lastChild && (this.lastChild._typeLevel === newChild._typeLevel)) {
			newChild.setBranchTypeIndentLevel(this.lastChild._typeIndentLevel);
		} else if (newChild._typeLevel === this._typeLevel) {
			newChild.setBranchTypeIndentLevel(this.typeIndentLevel() + 1);
		} else {
			newChild.setBranchTypeIndentLevel(0);
		}
	};

	/**
	 * Insert a child node, inserting before the provided reference sibling.
	 * Depending on the types involved this may update the child's text
	 * content. For example if you add one heading to another, then the added
	 * becomes a level 2 heading and an extra `#` is added to the childs text.
	 *
	 * You can temporarily create _invalid_ structures such as adding a
	 * heading to be a child of a body element. But if this happens the tree
	 * structure will soon be shuffled by the {@tutorial Classification}
	 * process to make it valid.
	 *
	 * @param {module:ft/core/node.Node} - New child to add.
	 * @param {module:ft/core/node.Node} - Sibling to insert before.
	 */
	Node.prototype.insertChildBefore = function insertChildBefore(aChild, referenceSibling, dontAdjustChildLevel) {
		var tree = this.tree,
			isTracked = this.isTracked;

		if (isTracked) {
			tree.beginUpdates();
		}

		aChild.removeFromParent();

		if (!dontAdjustChildLevel) {
			this._adjustNewChildBranchTypeIndentLevelToMatchSibling(aChild, referenceSibling);
		}

		if (isTracked) {
			tree._nodeWillBeInsertedBeforeNode(aChild, referenceSibling ? referenceSibling : this.nextBranch());
		}

		aChild.parent = this;

		if (referenceSibling) {
			aChild.nextSibling = referenceSibling;
			aChild.previousSibling = referenceSibling.previousSibling;
			referenceSibling.previousSibling = aChild;
			if (aChild.previousSibling) {
				aChild.previousSibling.nextSibling = aChild;
			}
		} else {
			aChild.previousSibling = this.lastChild;
			aChild.nextSibling = null;
			if (this.lastChild) {
				this.lastChild.nextSibling = aChild;
			}
		}

		if (!aChild.previousSibling) { this.firstChild = aChild; }
		if (!aChild.nextSibling) { this.lastChild = aChild;	}

		// Hack to fix inserting after an empty line. Empty line should adopt 
		var previousSibling = aChild.previousSibling;
		while (previousSibling && previousSibling._type === 'empty') {
			previousSibling._typeLevel = aChild._typeLevel;
			previousSibling.setTypeIndentLevel(aChild.typeIndentLevel());
			previousSibling = previousSibling.previousSibling;
		}

		if (isTracked) {
			if (aChild.type() === 'property') {
				var nonPropertyAncestor = this.nonPropertyAncestorOrSelf();
				if (nonPropertyAncestor && !nonPropertyAncestor.isRoot()) {
					nonPropertyAncestor.changed();
				}
			}
			tree._nodeWasInserted(aChild);
			tree.endUpdates();
		}
	};

	Node.prototype.insertChildrenBefore = function insertChildBefore(children, referenceSibling, dontAdjustChildLevel) {
		var tree = this.tree,
			isTracked = this.isTracked;

		if (isTracked) {
			tree.beginUpdates();
		}

		var length = children.length;
		for (var i = 0; i < length; i++) {
			this.insertChildBefore(children[i], referenceSibling, dontAdjustChildLevel);
		}

		if (isTracked) {
			tree.endUpdates();
		}
	};

	/**
	 * Append a new child node.
	 *
	 * @param {module:ft/core/node.Node} - Child to append.
	 */
	Node.prototype.appendChild = function appendChild(aChild, dontAdjustChildLevel) {
		this.insertChildBefore(aChild, null, dontAdjustChildLevel);
	};

	Node.prototype.appendChildren = function appendChildren(children, dontAdjustChildLevel) {
		this.insertChildrenBefore(children, null, dontAdjustChildLevel);
	};

	/**
	 * Remove a child from this node.
	 *
	 * @param {module:ft/core/node.Node} - Child to remove.
	 */
	Node.prototype.removeChild = function removeChild(aChild) {
		var tree = this.tree,
			isTracked = this.isTracked;

		if (isTracked) {
			tree.beginUpdates();
			tree._nodeWillBeRemoved(aChild);

			if (tree._previousParents[aChild.id] === undefined) {
				tree._previousParents[aChild.id] = aChild.parent;
			}

			if (aChild.type() === 'property') {
				var nonPropertyAncestor = this.nonPropertyAncestorOrSelf();
				if (nonPropertyAncestor && !nonPropertyAncestor.isRoot()) {
					nonPropertyAncestor.changed();
				}
			}
		}

		if (aChild.previousSibling) { aChild.previousSibling.nextSibling = aChild.nextSibling; }
		if (aChild.nextSibling) { aChild.nextSibling.previousSibling = aChild.previousSibling; }
		if (aChild === this.firstChild) { this.firstChild = aChild.nextSibling; }
		if (aChild === this.lastChild) { this.lastChild = aChild.previousSibling; }

		aChild.nextSibling = null;
		aChild.previousSibling = null;
		aChild.parent = null;

		if (isTracked) {
			tree.endUpdates();
		}
	};

	Node.prototype.removeChildren = function removeChildren(children) {
		var tree = this.tree,
			isTracked = this.isTracked;

		if (isTracked) {
			tree.beginUpdates();
		}

		var length = children.length;
		for (var i = 0; i < length; i++) {
			children[i].removeFromParent();
		}

		if (isTracked) {
			tree.endUpdates();
		}
	};

	/**
	 * Remove this node from it's parent.
	 */
	Node.prototype.removeFromParent = function removeFromParent() {
		if (this.parent) {
			this.parent.removeChild(this);
		}
	};

	//
	// Traversal
	//

	/**
	 * Each tree has a special "root" node that's the ancestor of all other
	 * nodes in the tree. It's special because unlike all other nodes the root
	 * node doesn't represent a line of text. You shouldn't try to modify the
	 * root node, it's just an implentation detail to make the tree structure
	 * work out.
	 *
	 * @returns {Boolean} - True if this is the root node.
	 * @methodgroup Traversal
	 */
	Node.prototype.isRoot = function isRoot() {
		return this._typeLevel === Taxonomy.ROOT_STRUCTURE_TYPE_LEVEL;
	};

	/**
	 * Return true if given node is an ancestor of this node.
	 *
	 * @param {module:ft/core/node.Node} - Posible ancestor node.
	 * @returns {Boolean}
	 */
	Node.prototype.isAncestorOfSelf = function isAncestor(node) {
		var each = this.parent;
		while (each) {
			if (each === node) {
				return true;
			}
			each = each.parent;
		}
		return false;
	};

	Node.prototype.root = function root() {
		if (this.isTracked) {
			return this.tree.root;
		}

		var parent = this;
		while (parent && !parent.isRoot()) {
			parent = parent.parent;
		}
		return parent;
	};

	Node.prototype.nonPropertyAncestorOrSelf = function nonPropertyAncestorOrSelf() {
		var each = this;
		while (each) {
			if (each.type() !== 'property') {
				return each;
			} else {
				each = each.parent;
			}
		}
		return null;
	};

	/**
	 * Returns the previous branch in the node tree. Previous node before all
	 * descendants of this node's previous sibling.
	 *
	 * @returns {module:ft/core/node.Node}
	 */
	Node.prototype.previousBranch = function previousBranch() {
		var previousSibling = this.previousSibling;
		if (previousSibling) {
			return previousSibling;
		}
		return this.previousLineNode();
	};

	/**
	 * Returns the next branch in node tree. Next node after all descendants
	 * of this node.
	 *
	 * @returns {module:ft/core/node.Node}
	 */
	Node.prototype.nextBranch = function nextBranch() {
		return this.lastDescendantOrSelf().nextLineNode();
	};

	/**
	 * Return all nodes descending from this node. This node's children,
	 * children's children, and so on.
	 *
	 * @return {Array.<module:ft/core/node.Node>}
	 */
	Node.prototype.descendants = function descendants() {
		var results = [],
			end = this.nextBranch(),
			each = this.nextLineNode();
		while (each !== end) {
			results.push(each);
			each = each.nextLineNode();
		}
		return results;
	};

	/**
	 * Return the last descendant of this node.
	 *
	 * @returns {module:ft/core/node.Node}
	 */
	Node.prototype.lastDescendant = function lastDescendant() {
		var each = this.lastChild;
		while (each && each.lastChild) {
			each = each.lastChild;
		}
		return each;
	};

	/**
	 * Return this node's last descendant, or if this node has no children
	 * then return this node.
	 *
	 * @returns {module:ft/core/node.Node}
	 */
	Node.prototype.lastDescendantOrSelf = function lastDescendantOrSelf() {
		var each = this,
			lastChild = each.lastChild;
		while (lastChild) {
			each = lastChild;
			lastChild = each.lastChild;
		}
		return each;
	};

	/**
	 * Return the node representing the previous line of text.
	 *
	 * @returns {module:ft/core/node.Node}
	 */
	Node.prototype.previousLineNode = function previousLineNode() {
		var parent,
			previousSibling = this.previousSibling;
		if (previousSibling) {
			return previousSibling.lastDescendantOrSelf();
		} else {
			parent = this.parent;
			if (!parent || parent.isRoot()) {
				return null;
			} else {
				return parent;
			}
		}
	};

	Node.prototype.previousLineNodeOrRoot = function previousLineNodeOrRoot() {
		var previousLineNode = this.previousLineNode();
		if (previousLineNode) {
			return previousLineNode;
		} else {
			return this.parent;
		}
	};

	/**
	 * Return the node representing the next line of text.
	 *
	 * @returns {module:ft/core/node.Node}
	 */
	Node.prototype.nextLineNode = function nextLineNode() {
		var firstChild = this.firstChild;
		if (firstChild) { return firstChild; }
		var nextSibling = this.nextSibling;
		if (nextSibling) { return nextSibling; }
		var parent = this.parent;
		while (parent) {
			nextSibling = parent.nextSibling;
			if (nextSibling) {
				return nextSibling;
			}
			parent = parent.parent;
		}
		return null;
	};

	/**
	 * Evaluate the node path starting with this node and return all matches.
	 *
	 * @tutorial Node Paths
	 * @param {String} - Node path
	 * @returns {Array.<module:ft/core/node.Node>}
	 */
	Node.prototype.evaluateNodePath = function evaluateNodePath(nodePath) {
		return NodePath.evaluate(nodePath, this);
	};

	Node.prototype.iterateIgnoringEmptyLines = function iterateIgnoringEmptyLines(callback) {
		var isString = typeof callback === 'string' || callback instanceof String,
			each = isString ? this[callback] : callback.call(this);
		while (each && each.type() === 'empty') {
			each = isString ? each[callback] : callback.call(each);
		}
		return each;
	};

	/**
	 * Returns the text represented by this node and all of its decendants.
	 *
	 * @returns {String}
	 */
	Node.prototype.branchToText = function branchToText() {
		var nextBranch = this.nextBranch();
		if (nextBranch) {
			return this.tree.createRangeFromNodes(this, 0, nextBranch, 0).textInRange();
		} else {
			var lastDescendantOrSelf = this.lastDescendantOrSelf();
			return this.tree.createRangeFromNodes(this, 0, lastDescendantOrSelf, lastDescendantOrSelf.line().length).textInRange();
		}
	};

	Node.prototype.branchNodes = function branchNodes() {
		var results = this.descendants();
		results.unshift(this);
		return results;
	};

	Node.prototype.branchStructureLevels = function branchStructureLevels() {
		var next = this.breadthFirstIterator(),
			structureLevels = [],
			each = this;

		while (each) {
			structureLevels.push(each.structureLevel());
			each = next();
		}

		structureLevels.sort();
		structureLevels.sortedUniquify();

		return structureLevels;
	};

	Node.prototype.depthFirstIterator = function depthFirstIterator(reverseSiblingOrder) {
		var stack = [this];
		return function next() {
			var current = statck.pop();
			if (current) {
				var each = reverseSiblingOrder ? current.lastChild : current.firstChild;
				while (each) {
					stack.push(each);
					each = reverseSiblingOrder ? each.previousSibling : each.nextSibling;
				}
			}
			return current;
		};
	};

	Node.prototype.breadthFirstIterator = function breadthFirstIterator(reverseSiblingOrder) {
		var queue = [this];
		return function next() {
			var current = queue.shift();
			if (current) {
				var each = reverseSiblingOrder ? current.lastChild : current.firstChild;
				while (each) {
					queue.push(each);
					each = reverseSiblingOrder ? each.previousSibling : each.nextSibling;
				}
			}
			return current;
		};
	};

	//
	// DOM
	//

	Node.prototype.renderNodeElementStyles = function renderNodeElementStyles(elementRenderer) {
		var prevNode = this.previousLineNode(),
			nextNode = this.nextLineNode(),
			modeContext = this.modeContext(),
			typeStarting = !prevNode || prevNode.type() !== this.type(),
			typeEnding = !nextNode || nextNode.type() !== this.type(),
			mode = this.mode(),
			tags = this.tags();

		if (typeStarting) {
			elementRenderer.addWrapClass('cm-' + this.type() + '-start-wrapper');
			elementRenderer.addBackgroundClass('cm-' + this.type() + '-start-background');
			elementRenderer.addTextClass('cm-' + this.type() + '-start');
		}

		elementRenderer.addWrapClass('cm-' + this.type() + '-wrapper');
		elementRenderer.addBackgroundClass('cm-' + this.type() + '-background');
		elementRenderer.addTextClass('cm-' + this.type());

		if (typeEnding) {
			elementRenderer.addWrapClass('cm-' + this.type() + '-end-wrapper');
			elementRenderer.addBackgroundClass('cm-' + this.type() + '-end-background');
			elementRenderer.addTextClass('cm-' + this.type() + '-end');
		}

		if (modeContext) {
			var modeContextStarting = !prevNode || prevNode.modeContext() !== this.modeContext(),
				modeContextEnding = !nextNode || nextNode.modeContext() !== this.modeContext();

			if (modeContextStarting) {
				elementRenderer.addTextClass('cm-' + modeContext + '-start-modecontext');				
			}

			elementRenderer.addTextClass('cm-' + modeContext + '-modecontext');

			if (modeContextEnding) {
				elementRenderer.addTextClass('cm-' + modeContext + '-end-modecontext');				
			}
		}

		if (mode) {
			elementRenderer.addTextClass('cm-' + mode + '-mode');
		}

		for (var name in tags) {
			if (name !== '') {
				elementRenderer.setAttribute('cm-' + name, tags[name]);
			}
		}

		elementRenderer.setAttribute('cm-level', this.typeIndentLevel());
	};

	//
	// Red Black interval tree used for line and position lookups.
	//

	Node.prototype.totalLineCount = function totalLineCount() {
		var result = this._RBTotalLineCount;
		if (!result) {
			this._RBFillCache();
			return this._RBTotalLineCount;
		} else {
			return result;
		}
	};

	Node.prototype.totalLength = function totalLength() {
		var result = this._RBTotalLength;
		if (!result) {
			this._RBFillCache();
			return this._RBTotalLength;
		} else {
			return result;
		}
	};

	Node.prototype._RBHasChildren = function _RBHasChildren() {
		return this._RBLeft !== null || this._RBRight !== null;
	};

	Node.prototype._RBIsLeftChild = function _RBIsLeftChild() {
		var _RBParent = this._RBParent;
		return _RBParent !== null && _RBParent._RBLeft === this;
	};

	Node.prototype._RBIsRightChild = function _RBIsRightChild() {
		var _RBParent = this._RBParent;
		return _RBParent !== null && _RBParent._RBRight === this;
	};

	Node.prototype._RBChildPredecessor = function _RBChildPredecessor() {
		if (!this._RBLeft) { return null; }
		return this._RBLeft._RBRightMost();
	};

	Node.prototype._RBChildSuccessor = function _RBChildSuccessor() {
		if (!this._RBRight) { return null; }
		return this._RBRight._RBLeftMost();
	};

	Node.prototype._RBNextLineNode = function _RBNextLineNode() {
		var node = this,
			childSuccessor = node._RBChildSuccessor();
		if (childSuccessor) {
			return childSuccessor;
		}
		var parent = node._RBParent;
		while (parent && node === parent._RBRight) {
			node = parent;
			parent = node._RBParent;
		}

		if (parent === this.root()) {
			return null;
		} else {
			return parent;
		}
	};

	Node.prototype._RBPreviousLineNode = function _RBPreviousLineNode() {
		var node = this,
			childPredecessor = node._RBChildPredecessor();
		if (childPredecessor) {
			return childPredecessor;
		}
		var parent = node._RBParent;
		while (parent && node === parent._RBLeft) {
			node = parent;
			parent = node._RBParent;
		}
		return parent;
	};

	Node.prototype._RBDepth = function _RBDepth() {
		var parent = this._RBParent,
			depth = 0;
		while (parent) {
			parent = parent._RBParent;
			depth++;
		}
		return depth;
	};

	Node.prototype._RBRightMost = function _RBRightMost() {
		var _RBRight = this;
		var r;
		while ((r = _RBRight._RBRight)) {
			_RBRight = r;
		}
		return _RBRight;
	};

	Node.prototype._RBLeftMost = function _RBLeftMost() {
		var _RBLeft = this;
		var l;
		while ((l = _RBLeft._RBLeft)) {
			_RBLeft = l;
		}
		return _RBLeft;
	};

	Node.prototype._RBSetLeft = function _RBSetLeft(child) {
		if (child) {
			child._RBParent = this;
			this._RBLeft = child;
		} else {
			this._RBLeft = null;
		}
	};

	Node.prototype._RBSetRight = function _RBSetRight(child) {
		if (child) {
			child._RBParent = this;
			this._RBRight = child;
		} else {
			this._RBRight = null;
		}
	};

	Node.prototype._RBFillCache = function _RBFillCache() {
		var totalLineCount = this._RBTotalLineCount;
		if (!totalLineCount) {
			var left = this._RBLeft,
				right = this._RBRight,
				totalLength = this.line().length + 1;

			totalLineCount = 1;

			if (left) {
				totalLineCount += left.totalLineCount();
				totalLength += left.totalLength();
			}

			if (right) {
				totalLineCount += right.totalLineCount();
				totalLength += right.totalLength();
			}

			this._RBTotalLineCount = totalLineCount;
			this._RBTotalLength = totalLength;
		}
	};

	Node.prototype._RBClearCache = function _RBClearCache() {
		var each = this;
		while (each && each._RBTotalLength !== null) {
			each._RBTotalLength = null;
			each._RBTotalLineCount = null;
			each = each._RBParent;
		}
		return;
	};

	Node.prototype._RBLeftRotate = function _RBLeftRotate(tree) {
		var y = this._RBRight;
		this._RBRight = y._RBLeft;
		if (y._RBLeft) {
			y._RBLeft._RBParent = this;
		}
		y._RBParent = this._RBParent;
		if (!this._RBParent) {
			tree._RBRoot = y;
		} else {
			if (this._RBIsLeftChild()) {
				this._RBParent._RBLeft = y;
			} else {
				this._RBParent._RBRight = y;
			}
		}
		y._RBLeft = this;
		this._RBParent = y;

		y._RBClearCache();
		this._RBClearCache();
	};

	Node.prototype._RBRightRotate = function _RBRightRotate(tree) {
		var y = this._RBLeft;
		this._RBLeft = y._RBRight;
		if (y._RBRight) {
			y._RBRight._RBParent = this;
		}
		y._RBParent = this._RBParent;
		if (!this._RBParent) {
			tree._RBRoot = y;
		} else {
			if (this._RBIsRightChild()) {
				this._RBParent._RBRight = y;
			} else {
				this._RBParent._RBLeft = y;
			}
		}
		y._RBRight = this;
		this._RBParent = y;

		y._RBClearCache();
		this._RBClearCache();
	};

	//
	// Debug
	//

	Node.prototype.toString = function toString(rbTree) {
		if (rbTree) {
			var leftRight = this._RBIsLeftChild() ? '< ' : '> ',
				depth = this._RBDepth(),
				depthString = '';

			for (var i = 0 ; i < depth; i++) {
				depthString += '| ';
			}

			return depthString + leftRight + this.line().replace(/(^[\s]+|[\s]+$)/g, '') + ' [' + this.id + '] {len: ' + this.totalLength() + ' cnt: ' + this.totalLineCount() + '}';
		} else {
			return '{node "' + this._lineAttributedString.toString() + '" }';
		}
	};

	Node.prototype.branchToString = function branchToString(depth, rbTree) {
		var each = this.firstChild,
			depthString = '',
			lines = [],
			i;

		for (i = 0; i < depth - 1; i++) {
			depthString += '\u2507';
		}

		lines.push(
			depthString + '\u2520' +
			(each ? '\u2530' : ' ') +
			(this.line().replace(/\t/g, '\u25B8').replace(/\n/g, '\\n') || '\u2205') +
			' [' + this.id + ', ' + this.type() + ']'
		);

		while (each) {
			lines.push(each.branchToString(depth + 1));
			each = each.nextSibling;
		}

		return lines.join('\n');
	};

	Node.BLACK = 0;
	Node.RED = 1;

	Node._RBColorOrBlack = function _RBColorOrBlack(nodeOrNull) {
		return (nodeOrNull) ? nodeOrNull._RBColor : Node.BLACK;
	};

	Node.commonDescendantsForNodes = function commonDescendantsForNodes(nodes) {
		var results = NodeSet.setFromArray(nodes),
			length = nodes.length;
		for (var i = 0; i < length; i++) {
			results.addNodesFromArray(nodes[i].descendants());
		}
		return results;
	};

	/**
	 * Return the common ancestors from the given nodes.
	 *
	 * @param {Array.<module:ft/core/node.Node>}
	 * @returns {module:ft/core/nodeset.NodeSet}
	 */
	Node.commonAncestorsForNodes = function commonAncestorsForNodes(nodes) {
		var length = nodes.length,
			i, each, eachParent,
			result = NodeSet.setFromArray(nodes);

		for (i = 0; i < length; i++) {
			each = nodes[i];
			eachParent = each.parent;

			while (eachParent) {
				if (result.containsNode(eachParent)) {
					result.removeNode(each);
					break;
				} else {
					eachParent = eachParent.parent;
				}
			}
		}

		return result;
	};

	/**
	 * Returns all ancestors of the given nodes.
	 *
	 * @param {Array.<module:ft/core/node.Node>}
	 * @returns {module:ft/core/nodeset.NodeSet}
	 */
	Node.ancestorsForNodes = function ancestorsForNodes(nodes) {
		var results = new NodeSet(),
			length = nodes.length,
			i;
		for (i = 0; i < length; i++) {
			var each = nodes[i];
			while ((each = each.parent)) {
				results.addNode(each);
			}
		}
		return results;
	};

	// remove all JSON Suport from this
	Node.formatNodes = function formatNodes(format, attributes, nodes, alwaysReturnList, dontStringify) {
		var length = nodes.length,
			result,
			i;

		if (format === 'ids') {
			result = [];
			for (i = 0; i < length; i++) {
				result.push(nodes[i].id);
			}
			return result.join(',');
		}

		if (attributes) {
			attributes = attributes.split(';');
		}

		if (!alwaysReturnList && length === 1) {
			if (format === 'txt') {
				result = nodes[0].line();
			} else {
				result = nodes[0].toJSON(attributes, true);
			}
		} else {
			result = [];
			for (i = 0; i < length; i++) {
				if (format === 'txt') {
					result.push(nodes[i].line());
				} else {
					result.push(nodes[i].toJSON(attributes, true));
				}
			}
		}

		if (format === 'txt') {
			if (result.join) {
				return result.join('\n');
			} else {
				return result;
			}
		} else {
			if (dontStringify) {
				return result;
			} else {
				return JSON.stringify(result);
			}
		}
	};

	exports.Node = Node;
});