//
// Copyright (c) 2014 Jesse Grosjean. All rights reserved.
//
// Permission is given for you to use this software ONLY while creating themes
// and plugins for the "FoldingText" software platform. Permission is NOT
// given for any other purpose.
//
define(function (require, exports, module) {
	'use strict';

	require('../util/array');

	var NodePathParser = require('./nodepathparser'),
		Cache = require('../util/cache').Cache;

	var NodePath = function NodePath(nodePathString, types) {
		this.types = types;
		this.nodePath = NodePath.parse(nodePathString, undefined, types).parsedPath;
		this.nodePathString = nodePathString;
	};

	(function () {

		this.isShowAllPath = function isShowAllPath() {
			return this.nodePathString === NodePath.SHOW_ALL_PATH;
		};

		//
		// Evaluation
		//

		this.evaluate = function evaluate(currentNode) {
			if (this.nodePath === undefined) {
				return [];
			}

			return this._evaluateNodePath(this.nodePath, currentNode);
		};

		this._evaluateNodePath = function _evaluateNodePath(nodePath, currentNode) {
			var union = nodePath.union,
				intersect = nodePath.intersect,
				except = nodePath.except,
				results;

			if (union) {
				results = this._evaluateUnion(union, currentNode);
			} else if (intersect) {
				results = this._evaluateIntersect(intersect, currentNode);
			} else if (except) {
				results = this._evaluateExcept(except, currentNode);
			} else {
				results = this._evaluatePath(nodePath, currentNode);
			}

			this._sliceResultsFrom(nodePath.slice, results, 0);

			return results;
		};

		this._unionTreeOrderedResults = function _unionTreeOrderedResults(results1, results2, tree) {
			var r1Index = -1, r2Index = -1,
				results = [],
				i = 0, j = 0,
				r1, r2;

			while (true) {
				r1 = results1[i];
				r2 = results2[j];

				if (!r1) {
					if (r2) {
						results.push.apply(results, results2.slice(j));
					}
					return results;
				} else if (!r2) {
					if (r1) {
						results.push.apply(results, results1.slice(i));
					}
					return results;
				} else if (r1 === r2) {
					results.push(r2);
					r1Index = -1;
					r2Index = -1;
					i++;
					j++;
				} else {
					r1Index = r1Index === -1 ? tree.nodeToLineNumber(r1) : r1Index;
					r2Index = r2Index === -1 ? tree.nodeToLineNumber(r2) : r2Index;
					if (r1Index < r2Index) {
						results.push(r1);
						r1Index = -1;
						i++;
					} else {
						results.push(r2);
						r2Index = -1;
						j++;
					}
				}
			}
		};

		this._evaluateUnion = function _evaluateUnion(paths, currentNode) {
			var results1 = this._evaluateNodePath(paths[0], currentNode),
				results2 = this._evaluateNodePath(paths[1], currentNode);
			return this._unionTreeOrderedResults(results1, results2, currentNode.tree);

			//results1.unionSet(results2);
			//return results1;*/
		};

		this._evaluateIntersect = function _evaluateIntersect(paths, currentNode) {
			var results1 = this._evaluateNodePath(paths[0], currentNode),
				results2 = this._evaluateNodePath(paths[1], currentNode),
				r1Index = -1, r2Index = -1,
				tree = currentNode.tree,
				results = [],
				i = 0, j = 0,
				r1, r2;

			while (true) {
				r1 = results1[i];
				r2 = results2[j];

				if (!r1) {
					return results;
				} else if (!r2) {
					return results;
				} else if (r1 === r2) {
					results.push(r2);
					r1Index = -1;
					r2Index = -1;
					i++;
					j++;
				} else {
					r1Index = r1Index === -1 ? tree.nodeToLineNumber(r1) : r1Index;
					r2Index = r2Index === -1 ? tree.nodeToLineNumber(r2) : r2Index;
					if (r1Index < r2Index) {
						r1Index = -1;
						i++;
					} else {
						r2Index = -1;
						j++;
					}
				}
			}

			//results1.intersectSet(results2);
			//return results1;
		};

		this._evaluateExcept = function _evaluateExcept(paths, currentNode) {
			var results1 = this._evaluateNodePath(paths[0], currentNode),
				results2 = this._evaluateNodePath(paths[1], currentNode),
				r1Index = -1, r2Index = -1,
				tree = currentNode.tree,
				results = [],
				i = 0, j = 0,
				r1, r2;

			while (true) {
				r1 = results1[i];
				r2 = results2[j];
				r1Index = r1Index === -1 ? tree.nodeToLineNumber(r1) : r1Index;
				r2Index = r2Index === -1 ? tree.nodeToLineNumber(r2) : r2Index;

				while (r2 && r1Index > r2Index) {
					j++;
					r2 = results2[j];
					r2Index = tree.nodeToLineNumber(r2);
				}

				if (!r1) {
					return results;
				} else if (!r2) {
					results.push.apply(results, results1.slice(i));
					return results;
				} else if (r1 === r2) {
					r1Index = -1;
					r2Index = -1;
					i++;
					j++;
				} else {
					results.push(r1);
					r1Index = -1;
					i++;
				}
			}

			//results1.minusSet(results2);
			//return results1;
		};

		this._evaluatePath = function _evaluatePath(path, currentNode) {
			var tree = currentNode.tree,
				absolute = path.absolute,
				contexts = [],
				eachContext,
				contextsLength,
				outerThis = this,
				results = null,
				i, j;

			if (absolute) {
				currentNode = currentNode.root();
			}

			contexts.push(currentNode);

			var steps = path.steps,
				stepsLength = steps.length,
				step;

			for (i = 0; i < stepsLength; i++) {
				step = steps[i];
				contextsLength = contexts.length;
				results = [];
				for (j = 0; j < contextsLength; j++) {
					eachContext = contexts[j];
					if (results.length) {
						// If evaluating from multiple contexts and we have some results already
						// make sure to merge the new set of context results in with the existing.
						var contextResults = [];
						this._evaluateStep(step, eachContext, contextResults);
						results = this._unionTreeOrderedResults(results, contextResults, tree);
					} else {
						this._evaluateStep(step, eachContext, results);
					}
				}
				contexts = results;
			}

			return results;
		};

		this._evaluateStep = function _evaluateStep(step, currentNode, results) {
			// Results should always be stored in tree order.
			var predicate = step.predicate,
				type = step.type,
				from = results.length,
				each,
				end;

			switch (step.axis) {

			case 'ancestor-or-self':
				each = currentNode;
				while (each) {
					if (this._evaluatePredicate(type, predicate, each)) {
						results.splice(from, 0, each);
					}
					each = each.parent;
				}
				break;

			case 'ancestor':
				each = currentNode.parent;
				while (each) {
					if (this._evaluatePredicate(type, predicate, each)) {
						results.splice(from, 0, each);
					}
					each = each.parent;
				}
				break;

			case 'child':
				each = currentNode.firstChild;
				while (each) {
					if (this._evaluatePredicate(type, predicate, each)) {
						results.push(each);
					}
					each = each.nextSibling;
				}
				break;

			case 'descendant-or-self':
				end = currentNode.nextBranch();
				each = currentNode;
				while (each && each !== end) {
					if (this._evaluatePredicate(type, predicate, each)) {
						results.push(each);
					}
					each = each.nextLineNode();
				}
				break;

			case 'descendant':
				end = currentNode.nextBranch();
				each = currentNode.firstChild;
				while (each && each !== end) {
					if (this._evaluatePredicate(type, predicate, each)) {
						results.push(each);
					}
					each = each.nextLineNode();
				}
				break;

			case 'filter-descendants':
				end = currentNode.nextBranch();
				each = currentNode.firstChild;
				var originalFrom = from;
				while (each && each !== end) {
					if (this._evaluatePredicate(type, predicate, each)) {
						var lastMatch = results.lastObject(),
							eachAncestor = each.parent;

						// splice in ancestors as needed... this can be optimized
						while (eachAncestor !== currentNode && results.indexOf(eachAncestor) === -1) {
							results.splice(from, 0, eachAncestor);
							eachAncestor = eachAncestor.parent;
						}

						results.push(each);
						from = results.length;
					}
					each = each.nextLineNode();
				}
				results.splice(originalFrom, 0, currentNode);
				break;

			case 'following-sibling':
				each = currentNode.nextSibling;
				while (each) {
					if (this._evaluatePredicate(type, predicate, each)) {
						results.push(each);
					}
					each = each.nextSibling;
				}
				break;

			case 'following':
				each = currentNode.nextLineNode();
				while (each) {
					if (this._evaluatePredicate(type, predicate, each)) {
						results.push(each);
					}
					each = each.nextLineNode();
				}
				break;

			case 'parent':
				each = currentNode.parent;
				if (each && this._evaluatePredicate(type, predicate, each)) {
					results.push(each);
				}
				break;

			case 'preceding-sibling':
				each = currentNode.previousSibling;
				while (each) {
					if (this._evaluatePredicate(type, predicate, each)) {
						results.splice(from, 0, each);
					}
					each = each.previousSibling;
				}
				break;

			case 'preceding':
				each = currentNode.previousLineNode();
				while (each) {
					if (this._evaluatePredicate(type, predicate, each)) {
						results.splice(from, 0, each);
					}
					each = each.previousLineNode();
				}
				break;

			case 'self':
				if (this._evaluatePredicate(type, predicate, currentNode)) {
					results.push(currentNode);
				}
				break;
			}

			this._sliceResultsFrom(step.slice, results, from);
		};

		this._evaluatePredicate = function _evaluatePredicate(type, predicate, currentNode) {
			if (type !== '*' && type !== currentNode.type()) {
				return false;
			}

			if (predicate === '*') {
				return true;
			}

			var and = predicate.and;
			if (and) {
				return this._evaluatePredicate('*', and[0], currentNode) && this._evaluatePredicate('*', and[1], currentNode);
			}

			var or = predicate.or;
			if (or) {
				return this._evaluatePredicate('*', or[0], currentNode) || this._evaluatePredicate('*', or[1], currentNode);
			}

			var not = predicate.not;
			if (not) {
				return !this._evaluatePredicate('*', not, currentNode);
			}

			var attributePath = predicate.attributePath,
				relation = predicate.relation,
				modifier = predicate.modifier,
				value = predicate.value;

			if (!relation && !value) {
				return currentNode.valueForAttributePath(predicate.attributePath) !== undefined;
			}

			if (!attributePath) {
				attributePath = NodePath.DEFAULT_ATTRIBUTE_PATH;
			}

			if (!relation) {
				relation = NodePath.DEFAULT_RELATIOIN;
			}

			if (!modifier) {
				modifier = NodePath.DEFAULT_MODIFIER;
			}

			var predicateValueCache = predicate.predicateValueCache;
			if (!predicateValueCache) {
				predicateValueCache = this._convertValueForModifier(value, modifier);
				predicate.predicateValueCache = predicateValueCache;
			}

			var attributeValue = currentNode.valueForAttributePath(attributePath);
			if (attributeValue !== undefined) {
				attributeValue = this._convertValueForModifier(attributeValue.toString(), modifier);
			}

			return this._evaluateRelation(attributeValue, relation, predicateValueCache, predicate);
		};

		this._convertValueForModifier = function _convertValueForModifier(value, modifier) {
			if (modifier === 'i') {
				return value.toLowerCase();
			} else if (modifier === 'n') {
				return parseFloat(value);
			} else if (modifier === 'd') {
				return Date.parse(value); // This is weak and implementation dependent, replace with good natural language parser.
			}

			// May want to test incomming values to see if they are both numeric... if so conver to number
			//function isNumeric(num){
			//	return !isNaN(num)
			//}

			return value;
		};

		this._evaluateRelation = function _evaluateRelation(left, relation, right, predicate) {
			switch (relation) {
			case '=':
				return left === right;
			case '!=':
				return left !== right;
			case '<':
				return left < right;
			case '>':
				return left > right;
			case '<=':
				return left <= right;
			case '>=':
				return left >= right;
			case 'beginswith':
				if (!left) {
					return false;
				}
				return left.startsWith(right);
			case 'contains':
				if (!left) {
					return false;
				}
				return left.indexOf(right) !== -1;
			case 'endswith':
				if (!left) {
					return false;
				}
				return left.endsWith(right);
			case 'matches':
				if (!left) {
					return false;
				}
				var joinedValueRegexCache = predicate.joinedValueRegexCache;
				if (joinedValueRegexCache === undefined) {
					try {
						joinedValueRegexCache = new RegExp(right.toString());
					} catch (e) {
						// Use null as a marker indicating that have a bad
						// regex and in that case the below logic will always
						// return false
						joinedValueRegexCache = null;
					}
					predicate.joinedValueRegexCache = joinedValueRegexCache; // cache
				}

				if (joinedValueRegexCache) {
					return left.toString().match(joinedValueRegexCache);
				} else {
					return false;
				}
			}
		};

		this._sliceResultsFrom = function _sliceResultsFrom(slice, results, from) {
			if (slice) {
				var length = results.length - from,
					start = slice.start,
					end = slice.end;

				if (length === 0) {
					return;
				}

				if (end > length) {
					end = length;
				}

				if (start !== 0 || end !== length) {
					var sliced;

					if (start < 0) {
						start += length;
						if (start < 0) {
							start = 0;
						}
					}

					if (start > length - 1) {
						start = length - 1;
					}

					if (end === null) {
						sliced = results[from + start];
					} else {
						if (end < 0) {
							end += length;
						}
						
						if (end < start) {
							end = start;
						}

						sliced = results.slice(from).slice(start, end);
					}

					Array.prototype.splice.apply(results, [from, results.length - from].concat(sliced));
				}
			}
		};

		//
		// Location and Filter Parts
		//

		this._leftmostNodePath = function _leftmostNodePath(nodePath) {
			if (nodePath.union) {
				return this._leftmostNodePath(nodePath.union[0]);
			} else if (nodePath.intersect) {
				return this._leftmostNodePath(nodePath.intersect[0]);
			} else if (nodePath.except) {
				return this._leftmostNodePath(nodePath.except[0]);
			} else {
				return nodePath;
			}
		};

		this.locationPathString = function locationPathString(minusLastPathStep) {
			var leftMosNodePath = this._leftmostNodePath(this.nodePath),
				path = JSON.parse(JSON.stringify(leftMosNodePath)),
				steps = path.steps,
				length = steps.length,
				locationSteps = [],
				each,
				i;

			for (i = 0; i < length; i++) {
				each = steps[i];
				if (each.axis === 'filter-descendants') {
					break;
				}
				locationSteps.push(each);
			}

			if (minusLastPathStep) {
				locationSteps.pop();
			}

			if (locationSteps.length === 0) {
				return '';
			} else {
				path.steps = locationSteps;
				return this._pathToString(path);
			}
		};

		this.setLocationPathString = function setLocationPathString(newLocationPathString) {
			var oldLocationPathString = this.locationPathString(),
				newNodePathString = newLocationPathString + this.toString().substr(oldLocationPathString.length);
			this.nodePath = NodePath.parse(newNodePathString).parsedPath;
			this.nodePathString = newNodePathString;
			return this;
		};

		this.filterStepString = function filterStepString() {
			var leftMosNodePath = this._leftmostNodePath(this.nodePath),
				steps = leftMosNodePath.steps,
				length = steps.length,
				each,
				i;

			for (i = 0; i < length; i++) {
				each = steps[i];
				if (each.axis === 'filter-descendants') {
					return this._stepToString(each).substr(2);
				}
			}

			return '';
		};

		this.setFilterStepString = function setFilterStepString(filterStepString) {
			var newNodePathString = this.locationPathString();

			if (filterStepString) {
				newNodePathString += '///' + filterStepString;
			}

			this.nodePath = NodePath.parse(newNodePathString).parsedPath;
			this.nodePathString = newNodePathString;

			return this;
		};

		//
		// Focus State
		//

		this.updateNodePathByFocusingOutOneLevel = function updateNodePathByFocusingOutOneLevel() {
			var filterStepString = this.filterStepString();
			if (filterStepString !== '*') {
				this.setFilterStepString('*');
			} else {
				var locationPathString = this.locationPathString(true);
				if (!locationPathString) {
					locationPathString = '';
				}
				this.setLocationPathString(locationPathString);
			}

			return this;
		};

		//
		// Convert to String
		//

		this._predicateToString = function _predicateToString(predicate, group) {
			if (predicate === '*') {
				return '*';
			}

			var openGroup = group ? '(' : '',
				closeGroup = group ? ')' : '';

			var and = predicate.and;
			if (and) {
				return openGroup + this._predicateToString(and[0], true) + ' and ' + this._predicateToString(and[1], true) + closeGroup;
			}

			var or = predicate.or;
			if (or) {
				return openGroup + this._predicateToString(or[0], true) + ' or ' + this._predicateToString(or[1], true) + closeGroup;
			}

			var not = predicate.not;
			if (not) {
				return 'not ' + this._predicateToString(not, true);
			}

			var attributePath = predicate.attributePath,
				relation = predicate.relation,
				modifier = predicate.modifier,
				value = predicate.value,
				result = [];

			if (attributePath) {
				result.push('@' + attributePath.join(':'));
			}

			if (relation) {
				result.push(relation);
			}

			if (modifier) {
				result.push('[' + modifier + ']');
			}

			if (value) {
				try {
					NodePathParser.parse(value,  { startRule : 'Value' });
				} catch (e) {
					value = '"' + value + '"';
				}

				result.push(value);
			}

			return result.join(' ');
		};

		this._stepToString = function _stepToString(step) {
			switch (step.axis) {
			case 'child':
				return this._predicateToString(step.predicate);
			case 'descendant':
				return '/' + this._predicateToString(step.predicate);
			case 'filter-descendants':
				return '//' + this._predicateToString(step.predicate);
			case 'parent':
				return '..' + this._predicateToString(step.predicate);
			default:
				return step.axis + '::' + this._predicateToString(step.predicate);
			}
		};

		this._pathToString = function _pathToString(path) {
			var stepStrings = [],
				steps = path.steps,
				stepsLength = steps.length,
				step,
				i;

			for (i = 0; i < stepsLength; i++) {
				stepStrings.push(this._stepToString(steps[i]));
			}

			if (path.absolute) {
				return '/' + stepStrings.join('/');
			} else {
				return stepStrings.join('/');
			}
		};

		this._nodePathToString = function _nodePathToString(nodePath, group) {
			var openGroup = group ? '(' : '',
				closeGroup = group ? ')' : '';

			var union = nodePath.union;
			if (union) {
				return openGroup + this._nodePathToString(union[0], true) + ' union ' + this._nodePathToString(union[1], true) + closeGroup;
			}

			var intersect = nodePath.intersect;
			if (intersect) {
				return openGroup + this._nodePathToString(intersect[0], true) + ' intersect ' + this._nodePathToString(intersect[1], true) + closeGroup;
			}

			var except = nodePath.except;
			if (except) {
				return openGroup + this._nodePathToString(except[0], true) + ' except ' + this._nodePathToString(except[1], true) + closeGroup;
			}

			return this._pathToString(nodePath);
		};

		this.toString = function toString() {
			return this._nodePathToString(this.nodePath);
		};

	}).call(NodePath.prototype);

	NodePath._cache = new Cache(3); // path, search, and combined

	function keywordCompare(a, b) {
		var aOffset = a.offset,
			bOffset = b.offset;

		if (aOffset !== bOffset) {
			return aOffset - bOffset;
		}

		if (a.text.length !== b.text.length) {
			return a.text.length - b.text.length;
		}

		if (a.label !== b.label) {
			if (a < b) {
				return -1;
			} else {
				return 1;
			}
		}

		return 0;
	}

	NodePath.parse = function parse(path, startRule, types) {
		if (!startRule) {
			startRule = 'NodePathExpression';
		}

		var cache = NodePath._cache,
			key = path + startRule;

		if (types) {
			for (var each in types) {
				key += each;
			}
		}

		var result = cache.getItem(key);

		if (!result) {
			var exception = null,
				keywords = [],
				parsedPath;

			try {
				parsedPath = NodePathParser.parse(path, { 
					startRule : startRule,
					types : types
				});
			} catch (e) {
				exception = e;
			}

			if (parsedPath) {
				keywords = parsedPath.keywords;
				keywords.sort(keywordCompare);
				keywords.sortedUniquify(keywordCompare);
			}

			result = {
				parsedPath: parsedPath,
				error: exception,
				keywords: keywords
			};

			cache.setItem(key, result);
		}

		return result;
	};

	NodePath.evaluate = function evaluate(nodePath, currentNode) {
		var isString = typeof nodePath === 'string' || nodePath instanceof String;
		if (isString) {
			nodePath = new NodePath(nodePath, currentNode.tree.taxonomy.types());
		}
		return nodePath.evaluate(currentNode);
	};

	NodePath.calculateMinNodePath = function calculateMinNodePath(node, includeDescendants) {
		var nodes = [];

		while (node) {
			nodes.push(node);
			node = node.parent;
		}

		nodes = nodes.reverse();

		var pathSegments = [],
			length = nodes.length,
			previousParent = nodes[0],
			i, j;

		for (i = 1; i < length; i++) {
			var targetChild = nodes[i],
				targetChildText = targetChild.text(),
				targetChildTextLength = targetChildText.length,
				candidateSegment;

			if (targetChildTextLength === 0) {
				candidateSegment = '@id = ' + targetChild.id;
			} else {
				var candidateSegmentLength = Math.min(4, targetChildTextLength),
					matches = Number.MAX_VALUE;

				while (candidateSegmentLength <= targetChildTextLength && matches > 1) {
					candidateSegment = targetChildText.substr(0, candidateSegmentLength).toLowerCase().replace(/^\s+|\s+$/g, '');
					matches = 0;

					var each = previousParent.firstChild;
					while (each) {
						if (each.text().toLowerCase().indexOf(candidateSegment) !== -1) {
							matches++;
						}
						each = each.nextSibling;
					}
					candidateSegmentLength++;
				}

				if (matches > 1) {
					candidateSegment = '@id = ' + targetChild.id;
				} else {
					// escape all quotes within the candidate segment with a \
					candidateSegment = candidateSegment.replace(/\"/g, '\\"');

					try {
						NodePathParser.parse(candidateSegment,  { startRule : 'Value' });
					} catch (e) {
						// Only add quotes when needed, otherwise leave the built path clean.
						candidateSegment = '"' + candidateSegment + '"';
					}
				}
			}

			pathSegments.push(candidateSegment);
			previousParent = targetChild;
		}

		var result = '/' + pathSegments.join('/');
		if (includeDescendants) {
			result += '///*';
		}
		return result;
	};

	NodePath.syntaxHighlight = function syntaxHighlight(attributedString, startRule, range) {
		if (!startRule) {
			startRule = 'NodePathExpression';
		}

		if (!range) {
			range = { location : 0, length: attributedString.string().length };
		}

		if (range.length === 0) {
			return attributedString;
		}

		var rangeString = attributedString.string().substr(range.location, range.length),
			results = NodePath.parse(rangeString, startRule),
			keywords = results.keywords,
			length = keywords.length,
			error = results.error,
			i;

		for (i = 0; i < length; i++) {
			var each = keywords[i],
				label = each.label,
				offset = each.offset,
				text = each.text;

			attributedString.addAttributeInRange(label, text, range.location + offset, text.length);
		}

		if (error) {
			var errorOffset = error.offset,
				errorLength = range.length - errorOffset;
			if (errorLength > 0) {
				var errorString = attributedString.string().substr(errorOffset + range.location, errorLength);
				attributedString.addAttributeInRange('invalid', errorString, errorOffset + range.location, errorLength);
			}
		}

		return attributedString;
	};

	NodePath.SHOW_ALL_PATH = '///*';
	NodePath.DEFAULT_ATTRIBUTE_PATH = ['line', 'text'];
	NodePath.DEFAULT_RELATIOIN = 'contains';
	NodePath.DEFAULT_MODIFIER = 'i';

	exports.NodePath = NodePath;
});
