//
// Copyright (c) 2014 Jesse Grosjean. All rights reserved.
//
// Permission is given for you to use this software ONLY while creating themes
// and plugins for the "FoldingText" software platform. Permission is NOT
// given for any other purpose.
//

/** @module ft/core/nodeset */
define(function (require, exports, module) {
	'use strict';

	var ObjectUtils = require('../util/objectutils');

	/**
	 * A "set" container for {@link
	 * module:ft/core/node.Node|Nodes}.
	 *
	 * @constructor
	 * @alias module:ft/core/nodeset.NodeSet
	 */
	var NodeSet = function NodeSet(nodeSet) {
		this._idsToNodes = {};
		this.count = 0;
		if (nodeSet) {
			this.unionSet(nodeSet);
		}
	};

	/**
	 * Return a copy of this set.
	 *
	 * @returns {module:ft/core/nodeset.NodeSet}
	 */
	NodeSet.prototype.copy = function copy() {
		return new NodeSet(this);
	};

	NodeSet.prototype.nodeForID = function nodeForID(id) {
		return this._idsToNodes[id];
	};

	NodeSet.prototype.anyNode = function anyNode() {
		if (this.count === 0) {
			return undefined;
		}

		var idsToNodes = this._idsToNodes,
			key;
		for (key in idsToNodes) {
			var node = idsToNodes[key];
			if (node) {
				return node;
			}
		}
		return undefined;
	};

	/**
	 * Return true if node is contained in this set.
	 *
	 * @param {module:ft/core/node.Node}
	 * @return {Boolean}
	 */
	NodeSet.prototype.containsNode = function containsNode(node) {
		if (this._idsToNodes[node.id]) {
			return true;
		}
		return false;
	};

	NodeSet.prototype.containsNodeForID = function containsNodeForID(id) {
		if (this._idsToNodes[id]) {
			return true;
		}
		return false;
	};

	NodeSet.prototype.containsNodes = function containsNodes(nodes) {
		var length = nodes.length,
			i;
		for (i = 0; i < length; i++) {
			if (!this.containsNode(nodes[i])) {
				return false;
			}
		}
		return true;
	};

	/**
	 * Add a new node to this set.
	 *
	 * @param {module:ft/core/node.Node}
	 */
	NodeSet.prototype.addNode = function addNode(node) {
		var map = this._idsToNodes,
			id = node.id;
		if (!map[id]) {
			map[id] = node;
			this.count++;
		}
	};

	NodeSet.prototype.addNodeIgnoringNull = function addNodeIgnoringNull(node) {
		if (node) {
			this.addNode(node);
		}
	};

	NodeSet.prototype.addNodesFromArray = function addNodesFromArray(nodes) {
		var length = nodes.length,
			i;
		for (i = 0; i < length; i++) {
			this.addNode(nodes[i]);
		}
	};

	/**
	 * Remove an existing node from this set.
	 *
	 * @param {module:ft/core/node.Node}
	 */
	NodeSet.prototype.removeNode = function removeNode(node) {
		var map = this._idsToNodes,
			id = node.id;
		if (map[id]) {
			map[id] = undefined;
			this.count--;
		}
	};

	NodeSet.prototype.removeNodeIgnoringNull = function removeNodeIgnoringNull(node) {
		if (node) {
			this.removeNode(node);
		}
	};

	NodeSet.prototype.removeNodesFromArray = function removeNodesFromArray(nodes) {
		var length = nodes.length,
			i;
		for (i = 0; i < length; i++) {
			this.removeNode(nodes[i]);
		}
	};

	NodeSet.prototype.removeAllNodes = function removeAllNodes() {
		this._idsToNodes = {};
		this.count = 0;
	};

	/**
	 * Return true if this set contains the same nodes as the provided
	 * set.
	 *
	 * @param {module:ft/core/nodeset.NodeSet}
	 * @returns {Boolean}
	 */
	NodeSet.prototype.equalsSet = function equalsSet(nodeSet) {
		if (this.count !== nodeSet.count) {
			return false;
		}

		var thisIdsToNodes = this._idsToNodes,
			thatIdsToNodes = nodeSet._idsToNodes,
			key;

		for (key in thatIdsToNodes) {
			if (thatIdsToNodes[key] !== thisIdsToNodes[key]) {
				return false;
			}
		}

		for (key in thisIdsToNodes) {
			if (thisIdsToNodes[key] !== thatIdsToNodes[key]) {
				return false;
			}
		}
		return true;
	};

	/**
	 * Add all nodes from the provided set to this set.
	 *
	 * @param {module:ft/core/nodeset.NodeSet}
	 */
	NodeSet.prototype.unionSet = function unionSet(nodeSet) {
		if (!nodeSet) { return; }

		var thisIdsToNodes = this._idsToNodes,
			thatIdsToNodes = nodeSet._idsToNodes;
		for (var key in thatIdsToNodes) {
			if (thatIdsToNodes[key] && !thisIdsToNodes[key]) {
				thisIdsToNodes[key] = thatIdsToNodes[key];
				this.count++;
			}
		}
		return this;
	};

	/**
	 * Remove all nodes from the provided set from this set.
	 *
	 * @param {module:ft/core/nodeset.NodeSet}
	 */
	NodeSet.prototype.minusSet = function minusSet(nodeSet) {
		var thisIdsToNodes = this._idsToNodes,
			thatIdsToNodes = nodeSet._idsToNodes;
		for (var key in thatIdsToNodes) {
			if (thatIdsToNodes[key] && thisIdsToNodes[key]) {
				thisIdsToNodes[key] = undefined;
				this.count--;
			}
		}
		return this;
	};

	/**
	 * Remove all nodes from this set that are not also included in the
	 * provided set.
	 *
	 * @param {module:ft/core/nodeset.NodeSet}
	 */
	NodeSet.prototype.intersectSet = function intersectSet(nodeSet) {
		var thisIdsToNodes = this._idsToNodes,
			thatIdsToNodes = nodeSet._idsToNodes,
			intersectionIDsToNodes = {},
			count = 0,
			key;

		for (key in thatIdsToNodes) {
			if (thatIdsToNodes[key]) {
				var each = thisIdsToNodes[key];
				if (each) {
					intersectionIDsToNodes[key] = each;
					count++;
				}
			}
		}

		this._idsToNodes = intersectionIDsToNodes;
		this.count = count;
		return this;
	};

	/**
	 * A callback for processing each node in a set.
	 *
	 *	nodeset.forEachNodeInSet(function (eachNode) {
	 *		console.log(eachNode.text());
	 *	});
	 *
	 * @callback forEachNodeInSetCallback
	 * @param {module:ft/core/node.Node} each - Each node in the set.
	 */

	/**
	 * Process each node in this set using the provided callback.
	 *
	 * @param {forEachNodeInSetCallback} - Callback to process each node in
	 *     the set.
	 */
	NodeSet.prototype.forEachNodeInSet = function forEachNodeInSet(callback) {
		var idsToNodes = this._idsToNodes,
			each,
			key;
		for (key in idsToNodes) {
			each = idsToNodes[key];
			if (each) {
				callback(each);
			}
		}
	};

	/**
	 * Return the nodes in this set sorted by line number.
	 *
	 * @return {Array.<module:ft/core/node.Node>}
	 */
	NodeSet.prototype.nodesInLineOrder = function nodesInLineOrder() {
		var count = this.count,
			anyNode = this.anyNode(),
			tree = anyNode ? anyNode.tree : null,
			idsToNodes;

		// 0. Simplest case, if only one node, no sorting needed.
		if (count === 1) {
			return [anyNode];
		} else if (tree) {
			var results = [],
				each;

			idsToNodes = this._idsToNodes;

			// 1. Count is less then 10% then lookup line numbers and sort.
			if (count < tree.nodeCount() / 10) {
				var nodeToLineNumberNumberTable = {},
					key;

				for (key in idsToNodes) {
					each = idsToNodes[key];
					if (each) {
						results.push(each);
						nodeToLineNumberNumberTable[each.id] = tree.nodeToLineNumber(each);
					}
				}

				var _lineOrderSorter = function (a, b) {
					return nodeToLineNumberNumberTable[a.id] - nodeToLineNumberNumberTable[b.id];
				};

				results = results.sort(_lineOrderSorter);
			// 2. Else iterate over entire model and sort that way.
			} else {
				each = tree.root.firstChild;
				while (each) {
					if (idsToNodes[each.id]) {
						results.push(each);
					}
					each = each.nextLineNode();
				}
			}
			return results;
		}

		return [];
	};

	NodeSet.prototype.nodesInReverseLineOrder = function nodesInReverseLineOrder() {
		return this.nodesInLineOrder().reverse();
	};

	NodeSet.prototype.nodesInLevelAscendingOrder = function nodesInLevelAscendingOrder() {
		var levelAscendingSorter = function (a, b) {
			return b.level - a.level;
		};
		return ObjectUtils.arrayOfValues(this._idsToNodes).sort(levelAscendingSorter);
	};

	NodeSet.prototype.toIDArray = function toIDArray() {
		var idsToNodes = this._idsToNodes,
			ids = [],
			id;

		for (id in idsToNodes) {
			ids.push(id);
		}

		return ids;
	};

	//
	// Debug
	//

	NodeSet.prototype.toString = function toString() {
		var ids = [],
			nodes = this.nodesInLineOrder(),
			length = nodes.length;
		for (var i = 0; i < length; i++) {
			ids.push(nodes[i].id);
		}
		return '[' + ids.join(', ') + ']';
	};

	NodeSet.EmptySet = new NodeSet();

	NodeSet.unionSets = function union(nodeSet1, nodeSet2) {
		var result = new NodeSet();
		result.unionSet(nodeSet1);
		result.unionSet(nodeSet2);
		return result;
	};

	NodeSet.minusSets = function union(nodeSet1, nodeSet2) {
		var result = nodeSet1.copy();
		result.minusSet(nodeSet2);
		return result;
	};

	NodeSet.intersectSets = function union(nodeSet1, nodeSet2) {
		var result = nodeSet1.copy();
		result.intersectSet(nodeSet2);
		return result;
	};

	NodeSet.setFromArray = function setFromArray(nodes) {
		var result = new NodeSet(),
			length = nodes.length;
		for (var i = 0; i < length; i++) {
			result.addNode(nodes[i]);
		}
		return result;
	};

	NodeSet.setFromNode = function setFromNode(node) {
		var result = new NodeSet();
		result.addNode(node);
		return result;
	};

	exports.NodeSet = NodeSet;
});
