//
// Copyright (c) 2014 Jesse Grosjean. All rights reserved.
//
// Permission is given for you to use this software ONLY while creating themes
// and plugins for the "FoldingText" software platform. Permission is NOT
// given for any other purpose.
//

/** @module ft/core/range */
define(function (require, exports, module) {
	'use strict';

	var NodeSet = require('./nodeset').NodeSet,
		Node = require('./node').Node;

	/**
	 * A range of content in a {@link
	 * module:ft/core/tree.Tree|Tree}
	 *
	 * A range is defined by two points: A start and end node, and text
	 * offsets into the line of those nodes. Ranges provides conversion
	 * functions to efficiently get line numbers and document text offsets
	 * from these base positions.
	 *
	 * To create a new range use one of the {@link
	 * module:ft/core/tree.Tree#createRangeFromNodes|Tree#createRange}
	 * methods.
	 *
	 * @private
	 * @constructor
	 * @alias module:ft/core/range.Range
	 */
	var Range = function Range(tree, startNode, startOffset, endNode, endOffset, anchoredAtEnd) {
		this._tree = tree;

		if (!startNode) {
			startNode = tree.root.firstChild;
		}

		if (!endNode) {
			endNode = startNode;
		}

		if (startOffset === undefined) {
			startOffset = 0;
			if (endOffset === undefined) {
				endOffset = endNode.line().length;
			}
		} else if (endOffset === undefined) {
			endOffset = startOffset;
		}

		/**
		 * @type module:ft/core/node.Node
		 */
		this.startNode = startNode;

		/**
		 * @type Number
		 */
		this.startOffset = startOffset;

		/**
		 * @type module:ft/core/node.Node
		 */
		this.endNode = endNode;

		/**
		 * @type Number
		 */
		this.endOffset = endOffset;
		this.anchoredAtStart = !anchoredAtEnd;
	};

	/**
	 * Return copy of this range.
	 *
	 * @returns {module:ft/core/range.Range}
	 */
	Range.prototype.copy = function copy() {
		return new Range(this._tree, this.startNode, this.startOffset, this.endNode, this.endOffset, true);
	};

	/**
	 * Return true if this range has zero length.
	 *
	 * @returns {Boolean}
	 * @methodgroup Testing
	 */
	Range.prototype.isCollapsed = function isCollapsed() {
		return this.startNode === this.endNode && this.startOffset === this.endOffset;
	};

	Range.prototype.isSelectingAll = function isSelectingAll() {
		return this.length() === this._tree.textLength();
	};

	/**
	 * Return true if this range is equal to the range.
	 *
	 * @param {Range}
	 * @returns {Boolean}
	 */
	Range.prototype.isEqualToRange = function isEqualToRange(range) {
		return this.startNode === range.startNode &&
			this.startOffset === range.startOffset &&
			this.endNode === range.endNode &&
			this.endOffset === range.endOffset;
	};

	Range.prototype.containsRange = function containsRange(range) {
		var thisStart = this.location(),
			thisEnd = thisStart + this.length(),
			rangeStart = range.location(),
			rangeEnd = rangeStart + range.length();
		return thisStart <= rangeStart && thisEnd >= rangeEnd;
	};
	
	/**
	 * Return the tree text offset location of this range.
	 *
	 * @returns {Number}
	 * @methodgroup Text
	 */
	Range.prototype.location = function location() {
		return this._tree.nodeOffsetToTextOffset(this.startNode, this.startOffset);
	};

	/**
	 * The number of characters covered by this range.
	 *
	 * @returns {Number}
	 */
	Range.prototype.length = function length() {
		var startNode = this.startNode,
			endNode = this.endNode;

		if (startNode === endNode) {
			return this.endOffset - this.startOffset;
		} else {
			return this._tree.nodeOffsetToTextOffset(endNode, this.endOffset) - this.location();
		}
	};

	/**
	 * Return text in this range.
	 *
	 * @returns {String}
	 */
	Range.prototype.textInRange = function textInRange() {
		var lines = [];
		this.forEachNodeInRange(function (node, line, location, length, fullySelected) {
			if (fullySelected) {
				lines.push(node.line());
			} else {
				lines.push(node.lineInRange(location, length));
			}
		});
		return lines.join('\n');
	};

	Range.prototype.linesInRange = function linesInRange() {
		var lines = [];
		this.forEachNodeInRange(function (node, line, location, length, fullySelected) {
			lines.push(node.line());
		});
		return lines;
	};

	/**
	 * Return starting line number of this range.
	 *
	 * @returns {Number}
	 * @methodgroup Lines
	 */
	Range.prototype.startLine = function startLine() {
		return this._tree.nodeToLineNumber(this.startNode);
	};

	/**
	 * Return starting LineChar of this range.
	 *
	 * @returns {LineChar}
	 */
	Range.prototype.startLineCh = function startLineCh() {
		return {line: this.startLine(), ch: this.startOffset};
	};

	/**
	 * Return ending line number of this range.
	 *
	 * @returns {Number}
	 */
	Range.prototype.endLine = function endLine() {
		return this._tree.nodeToLineNumber(this.endNode);
	};

	/**
	 * Return ending LineChar of this range.
	 *
	 * @returns {LineChar}
	 */
	Range.prototype.endLineCh = function endLineCh() {
		return {line: this.endLine(), ch: this.endOffset};
	};

	Range.prototype.anchorNode = function anchorNode() {
		return this.anchoredAtStart ? this.startNode : this.endNode;
	};

	Range.prototype.anchorOffset = function anchorOffset() {
		return this.anchoredAtStart ? this.startOffset : this.endOffset;
	};

	Range.prototype.anchorLineCh = function anchorLineCh() {
		return this.anchoredAtStart ? this.startLineCh() : this.endLineCh();
	};

	Range.prototype.headNode = function headNode() {
		return !this.anchoredAtStart ? this.startNode : this.endNode;
	};

	Range.prototype.headOffset = function headOffset() {
		return !this.anchoredAtStart ? this.startOffset : this.endOffset;
	};

	Range.prototype.headLineCh = function headLineCh() {
		return !this.anchoredAtStart ? this.startLineCh() : this.endLineCh();
	};

	/**
	 * Return nodes in this range.
	 *
	 * @returns {Array.<module:ft/core/node.Node>}
	 * @methodgroup Nodes
	 */
	Range.prototype.nodesInRange = function nodesInRange() {
		var nodes = [];
		this.forEachNodeInRange(function (node, line, location, length, fullySelected) {
			nodes.push(node);
		});
		return nodes;
	};

	Range.prototype.nodeAfterEndNode = function nodeAfterEndNode() {
		return this.endNode.nextLineNode();
	};

	Range.prototype.startNodeTextBeforeRange = function startNodeTextBeforeRange() {
		return this.startNode.line().substr(0, this.startOffset);
	};

	Range.prototype.endNodeTextAfterRange = function endNodeTextAfterRange() {
		return this.endNode.line().substr(this.endOffset);
	};

	Range.prototype.beginTrackingOffsets = function beginTracking() {
		if (this.startNode === this.endNode) {
			this.startNode.addTrackedRange(this);
		} else {
			this.startNode.addTrackedRange(this);
			this.endNode.addTrackedRange(this);
		}
	};

	Range.prototype.endTrackingOffsets = function endTracking() {
		if (this.startNode === this.endNode) {
			this.startNode.removeTrackedRange(this);
		} else {
			this.startNode.removeTrackedRange(this);
			this.endNode.removeTrackedRange(this);
		}
	};

	/**
	 * A callback for processing each line in a range.
	 *
	 *	range.forEachNodeInRange(function (node, line, location, length, fullySelected) {
	 *		console.log(eachNode.text());
	 *	});
	 *
	 * @callback forEachNodeInRangeCallback
	 * @param {module:ft/core/node.Node} node - Each node.
	 * @param {Number} line - Line number.
	 * @param {Number} location - Node relative range start location.
	 * @param {Number} length - Node relative range length.
	 * @param {Boolean} fullySelected - True if the node is fully selected by
	 *     the range.
	 */

	/**
	 * Process each line in the range.
	 *
	 * @param {forEachNodeInRangeCallback} - Callback to process each node in the
	 *     set.
	 */
	Range.prototype.forEachNodeInRange = function forEachNodeInRange(callback) {
		var startNode = this.startNode,
			startLine = this.startLine(),
			startOffset = this.startOffset,
			endNode = this.endNode,
			endOffset = this.endOffset;

		if (startNode === endNode) {
			callback(startNode, startLine, startOffset, endOffset - startOffset, false);
		} else {
			var each = startNode.nextLineNode();

			callback(startNode, startLine, startOffset, startNode.line().length - startOffset, false);
			while (each !== endNode) {
				callback(each, ++startLine, 0, each.line().length, true);
				each = each.nextLineNode(each);
			}
			callback(endNode, ++startLine, 0, endOffset, false);
		}
	};

	/**
	 * Return a new range resulting from offsetting this range.
	 *
	 * @param {Boolean} [toStart=false] - Collapse to start or end.
	 * @returns {module:ft/core/range.Range}
	 * @methodgroup Creating Ranges
	 */
	Range.prototype.rangeByCollapsing = function rangeByCollapsing(toStart) {
		var result = this.copy();
		if (toStart) {
			result.endNode = result.startNode;
			result.endOffset = result.startOffset;
		} else {
			result.startNode = result.endNode;
			result.startOffset = result.endOffset;
		}
		return result;
	};

	/**
	 * Return a new range resulting from offsetting this range.
	 *
	 * @param {Number} - Offset start location by this amount.
	 * @param {Number} - Offset length by this amount.
	 * @returns {module:ft/core/range.Range}
	 */
	Range.prototype.rangeByOffsetting = function rangeByOffsetting(locationOffset, lengthOffset) {
		var location = this.location(),
			length = this.length(),
			tree = this._tree,
			textLength = tree.textLength();

		location += locationOffset;
		length += lengthOffset;

		if (location < 0) { location = 0; }
		if (location > textLength) { location = textLength; }
		if (length < 0) { length = 0; }
		if (location + length > textLength) { length = textLength - location; }

		return tree.createRangeFromLocation(location, length);
	};

	Range.prototype.rangeByAdjustingToEffectedNodesInEditor = function rangeByAdjustingToEffectedNodesInEditor(editor) {
		var startNode = this.startNode,
			startOffset = this.startOffset,
			endNode = this.endNode,
			endOffset = this.endOffset,
			tree = this._tree;

		if (editor) {
			var firstVisible = startNode;
			while (firstVisible && editor.nodeIsHiddenInFold(firstVisible, true)) {
				firstVisible = firstVisible.previousLineNode();
			}

			if (!firstVisible) {
				firstVisible = tree.firstLineNode();
			}

			if (firstVisible !== startNode) {
				startNode = firstVisible;
				startOffset = 0;
			}
		}

		if (startNode !== endNode && endOffset === 0) {
			endNode = endNode.previousLineNode();
			endOffset = endNode.line().length;
		}

		if (editor && (editor.nodeIsHiddenInFold(endNode) || editor.isCollapsed(endNode))) {
			var lastHidden = endNode.nextLineNode();
			while (lastHidden && editor.nodeIsHiddenInFold(lastHidden)) {
				lastHidden = lastHidden.nextLineNode();
			}

			if (lastHidden) {
				lastHidden = lastHidden.previousLineNode();
			} else {
				lastHidden = tree.lastLineNode();
			}

			if (lastHidden !== endNode) {
				endNode = lastHidden;
				endOffset = endNode.line().length;
			}
		}

		return tree.createRangeFromNodes(startNode, startOffset, endNode, endOffset);
	};

	/**
	 * Return a new range resulting from extending this range to word
	 * boundaries.
	 *
	 * @returns {module:ft/core/range.Range}
	 */
	Range.prototype.rangeByExtendingToWord = function rangeByExtendingToWord(justReturnPseudoRange) {
		var startNode = this.startNode,
			endNode = this.endNode,
			taxonomy = this._tree.taxonomy,
			startRange = taxonomy.wordRange(startNode, this.startOffset),
			endRange = taxonomy.wordRange(endNode, this.endOffset);

		if (justReturnPseudoRange) {
			return { location: startRange.location, length: endRange.location + endRange.length - startRange.location };
		}

		var newRange = this._tree.createRangeFromNodes(startNode, startRange.location, endNode, endRange.location + endRange.length);
		if (newRange.textInRange().replace(/^[.?!; ]+$/g, '').length === 0) {
			return this;	// don't create range if you can't
		}
		return newRange;
	};

	/**
	 * Return a new range resulting from extending this range to sentence
	 * boundaries.
	 *
	 * @returns {module:ft/core/range.Range}
	 */
	Range.prototype.rangeByExtendingToSentence = function rangeByExtendingToSentence(justReturnPseudoRange) {
		var startNode = this.startNode,
			endNode = this.endNode,
			taxonomy = this._tree.taxonomy,
			startRange = taxonomy.sentenceRange(startNode, this.startOffset),
			endRange = taxonomy.sentenceRange(endNode, this.endOffset);

		if (this.length() > 0 && endRange.location === this.endOffset) {
			if (justReturnPseudoRange) {
				return { location: startRange.location, length: this.endOffset - startRange.location };
			}
			return this._tree.createRangeFromNodes(startNode, startRange.location, endNode, this.endOffset);
		} else {
			if (justReturnPseudoRange) {
				return { location: startRange.location, length: endRange.length + endRange.location - startRange.location };
			}
			return this._tree.createRangeFromNodes(startNode, startRange.location, endNode, endRange.location + endRange.length);
		}
	};

	Range.prototype.rangeByExtendingToLine = function rangeByExtendingToLine() {
		var endNode = this.endNode;
		return this._tree.createRangeFromNodes(this.startNode, 0, endNode, endNode.line().length);
	};

	Range.prototype.rangeByExtendingToEndOfLine = function rangeByExtendingToEndOfLine() {
		var tree = this._tree,
			startNode = this.startNode,
			startOffset = this.startOffset,
			endNode = this.endNode;
		return tree.createRangeFromNodes(startNode, startOffset, endNode, endNode.line().length);
	};

	Range.prototype.rangeByExtendingToParagraph = function rangeByExtendingToParagraph() {
		var tree = this._tree,
			startNode = this.startNode,
			endNode = this.endNode,
			endOffset = endNode.line().length;

		if (this.length() > 0 && this.endOffset === 0) {
			endOffset = 0;
		} else {
			var lineAfterNode = endNode.nextLineNode();
			if (lineAfterNode) {
				endOffset = 0;
				endNode = lineAfterNode;
			}
		}
		return tree.createRangeFromNodes(startNode, 0, endNode, endOffset);
	};

	Range.prototype.rangeByExtendingToBeginningOfParagraph = function rangeByExtendingToBeginningOfParagraph() {
		var tree = this._tree,
			startNode = this.startNode,
			endNode = this.endNode,
			endOffset = this.endOffset;
		return tree.createRangeFromNodes(startNode, 0, endNode, endOffset);
	};

	Range.prototype.rangeByExtendingToEndOfParagraph = function rangeByExtendingToEndOfParagraph() {
		var tree = this._tree,
			startNode = this.startNode,
			startOffset = this.startOffset,
			extendedRange = this.rangeByExtendingToParagraph();
		return tree.createRangeFromNodes(startNode, startOffset, extendedRange.endNode, extendedRange.endOffset);
	};

	Range.prototype.rangeByExtendingToBranch = function rangeByExtendingToBranch() {
		var branchNodes = new NodeSet(),
			rangeNodes = this.rangeByAdjustingToEffectedNodesInEditor().nodesInRange(),
			commonAncestors = Node.commonAncestorsForNodes(rangeNodes);

		commonAncestors.forEachNodeInSet(function (each) {
			var descendants = each.descendants(),
				length = descendants.length,
				i;

			branchNodes.addNode(each);

			for (i = 0; i < length; i++) {
				branchNodes.addNode(descendants[i]);
			}
		});

		var orderedBranchNodes = branchNodes.nodesInLineOrder(),
			firstNode = orderedBranchNodes[0],
			lastNode = orderedBranchNodes.lastObject(),
			lineAfterNode = lastNode.nextLineNode();

		if (lineAfterNode) {
			return this._tree.createRangeFromNodes(firstNode, 0, lineAfterNode, 0);
		} else {
			return this._tree.createRangeFromNodes(firstNode, 0, lastNode, lastNode.line().length);
		}
	};

	Range.prototype.rangeByExtendingToSiblingBranches = function rangeByExtendingToSiblingBranches() {
		var range = this.rangeByExtendingToBranch(),
			startNode = range.startNode.parent.nextLineNode(),
			endNode = range.startNode.parent.nextBranch();

		if (endNode === null) {
			endNode = this._tree.lastLineNode();
			return this._tree.createRangeFromNodes(startNode, 0, endNode, endNode.line().length);
		} else {
			return this._tree.createRangeFromNodes(startNode, 0, endNode, 0);
		}
	};

	Range.prototype.rangeByExtendingToParentBranch = function rangeByExtendingToParentBranch() {
		var range = this.rangeByExtendingToBranch(),
			startNode = range.startNode.parent.nextLineNode(),
			startNodeParent = startNode.parent,
			endNode = range.startNode.parent.nextBranch(),
			tree = this._tree,
			root = tree.root;

		return tree.createRangeFromNodes(startNodeParent, 0).rangeByExtendingToBranch();
	};

	Range.prototype.rangeByLevelUpExtension = function rangeByLevelUpExtension() {
		var startNode = this.startNode,
			endNode = this.endNode,
			attributedString = startNode.lineAttributedString(),
			origRange = { location : this.startOffset },
			multiLine = startNode !== endNode,
			tree = this._tree,
			startOffset = this.startOffset,
			endOffset = this.endOffset;

		function narrowerRange(r1, r2) {
			// following two tests require that the selection souldn't be narrower than the current range
			if (r1.location >= startOffset && r1.length + r1.location <= endOffset) { return r2; }
			if (r2.location >= startOffset && r2.length + r2.location <= endOffset) { return r1; }
			if (r1.location <= r2.location && r1.location + r1.length >= r2.location + r2.length) { return r2; }
			return r1;
		}

		if (!multiLine) {
			if (startOffset === 0 && endOffset === this.endNode.line().length) {
				if (endNode.nextLineNode()) {
					return this.rangeByExtendingToParagraph();
				} else {
					return this.rangeByExtendingToParentBranch();
				}
			}

			origRange.length = endOffset - startOffset;

			var extendedRange = attributedString.extendedRangeFromRange(origRange),
				narrower;

			if (this.isCollapsed()) {
				var wordRange = this.rangeByExtendingToWord(true);
				narrower = narrowerRange(wordRange, extendedRange);
				return tree.createRangeFromNodes(startNode, narrower.location, startNode, narrower.location + narrower.length);
			} else {
				var sentenceRange = this.rangeByExtendingToSentence(true);
				narrower = narrowerRange(sentenceRange, extendedRange);
				return tree.createRangeFromNodes(startNode, narrower.location, startNode, narrower.location + narrower.length);
			}
		} else {
			var parentBranchRange = this.rangeByExtendingToBranch();
			if (this.isEqualToRange(parentBranchRange)) {
				return this.rangeByExtendingToParentBranch();
			} else {
				return parentBranchRange;
			}
		}
	};

	Range.prototype.rangeByExtendingToLevel = function rangeByExtendingToLevel(level) {
		if (level === 0) {
			return this;
		}

		if (level === 1) {
			return this.rangeByExtendingToWord();
		}

		if (level === 2) {
			return this.rangeByExtendingToSentence();
		}

		if (level === 3) {
			return this.rangeByExtendingToLine();
		}

		if (level === 4) {
			return this.rangeByExtendingToParagraph();
		}

		if (level === 5) {
			return this.rangeByExtendingToBranch();
		}

		var range = this,
			doSiblings = true;

			// in progress trying to make extend to sibling branches working...
		while (level > 5) {
			if (doSiblings) {
				range = range.rangeByExtendingToSiblingBranches();
			} else {
				range = range.rangeByExtendingToParentBranch();
			}
			doSiblings = !doSiblings;
			level--;
		}
		return range;
	};

	Range.prototype.toString = function toString() {
		var startLine = this.startNode.line(),
			startOffset = this.startOffset,
			endLine = this.endNode.line(),
			endOffset = this.endOffset;

		if (this.startNode === this.endNode) {
			var line = startLine.substr(0, startOffset) + '^' + startLine.substr(startOffset);
			if (startOffset !== endOffset) {
				line = line.substr(0, endOffset + 1) + '^' + line.substr(endOffset + 1);
			}
			return '("' + line + '")';
		} else {
			var join = [];
			join.push(startLine.substr(0, startOffset) + '^' + startLine.substr(startOffset));
			join.push(endLine.substr(0, endOffset) + '^' + endLine.substr(endOffset));
			return '("' + join.join('", "') + '")';
		}
	};

	Range.copyRanges = function copyRanges(ranges) {
		var copiedRanges = [];
		var length = ranges.length;
		var i;
		for (i = 0; i < length; i++) {
			copiedRanges.push(ranges[i].copy());
		}
		return copiedRanges;
	};

	Range.areRangesEqual = function areRangesEqual(range1, range2) {
		if (range1 === null) { return range2 === null; }
		if (range2 === null) { return range1 === null; }
		return range1.isEqualToRange(range2);
	};

	Range.areRangeArraysEqual = function areRangeArraysEqual(ranges1, ranges2) {
		if (ranges1 === null) { return ranges2 === null; }
		if (ranges2 === null) { return ranges1 === null; }

		var length1 = ranges1.length;
		var length2 = ranges2.length;

		if (length1 !== length2) { return false; }

		var i;
		for (i = 0; i < length1; i++) {
			var each1 = ranges1[i];
			var each2 = ranges2[i];
			if (!each1.isEqualToRange(each2)) {
				return false;
			}
		}

		return true;
	};

	Range.nodesInRanges = function nodesInRanges(ranges) {
		var nodes = [];
		var length = ranges.length;
		var i;
		for (i = 0; i < length; i++) {
			nodes = nodes.concat(ranges[i].nodesInRange());
		}
		return nodes;
	};

	Range.textInRanges = function textInRanges(ranges) {
		var lines = [];
		var length = ranges.length;
		var i;
		for (i = 0; i < length; i++) {
			lines.push(ranges[i].textInRange());
		}
		return lines.join('');
	};

	Range.forEachNodeInRanges = function forEachNodeInRanges(ranges, block) {
		var length = ranges.length;
		var i;
		for (i = 0; i < length; i++) {
			ranges[i].forEachNodeInRange(block);
		}
	};

	Range.isLessThanLineCh = function isLessThanLineCh(a, b) {
		var aLine = a.line,
			bLine = b.line;
		if (aLine < bLine) {
			return true;
		} else if (aLine === bLine) {
			return a.ch < b.ch;
		}
		return false;
	};

	Range.isLessThanOrEqualLineCh = function isLessThanOrEqualLineCh(a, b) {
		var aLine = a.line,
			bLine = b.line;
		if (aLine < bLine) {
			return true;
		} else if (aLine === bLine) {
			return a.ch <= b.ch;
		}
		return false;
	};

	Range.rangeFromNodes = function rangeFromNodes(startNode, endNode) {
		return new Range(startNode.tree, startNode, 0, endNode, endNode.line().length);
	};

	exports.Range = Range;
});