//
// Copyright (c) 2014 Jesse Grosjean. All rights reserved.
//
// Permission is given for you to use this software ONLY while creating themes
// and plugins for the "FoldingText" software platform. Permission is NOT
// given for any other purpose.
//
define(function (require, exports, module) {
	'use strict';

	var Classifier = require('./classifier').Classifier,
		ObjectUtils = require('../util/objectutils'),
		ROOT_STRUCTURE_TYPE_LEVEL = -100,
		ESCAPE_SYNTAX_CHAR = '\uE002';

	var nonASCIISingleCaseWordChar = /[\u3040-\u309f\u30a0-\u30ff\u3400-\u4db5\u4e00-\u9fcc]/,
		whiteSpaceChar = /\s/,
		wordChar = /\w/;

	function isWordChar(ch) {
		return wordChar.test(ch) || ch > '\x80' && (ch.toUpperCase() !== ch.toLowerCase() || nonASCIISingleCaseWordChar.test(ch));
	}

	function isWhitespaceChar(ch) {
		return whiteSpaceChar.test(ch);
	}

	function isNotWhiteSpaceAndNotWordChar(ch) {
		return !isWhitespaceChar(ch) && !isWordChar(ch);
	}

	var Taxonomy = function Taxonomy(spec) {
		this.spec = spec || {};
	};

	(function () {

		this.wordRange = function wordRange(node, index) {
			var string = node.line(),
				length = string.length,
				foundStart = index,
				foundEnd = index,
				flip = false,
				test,
				i;

			if (length === 0) {
				return { location : index, length : 0 };
			}

			var beforeCh = string[index - 1],
				afterCh = string[index];

			if ((beforeCh && isWordChar(beforeCh)) || (afterCh && isWordChar(afterCh))) {
				test = isWordChar;
			} else if ((beforeCh && /\s/.test(beforeCh)) || (afterCh && /\s/.test(afterCh))) {
				test = isWhitespaceChar;
			} else {
				test = isNotWhiteSpaceAndNotWordChar;
			}

			i = index - 1;
			while (i >= 0 && test(string[i])) {
				i--;
				foundStart = i + 1;
			}

			i = index;
			while (i < length && test(string[i])) {
				i++;
				foundEnd = i;
			}

			return { location : foundStart, length : foundEnd - foundStart };
		};

		this.sentenceRange = function sentenceRange(node, index) {
			var string = node.line(),
				length = string.length;
			if (length === 0) {
				return { location : index, length : 0 };
			}

			var start = 0,
				end = length,
				skipsRegex = /(?:^\s*\d\.)|Mr\.|Mrs\.|Ms\.|Jr\.|Dr\.|Prof\.|Sr\./g,
				sentenceBoundaryRegex = /([.!?]|(?:[.!?][\'"]))(\s+|$)/g,
				match;

			skipsRegex.exec(string);

			while ((match = sentenceBoundaryRegex.exec(string))) {
				if (skipsRegex.lastIndex - 1 === match.index) {
					skipsRegex.exec(string);
				} else {
					end = match.index + match[0].length;
					if (end > index || (end === index && end === length)) {
						break;
					} else {
						start = end;
					}
				}
			}

			// Catch sentance at end that doesn't end with punctuation
			if (start === end && end < length) {
				end = length;
			}

			return { location : start, length : end - start };
		};

		//
		// Modification
		//

		this.setType = function setType(node, type) {
			return type;
		};

		this.typeLevelForType = function typeLevelForType(type) {
			return 0;
		};

		this.setTypeIndentLevel = function setTypeIndentLevel(node, level) {
			if (level >= 0) {
				return level;
			}
			return 0;
		};

		this.validatedTypeIndentLevel = function validatedTypeIndentLevel(node, level) {
			if (level < 0) {
				return 0;
			}
			return level;
		};

		this.setMode = function setMode(node, mode) {
			return mode;
		};

		this.removeTag = function removeTag(node, tagName) {
			return true;
		};

		this.addTag = function addTag(node, tagName, tagValue) {
			return true;
		};

		this.setPropertyNodeValue = function setPropertyNodeValue(node, propertyValue) {
			throw 'Unsupported';
		};

		this.setText = function setText(node, text) {
			var oldTextStart = node._textStart || 0,
				oldTextLength = node._text.length;
			if (oldTextLength === 0 && oldTextStart < node.line().length) {
				node.replaceLineInRange(text + ' ', oldTextStart, oldTextLength); // pad new text from what follows.
			} else {
				node.replaceLineInRange(text, oldTextStart, oldTextLength);
			}
			return text;
		};

		//
		// Classification
		//

		this.types = function types() {
			return {};
		};

		this.classify = function classify(text, state, previousState) {
			state.text = text.text();
			state.nextAction = Classifier.CLASSIFY_NEXT;
		};

		this.reclassifyUp = function reclassifyUp(text, state, nextState) {
			throw 'Unsupported';
		};

		this.reclassifyDown = function reclassifyDown(text, state, previousState) {
			throw 'Unsupported';
		};

		this.textAttributesToClearBeforeClassify = function textAttributesToClearBeforeClassify() {
			return [];
		};

		this.startState = function startState() {
			return {
				text: '',
				typeLevel: 0,
				typeIndentLevel: 0,
				tags: {},
				nextAction: Classifier.CLASSIFY_NEXT
			};
		};

		this.copyState = function copyState(state) {
			return {
				text: state.text,
				typeLevel: state.typeLevel,
				typeIndentLevel: state.typeIndentLevel,
				tags: ObjectUtils.shallowClone(state.tags),
				nextAction: state.nextAction
			};
		};

		this.equalStates = function equalStates(a, b) {
			return a.text === b.text &&
				a.typeLevel === b.typeLevel &&
				a.typeIndentLevel === b.typeIndentLevel &&
				ObjectUtils.shallowEquals(a.tags, b.tags) &&
				a.nextAction === b.nextAction;
		};

		//
		// Leads
		//

		var spaceRegex = /^\s*/;

		this.matchLeadForNode = function matchLeadForNode(node) {
			return spaceRegex.exec(node.line());
		};

		this.nextLeadFromNode = function nextLeadFromNode(node, after) {
			return this.matchLeadForNode(node)[0];
		};

		//
		// Rendering
		//

		this.styles = function styles() {
			return '';
		};

		this.renderNode = function renderNode(editor, node, nodeRenderer) {
		};

		this.renderNodeElementStyles = function renderNodeElementStyles(editor, node, elementRenderer) {
			node.renderNodeElementStyles(elementRenderer);
		};

		//
		// Export
		//

		this.rangeToHTML = function rangeToHTML(range) {
			var pre = document.createElement('pre');
			pre.textContent = range.textInRange();
			return pre.outerHTML;
		};

	}).call(Taxonomy.prototype);

	Taxonomy.isWordChar = isWordChar;
	Taxonomy.ROOT_STRUCTURE_TYPE_LEVEL = ROOT_STRUCTURE_TYPE_LEVEL;
	Taxonomy.ESCAPE_SYNTAX_CHAR = ESCAPE_SYNTAX_CHAR;

	Taxonomy.escapeText = function escapeText(text) {
		if (text.indexOf('\\') !== -1) {
			/*jshint regexp: false */
			text = text.replace(/\\./, ESCAPE_SYNTAX_CHAR + ESCAPE_SYNTAX_CHAR);
			/*jshint regexp: true */
		}
		return text;
	};

	exports.Taxonomy = Taxonomy;
});