//
// Copyright (c) 2014 Jesse Grosjean. All rights reserved.
//
// Permission is given for you to use this software ONLY while creating themes
// and plugins for the "FoldingText" software platform. Permission is NOT
// given for any other purpose.
//

/** @module ft/core/tree */
define(function (require, exports, module) {
	'use strict';

	var Node = require('./node').Node,
		Range = require('./range').Range,
		Delta = require('./delta').Delta,
		Taxonomies = require('./taxonomies'),
		NodeSet = require('./nodeset').NodeSet,
		Taxonomy = require('./taxonomy').Taxonomy,
		Classifier = require('./classifier').Classifier,
		Extensions = require('./extensions').Extensions,
		ClassifierText = require('./classifiertext').ClassifierText,
		TreeChangedEvent = require('./treechangedevent').TreeChangedEvent,

		oop = require('ft/util/oop'),
		assert = require('ft/util/assert'),
		UUID = require('ft/util/uuid').UUID,
		ObjectUtils = require('ft/util/objectutils'),
		Environment = require('ft/util/environment'),
		EventEmitter = require('ft/util/eventemitter').EventEmitter,
		queryParameter = require('ft/util/queryparameter').QueryParameter;

	/**
	 * Maps lines of text to {@link
	 * module:ft/core/node.Node|Nodes}.
	 *
	 * Allows you to efficiently access and manipulate the tree using Nodes or
	 * using text ranges. Changes can be grouped for efficiency using {@link
	 * module:ft/core/tree.Tree#beginUpdates|Tree#beginUpdates}
	 *
	 * Manages the {@tutorial Classification} process that determins node
	 * attributes and structure (parent, child, etc) based on the node's line
	 * of text. Use {@link
	 * module:ft/core/tree.Tree#ensureClassified|Tree#ensureClassified} to
	 * make sure the classification process is up to date.
	 *
	 * @private
	 * @constructor
	 * @alias module:ft/core/tree.Tree
	 */
	var Tree = function Tree(text, taxonomy) {
		// Init Fields
		this.id = UUID();
		this._lastNodeID = 0;
		this._updateCount = 0;
		this._RBRoot = null;
		this._tagsCache = null;
		this._parsedTagsCache = null;
		this.classifier = null;
		this.root = null;

		if (!taxonomy) {
			this.taxonomy = Taxonomies.taxonomy();
		} else if (typeof taxonomy === 'string') {
			this.taxonomy = Taxonomies.taxonomy({}, taxonomy);
		} else {
			this.taxonomy = taxonomy;
		}

		// Setup root
		var root = this.createNode('', true);
		root._type = 'root';
		root._typeLevel = Taxonomy.ROOT_STRUCTURE_TYPE_LEVEL;
		root._typeIndentLevel = 0;
		root._structureLevel = root._typeLevel + root._typeIndentLevel;

		this.root = root;
		this._deltaRegistrationDisabled = 0;
		this._deltas = null;
		this.beginUpdates();
		this.disableDeltaRegistration();
		this._nodeWillBeInsertedBeforeNode(root, null);
		this._nodeWasInserted(root);
		this.setText(text);
		this.enableDeltaRegistration();
		this.endUpdates();

		// Setup classifier
		this.classifier = new Classifier(this);
	};

	oop.implement(Tree.prototype, EventEmitter);

	//
	// Node Access
	//

	/**
	 * Return all nodes in tree (except root) in line order.
	 *
	 * @returns {Array.<module:ft/core/node.Node>}
	 * @methodgroup Nodes
	 */
	Tree.prototype.nodes = function nodes() {
		var results = [];
		var each = this.root.firstChild;
		while (each) {
			results.push(each);
			each = each.nextLineNode();
		}
		return results;
	};

	/**
	 * Return number of nodes in this tree.
	 *
	 * @returns {Number}
	 */
	Tree.prototype.nodeCount = function nodeCount() {
		if (!this._RBRoot) {
			return 0;
		}
		return this._RBRoot.totalLineCount();
	};

	/**
	 * Return the node associated with the given node ID in this tree.
	 *
	 * @param {String} - Node ID
	 * @returns {module:ft/core/node.Node}
	 */
	Tree.prototype.nodeForID = function nodeForID(id) {
		var each = this.root.nextLineNode();
		while (each) {
			if (each.id === id) {
				return each;
			}
			each = each.nextLineNode();
		}
		return undefined;
	};

	Tree.prototype.nodesForIDs = function nodesForIDs(ids) {
		var nodes = [];
		var length = ids.length;
		for (var i = 0; i < length; i++) {
			var node = this.nodeForID(ids[i]);
			if (node) {
				nodes.push(node);
			}
		}
		return nodes;
	};

	/**
	 * Return node representing the first line in this tree.
	 *
	 * @returns {module:ft/core/node.Node}
	 */
	Tree.prototype.firstLineNode = function firstLineNode() {
		return this.root.firstChild;
	};

	/**
	 * Return the node representing the last line in this tree.
	 *
	 * @returns {module:ft/core/node.Node}
	 */
	Tree.prototype.lastLineNode = function lastLineNode() {
		return this.root.lastDescendant();
	};

	Tree.prototype._nextNodeID = function _nextNodeID() {
		return '' + this._lastNodeID++;
	};

	var _createNodeClassifierText = new ClassifierText();

	/**
	 * Create a new node not yet inserted into the tree.
	 *
	 * @param {String} - Node's line of text.
	 * @param {Boolean=} - Normally when you create a node the classification
	 *     process will be run so that the node's structure information (type,
	 *     etc) is up-to-date. Use this flag to disable that default behavior.
	 * @returns {module:ft/core/node.Node}
	 */
	Tree.prototype.createNode = function createNode(line, dontPreclassifiy) {
		var node = new Node(line, this);

		if (!dontPreclassifiy) {
			// By default classify each node as it's created. So that it will
			// have property type level etc. And be easier to insert
			// correctly into the tree.
			var taxonomy = this.taxonomy,
				startState = taxonomy.startState();

			_createNodeClassifierText._setNode(node, null);
			taxonomy.classify(_createNodeClassifierText, startState, startState);
			node._updateFromTaxonomyState(startState);

			// Make sure to null taxonomyState, since that's the marker that
			// the classifier process uses to determine if a node is properly
			// classified or not. If we don't set to null... then classifier
			// will stop at this node since its state is correct. But this
			// state wasn't calculated taking surrounding node context into
			// consideration, so we null this to make sure that this node will
			// be fully reclassified after being inserted into the tree.
			node._taxonomyState = null;
		}

		return node;
	};

	Tree.prototype.createNodesFromLines = function createNodesFromLines(lines) {
		if (!Array.isArray(lines)) {
			lines = lines.split(Tree.NEW_LINE_REGEX);
		}

		var tree = new Tree(null, this.taxonomy),
			length = lines.length,
			i;

		tree.beginUpdates();
		for (i = 0; i < length; i++) {
			tree.insertNodeBefore(tree.createNode(lines[i], true), null);
		}
		tree.endUpdates();
		tree.ensureClassified();

		var nodes = tree.root.descendants(),
			children = tree.root.children();

		length = children.length;
		for (i = 0; i < length; i++) {
			this.adoptNode(children[i]);
		}

		tree.cleanupForCollection();

		return {'roots' : children, 'all' : nodes};
	};

	Tree.prototype.adoptNode = function adoptNode(node) {
		if (node.tree !== this) {
			node.removeFromParent();
			while (node) {
				node.tree = this;
				node.id = this._nextNodeID();
				node._taxonomyState = null; // see createNode, new tree needs to reclassify
				node = node.nextLineNode();
			}
		}
	};

	/**
	 * Evaluate the NodePath starting with this tree's root node and return
	 * all matches.
	 *
	 * @tutorial Node Paths
	 * @param {String} - Node path
	 * @returns {Array.<module:ft/core/node.Node>}
	 */
	Tree.prototype.evaluateNodePath = function evaluateNodePath(nodePath) {
		if (!nodePath) {
			return this.root.descendants();
		} else {
			return this.root.evaluateNodePath(nodePath);
		}
	};

	/**
	 * Return the line number of the given node in this tree.
	 *
	 * @param {module:ft/core/node.Node}
	 * @returns {Number}
	 */
	Tree.prototype.nodeToLineNumber = function nodeToLineNumber(node) {
		if (!node || !node.isTracked || node === this.root) {
			return -1;
		}

		var sz = (node._RBLeft) ? node._RBLeft.totalLineCount() : 0,
			curr = node;
		while (curr !== this._RBRoot) {
			if (curr._RBIsRightChild()) {
				sz += (curr._RBParent._RBLeft) ? curr._RBParent._RBLeft.totalLineCount() + 1 : 1;
			}
			curr = curr._RBParent;
		}
		return sz;
	};

	/**
	 * Return the node at the given line number.
	 *
	 * @param {Number}
	 * @returns {module:ft/core/node.Node}
	 */
	Tree.prototype.lineNumberToNode = function lineNumberToNode(line) {
		if (line < 0 || !this._RBRoot || line >= this._RBRoot.totalLineCount()) {
			return null;
		}
		return this._RBNodeAtSubtreeIndex(this._RBRoot, line);
	};

	Tree.prototype.lineAndCharForTextOffset = function lineAndOffsetForTextOffset(textOffset) {
		if (textOffset === 0) {
			return {line: 0, ch: 0};
		}
		var nodeOffset = this.textOffsetToNodeOffset(textOffset),
			line = this.nodeToLineNumber(nodeOffset.node);
		return {line: line, ch: nodeOffset.offset};
	};

	Tree.prototype.textOffsetForLineChar = function textOffsetForLineChar(line, char) {
		return this.nodeOffsetToTextOffset(this.lineNumberToNode(line), char);
	};

	/**
	 * Insert the given node (and its children) before the given reference
	 * node in this tree.
	 *
	 * @param {module:ft/core/node.Node}
	 * @param {module:ft/core/node.Node}
	 */
	Tree.prototype.insertNodeBefore = function insertNodeBefore(node, referenceNode) {
		var isDeltaRegistrationEnabled = this.isDeltaRegistrationEnabled();

		this.beginUpdates();

		node.removeFromParent();

		this.disableDeltaRegistration();

		var insertedNodes,
			insertingAtEnd = !referenceNode;

		if (isDeltaRegistrationEnabled) {
			insertedNodes = node.descendants();
			insertedNodes.splice(0, 0, node);
		}

		while (node.lastChild) {
			var lastChild = node.lastChild;
			this.insertNodeBefore(lastChild, referenceNode);
			referenceNode = lastChild;
		}

		// 1. Start values
		var nodeParent = referenceNode ? referenceNode.previousLineNodeOrRoot() : this.root.lastDescendantOrSelf(),
			nodeNextSibling = nodeParent.firstChild,
			nodeParentStructureLevel = nodeParent._structureLevel,
			nodeStructureLevel = node._structureLevel,
			nextBranch = referenceNode;

		// 2. Find parent for inserted node
		while (nodeParentStructureLevel >= nodeStructureLevel) {
			nodeNextSibling = nodeParent.nextSibling;
			nodeParent = nodeParent.parent;
			nodeParentStructureLevel = nodeParent._structureLevel;
		}

		// 3. Reparent trailing sub-branches to new node.
		if (nextBranch) {
			while (nextBranch && (nextBranch._structureLevel > nodeStructureLevel)) {
				if (nextBranch === nodeNextSibling) {
					nodeNextSibling = nextBranch.nextSibling;
				}
				var nextNextBranch = nextBranch.nextBranch();
				node.appendChild(nextBranch, true);
				nextBranch = nextNextBranch;
			}
		}

		// 4. Insert the node.
		nodeParent.insertChildBefore(node, nodeNextSibling, true);

		this.enableDeltaRegistration();

		if (isDeltaRegistrationEnabled) {
			var previous = node.previousLineNodeOrRoot(),
				firstNode = this.root.firstChild,
				lastNode = this.root.lastDescendantOrSelf(),
				insertedText = Node.formatNodes('txt', null, insertedNodes, false, false),
				fromLineCh,
				toLineCh,
				location;

			if (this.nodeCount() === 1) {
				location = 0;
			} else if (node === firstNode) {
				location = 0;
				insertedText = insertedText + '\n';
			} else if (insertingAtEnd) {
				location = this.nodeOffsetToTextOffset(previous, previous.line().length);
				insertedText = '\n' + insertedText;
			} else {
				location = this.nodeOffsetToTextOffset(previous, previous.line().length + 1);
				insertedText = insertedText + '\n';
			}

			fromLineCh = this.lineAndCharForTextOffset(location);
			toLineCh = fromLineCh;

			this.recordDelta(new Delta(location, '', insertedText, fromLineCh, toLineCh, null, null, insertedNodes));
		}

		this.validate();
		this.endUpdates();
	};

	/**
	 * Add the given node to the end of this tree's nodes.
	 *
	 * @param {module:ft/core/node.Node}
	 */
	Tree.prototype.appendNode = function appendNode(node) {
		this.insertNodeBefore(node, null);
	};

	/**
	 * Add the given nodes to the end of this tree's nodes.
	 *
	 * @param {Array.<module:ft/core/node.Node>}
	 */
	Tree.prototype.appendNodes = function appendNodes(nodes) {
		this.beginUpdates();
		var length = nodes.length;
		for (var i = 0; i < length; i++) {
			this.appendNode(nodes[i]);
		}
		this.endUpdates();
	};

	/**
	 * Insert the given nodes (and their children) before the given reference
	 * node in this tree.
	 *
	 * @param {Array.<module:ft/core/node.Node>}
	 * @param {module:ft/core/node.Node}
	 */
	Tree.prototype.insertNodesBefore = function insertNodesBefore(nodes, referenceNode) {
		var length = nodes.length;

		this.beginUpdates();

		for (var i = 0; i < length; i++) {
			this.insertNodeBefore(nodes[i], referenceNode);
		}

		this.endUpdates();
	};

	/**
	 * Remove the given node from this tree's nodes.
	 *
	 * @param {module:ft/core/node.Node}
	 */
	Tree.prototype.removeNode = function removeNode(node) {
		var isDeltaRegistrationEnabled = this.isDeltaRegistrationEnabled(),
			previousLineNode = node.previousLineNode(),
			nextLineNode = node.nextLineNode(),
			nextBranch = node.nextBranch(),
			fromLineCh,
			toLineCh,
			location;

		if (isDeltaRegistrationEnabled) {
			location = this.nodeOffsetToTextOffset(node, 0);
			fromLineCh = this.lineAndCharForTextOffset(location);
			toLineCh = {line: fromLineCh.line + 1, ch: 0};
		}

		this.beginUpdates();
		this.disableDeltaRegistration();

		node.removeFromParent();

		while (node.firstChild) {
			this.insertNodeBefore(node.firstChild, nextBranch);
		}

		this.enableDeltaRegistration();

		if (isDeltaRegistrationEnabled) {
			var line = node.line();
			if (nextLineNode) {
				this.recordDelta(new Delta(location, line + '\n', '', fromLineCh, toLineCh, null, [node], null));
			} else if (previousLineNode) {
				fromLineCh.line--;
				fromLineCh.ch = previousLineNode.line().length;
				toLineCh.line--;
				toLineCh.ch = node.line().length;
				this.recordDelta(new Delta(location - 1,  '\n' + line, '', fromLineCh, toLineCh, null, [node], null));
			} else {
				toLineCh.line--;
				toLineCh.ch = node.line().length;
				this.recordDelta(new Delta(location, line, '', fromLineCh, toLineCh, null, [node], null));
			}
		}

		this.endUpdates();
		this.validate();
	};

	/**
	 * Remove the given nodes from this tree's nodes.
	 *
	 * @param {Array.<module:ft/core/node.Node>}
	 */
	Tree.prototype.removeNodes = function removeNodes(nodes) {
		var length = nodes.length;

		this.beginUpdates();

		for (var i = 0; i < length; i++) {
			this.removeNode(nodes[i]);
		}

		this.endUpdates();
	};

	//
	// Text - Tree as a string.
	//

	/**
	 * Tree's entire text content.
	 *
	 * @returns {String}
	 * @methodgroup Text
	 */
	Tree.prototype.text = function text() {
		var textLength = this.textLength();
		if (textLength === 0) {
			return '';
		} else {
			return this.textInRange(this.createRangeFromLocation(0, textLength));
		}
	};

	/**
	 * Length of tree's text content.
	 *
	 * @returns {Number}
	 */
	Tree.prototype.textLength = function textLength() {
		if (!this._RBRoot) {
			return 0;
		}
		return this._RBRoot.totalLength() - 1;
	};

	/**
	 * Return tree's text content in the given range.
	 *
	 * @param {module:ft/core/range.Range}
	 * @returns {String}
	 */
	Tree.prototype.textInRange = function textInRange(range) {
		return this._nodesAndTextSegmentsInRange(range).lines.join('\n');
	};

	/**
	 * Replace tree's entire text content with new text.
	 *
	 * @param {String}
	 */
	Tree.prototype.setText = function setText(newText) {
		if (newText !== null) {
			return this.replaceTextInRange(this.createRangeFromLocation(0, this.textLength()), newText);
		} else {
			var isDeltaRegistrationEnabled = this.isDeltaRegistrationEnabled(),
				root = this.root,
				lastLineNode = this.lastLineNode(),
				removedNodes,
				fromLineCh,
				toLineCh,
				oldText;

			if (isDeltaRegistrationEnabled) {
				removedNodes = this.root.descendants();
				fromLineCh = {line: 0, ch: 0};
				toLineCh = {line: this.nodeCount() - 1, ch: lastLineNode ? lastLineNode.line().length : 0 };
				oldText = this.text();
			}

			this.beginUpdates();
			this.disableDeltaRegistration();
			while (root.firstChild) {
				root.firstChild.removeFromParent();
			}
			this.enableDeltaRegistration();
			if (isDeltaRegistrationEnabled) {
				this.recordDelta(new Delta(0, oldText, '', fromLineCh, toLineCh, null, removedNodes, null));
			}
			this.endUpdates();
		}
	};

	Tree.prototype._nodesAndTextSegmentsInRange = function _nodesAndTextSegmentsInRange(range) {
		if (!this.root.firstChild) {
			return { lines : [], nodes : [] };
		}

		var startNode = range.startNode,
			endNode = range.endNode,
			nodes = [ startNode ],
			lines = [];

		if (startNode === endNode) {
			var startLine = startNode.line();
			if (range.length() === startLine.length + 1) {
				lines.push(startLine + '\n');
			} else {
				lines.push(startLine.substr(range.startOffset, range.length()));
			}
		} else {
			var head = startNode.line().substr(range.startOffset);
			var tail = endNode.lineInRange(0, range.endOffset);
			var each = startNode.nextLineNode();

			lines.push(head);

			while (each !== endNode) {
				lines.push(each.line());
				nodes.push(each);
				each = each.nextLineNode();
			}

			lines.push(tail);
			nodes.push(endNode);
		}

		return { lines : lines, nodes : nodes };
	};

	/**
	 * Replace the tree's text content in the given range.
	 *
	 * @param {module:ft/core/range.Range}
	 * @param {String}
	 */
	Tree.prototype.replaceTextInRange = function replaceTextInRange(range, insertedText) {
		this.replaceTextInRanges([range], insertedText);
	};

	Tree.prototype.replaceCharactersInRange = function replaceCharactersInRange(location, length, insertedText) {
		this.replaceTextInRange(insertedText, this.createRangeFromLocation(location, length));
	};

	Tree.prototype.replaceTextInRanges = function replaceTextInRanges(ranges, insertedText) {
		this.beginUpdates();

		var length = ranges.length;
		for (var i = 0; i < length; i++) {
			var eachRange = ranges[i];
			if (i === 0) {
				this._replaceTextInRangeImplementation(eachRange, insertedText);
			} else {
				this._replaceTextInRangeImplementation(eachRange, '');
			}
		}

		this.endUpdates();
	};

	Tree.prototype._replaceTextInRangeImplementation = function _replaceTextInRangeImplementation(range, insertedText) {
		var isDeltaRegistrationEnabled = this.isDeltaRegistrationEnabled(),
			rangeLength = range.length(),
			rangeLocation = range.location(),
			fromLineCh = this.lineAndCharForTextOffset(rangeLocation),
			toLineCh = this.lineAndCharForTextOffset(rangeLocation + rangeLength),
			startNode = range.startNode,
			startLine = startNode.line(),
			startOffset = range.startOffset,
			effectedNodes = range.nodesInRange(),
			effectedNodesLength = effectedNodes.length,
			nodeAfterEndNode = range.nodeAfterEndNode(),
			oldLines = range.linesInRange(),
			oldLinesText = oldLines.join('\n'),
			newLinesText = oldLinesText.slice(0, startOffset) + insertedText + oldLinesText.slice(startOffset + rangeLength),
			newLines = newLinesText.split(Tree.NEW_LINE_REGEX),
			newLinesLength = newLines.length,
			updatedNode = null,
			insertedNodes = [],
			each,
			j;

		this.disableDeltaRegistration();

		// 1. Minimally update first affected node if it's changed
		updatedNode = startNode;
		if (newLines.length === 1 && effectedNodes.length === 1) {
			startNode.replaceLineInRange(insertedText, startOffset, range.endOffset - startOffset);
		} else {
			startNode.replaceLineInRange(newLines[0].substr(startOffset), startOffset, startLine.length - startOffset);
		}

		// 2. Remove all trailing lines in replace range. (reverse order faster in this case)
		for (j = effectedNodesLength - 1; j > 0; j--) {
			this.removeNode(effectedNodes[j]);
		}

		// 3. Insert new lines
		for (j = 1; j < newLinesLength; j++) {
			each = this.createNode(newLines[j], true);
			insertedNodes.push(each);
			this.insertNodeBefore(each, nodeAfterEndNode);
		}

		this.enableDeltaRegistration();

		if (isDeltaRegistrationEnabled) {
			effectedNodes.shift();
			this.recordDelta(new Delta(rangeLocation, oldLinesText.substr(startOffset, rangeLength), insertedText, fromLineCh, toLineCh, updatedNode, effectedNodes, insertedNodes));
		}
	};

	/**
	 * Append text to the end of this tree.
	 *
	 * @param {String} - Text to append.
	 */
	Tree.prototype.appendText = function appendText(text) {
		var range = this.createRangeFromLocation(this.textLength(), 0);
		this.replaceTextInRange(range, text);		
	};

	/**
	 * Returns the text offset in this tree's text for the given node, and
	 * offset into that node's line of text.
	 *
	 * @param {module:ft/core/node.Node}
	 * @param {Number} - Offset into node's line of text.
	 * @returns {Number}
	 */
	Tree.prototype.nodeOffsetToTextOffset = function nodeOffsetToTextOffset(node, offset) {
		if (offset === -1) {
			offset = node.line().length + 1;
			if (node === this.lastLineNode()) {
				offset--;
			}
		}
		return this._RBTextOffsetOfNode(node) + offset;
	};

	/**
	 * Returns an object with node and offset properties for the given text
	 * offset into the tree's text.
	 *
	 * @param {Number} - Offset into tree's text content.
	 * @returns {Object} - node, offset
	 */
	Tree.prototype.textOffsetToNodeOffset = function textOffsetToNodeOffset(textOffset) {
		var _RBRoot = this._RBRoot;
		if (!_RBRoot || (textOffset > _RBRoot.totalLength() || textOffset < 0)) {
			return null;
		}

		var node = this._RBNodeAtTextOffset(textOffset);
		return {
			node: node,
			offset: textOffset - this._RBTextOffsetOfNode(node)
		};
	};

	//
	// Node updates - Tell when tree is updating and track what has changed.
	//

	/**
	 * Return true if tree is currently updating.
	 *
	 * @methodgroup Grouping Changes
	 * @returns {Boolean}
	 */
	Tree.prototype.isUpdating = function isUpdating() {
		return this._updateCount !== 0;
	};

	/**
	 * Call before making changes to the tree. All changes that you make
	 * between calls to `beginUpdates` and `endUpdates` are grouped into a
	 * single
	 * module:ft/core/treechangedevent.TreeChangedEvent|TreeChangedEvent}.
	 */
	Tree.prototype.beginUpdates = function beginUpdates() {
		if (this._updateCount === 0) {
			this._deltas = [];
			this._previousParents = {};
		}
		this._updateCount++;
	};

	/**
	 * Call after you've finished making changes to the tree. All changes that
	 * you make between calls to `beginUpdates` and `endUpdates` are grouped
	 * into a single
	 * module:ft/core/treechangedevent.TreeChangedEvent|TreeChangedEvent}.
	 */
	Tree.prototype.endUpdates = function endUpdates() {
		this._updateCount--;
		if (this._updateCount === 0) {
			var previousParents = this._previousParents,
				deltas = this._deltas,
				length = deltas.length,
				eachDelta,
				eachNode,
				i, j;

			this._tagsCache = null;
			this._parsedTagsCache = null;
			this._deltas = undefined;
			this._previousParents = undefined;
			this.validate();

			if (length !== 0) {
				this._dispatchEvent('treeChanged', new TreeChangedEvent(this, deltas, previousParents));
			}
		}
	};

	Tree.prototype.isDeltaRegistrationEnabled = function isDeltaRegistrationEnabled() {
		return this._deltaRegistrationDisabled === 0;
	};

	Tree.prototype.disableDeltaRegistration = function disableDeltaRegistration() {
		this._deltaRegistrationDisabled++;
	};

	Tree.prototype.enableDeltaRegistration = function enableDeltaRegistration() {
		this._deltaRegistrationDisabled--;
	};

	Tree.prototype.recordDelta = function recordDelta(delta) {
		if (this.isDeltaRegistrationEnabled()) {
			this._deltas.push(delta);
			this.classifier._invalidateLinesForDelta(delta);
		}
	};

	Tree.prototype._nodeWillBeInsertedBeforeNode = function _nodeWillBeInsertedBeforeNode(node, referenceNode) {
		var isDeltaRegistrationEnabled = this.isDeltaRegistrationEnabled(),
			insertedNodes = isDeltaRegistrationEnabled ? [] : null,
			each = node;

		while (each) {
			if (insertedNodes) {
				insertedNodes.push(each);
			}
			each.isTracked = true;
			each = each.nextLineNode();
		}

		if (isDeltaRegistrationEnabled) {
			var insertedText = Node.formatNodes('txt', null, insertedNodes, false, false),
				replacedText = '',
				fromLineCh,
				toLineCh,
				location;

			if (referenceNode) {
				location = this.nodeOffsetToTextOffset(referenceNode, 0);
				insertedText = insertedText + '\n';
			} else {
				var insertAfter = this.root.lastDescendant();
				if (insertAfter) {
					location = this.nodeOffsetToTextOffset(insertAfter, -1);
					insertedText = '\n' + insertedText;
				} else {
					location = 0;
				}
			}

			fromLineCh = this.lineAndCharForTextOffset(location);
			toLineCh = fromLineCh;

			this.recordDelta(new Delta(location, replacedText, insertedText, fromLineCh, toLineCh, null, null, insertedNodes));
		}

		if (this.root !== node) {
			each = node;
			while (each) {
				this._RBInsertLineBefore(each, referenceNode);
				each = each.nextLineNode();
			}
		}
	};

	Tree.prototype._nodeWasInserted = function _nodeWasInserted(node) {
	};

	Tree.prototype._nodeWillBeRemoved = function _nodeWillBeRemoved(node) {
		var isDeltaRegistrationEnabled = this.isDeltaRegistrationEnabled(),
			removedNodes = isDeltaRegistrationEnabled ? [] : null,
			lastDescendantOrSelf = node.lastDescendantOrSelf(),
			each = node,
			fromLineCh,
			toLineCh,
			fromLocation,
			toLocation;

		if (isDeltaRegistrationEnabled) {
			fromLocation = this.nodeOffsetToTextOffset(node, 0);
			fromLineCh = this.lineAndCharForTextOffset(fromLocation);
			toLocation = this.nodeOffsetToTextOffset(lastDescendantOrSelf, lastDescendantOrSelf.line().length);
			toLineCh = this.lineAndCharForTextOffset(toLocation);
		}

		while (each) {
			if (removedNodes) {
				removedNodes.push(each);
			}
			this._RBRemoveLine(each);
			each.isTracked = false;
			if (each === lastDescendantOrSelf) {
				each = null;
			} else {
				each = each.nextLineNode();
			}
		}

		if (isDeltaRegistrationEnabled) {
			var lastNode = this.root.lastDescendantOrSelf(),
				lastRemovedNode = node.lastDescendantOrSelf(),
				replacedText = Node.formatNodes('txt', null, removedNodes, false, false),
				firstNodeRemoved = node === this.root.firstChild,
				lastNodeRemoved = lastNode === lastRemovedNode,
				insertedText = '';

			if (lastNodeRemoved) {
				if (!firstNodeRemoved) {
					replacedText = '\n' + replacedText;
					fromLineCh.line--;
					fromLineCh.ch = node.previousLineNode().line().length;
					fromLocation--;
				}
			} else {
				replacedText = replacedText + '\n';
				toLineCh.line++;
				toLineCh.ch = 0;
				toLocation++;
			}

			this.recordDelta(new Delta(fromLocation, replacedText, insertedText, fromLineCh, toLineCh, null, removedNodes, null));
		}
	};

	Tree.prototype._nodeLineWasUpdated = function _nodeLineWasUpdated(node, location, replacedText, insertedText) {
		if (this.isDeltaRegistrationEnabled()) {
			location = location || 0;

			var textStart = node.lineTextStart() + location,
				length = replacedText === undefined ? node.line().length : replacedText.length,
				fromLineCh = {line: node.lineNumber(), ch: location},
				toLineCh = {line: fromLineCh.line, ch: location + length};

			this.recordDelta(new Delta(textStart, replacedText, insertedText, fromLineCh, toLineCh, node, null, null));
		}
	};

	/**
	 * Create a range from a position in a start node and position in an end
	 * node.
	 *
	 * @param {module:ft/core/node.Node} - Range start node.
	 * @param {Number} - Offset into range start node, -1 to start at end of
	 *     line.
	 * @param {module:ft/core/node.Node} [endNode=startNode] - Range end node.
	 * @param {Number} [endOffset=startOffset] - Offset into range end node,
	 *     -1 to end at end of line.
	 * @param {Boolean} [anchoredAtEnd=false] - Is range anchored at end.
	 * @returns {module:ft/core/range.Range}
	 * @methodgroup Creating Ranges
	 */
	Tree.prototype.createRangeFromNodes = function createRangeFromNodes(startNode, startOffset, endNode, endOffset, anchoredAtEnd) {
		if (startOffset === -1) {
			startOffset = startNode.line().length;
		}
		if (endOffset === -1) {
			endOffset = endNode.line().length;
		}
		return new Range(this, startNode, startOffset, endNode, endOffset, anchoredAtEnd);
	};

	Tree.prototype.createRangeFromString = function createRangeFromString(rangeString, defaultLocation, defaultLength, anchoredAtEnd) {
		var rangeComponents = rangeString.split(',');
		if (defaultLocation === undefined) { defaultLocation = '0'; }
		if (defaultLength === undefined) { defaultLength = '0'; }
		return this.createRangeFromLocation(parseInt(rangeComponents[0] || defaultLocation, 10), parseInt(rangeComponents[1] || defaultLength, 10), anchoredAtEnd);
	};

	Tree.prototype.createRangeFromLineChs = function createRangeFromLineChs(startLineCh, endLineCh, anchoredAtEnd) {
		var startNode = this.lineNumberToNode(startLineCh.line),
			endNode;

		if (!endLineCh) {
			endLineCh = startLineCh;
		}

		if (startLineCh.line === endLineCh.line) {
			endNode = startNode;
		} else {
			endNode = this.lineNumberToNode(endLineCh.line);
			if (!endNode) {
				endNode = this.lastLineNode();
				endLineCh.ch = endNode.line().length;
			}
		}
		return this.createRangeFromNodes(startNode, startLineCh.ch, endNode, endLineCh.ch, anchoredAtEnd);
	};

	Tree.prototype.createRangeFromAnchoreAndHeadLineChs = function createRangeFromAnchoreAndHeadLineChs(anchor, head) {
		var headNode = this.lineNumberToNode(head.line),
			anchoredAtStart,
			anchorNode;

		if (head.line === anchor.line) {
			anchorNode = headNode;
		} else {
			anchorNode = this.lineNumberToNode(anchor.line);
			if (!anchorNode) {
				anchorNode = this.lastLineNode();
				anchor.ch = anchorNode.line().length;
			}
		}

		if (anchor.line < head.line || (anchor.line === head.line && anchor.ch < head.ch)) {
			return this.createRangeFromNodes(anchorNode, anchor.ch, headNode, head.ch, false);
		} else {
			return this.createRangeFromNodes(headNode, head.ch, anchorNode, anchor.ch, true);
		}
	};

	/**
	 * Create a range from a text offset location and length of characters to
	 * include.
	 *
	 * @param {Number} - Use -1 for end of document.
	 * @param {Number} - Use -1 for end of document.
	 * @param {Boolean} [anchoredAtEnd=false] - Is range anchored at end.
	 * @returns {module:ft/core/range.Range}
	 */
	Tree.prototype.createRangeFromLocation = function createRangeFromLocation(location, length, anchoredAtEnd) {
		if (!this.root.firstChild) {
			this.root.appendChild(this.createNode('', true));
		}

		if (location === undefined) {
			location = 0;
		}

		if (location === -1) {
			location = this.textLength();
		}

		if (length === undefined) {
			length = 0;
		}

		if (length === -1) {
			length = this.textLength() - location;
		}

		var startNodeOffset = this.textOffsetToNodeOffset(location),
			endNodeOffset = this.textOffsetToNodeOffset(location + length);

		if (startNodeOffset && endNodeOffset) {
			return new Range(this, startNodeOffset.node, startNodeOffset.offset, endNodeOffset.node, endNodeOffset.offset, anchoredAtEnd);
		} else {
			return null;
		}
	};

	//
	// Classification
	//

	/**
	 * Ensure that the classifier process has run.
	 *
	 * For example if you've replaced some text then you might need to call
	 * this if you immediately need to get right node types and other parsed
	 * information. If you don't call this method the classifier will run in
	 * background and eventually get everything classified, but maybe only
	 * after your code has run.
	 *
	 * @tutorial Classification
	 * @param {module:ft/core/node.Node=} - From node, if undefined then will start on first line in document.
	 * @param {module:ft/core/node.Node=} - To node, if undefined then will end on last line in document.
	 * @methodgroup Classification
	 */
	Tree.prototype.ensureClassified = function ensureClassified(fromNode, toNode) {
		this.classifier.ensureClassified(fromNode, toNode);
		this.validate();
	};


	//
	// Tags Access
	//

	/**
	 * Return list of the names of all tags used in this tree.
	 *
	 * @param {Boolean} [onlyIncludeParsedTags=false] - Only inclused parsed
	 *     `@` style tags.
	 * @returns {Array.<string>}
	 * @methodgroup Tags
	 */
	Tree.prototype.tags = function tags(onlyIncludeParsedTags) {
		var tagsCache = onlyIncludeParsedTags ? this._parsedTagsCache : this._tagsCache;
		if (!tagsCache) {
			var each = this.root.firstChild,
				treeTags = {},
				eachTags,
				eachKeys,
				eachKey,
				length,
				i;
			while (each) {
				eachTags = each.tags();
				if (eachTags) {
					var attributeRuns = each.lineAttributedString().attributeRuns(),
						attributeRunsLength = attributeRuns.length,
						eachTag,
						j;

					eachKeys = Object.keys(eachTags);
					length = eachKeys.length;
					for (i = 0; i < length; i++) {
						eachKey = eachKeys[i];
						if (onlyIncludeParsedTags) {
							for (j = 0; j < attributeRunsLength; j++) {
								if (attributeRuns[j].attributes.tagname === eachKey) {
									treeTags[eachKey] = eachKey;
									break;
								}
							}
						} else {
							treeTags[eachKey] = eachKey;
						}
					}
				}
				each = each.nextLineNode();
			}
			tagsCache = Object.keys(treeTags);

			if (onlyIncludeParsedTags) {
				this._parsedTagsCache = tagsCache;
			} else {
				this._tagsCache = tagsCache;
			}
		}
		return tagsCache;
	};


	//
	// RB Tree Interval Searches
	//

	Tree.prototype._RBNodeAtTextOffset = function _RBNodeAtTextOffset(offset) {
		if (offset === this._RBRoot.totalLength()) {
			return this._RBRoot._RBRightMost();
		}

		var node = this._RBRoot;
		while (node) {
			if (node._RBLeft !== null && offset < node._RBLeft.totalLength()) {
				node = node._RBLeft;
			} else {
				if (node._RBLeft !== null) {
					offset -= node._RBLeft.totalLength();
				}
				offset -= node.line().length + 1;
				if (offset < 0) {
					return node;
				}
				node = node._RBRight;
			}
		}
		return null;
	};

	Tree.prototype._RBTextOffsetOfNode = function _RBTextOffsetOfNode(node) {
		var offset = (node._RBLeft !== null) ? node._RBLeft.totalLength() : 0,
			RBParent;

		while ((RBParent = node._RBParent)) {
			if (node === RBParent._RBRight) {
				if (RBParent._RBLeft) {
					offset += RBParent._RBLeft.totalLength();
				}
				offset += RBParent.line().length + 1;
			}
			node = RBParent;
		}

		return offset;
	};

	Tree.prototype._RBNodeAtSubtreeIndex = function _RBNodeAtSubtreeIndex(current, index) {
		var sz = (current._RBLeft) ? current._RBLeft.totalLineCount() : 0;
		if (index === sz) {
			return current;
		} else if (index < sz) {
			return this._RBNodeAtSubtreeIndex(current._RBLeft, index);
		}
		return this._RBNodeAtSubtreeIndex(current._RBRight, index - (sz + 1));
	};

	Tree.prototype._RBNodeAtYOffset = function _RBNodeAtYOffset(offset) {
		if (offset <= 0) {
			return this.firstLineNode();
		} else {
			var textHeight = this.textHeight();
			if (offset >= textHeight) {
				return this.lastLineNode();
			}

			var node = this._RBRoot;
			while (node) {
				if (node._RBLeft !== null && offset < node._RBLeft.totalHeight()) {
					node = node._RBLeft;
				} else {
					if (node._RBLeft !== null) {
						offset -= node._RBLeft.totalHeight();
					}
					offset -= node.height();
					if (offset < 0) {
						return node;
					}
					node = node._RBRight;
				}
			}
			return null;
		}
	};

	Tree.prototype._RBYOffsetOfNode = function _RBYOffsetOfNode(node) {
		var offset = (node._RBLeft !== null) ? node._RBLeft.totalHeight() : 0;
		while (node._RBParent) {
			if (node === node._RBParent._RBRight) {
				if (node._RBParent._RBLeft) {
					offset += node._RBParent._RBLeft.totalHeight();
				}
				offset += node._RBParent.height();
			}
			node = node._RBParent;
		}
		return offset;
	};

	//
	// RB Updates
	//

	Tree.prototype._RBPreorderTraverse = function _RBPreorderTraverse(current, block) {
		if (current) {
			block(current);
			this._RBPreorderTraverse(current._RBLeft, block);
			this._RBPreorderTraverse(current._RBRight, block);
		}
	};

	Tree.prototype._RBInsertLineBefore = function _RBInsertLineBefore(node, referenceNode) {
		if (this._RBRoot === null) {
			this._RBRoot = node;
			return;
		}

		node._RBColor = Node.RED;
		if (referenceNode === null) {
			var last = this._RBRoot._RBRightMost();
			last._RBRight = node;
			node._RBParent = last;
			last._RBClearCache();
		} else if (referenceNode._RBLeft === null) {
			referenceNode._RBLeft = node;
			node._RBParent = referenceNode;
			referenceNode._RBClearCache();
		} else {
			var insertPoint = referenceNode._RBChildPredecessor();
			insertPoint._RBRight = node;
			node._RBParent = insertPoint;
			insertPoint._RBClearCache();
		}
		this._RBFixTreeOnInsert(node);
	};

	Tree.prototype._RBRemoveLine = function _RBRemoveLine(node) {
		if (!this._RBRoot._RBHasChildren()) {
			this._RBRoot = null;
			node._RBLeft = null;
			node._RBRight = null;
			node._RBParent = null;
			node._RBColor = Node.BLACK;
			return;
		}

		var y = (!node._RBLeft || !node._RBRight) ? node : node._RBChildSuccessor(),
			x = (y._RBLeft) ? y._RBLeft : y._RBRight;

		if (x) {
			x._RBParent = y._RBParent;
		}

		if (!y._RBParent) {
			// we're going to replace the root node
			this._RBRoot = x;
		} else {
			if (y._RBIsLeftChild()) {
				y._RBParent._RBSetLeft(x);
			} else {
				y._RBParent._RBSetRight(x);
			}
		}

		y._RBClearCache();

		if (y !== node) {
			// make y take the place of node
			if (node._RBLeft && node._RBLeft !== y) {
				y._RBLeft = node._RBLeft;
				y._RBLeft._RBParent = y;
			}

			if (node._RBRight && node._RBRight !== y) {
				y._RBRight = node._RBRight;
				y._RBRight._RBParent = y;
			}

			y._RBParent = node._RBParent;
			if (y._RBParent) {
				if (y._RBParent._RBLeft === node) {
					y._RBParent._RBLeft = y;
				} else {
					y._RBParent._RBRight = y;
				}
			} else {
				this._RBRoot = y;
			}

			y._RBClearCache();
		}

		if (y._RBParent) {
			y._RBParent._RBClearCache();
		}

		y._RBColor = node._RBColor;

		if (Node._RBColorOrBlack(y) === Node.BLACK && x) {
			this._RBFixTreeOnRemove(x);
		}

		node._RBLeft = null;
		node._RBRight = null;
		node._RBParent = null;
		node._RBColor = Node.BLACK;
	};

	Tree.prototype._RBFixTreeOnInsert = function _RBFixTreeOnInsert(node) {
		var BLACK = Node.BLACK,
			RED = Node.RED,
			y;

		while (node !== this._RBRoot && Node._RBColorOrBlack(node._RBParent) === RED) {
			var par = node._RBParent;
			if (par === par._RBParent._RBLeft) {
				y = par._RBParent._RBRight;
				if (Node._RBColorOrBlack(y) === RED) {
					par._RBColor = BLACK;
					y._RBColor = BLACK;
					par._RBParent._RBColor = RED;
					node = par._RBParent;
				} else {
					if (node === node._RBParent._RBRight) {
						node = par;
						node._RBLeftRotate(this);
					}
					par._RBColor = BLACK;
					par._RBParent._RBColor = RED;
					par._RBParent._RBRightRotate(this);
				}
			} else {
				y = par._RBParent._RBLeft;
				if (Node._RBColorOrBlack(y) === RED) {
					par._RBColor = BLACK;
					y._RBColor = BLACK;
					par._RBParent._RBColor = RED;
					node = par._RBParent;
				} else {
					if (node === node._RBParent._RBLeft) {
						node = par;
						node._RBRightRotate(this);
					}
					par._RBColor = BLACK;
					par._RBParent._RBColor = RED;
					par._RBParent._RBLeftRotate(this);
				}
			}
		}
		this._RBRoot._RBColor = BLACK;
	};

	Tree.prototype._RBFixTreeOnRemove = function _RBFixTreeOnRemove(x) {
		var BLACK = Node.BLACK,
			RED = Node.RED,
			w;

		while (x !== this._RBRoot && Node._RBColorOrBlack(x) === BLACK) {
			if (x._RBIsLeftChild()) {
				w = x._RBParent._RBRight;
				if (Node._RBColorOrBlack(w) === RED)  {
					w._RBColor = BLACK;
					x._RBParent._RBColor = RED;
					x._RBParent._RBLeftRotate(this);
					w = x._RBParent._RBRight;
				}
				if (w && Node._RBColorOrBlack(w._RBLeft) === BLACK &&
						Node._RBColorOrBlack(w._RBRight) === BLACK) {
					w._RBColor = RED;
					x = x._RBParent;
				} else {
					if (w && Node._RBColorOrBlack(w._RBRight) === BLACK) {
						if (w._RBLeft) {
							w._RBLeft._RBColor = BLACK;
						}
						w._RBColor = RED;
						w._RBRightRotate(this);
						w = x._RBParent._RBRight;
					}
					if (w) {
						w._RBColor = x._RBParent._RBColor;
					}
					x._RBParent._RBColor = BLACK;
					if (w && w._RBRight) {
						w._RBRight._RBColor = BLACK;
					}
					x._RBParent._RBRightRotate(this);
					x = this._RBRoot;
				}
			} else {
				w = x._RBParent._RBLeft;
				if (Node._RBColorOrBlack(w) === RED) {
					if (w) {
						w._RBColor = BLACK;
					}
					x._RBParent._RBColor = RED;
					x._RBParent._RBRightRotate(this);
					w = x._RBParent._RBLeft;
				}
				if (w && Node._RBColorOrBlack(w._RBRight) === BLACK &&
						Node._RBColorOrBlack(w._RBLeft) === BLACK) {
					w._RBColor = RED;
					x = x._RBParent;
				} else {
					if (w && Node._RBColorOrBlack(w._RBLeft) === BLACK) {
						if (w._RBRight) {
							w._RBRight._RBColor = BLACK;
						}
						w._RBColor = RED;
						w._RBLeftRotate(this);
						w = x._RBParent._RBLeft;
					}
					if (w) {
						w._RBColor = x._RBParent._RBColor;
					}
					x._RBParent._RBColor = BLACK;
					if (w && w._RBLeft) {
						w._RBLeft._RBColor = BLACK;
					}
					x._RBParent._RBLeftRotate(this);
					x = this._RBRoot;
				}
			}
		}
		x._RBColor = BLACK;
	};

	//
	// Cleanup
	//

	/**
	 * Used when writing specs. Removes scheduled timers and things like that.
	 * @methodgroup Util
	 */
	Tree.prototype.cleanupForCollection = function cleanupForCollection() {
		this.removeAllListeners('treeChanged');
		if (this.classifier) {
			this.classifier.cleanupForCollection();
		}
	};

	//
	// Debug
	//

	Tree.prototype.validate = function validate() {
		if (Environment.debug) {
			var each = this.firstLineNode(),
				nodeCount = 0,
				textLength = 0;

			if (!each) {
				textLength++;
			}

			while (each) {
				nodeCount++;
				textLength += each.line().length + 1;
				assert.ok(each._RBNextLineNode() === each.nextLineNode(), 'Next line traversal should be same in both node and RBTrees');
				assert.ok(each._RBPreviousLineNode() === each.previousLineNode(), 'Previous line traversal should be same in both node and RBTrees');
				assert.ok(each._structureLevel > each.parent._structureLevel, 'Structure level must be greater then parents structure level');
				each = each.nextLineNode();
			}

			textLength--;

			assert.ok(textLength === this.textLength(), 'Actual text length must match RBTree text length');
			assert.ok(nodeCount === this.nodeCount(), 'Actual lines length must match RBTree lines length');
		}
	};

	Tree.prototype.toString = function toString(rbTree) {
		var lines = [];

		if (rbTree) {
			var root = this._RBRoot;
			if (root) {
				lines.push('Binary Search Tree [lineCount: ' + root.totalLineCount() + ']');
				this._RBPreorderTraverse(root, function (each) {
					lines.push(each.toString(true));
				});
			}
		} else {
			lines.push('\n\u250C Tree');
			lines.push(this.root.branchToString(1));

			if (this._deltas && this._deltas.length > 0) {
				lines.push('Text Deltas [');
				var length = this._deltas.length,
					i;
				for (i = 0; i < length; i++) {
					lines.push(this._deltas[i].toString());
				}
				lines.push(']');
			}
		}

		return lines.join('\n');
	};

	Tree.NEW_LINE_REGEX = /\u000d(?:\u000a)?|\u000a|\u2028|\u2029|\u000c|\u0085/;

	Tree.stringByNormalizingLineEndings = function stringByNormalizingLineEndings(string) {
		if (string === null || string === undefined) {
			return string;
		}
		return string.split(Tree.NEW_LINE_REGEX).join('\n');
	};

	exports.Tree = Tree;
});