//
// Copyright (c) 2014 Jesse Grosjean. All rights reserved.
//
// Permission is given for you to use this software ONLY while creating themes
// and plugins for the "FoldingText" software platform. Permission is NOT
// given for any other purpose.
//

/** @module ft/core/treechangedevent */
define(function (require, exports, module) {
	'use strict';

	var NodeSet = require('./nodeset').NodeSet;

	/**
	 * Reports changes in a {@link
	 * module:ft/core/tree.Tree|Tree}.
	 * 
	 * The individual regions of change are
	 * stored in {@link module:ft/core/delta.Delta|Delta's}. This example
       shows how to use the `com.foldingtext.editor.treeChange` extension
       point to listen for TreeChangedEvents.
	 *
	 *	Extensions.addTreeChanged(function (editor, e) {
	 *		e.deltas.forEach(function (eachDelta) {
	 *			// process each delta
	 *		})
	 *	});
	 *
	 * @private
	 * @constructor
	 * @alias module:ft/core/treechangedevent.TreeChangedEvent
	 */
	var TreeChangedEvent = function TreeChangedEvent(tree, deltas, previousParents) {

		/**
		 * The tree where the changes took place.
		 * @type module:ft/core/tree.Tree
		 */
		this.tree = tree;

		/**
		 * The change deltas in the order in which they happened.
		 * @type module:ft/core/delta.Delta
		 */
		this.deltas = deltas;

		this.previousParents = previousParents;
	};

	/**
	 * Return this node's parent before the reported changes took place.
	 *
	 * @param {module:ft/core/node.Node}
	 * @returns {module:ft/core/node.Node}
	 */
	TreeChangedEvent.prototype.previousParent = function previousParent(node) {
		return this.previousParents[node.id];
	};

	TreeChangedEvent.prototype.movedNodes = function movedNodes() {
		var deltas = this.deltas,
			length = deltas.length,
			results = [];

		if (length === 1) {
			return results;
		}

		var allRemovedNodes = [],
			tree = this.tree,
			each,
			i;

		for (i = 1; i < length; i++) {
			allRemovedNodes.concat(deltas[i].removedNodes);
		}

		length = allRemovedNodes.length;
		for (i = 1; i < length; i++) {
			each = allRemovedNodes[i];
			if (each.tree === tree && each.isTracked) {
				results.push(each);
			}
		}

		return results;
	};

	exports.TreeChangedEvent = TreeChangedEvent;
});