//
// Copyright (c) 2014 Jesse Grosjean. All rights reserved.
//
// Permission is given for you to use this software ONLY while creating themes
// and plugins for the "FoldingText" software platform. Permission is NOT
// given for any other purpose.
//
define(function (require, exports, module) {
	'use strict';

	var CodeMirror = require('codemirror').CodeMirror;

	var Extensions = require('ft/core/extensions').Extensions,
		removeDiacritics = require('ft/editor/diacritics').removeDiacritics,
		killLocation = 0,
		killBuffer = '';

	Extensions.addCommand({
		name: 'moveForward',
		performCommand: function (editor) {
			return CodeMirror.commands.goCharRight(editor.cm());
		}
	});

	Extensions.addCommand({
		name: 'moveRight',
		performCommand: function (editor) {
			return CodeMirror.commands.goCharRight(editor.cm());
		}
	});

	Extensions.addCommand({
		name: 'moveBackward',
		performCommand: function (editor) {
			return CodeMirror.commands.goCharLeft(editor.cm());
		}
	});

	Extensions.addCommand({
		name: 'moveLeft',
		performCommand: function (editor) {
			return CodeMirror.commands.goCharLeft(editor.cm());
		}
	});

	Extensions.addCommand({
		name: 'moveUp',
		performCommand: function (editor) {
			var range = editor.selectedRange();
			if (!range.isCollapsed()) {
				editor.setSelectedRange(range.rangeByCollapsing(true));
			}
			return CodeMirror.commands.goLineUp(editor.cm());
		}
	});

	Extensions.addCommand({
		name: 'moveDown',
		performCommand: function (editor) {
			var range = editor.selectedRange();
			if (!range.isCollapsed()) {
				editor.setSelectedRange(range.rangeByCollapsing(false));
			}
			return CodeMirror.commands.goLineDown(editor.cm());
		}
	});

	Extensions.addCommand({
		name: 'moveWordForward',
		performCommand: function (editor) {
			return CodeMirror.commands.goWordRight(editor.cm());
		}
	});

	Extensions.addCommand({
		name: 'moveWordBackward',
		performCommand: function (editor) {
			return CodeMirror.commands.goWordLeft(editor.cm());
		}
	});

	Extensions.addCommand({
		name: 'moveToBeginningOfLine',
		performCommand: function (editor) {
			return CodeMirror.commands.goLineLeft(editor.cm());
		}
	});

	Extensions.addCommand({
		name: 'moveToEndOfLine',
		performCommand: function (editor) {
			return CodeMirror.commands.goLineRight(editor.cm());
		}
	});

	Extensions.addCommand({
		name: 'moveToBeginningOfParagraph',
		performCommand: function (editor) {
			return CodeMirror.commands.goLineStart(editor.cm());
		}
	});

	Extensions.addCommand({
		name: 'moveToEndOfParagraph',
		performCommand: function (editor) {
			return CodeMirror.commands.goLineEnd(editor.cm());
		}
	});

	Extensions.addCommand({
		name: 'moveToEndOfDocument',
		performCommand: function (editor) {
			return CodeMirror.commands.goDocEnd(editor.cm());
		}
	});

	Extensions.addCommand({
		name: 'moveToBeginningOfDocument',
		performCommand: function (editor) {
			return CodeMirror.commands.goDocStart(editor.cm());
		}
	});

	Extensions.addCommand({
		name: 'pageDown',
		performCommand: function (editor) {
			return CodeMirror.commands.goPageDown(editor.cm());
		}
	});

	Extensions.addCommand({
		name: 'pageUp',
		performCommand: function (editor) {
			return CodeMirror.commands.goPageUp(editor.cm());
		}
	});

	Extensions.addCommand({
		name: 'centerSelectionInVisibleArea',
		performCommand: function (editor) {
			var selectedRange = editor.selectedRange(),
				startCoords = editor.charCoords(selectedRange.startLineCh(), 'local'),
				endCoords = selectedRange.isCollapsed() ? startCoords : editor.charCoords(selectedRange.endLineCh(), 'local'),
				halfWindowHeight = editor.wrapperElement().offsetHeight / 2,
				top = startCoords.top,
				bottom = endCoords.bottom,
				halfSelectionHeight = (bottom - top) / 2,
				scrollTo = Math.round((top - halfWindowHeight) + halfSelectionHeight);
			editor.scrollTo(null, scrollTo);
		}
	});

	//
	//
	//

	function extendSelectionWithMovement(editor, movement) {
		var cm = editor.cm();
		cm.setExtending(true);
		movement();
		cm.setExtending(false);
	}

	Extensions.addCommand({
		name: 'moveBackwardAndModifySelection',
		performCommand: function (editor) {
			extendSelectionWithMovement(editor, function () {
				return CodeMirror.commands.goCharLeft(editor.cm());
			});
		}
	});

	Extensions.addCommand({
		name: 'moveForwardAndModifySelection',
		performCommand: function (editor) {
			extendSelectionWithMovement(editor, function () {
				return CodeMirror.commands.goCharRight(editor.cm());
			});
		}
	});

	Extensions.addCommand({
		name: 'moveWordForwardAndModifySelection',
		performCommand: function (editor) {
			extendSelectionWithMovement(editor, function () {
				return CodeMirror.commands.goWordRight(editor.cm());
			});
		}
	});

	Extensions.addCommand({
		name: 'moveWordBackwardAndModifySelection',
		performCommand: function (editor) {
			extendSelectionWithMovement(editor, function () {
				return CodeMirror.commands.goWordLeft(editor.cm());
			});
		}
	});

	Extensions.addCommand({
		name: 'moveUpAndModifySelection',
		performCommand: function (editor) {
			extendSelectionWithMovement(editor, function () {
				return CodeMirror.commands.goLineUp(editor.cm());
			});
		}
	});

	Extensions.addCommand({
		name: 'moveDownAndModifySelection',
		performCommand: function (editor) {
			extendSelectionWithMovement(editor, function () {
				return CodeMirror.commands.goLineDown(editor.cm());
			});
		}
	});

	//
	//
	//

	Extensions.addCommand({
		name: 'moveToBeginningOfLineAndModifySelection',
		performCommand: function (editor) {
			extendSelectionWithMovement(editor, function () {
				return CodeMirror.commands.goLineLeft(editor.cm());
			});
		}
	});

	Extensions.addCommand({
		name: 'moveToEndOfLineAndModifySelection',
		performCommand: function (editor) {
			extendSelectionWithMovement(editor, function () {
				return CodeMirror.commands.goLineRight(editor.cm());
			});
		}
	});

	Extensions.addCommand({
		name: 'moveToBeginningOfParagraphAndModifySelection',
		performCommand: function (editor) {
			extendSelectionWithMovement(editor, function () {
				return CodeMirror.commands.goLineStart(editor.cm());
			});
		}
	});

	Extensions.addCommand({
		name: 'moveToEndOfParagraphAndModifySelection',
		performCommand: function (editor) {
			extendSelectionWithMovement(editor, function () {
				return CodeMirror.commands.goLineEnd(editor.cm());
			});
		}
	});

	Extensions.addCommand({
		name: 'moveToEndOfDocumentAndModifySelection',
		performCommand: function (editor) {
			extendSelectionWithMovement(editor, function () {
				return CodeMirror.commands.goDocEnd(editor.cm());
			});
		}
	});

	Extensions.addCommand({
		name: 'moveToBeginningOfDocumentAndModifySelection',
		performCommand: function (editor) {
			extendSelectionWithMovement(editor, function () {
				return CodeMirror.commands.goDocStart(editor.cm());
			});
		}
	});

	Extensions.addCommand({
		name: 'pageDownAndModifySelection',
		performCommand: function (editor) {
			extendSelectionWithMovement(editor, function () {
				return CodeMirror.commands.goPageDown(editor.cm());
			});
		}
	});

	Extensions.addCommand({
		name: 'pageUpAndModifySelection',
		performCommand: function (editor) {
			extendSelectionWithMovement(editor, function () {
				return CodeMirror.commands.goPageUp(editor.cm());
			});
		}
	});

	Extensions.addCommand({
		name: 'moveParagraphForwardAndModifySelection',
		performCommand: function (editor) {
			extendSelectionWithMovement(editor, function () {
				CodeMirror.commands.goLineEnd(editor.cm());
				editor.performCommand('moveForward');
			});
		}
	});

	Extensions.addCommand({
		name: 'moveParagraphBackwardAndModifySelection',
		performCommand: function (editor) {
			extendSelectionWithMovement(editor, function () {
				editor.performCommand('moveBackward');
				CodeMirror.commands.goLineStart(editor.cm());
			});
		}
	});

	//
	//
	//

	Extensions.addCommand({
		name: 'moveWordLeft',
		performCommand: function (editor) {
			return CodeMirror.commands.goWordLeft(editor.cm());
		}
	});

	Extensions.addCommand({
		name: 'moveWordRight',
		performCommand: function (editor) {
			return CodeMirror.commands.goWordRight(editor.cm());
		}
	});

	Extensions.addCommand({
		name: 'moveRightAndModifySelection',
		performCommand: function (editor) {
			extendSelectionWithMovement(editor, function () {
				return CodeMirror.commands.goCharRight(editor.cm());
			});
		}
	});

	Extensions.addCommand({
		name: 'moveLeftAndModifySelection',
		performCommand: function (editor) {
			extendSelectionWithMovement(editor, function () {
				return CodeMirror.commands.goCharLeft(editor.cm());
			});
		}
	});

	Extensions.addCommand({
		name: 'moveWordRightAndModifySelection',
		performCommand: function (editor) {
			extendSelectionWithMovement(editor, function () {
				return CodeMirror.commands.goWordRight(editor.cm());
			});
		}
	});

	Extensions.addCommand({
		name: 'moveWordLeftAndModifySelection',
		performCommand: function (editor) {
			extendSelectionWithMovement(editor, function () {
				return CodeMirror.commands.goWordLeft(editor.cm());
			});
		}
	});

	//
	//
	//

	Extensions.addCommand({
		name: 'moveToLeftEndOfLine',
		performCommand: function (editor) {
			return CodeMirror.commands.goLineLeft(editor.cm());
		}
	});

	Extensions.addCommand({
		name: 'moveToRightEndOfLine',
		performCommand: function (editor) {
			return CodeMirror.commands.goLineRight(editor.cm());
		}
	});

	Extensions.addCommand({
		name: 'moveToLeftEndOfLineAndModifySelection',
		performCommand: function (editor) {
			extendSelectionWithMovement(editor, function () {
				return CodeMirror.commands.goLineLeft(editor.cm());
			});
		}
	});

	Extensions.addCommand({
		name: 'moveToRightEndOfLineAndModifySelection',
		performCommand: function (editor) {
			extendSelectionWithMovement(editor, function () {
				return CodeMirror.commands.goLineRight(editor.cm());
			});
		}
	});

	//
	//
	//

	function offsetScroll(editor, offset) {
		var cm = editor.cm(),
			scrollInfo = editor.scrollInfo();
		scrollInfo.top += offset;
		cm.scrollIntoView({
			left: 0,
			top: scrollInfo.top,
			right: 0,
			bottom: scrollInfo.top + scrollInfo.clientHeight
		});
	}

	Extensions.addCommand({
		name: 'scrollPageUp',
		performCommand: function (editor) {
			offsetScroll(editor, -editor.scrollInfo().clientHeight);
		}
	});

	Extensions.addCommand({
		name: 'scrollPageDown',
		performCommand: function (editor) {
			offsetScroll(editor, editor.scrollInfo().clientHeight);
		}
	});

	Extensions.addCommand({
		name: 'scrollLineUp',
		performCommand: function (editor) {
			offsetScroll(editor, editor.cm().defaultTextHeight());
		}
	});

	Extensions.addCommand({
		name: 'scrollLineDown',
		performCommand: function (editor) {
			offsetScroll(editor, editor.cm().defaultTextHeight());
		}
	});

	Extensions.addCommand({
		name: 'scrollToBeginningOfDocument',
		performCommand: function (editor) {
			editor.cm().scrollTo(0, 0);
		}
	});

	Extensions.addCommand({
		name: 'scrollToEndOfDocument',
		performCommand: function (editor) {
			editor.cm().scrollTo(0, Number.MAX_VALUE);
		}
	});

	//
	// Graphical Element transposition
	//

	Extensions.addCommand({
		name: 'transpose',
		performCommand: function (editor) {
			return CodeMirror.commands.transposeChars(editor.cm());
		}
	});

	Extensions.addCommand({
		name: 'transposeWords',
		performCommand: function (editor) {
			return true;
		}
	});

	//
	// Selections
	//

	Extensions.addCommand({
		name: 'selectAll',
		performCommand: function (editor) {
			var cm = editor.cm(),
				scrollInfo = editor.scrollInfo();
			CodeMirror.commands.selectAll(editor.cm());
			cm.scrollTo(0, scrollInfo.top);
		}
	});

	Extensions.addCommand({
		name: 'selectParagraph',
		performCommand: function (editor) {
			return editor.setSelectedRange(editor.selectedRange().rangeByExtendingToParagraph());
		}
	});

	Extensions.addCommand({
		name: 'selectLine',
		performCommand: function (editor) {
			return editor.setSelectedRange(editor.selectedRange().rangeByExtendingToLine());
		}
	});

	Extensions.addCommand({
		name: 'selectWord',
		performCommand: function (editor) {
			return editor.setSelectedRange(editor.selectedRange().rangeByExtendingToWord());
		}
	});

	//
	// Insertions and Indentations
	//

	Extensions.addCommand({
		name: 'indent',
		performCommand: function (editor) {
			return CodeMirror.commands.indentMore(editor.cm());
		}
	});

	Extensions.addCommand({
		name: 'insertTab',
		performCommand: function (editor) {
			return CodeMirror.commands.insertTab(editor.cm());
		}
	});

	Extensions.addCommand({
		name: 'insertBacktab',
		performCommand: function (editor) {
			return CodeMirror.commands.indentLess(editor.cm());
		}
	});

	Extensions.addCommand({
		name: 'insertNewline',
		performCommand: function (editor) {
			editor.replaceSelection('\n', 'end', '+input');
		}
	});

	Extensions.addCommand({
		name: 'insertNewlineBelow',
		performCommand: function (editor) {
			editor.performCommand('moveToEndOfParagraph');
			editor.performCommand('insertNewline');
		}
	});

	Extensions.addCommand({
		name: 'insertNewlineAbove',
		performCommand: function (editor) {
			editor.performCommand('moveToBeginningOfParagraph');
			editor.performCommand('insertNewline');
			editor.performCommand('moveLeft');
		}
	});

	Extensions.addCommand({
		name: 'insertParagraphSeparator',
		performCommand: function (editor) {
			editor.replaceSelection('\n\n', 'end', '+input');
		}
	});

	Extensions.addCommand({
		name: 'insertNewlineIgnoringFieldEditor',
		performCommand: function (editor) {
			editor.replaceSelection('\n', 'end', '+input');
		}
	});

	Extensions.addCommand({
		name: 'insertTabIgnoringFieldEditor',
		performCommand: function (editor) {
			return CodeMirror.commands.insertTab(editor.cm());
		}
	});

	Extensions.addCommand({
		name: 'insertLineBreak',
		performCommand: function (editor) {
		}
	});

	Extensions.addCommand({
		name: 'insertContainerBreak',
		performCommand: function (editor) {
		}
	});

	Extensions.addCommand({
		name: 'insertSingleQuoteIgnoringSubstitution',
		performCommand: function (editor) {
		}
	});

	Extensions.addCommand({
		name: 'insertDoubleQuoteIgnoringSubstitution',
		performCommand: function (editor) {
		}
	});

	//
	// Case changes
	//

	function transformWord(editor, t) {
		var selectedRange = editor.selectedRange(),
			selectedText = selectedRange.textInRange(),
			transformedText = t(selectedText);
		editor.replaceSelection(transformedText, 'around');
	}

	Extensions.addCommand({
		name: 'changeCaseOfLetter',
		performCommand: function (editor) {
		}
	});

	Extensions.addCommand({
		name: 'uppercaseWord',
		performCommand: function (editor) {
			editor.performCommand('selectWord');
			transformWord(editor, function (text) {
				return text.toUpperCase();
			});
		}
	});

	Extensions.addCommand({
		name: 'lowercaseWord',
		performCommand: function (editor) {
			editor.performCommand('selectWord');
			transformWord(editor, function (text) {
				return text.toLowerCase();
			});
		}
	});

	Extensions.addCommand({
		name: 'capitalizeWord',
		performCommand: function (editor) {
			editor.performCommand('selectWord');
			transformWord(editor, function (text) {
				return text.charAt(0).toUpperCase() + text.slice(1);
			});
		}
	});

	//
	// Deletions
	//

	Extensions.addCommand({
		name: 'deleteForward',
		performCommand: function (editor) {
			return CodeMirror.commands.delCharAfter(editor.cm());
		}
	});

	Extensions.addCommand({
		name: 'deleteBackward',
		performCommand: function (editor) {
			return CodeMirror.commands.delCharBefore(editor.cm());
		}
	});

	Extensions.addCommand({
		name: 'deleteBackwardByDecomposingPreviousCharacter',
		performCommand: function (editor) {
			var range = editor.selectedRange();
			if (range.isCollapsed()) {
				var location = range.location();
				if (location > 0) {
					range = range.rangeByOffsetting(-1, 1);
					var text = range.textInRange(),
						textWithoutDiacritics = removeDiacritics(text);
					if (text === textWithoutDiacritics) {
						editor.performCommand('deleteBackward');
					} else {
						editor.setSelectedRange(range);
						editor.replaceSelection(textWithoutDiacritics, 'end', '+delete');
					}
				}
			} else {
				editor.performCommand('deleteBackward');
			}
		}
	});

	Extensions.addCommand({
		name: 'deleteWordForward',
		performCommand: function (editor) {
			return CodeMirror.commands.delWordAfter(editor.cm());
		}
	});

	Extensions.addCommand({
		name: 'deleteWordBackward',
		performCommand: function (editor) {
			return CodeMirror.commands.delWordBefore(editor.cm());
		}
	});

	Extensions.addCommand({
		name: 'deleteToBeginningOfLine',
		performCommand: function (editor) {
			var range = editor.selectedRange();
			if (range.isCollapsed()) {
				editor.performCommand('moveToBeginningOfLineAndModifySelection');
			}
			editor.replaceSelection('', 'end', '+delete');
		}
	});

	Extensions.addCommand({
		name: 'deleteToEndOfLine',
		performCommand: function (editor) {
			var range = editor.selectedRange();
			if (range.isCollapsed()) {
				editor.performCommand('moveToEndOfLineAndModifySelection');
			}
			editor.replaceSelection('', 'end', '+delete');
		}
	});

	Extensions.addCommand({
		name: 'deleteToBeginningOfParagraph',
		performCommand: function (editor) {
			var range = editor.selectedRange();
			if (range.isCollapsed()) {
				editor.setSelectedRange(range.rangeByExtendingToBeginningOfParagraph());
			}
			editor.replaceSelection('', 'end', '+delete');
		}
	});

	Extensions.addCommand({
		name: 'deleteToEndOfParagraph',
		performCommand: function (editor) {
			var range = editor.selectedRange(),
				location = range.location();

			if (range.isCollapsed()) {
				range = range.rangeByExtendingToEndOfLine();
				if (range.isCollapsed()) {
					range = range.rangeByExtendingToEndOfParagraph();
				}
			}

			var killText = range.textInRange();
			if (killLocation === location) {
				killBuffer += killText;
			} else {
				killLocation = location;
				killBuffer = killText;
			}

			editor.setSelectedRange(range);
			editor.replaceSelection('', 'end', '+delete');
		}
	});

	Extensions.addCommand({
		name: 'yank',
		performCommand: function (editor) {
			editor.replaceSelection(killBuffer, 'end', '+input');
		}
	});

	//
	// Completion
	//

	Extensions.addCommand({
		name: 'complete',
		performCommand: function (editor) {
			if (editor.complete) {
				editor.complete();
			}
			return true;
		}
	});

	//
	// Mark/Point manipulation
	//

	Extensions.addCommand({
		name: 'setMark',
		performCommand: function (editor) {
			return true;
		}
	});

	Extensions.addCommand({
		name: 'deleteToMark',
		performCommand: function (editor) {
			return true;
		}
	});

	Extensions.addCommand({
		name: 'selectToMark',
		performCommand: function (editor) {
			return true;
		}
	});

	Extensions.addCommand({
		name: 'swapWithMark',
		performCommand: function (editor) {
			return true;
		}
	});

	//
	// Cancellation
	//

	Extensions.addCommand({
		name: 'cancelOperation',
		performCommand: function (editor) {
			editor.performCommand('complete');
			return true;
		}
	});
});