//
// Copyright (c) 2014 Jesse Grosjean. All rights reserved.
//
// Permission is given for you to use this software ONLY while creating themes
// and plugins for the "FoldingText" software platform. Permission is NOT
// given for any other purpose.
//

/*
 * Composition Mod for CodeMirror
 * v2.0.1
 * Zhusee <zhusee2@gmail.com>
 *
 * Additional instance properties added to CodeMirror:
 * - cm.display.inCompositionMode (Boolen)
 * - cm.display.compositionHead (Pos)
 * - cm.display.textMarkerInComposition (TextMarker)
 */

define(function (require, exports, module) {
	'use strict';

	var CodeMirror = require('codemirror').CodeMirror,
		TEXT_MARKER_CLASS_NAME = 'CodeMirror-text-in-composition',
		PREFIX_LIST = ['webkit', 'moz', 'o'],
		modInitialized = false,
		TEXT_MARKER_OPTIONS = {
			inclusiveLeft: true,
			inclusiveRight: true,
			className: TEXT_MARKER_CLASS_NAME
		};


	function capitalizeString(string) {
		return string.charAt(0).toUpperCase() + string.slice(1);
	}

	function getPrefixedPropertyName(propertyName) {
		var prefix, prefixedPropertyName, tempElem, _i, _len;
		tempElem = document.createElement('div');
		if (tempElem.style[propertyName] !== undefined) {
			return propertyName;
		}
		for (_i = 0, _len = PREFIX_LIST.length; _i < _len; _i++) {
			prefix = PREFIX_LIST[_i];
			prefixedPropertyName = prefix + capitalizeString(propertyName);
			if (tempElem.style[prefixedPropertyName] !== undefined) {
				return prefixedPropertyName;
			}
		}
		return false;
	}

	CodeMirror.defineOption('enableCompositionMod', false, function(cm, newVal, oldVal) {
		if (newVal && !modInitialized) {
			if (window.CompositionEvent) {
				return initCompositionMode(cm);
			} else {
				console.warn("Your browser doesn't support CompositionEvent.");
				return cm.setOption('enableCompositionMod', false);
			}
		}
	});

	CodeMirror.defineOption('debugCompositionMod', false, function(cm, newVal, oldVal) {
		var inputField;
		inputField = cm.display.input;
		if (typeof jQuery === "undefined" || jQuery === null) {
			return;
		}
		if (newVal) {
			$(inputField).on('input.composition-debug keypress.composition-debug compositionupdate.composition-debug', function(e) {
				return console.log("[" + e.type + "]", e.originalEvent.data, inputField.value, e.timeStamp);
			});
			return $(inputField).on('compositionstart.composition-debug compositionend.composition-debug', function(e) {
				return console.warn("[" + e.type + "]", e.originalEvent.data, inputField.value, cm.getCursor(), e.timeStamp);
			});
		} else {
			return $(inputField).off('.composition-debug');
		}
	});

	function setInputTranslate(cm, translateValue) {
		var transformProperty;
		transformProperty = getPrefixedPropertyName("transform");
		cm.display.input.style[transformProperty] = translateValue;
		return translateValue;
	}

	function resetInputTranslate(cm) {
		return setInputTranslate("");
	}

	function clearCompositionTextMarkers(cm) {
		var textMarker, textMarkersArray, _i, _len;
		textMarkersArray = cm.getAllMarks();
		for (_i = 0, _len = textMarkersArray.length; _i < _len; _i++) {
			textMarker = textMarkersArray[_i];
			if (textMarker && textMarker.className === TEXT_MARKER_CLASS_NAME) {
				textMarker.clear();
				if (cm.options.debugCompositionMod) {
					console.log("[TextMarker] Cleared");
				}
			}
		}
		return true;
	}

	function initCompositionMode(cm) {
		var inputField, inputWrapper, moveInputWithCursor;
		inputField = cm.display.input;
		inputWrapper = cm.display.inputDiv;
		moveInputWithCursor = cm.options.moveInputWithCursor;

		CodeMirror.on(inputField, 'compositionstart', function(event) {
			if (!cm.options.enableCompositionMod) {
				return;
			}
			
			cm.display.inCompositionMode = true;
			cm.setOption('readOnly', true);
			cm.setOption('moveInputWithCursor', true);

			if (cm.somethingSelected()) {
				cm.replaceSelection("");
			}
			cm.display.compositionHead = cm.getCursor();
			if (cm.options.debugCompositionMod) {
				console.log("[compositionstart] Update Composition Head", cm.display.compositionHead);
			}
			
			inputField.value = "";
			// should assing current line styling here

			if (cm.options.debugCompositionMod) {
				console.log("[compositionstart] Clear cm.display.input", cm.display.compositionHead);
			}
			inputWrapper.classList.add('CodeMirror-input-wrapper');
			return inputWrapper.classList.add('in-composition');
		});

		CodeMirror.on(inputField, 'compositionupdate', function(event) {
			var endPos, headPos, markerRange, pixelToTranslate;
			if (!cm.options.enableCompositionMod) {
				return;
			}
			headPos = cm.display.compositionHead;
			if (cm.display.textMarkerInComposition) {
				markerRange = cm.display.textMarkerInComposition.find();
				cm.replaceRange(event.data, headPos, markerRange.to);
				cm.display.textMarkerInComposition.clear();
				cm.display.textMarkerInComposition = void 0;
			} else {
				cm.replaceRange(event.data, headPos, headPos);
			}
			endPos = cm.getCursor();
			cm.display.textMarkerInComposition = cm.markText(headPos, endPos, TEXT_MARKER_OPTIONS);
			var headCoords = cm.charCoords(headPos, 'local');
			var endCoords = cm.charCoords(endPos, 'local');
			if (headCoords.top === endCoords.top) {
				pixelToTranslate = endCoords.left - headCoords.left;
			} else {
				var marginLeft = cm.getOption('ftEditor').defaultPREStyling().marginLeft;
				pixelToTranslate = (endCoords.left - marginLeft);
			}
			return setInputTranslate(cm, "translateX(-" + pixelToTranslate + "px)");
		});

		return CodeMirror.on(inputField, 'compositionend', function(event) {
			var endPos, headPos, postCompositionEnd, textLeftComposition, _ref;
			if (!cm.options.enableCompositionMod) {
				return;
			}
			textLeftComposition = event.data;
			headPos = cm.display.compositionHead;
			endPos = cm.getCursor();
			cm.replaceRange(textLeftComposition, headPos, endPos);
			cm.display.inCompositionMode = false;
			cm.display.compositionHead = void 0;
			if ((_ref = cm.display.textMarkerInComposition)) {
				_ref.clear();
			}
			cm.display.textMarkerInComposition = void 0;
			cm.setOption('readOnly', false);
			cm.setOption('moveInputWithCursor', moveInputWithCursor);

			inputWrapper.classList.remove('CodeMirror-input-wrapper');
			inputWrapper.classList.remove('in-composition');
			clearCompositionTextMarkers(cm);
			postCompositionEnd = function() {
				if (cm.display.inCompositionMode) {
					return false;
				}
				inputField.value = "";
				if (cm.options.debugCompositionMod) {
					console.warn("[postCompositionEnd] Input Cleared");
				}
				CodeMirror.off(inputField, 'input', postCompositionEnd);
				if (cm.options.debugCompositionMod) {
					return console.log("[postCompositionEnd] Handler unregistered for future input events");
				}
			};
			return CodeMirror.on(inputField, 'input', postCompositionEnd);
		});
	}
});
