//
// Copyright (c) 2014 Jesse Grosjean. All rights reserved.
//
// Permission is given for you to use this software ONLY while creating themes
// and plugins for the "FoldingText" software platform. Permission is NOT
// given for any other purpose.
//

/**
 * A location represented by a line number and a character offset into that
 * line.
 *
 * @typedef LineChar
 * @type {Object}
 * @property {Number} line - Zero based line number.
 * @property {Number} ch - Zero based character offset into the line.
 */

/**
 * A rectangle.
 *
 * @typedef Rect
 * @type {Object}
 * @property {Number} top
 * @property {Number} bottom
 * @property {Number} left
 * @property {Number} right
 * @property {Number} width
 * @property {Number} height
 */

/**
 * Key maps are ways to associate keys with functionality. A key map is an
 * object mapping strings that identify the keys to functions that implement
 * their functionality.
 *
 * This editor uses the CodeMirror [key
 * map](http://codemirror.net/doc/manual.html#keymaps) implementation with two
 * changes:
 *
 * 1. CodeMirror keymaps pass a CodeMirror instance as the first parameter to
 *    callback functions, while FoldingText key maps pass the FoldingText
 *    editor instance.
 *
 * 2. FoldingText keymaps will first try to perform any named command declared
 *    in the FoldingText editor. If no matching command is found then commands
 *    declared at the CodeMirror level will be tried.
 *
 * @typedef KeyMap
 * @type {Object}
 * @property {string=} name
 */

/**
 * Command declared in the editor using the 'com.foldingtext.editor.commands'
 * extension point.
 *
 * @typedef Command
 * @type {Object}
 * @property {String} name - Command's name.
 * @property {String} description - Command's description.
 */

/**
 * @typedef TargetResult
 * @type {Object}
 * @property {module:ft/core/node.Node} node
 * @property {LineChar} lineCh
 * @property {Object} attributes
 */

/** @module ft/editor/editor */
define(function (require, exports, module) {
	'use strict';

	require('ft/util/string');
	require('ft/editor/commands');
	require('ft/editor/keybindings');
	require('ft/editor/composition');
	require('ft/extensions/folding');
	require('ft/extensions/layout');
	require('ft/extensions/measuretext');


	var CodeMirror = require('codemirror').CodeMirror,
		AttributedString = require('ft/core/attributedstring').AttributedString,
		ElementRenderer = require('ft/editor/elementrenderer').ElementRenderer,
		queryParameter = require('ft/util/queryparameter').QueryParameter,
		NodeRenderer = require('ft/editor/noderenderer').NodeRenderer,
		Pasteboard = require('ft/system/pasteboard').Pasteboard,
		ObjectUtils = require('ft/util/objectutils'),
		Environment = require('ft/util/environment'),
		NodeSet = require('ft/core/nodeset').NodeSet,
		Extensions = require('ft/core/extensions').Extensions,
		Delay = require('../util/delay').Delay,
		Tree = require('ft/core/tree').Tree,
		assert = require('ft/util/assert'),
		dom = require('ft/util/dom'),
		oop = require('ft/util/oop');

	/**
	 * A text editor using a {@link module:ft/core/tree.Tree|Tree} to hold
	 * it's text and [CodeMirror](http://codemirror.net/) to display it.
	 *
	 * In some ways this editor tries to wrap the underlying CodeMirror
	 * instance. When there's similar functionality, such as get selection, I
	 * would encourage you to use this editor's methods. With that said,
	 * CodeMirror is considered a public part of this Editor's API. Use it as
	 * needed, especially if you are trying to reuse some existing CodeMirror
	 * code.
	 *
	 * @private
	 * @constructor
	 * @alias module:ft/editor/editor.Editor
	 */
	var Editor = function Editor(value, taxonomy, options) {
		Extensions.processExtensions('com.foldingtext.editor.methods', function (each) {
			for (var methodName in each) {
				Editor.prototype[methodName] = each[methodName];
			}
		});

		if (!options) {
			options = {};
		}

		var outerThis = this;

		CodeMirror.keyMap.empty = {};
		options.keyMap = 'empty';
		options.value = '';

		//options.styleActiveLine = true;
		options.mode = null;
		options.AttributedString = AttributedString;
		options.styleSelectedText = true;
		options.textForDropEvent = function (e) {
			var text = outerThis._droppedFilePaths,
				dataTransfer = e.dataTransfer;

			outerThis._droppedFilePaths = null;

			if (!text) {
				text = dataTransfer.getData('public.url');
				var name = dataTransfer.getData('public.url-name');
				if (name) {
					text = '[' + name + '](' + text + ')';
				}
			}

			if (!text) {
				text = dataTransfer.getData('text/uri-list');
			}

			if (text) {
				return text;
			}

			return null;
		};
		options.ftEditor = this;

		this.initialized = false;
		this._keyMaps = [];
		this._options = {};
		this._commands = {};
		this._modes = {};
		this._optionHandlers = {};
		this._reverseKeyMapLookup = null;
		this._tree = new Tree('', taxonomy);
		this._viewPortNodesChangedDelay = new Delay();

		var place = document.querySelector('.editor');
		if (!place) {
			place = document.createElement('div');
			place.className = 'editor';
			document.body.appendChild(place);
		}

		this._cm = new CodeMirror(place, options);
		this._viewportNodes = [];
		this._viewportNodesSet = new NodeSet();
		this._needsRenderNodes = new NodeSet();
		this._nodeRenderer = new NodeRenderer(this, this._cm);
		this._elementRenderer = new ElementRenderer(this, this._cm);
		this._setupSync();

//		this._textView = new CodeMirrorTextView(this, options);
		this._beginTimer();

		var extensions = queryParameter('extensions', '');
		if (extensions.length > 0) {
			require(extensions.split(';'), function () {
				outerThis._loadExtensionPoints(options, value);
			});
		} else {
			this._loadExtensionPoints(options, value);
		}

		window.isFoldingTextReady = function isFoldingTextReady() {
			// dummy method called by native so that it can determine when this code is loaded.
			return true;
		};

		window.textContent = function textContent() {
			return outerThis.textContent();
		};

		window.setTextContent = function setTextContent(textContent) {
			outerThis.setTextContent(textContent);
		};

		window.replaceCharactersInRangeWithString = function setTextContent(rangeJSON, string) {
			var tree = outerThis.tree(),
				parsedRange = JSON.parse(rangeJSON),
				range = tree.createRangeFromLocation(parsedRange.location, parsedRange.length);
			outerThis.replaceTextInRange(range, string);
		};

		window.scrollRangeToVisible = function scrollRangeToVisible(rangeJSON) {
			var tree = outerThis.tree(),
				parsedRange = JSON.parse(rangeJSON),
				range = tree.createRangeFromLocation(parsedRange.location, parsedRange.length);

			outerThis.scrollRangeToVisible(range);
		};

		window.visibleCharacterRange = function visibleCharacterRange() {
			var viewPort = outerThis.viewport(),
				from = {line: viewPort.from, ch: 0},
				to = {line: viewPort.to, ch: 0},
				tree = outerThis.tree(),
				range = tree.createRangeFromLineChs(from, to),
				rangeJSON = JSON.stringify({ location: range.location(), length: range.length() });
			return rangeJSON;
		};

		window.rectsForCharacterRange = function rectsForCharacterRange(rangeJSON) {
			var tree = outerThis.tree(),
				parsedRange = JSON.parse(rangeJSON),
				range = tree.createRangeFromLocation(parsedRange.location, parsedRange.length);
			return JSON.stringify(outerThis.rectsForCharacterRange(range));
		};

		window.suggestedNameFromContent = function suggestedNameFromContent() {
			return outerThis.suggestedNameFromContent();
		};

		window.restorableStateJSON = function restorableStateJSON() {
			return outerThis.restorableStateJSON();
		};

		window.setRestorableStateJSON = function setRestorableStateJSON(stateJSON, fileChanged) {
			return outerThis.setRestorableStateJSON(stateJSON, fileChanged);
		};

		window.getOption = function getOption(option) {
			return outerThis.option(option);
		};

		window.setOption = function setOption(name, value) {
			return outerThis.setOption(name, value);
		};

		window.refresh = function refresh() {
			outerThis.refresh();
		};

		window.performAction = function performAction(actionName, optionalTagValue) {
			outerThis.performAction(actionName, optionalTagValue);
		};

		window.validateMenuItem = function validateMenuItem(menuItem) {
			return outerThis.validateMenuItem(menuItem);
		};

		window.printMode = function printMode() {
			CodeMirror.runMode(outerThis.textContent(), 'foldingtext', document.getElementById('print'));
			var element = document.querySelector('.CodeMirror');
			if (element && element.parentNode) {
				element.parentNode.removeChild(element);
			}
		};

		window.printHeight = function printHeight() {
			return document.getElementById('print').clientHeight;
		};

		window.evaluateScript = function evaluateScript(script, options) {
			var result = { '_wrappedValue' : null };

			try {
				if (options) {
					options = JSON.parse(options)._wrappedValue;
				}

				var wrappedScript = '(' + script + ')';
				/*jshint evil:true */
				var f = eval(wrappedScript);
				/*jshint evil:false */
				var r = f(outerThis, options);
				if (r === undefined) {
					r = null; // survive JSON round trip
				}
				result._wrappedValue = r;
			} catch (e) {
				result._wrappedValue = e.toString() + '\n\tUse the Help > SDKRunner to debug';
			}

			return JSON.stringify(result);
		};
	};

	/**
	 * Return the CodeMirror instance associated with this editor.
	 *
	 * @returns {CodeMirror}
	 */
	Editor.prototype.cm = function cm() {
		return this._cm;
	};

	/**
	 * Return the Tree that holds this editor's text content.
	 *
	 * @returns {module:ft/core/tree.Tree}
	 */
	Editor.prototype.tree = function tree() {
		return this._tree;
	};

	/**
	 * Open the given link. When running natively this method will try to use
	 * native methods to open the line.
	 *
	 * @param {String} - URL text.
	 */
	Editor.prototype.openLink = function openLink(link, inBackground) {
		if (!link.match(/[a-z][\w\-]+:/)) {
			var c = link[0];
			if (c === '/') {
				link = 'file://' + encodeURI(link);
			} else if (c === '.' || c === '~') {
				if (window.nativeViewController) {
					/*jshint camelcase: false */
					window.nativeViewController.javascriptOpenRelativePath_inBackground_(encodeURI(link), inBackground);
					/*jshint camelcase: true */
					return;
				}
			} else {
				link = 'http://' + link;
			}
		}
		if (window.nativeViewController) {
			/*jshint camelcase: false */
			window.nativeViewController.javascriptOpenURL_inBackground_(link, inBackground);
			/*jshint camelcase: true */
		} else {
			window.open(link, inBackground ? '_blank' : '_self');
		}
	};

	Editor.prototype.operation = function operation(callback) {
		this.cm().operation(callback);
	};

	//
	// Options
	//

	/**
	 * Return the value of the given option.
	 *
	 * @param {String}
	 * @returns {Object}
	 * @methodgroup Options
	 */
	Editor.prototype.option = function option(name) {
		var value = this._options[name];
		if (value === undefined) {
			return this.cm().getOption(name);
		}
		return value;
	};

	/**
	 * Set a new value for the given option.
	 *
	 * @param {String}
	 * @param {Object}
	 */
	Editor.prototype.setOption = function setOption(name, value) {
		var handler = this._optionHandlers[name],
			options = this._options,
			old = options[name];

		if (old !== value) {
			options[name] = value;
			if (handler) {
				handler(this, old, value);
			}
		}
	};

	Extensions.addState({
		readState: function (editor, state) {
			var options = state.options;
			if (options !== undefined) {
				for (var name in options) {
					editor.setOption(name, options[name]);
				}
			}
		},
		writeState: function (editor, state) {
			var options = {};
			Extensions.processExtensions('com.foldingtext.editor.options', function (each) {
				var name = each.name,
					value = editor.option(name);
				if (value !== each.defaultValue) {
					options[name] = value;
				}
			});
			state.options = options;
		}
	});

	Extensions.add('com.foldingtext.editor.options', {
		name: 'hideSyntax',
		defaultValue: true,
		onChange: function (editor, oldValue, newValue) {
			var linesElement = editor.linesElement();
			if (oldValue) {
				dom.removeClass(linesElement, 'cm-hidesyntax');
			}
			if (newValue) {
				dom.addClass(linesElement, 'cm-hidesyntax');
			}
			editor.refresh();
		}
	});

	Extensions.addCommand({
		name: 'toggleHideSyntax',
		validate: function (editor, options) {
			options.interfaceItemState = editor.option('hideSyntax');
			return true;
		},
		performCommand: function (editor) {
			editor.setOption('hideSyntax', !editor.option('hideSyntax'));
		}
	});

	//
	// Text
	//

	/**
	 * Return the text content of this editor.
	 *
	 * @returns {String}
	 * @methodgroup Text Content
	 */
	Editor.prototype.textContent = function textContent() {
		return this.tree().text();
	};

	/**
	 * Replace the text content of this editor with new text.
	 *
	 * @param {String}
	 */
	Editor.prototype.setTextContent = function setTextContent(textContent) {
		this._cm.setValue(textContent);
	};

	/**
	 * Replace this editor's text in the specified range with new text.
	 *
	 * @param {module:ft/core/range.Range}
	 * @param {String}
	 */
	Editor.prototype.replaceTextInRange = function replaceTextInRange(range, insertedText, origin) {
		this.cm().replaceRange(insertedText, range.startLineCh(), range.endLineCh(), origin);
	};

	//
	// Selection
	//

	/**
	 * Return true if there is any selected text.
	 *
	 * @returns {Boolean}
	 * @methodgroup Selection
	 */
	Editor.prototype.hasSelection = function hasSelection() {
		return this._cm.somethingSelected();
	};

	/**
	 * Return all text selected.
	 *
	 * @returns {String}
	 */
	Editor.prototype.selectedText = function getSelectedText() {
		return this.selectedRange().textInRange();
	};

	/**
	 * Return the primary selected range.
	 *
	 * @returns {module:ft/core/range.Range}
	 */
	Editor.prototype.selectedRange = function selectedRange() {
		var cm = this._cm,
			anchore = cm.getCursor('anchor'),
			head = cm.getCursor('head');

		return this.tree().createRangeFromAnchoreAndHeadLineChs(anchore, head);
	};

	/**
	 * Return all selected ranges in this editor.
	 *
	 * @returns {Array.<module:ft/core/range.Range>}
	 */
	Editor.prototype.selectedRanges = function selectedRanges() {
		var tree = this.tree(),
			selections = this.cm().listSelections(),
			length = selections.length,
			ranges = [],
			i;

		for (i = 0; i < length; i++) {
			var selection = selections[i],
				anchor = selection.anchor,
				head = selection.head;

			ranges.push(tree.createRangeFromAnchoreAndHeadLineChs(anchor, head));
		}
		return ranges;
	};

	/**
	 * Replace the selection(s) with the given string. By default, the new
	 * selection ends up after the inserted text. The optional select argument
	 * can be used to change this — passing "around" will cause the new text to
	 * be selected, passing "start" will collapse the selection to the start
	 * of the inserted text.
	 *
	 * @param {String} - Replacement text.
	 * @param {String} [select="end"] start|around|end
	 */
	Editor.prototype.replaceSelection = function replaceSelection(text, select, origin) {
		this.cm().replaceSelection(text, select, origin);
	};

	Editor.prototype.replaceSelections = function replaceSelections(texts, select, origin) {
		texts = [].concat(texts);

		var selectedRanges = this.selectedRanges(),
			replacements = [],
			cm = this.cm();

		if (selectedRanges.length === texts.length) {
			replacements = texts;
		} else {
			selectedRanges.forEach(function (each) {
				replacements.push(texts.join('\n'));
			});
		}

		this.operation(function () {
			cm.replaceSelections(replacements, select, origin);
		});
	};

	/**
	 * Set selected range of text.
	 *
	 * @param {module:ft/core/range.Range}
	 */
	Editor.prototype.setSelectedRange = function setSelectedRange(range, options) {
		this.cm().setSelection(range.anchorLineCh(), range.headLineCh(), options);
	};

	Editor.prototype.setSelectedLineCh = function setSelectedLineCh(anchor, head, options) {
		this.cm().setSelection(anchor, head, options);
	};

	/**
	 * Set multiple selected ranges of text.
	 *
	 * @param {Array.<module:ft/core/range.Range>}
	 */
	Editor.prototype.setSelectedRanges = function setSelectedRanges(ranges, options) {
		var length = ranges.length,
			selections = [];
		for (var i = 0; i < length; i++) {
			selections.push({
				anchor: ranges[i].anchorLineCh(),
				head: ranges[i].headLineCh()
			});
		}
		this.cm().setSelections(selections, options);
	};

	Editor.prototype.cursorRange = function cursorRange(start) {
		var cm = this.cm(),
			pos = cm.getCursor(start || 'head');
		return this.tree().createRangeFromLineChs(pos, pos);
	};

	/**
	 * Give the editor keyboard focus.
	 * @methodgroup Focus
	 */
	Editor.prototype.focus = function focus() {
		this.cm().focus();
	};

	//
	// Commands
	//

	/**
	 * Returns all commands installed in this editor.
	 *
	 * @returns {Array.<Command>}
	 * @methodgroup Commands
	 */
	Editor.prototype.commands = function commands() {
		var reverseKeyMapLookup = this.reverseKeyMapLookup(),
			commandsList = this._commands,
			results = [],
			each;

		for (each in commandsList) {
			if (commandsList.hasOwnProperty(each)) {
				each = commandsList[each];
				results.push({
					name: each.name,
					description: each.description,
					shortcut: reverseKeyMapLookup[each.name]
				});
			}
		}

		return results;
	};

	/**
	 * Perform the given command.
	 *
	 * @param {String} - Command's name.
	 * @param {Object=} - Command options.
	 */
	Editor.prototype.performCommand = function performCommand(commandName, options) {
		if (window._gaq) {
			window._gaq.push(['_trackEvent', 'editor', 'performCommand', commandName]);
		}

		var outerThis = this;
		Extensions.processExtensions('com.foldingtext.editor.willPerformCommand', function (each) {
			each(outerThis, commandName);
		});

		var command = this._commands[commandName];
		if (command) {
			return command.performCommand(this, options);
		} else if (CodeMirror.commands[commandName]) {
			return CodeMirror.commands[commandName](this.cm(), options);
		} else {
			console.log('Command not found: ' + commandName);
		}
		return false;
	};

	Editor.prototype.validateCommand = function performCommand(commandName, options) {
		var command = this._commands[commandName];

		if (command) {
			if (command.validate) {
				return command.validate(this, options);
			}
		}
		return true;
	};

	Editor.prototype.performAction = function performAction(actionName, optionalTagValue) {
		this.performCommand(actionName, {interfaceItemTag: optionalTagValue});
	};

	var NSOffState = 0;
	var NSOnState = 1;

	Editor.prototype.validateMenuItem = function validateMenuItem(menuItem) {
		var actionName = menuItem.actionAsString(),
			command = this._commands[actionName];

		if (command) {
			if (command.validate) {
				var options = {interfaceItemTag: menuItem.tag()},
					result = command.validate(this, options);

				if (options.interfaceItemState !== undefined) {
					if (options.interfaceItemState) {
						/*jshint camelcase: false */
						menuItem.setState_(NSOnState);
						/*jshint camelcase: true */
					} else {
						/*jshint camelcase: false */
						menuItem.setState_(NSOffState);
						/*jshint camelcase: true */
					}
				}

				return result;
			}
		}
		return true;
	};

	//
	// Geometry
	//

	/**
	 * Return character coordinates for the given LineChar.
	 *
	 * @param {LineChar}
	 * @returns {Rect}
	 * @methodgroup Geometry
	 */
	Editor.prototype.charCoords = function charCoords(lineCh, mode) {
		return this.cm().charCoords(lineCh, mode);
	};

	/**
	 * Return cursor coordinates for the given LineChar.
	 *
	 * @param {LineChar}
	 * @returns {Rect}
	 */
	Editor.prototype.cursorCoords = function cursorCoords(lineCh, mode) {
		return this.cm().cursorCoords(lineCh, mode);
	};

	Editor.prototype.coordsChar = function coordsChar(left, top) {
		return this.cm().coordsChar({ left: left, top : top});
	};

	Editor.prototype.rectsForCharacterRange = function rectsForCharacterRange(range) {
		var pos = { line : 0, ch : 0 },
			outerThis = this,
			rects = [],
			lineEnd,
			i;

		range.forEachNodeInRange(function (node, line, location, length, fullySelected) {
			lineEnd = location + length;
			pos.line = line;

			for (i = location; i < lineEnd; i++) {
				pos.ch = i;
				rects.push(outerThis.charCoords(pos));
			}
		});

		return rects;
	};

	/**
	 * Return the node, text position, and text attributes at the given x,y
	 * position.
	 *
	 * @param {Number}
	 * @param {Number}
	 * @returns {TargetResult}
	 */
	Editor.prototype.targetsAt = function targetsAt(x, y) {
		var cm = this.cm(),
			pos = cm.coordsChar({left: x, top: y}),
			node = this.tree().lineNumberToNode(pos.line),
			rect = cm.charCoords(pos),
			attributes;

		if (rect.left > x && pos.ch > 0) {
			pos.ch--;
			rect = cm.charCoords(pos);
		}

		if (x >= rect.left && x <= rect.right && y >= rect.top && y <= rect.bottom) {
			attributes = node.lineAttributesAtIndex(pos.ch, null);
		}

		return {
			node: node,
			lineCh: pos,
			attributes: attributes || {}
		};
	};

	/**
	 * Scroll the given range to visible in the editor's viewport.
	 *
	 * @param {module:ft/core/range.Range}
	 * @methodgroup Scrolling
	 */
	Editor.prototype.scrollRangeToVisible = function scrollRangeToVisible(range, topMargin, bottomMargin) {
		this.cm().scrollIntoView({ from: range.startLineCh(), to: range.endLineCh() }, topMargin, bottomMargin);
	};

	/**
	 * Scroll the given rect to visible in the editor's viewport.
	 *
	 * @param {Rect}
	 */
	Editor.prototype.scrollRectToVisible = function scrollRectToVisible(rect) {
		this.cm().scrollIntoView(rect);
	};

	/**
	 * Scroll the editor to a given (pixel) position.
	 *
	 * @param {Number}
	 * @param {Number}
	 */
	Editor.prototype.scrollTo = function scrollTo(x, y) {
		this.cm().scrollTo(x, y);
	};

	Editor.prototype.scrollToTop = function scrollToTop() {
		this.cm().scrollTo(0, -Infinity);
	};

	/**
	 * Scroll the editor to a given line and optional pixel offset.
	 *
	 * @param {Number}
	 * @param {Number}
	 */
	Editor.prototype.scrollToLine = function scrollToLine(line, offset) {
		if (line < 0) line = 0;
		if (!offset) offset = 0;
		this.cm().scrollIntoView(line, offset, this.scrollInfo().clientHeight);
	};

	//
	// Viewport
	//

	Editor.prototype.viewport = function viewport() {
		var tree = this.tree(),
			vp = this._cm.getViewport(),
			numberOfLines = tree.nodeCount(),
			topLineChar = this.coordsChar(0, 0);

		while (topLineChar.line < (numberOfLines - 1) && this.lineIsHiddenInFold(topLineChar.line)) {
			topLineChar.line++;
		}

		vp.firstVisibleLine = topLineChar.line;
		vp.firstVisibleLineOffset = this.charCoords(topLineChar).top;

		return vp;
	};

	/**
	 * Return range covering visible editor viewport.
	 *
	 * @returns {module:ft/core/range.Range}
	 * @methodgroup Viewport
	 */
	Editor.prototype.viewportRange = function viewportRange() {
		var viewport = this.viewport(),
			range = this.tree().createRangeFromLineChs({ line: viewport.from }, { line: viewport.to });
		return range;
	};

	Editor.prototype.scrollInfo = function scrollInfo() {
		return this._cm.getScrollInfo();
	};

	Editor.prototype.setNeedsRender = function setNeedsRender(node) {
		assert.ok(this._cm.curOp !== null, 'Need to start operation before calling setNeedsRender');
		this._needsRenderNodes.addNode(node);
	};

	//
	// Keymaps
	//

	function _mapToFunctionWithEditorParameter(func, cm) {
		return func(cm.getOption('ftEditor'));
	}

	function _mapToExecuteEditorCommand(command, cm) {
		return cm.getOption('ftEditor').performCommand(command);
	}

	/**
	 * Attach a CodeMirror style [key
	 * map](http://codemirror.net/doc/manual.html#keymaps) to the editor.
	 *
	 * @param {KeyMap} - KeyMap to add.
	 * @param {Object} [bottom=false] - If true map will be added to bottom of
	 *     the list of keymaps and be procesed last.
	 * @methodgroup Keymaps	 
	 */
	Editor.prototype.addKeyMap = function addKeyMap(map, bottom) {
		// Editor maintains its own set of keymaps, primarily to implement
		// reverseKeyMapLookup. Real keymaps stored in CM, in wrapper
		// functions so that editor is first parameter to callback instead of
		// cm instance.

		var outerThis = this,
			isMac = CodeMirror.mac,
			cmMap = {},
			each,
			key;

		if (!isMac) {
			var newMap = {};
			for (key in map) {
				if (map.hasOwnProperty(key)) {
					var remappedKey = key;
					if (remappedKey.indexOf('Cmd') !== -1) {
						if (remappedKey.indexOf('Ctrl') !== -1) {
							remappedKey = remappedKey.replace('Ctrl', 'Shift');
						}
						remappedKey = remappedKey.replace('Cmd', 'Ctrl');
						newMap[remappedKey] = map[key];
					}
				}
			}
			map = newMap;
		}

		this._keyMaps[bottom ? 'push' : 'unshift'](map);
		this._reverseKeyMapLookup = null;

		for (key in map) {
			if (map.hasOwnProperty(key)) {
				each = map[key];
				if (key === 'name' || key === 'nofallthrough' || key === 'disableInput') {
					cmMap[key] = each;
				} else {
					if (typeof(each) === 'function') {
						cmMap[key] = _mapToFunctionWithEditorParameter.bind(outerThis, each);
					} else if (typeof(each) === 'string') {
						cmMap[key] = _mapToExecuteEditorCommand.bind(outerThis, each);
					}
				}
			}
		}

		this.cm().addKeyMap(cmMap, bottom);
	};

	/**
	 * Remove a keymap added with `addKeyMap`. Either pass in the key map
	 * object itself, or a string, which will be compared against the name
	 * property of the active key maps.
	 *
	 * @param {KeyMap} - KeyMap to remove.
	 */
	Editor.prototype.removeKeyMap = function removeKeyMap(map) {
		var keyMaps = this._keyMaps;
		for (var i = 0; i < keyMaps.length; ++i) {
			if ((typeof map === 'string' ? keyMaps[i].name : keyMaps[i]) === map) {
				keyMaps.splice(i, 1);
			}
		}
		this._reverseKeyMapLookup = null;
		this.cm().removeKeyMap(map);
	};

	Editor.prototype.reverseKeyMapLookup = function reverseKeyMapLookup() {
		function keySymbols(shortcut) {
			var components = shortcut.split('-'),
				result = components.lastObject() || '-';

			if (components.indexOf('Cmd') !== -1) result = '⌘' + result;
			if (components.indexOf('Shift') !== -1) result = '⇧' + result;
			if (components.indexOf('Alt') !== -1) result = '⌥' + result;
			if (components.indexOf('Ctrl') !== -1) result = '^' + result;
			
			return result;
		}

		var reverseMap = this._reverseKeyMapLookup;
		if (!reverseMap) {
			var commands = this._commands,
				keyMaps = this._keyMaps,
				length = keyMaps.length,
				i;

			reverseMap = {};
			for (i = 0; i < length; i++) {
				var map = keyMaps[i];
				for (var key in map) {
					var command = map[key];
					if (commands[command] && !reverseMap[command]) {
						reverseMap[command] = keySymbols(key);
					}
				}
			}

			this._reverseKeyMapLookup = reverseMap;
		}
		return reverseMap;
	};

	//
	// Restorable State
	//

	Editor.prototype.invalidateRestorableState = function invalidateRestorableState() {
		if (window.nativeViewController) {
			window.nativeViewController.javascriptInvalidateRestorableState();
		}
	};

	Editor.prototype.restorableStateJSON = function restorableStateJSON() {
		var state = {},
			outerThis = this;

		/**
		 * Extension point!
		 */
		Extensions.processExtensions('com.foldingtext.editor.state', function (each) {
			each.writeState(outerThis, state);
		});

		return JSON.stringify(state);
	};

	Editor.prototype.setRestorableStateJSON = function setRestorableStateJSON(stateJSON, fileChanged) {
		if (!stateJSON) {
			return;
		}

		try {
			var state = JSON.parse(stateJSON),
				outerThis = this;
			state.fileChangedSinceLastStateSync = fileChanged;
			Extensions.processExtensions('com.foldingtext.editor.state', function (each) {
				each.readState(outerThis, state);
			});
		} catch (e) {
			console.log(e);
		}
	};

	//
	//
	//

	Editor.prototype.beep = function beep() {
		if (window.nativeViewController) {
			window.nativeViewController.javascriptBeep();
		}
	};

	Editor.prototype.suggestedNameFromContent = function suggestedNameFromContent() {
		var each = this._tree.root.firstChild;
		var backup = null;
		while (each) {
			var eachText = each.text();
			if (eachText.length > 0) {
				if (each.type() === 'heading') {
					return eachText;
				} else if (!backup) {
					backup = eachText;
				}
			}
			each = each.nextSibling;
		}
		return backup;
	};

	/**
	 * Used when writing specs. Removes scheduled timers and things like that.
	 * @methodgroup Util
	 */
	Editor.prototype.removeAndCleanupForCollection = function removeAndCleanupForCollection() {
		if (this._timer) {
			clearTimeout(this._timer);
		}

		this._viewPortNodesChangedDelay.clear();

		var cm = this.cm(),
			wrapperElement = this.wrapperElement();

		if (cm.getPanel !== undefined) {
			cm.getPanel().remove();
		}

		if (wrapperElement.parentNode) {
			wrapperElement.parentNode.removeChild(wrapperElement);
		}

		this.tree().cleanupForCollection();
	};

	Editor.prototype._beginTimer = function _beginTimer() {
		var outerThis =  this,
			onTimeout = function onTimeout(e) {
				if (Extensions.hasExtensions('com.foldingtext.editor.clockTick')) {
					var date = new Date();
					outerThis.operation(function () {
						Extensions.processExtensions('com.foldingtext.editor.clockTick', function (each) {
							each(outerThis, date);
						});
					});
				}
				outerThis._timer = setTimeout(onTimeout, 1000);
			};
		this._timer = setTimeout(onTimeout, 1000);
	};

	Editor.prototype._loadExtensionPoints = function _loadExtensionPoints(overrideOptions, value) {
		var cm = this.cm(),
			outerThis = this,
			eachOverrideOption;

		this.operation(function () {
			Extensions.processExtensions('com.foldingtext.editor.options', function (each) {
				eachOverrideOption = overrideOptions[each.name];
				outerThis._optionHandlers[each.name] = each.onChange;
				outerThis.setOption(each.name, eachOverrideOption !== undefined ? eachOverrideOption : each.defaultValue);
			});
			Extensions.processExtensions('com.foldingtext.editor.commands', function (each) {
				outerThis._commands[each.name] = each;
			});
			Extensions.processExtensions('com.foldingtext.editor.init', function (each) {
				each(outerThis);
			});

			if (value && outerThis.tree().textLength() === 0) {
				outerThis.tree().setText(value);
				cm.setCursor(new CodeMirror.Pos(0, 0));
			}

			outerThis.initialized = true;
			outerThis.setRestorableStateJSON(queryParameter('defaultState', ''));

			Extensions.processExtensions('com.foldingtext.editor.selectionDidChange', function (each) {
				each(outerThis);
			});

			outerThis.refresh();
		});
	};

	Editor.prototype._setupSync = function _setupSync() {
		var tree = this.tree(),
			changeIsCommingFromFoldingText = 0,
			changeIsCommingFromCodeMirror = 0,
			nodeRenderer = this._nodeRenderer,
			elementRenderer = this._elementRenderer,
			outerThis = this,
			cm = this.cm();

		cm.on('contextmenu', function (cm, e) {
			//var range = outerThis.selectedRange();
			//if (range.isCollapsed()) {
			//	outerThis.performCommand('expandSelectionToWord');
			//}
			//e.preventDefault();
			//return !! Extensions.processExtensions('com.foldingtext.editor.nativecontextmenu', function (each) {
			//	return each(editor, e);
			//}, true);
		});

		var wrapper = this.wrapperElement(),
			lastMouseTimestamp = new Date().getTime();

		CodeMirror.on(wrapper, 'mouseenter', function (e) {
			Extensions.processExtensions('com.foldingtext.editor.mouseenter', function (each) {
				each(outerThis, e);
			});			
		});

		CodeMirror.on(wrapper, 'mouseover', function (e) {
			lastMouseTimestamp = e.timeStamp;
			Extensions.processExtensions('com.foldingtext.editor.mouseover', function (each) {
				each(outerThis, e);
			});			
		});

		CodeMirror.on(wrapper, 'mouseleave', function (e) {
			Extensions.processExtensions('com.foldingtext.editor.mouseleave', function (each) {
				each(outerThis, e);
			});			
		});

		cm.on('mousedown', function (cm, e) {
			lastMouseTimestamp = e.timeStamp;

			if (e.altKey) {
				return;
			}

			e.codemirrorIgnore = !! Extensions.processExtensions('com.foldingtext.editor.mouse', function (each) {
				if (each.mouseDown(outerThis, e)) {
					if (each.mouseUp) {
						var mouseUpHandler = function (e2) {
							CodeMirror.off(document, 'mouseup', mouseUpHandler);
							if (each.mouseUp(outerThis, e2)) {
								e2.preventDefault();
							}
						};
						e.preventDefault();
						CodeMirror.on(document, 'mouseup', mouseUpHandler);
					}
					return true;
				}
			}, true);
		});

		cm.on('contextmenu', function (cm, e) {
			//console.log(e);
		});


		cm.on('beforeSelectionChange', function (cm, change) {
			Extensions.processExtensions('com.foldingtext.editor.beforeSelectionChange', function (each) {
				each(outerThis, change);
			});
		});

		cm.on('cursorActivity', function () {
			var lastMouseAgo = new Date().getTime() - lastMouseTimestamp;

			if (lastMouseAgo > 100 && window.nativeViewController) {
				// For some reason cursor isn't hidden (only when embedding CM
				// in native app) when the user moves cursor with arrow keys.
				// This does work in normal web browser without requriing hack.
				/*jshint camelcase: false */
				window.nativeViewController.javascriptHideCursorUntilMouseMoves();
				/*jshint camelcase: true */
			}

		//	Extensions.processExtensions('com.foldingtext.editor.selectionDidChange', function (each) {
		//		each(outerThis);
		//	});
//				invalidateRestorableState();
		});

		cm.on('scroll', function () {
			/*if (window.nativeViewController) {
				window.nativeViewController.javascriptFindIndicatorNeedsUpdate();
			}
			Extensions.processExtensions('com.foldingtext.editor.viewportDidScroll', function (each) {
				each(outerThis);
			});*/
		});

		cm.on('beforeChange', function (doc, changeObj) {
			Extensions.processExtensions('com.foldingtext.editor.beforeChange', function (each) {
				each(outerThis, changeObj);
			});

			if (!changeObj.next && changeObj.origin === '+input') {
				Extensions.processExtensions('com.foldingtext.editor.insertText', function (each) {
					each(outerThis, changeObj);
				}, true);
			}

			if (!changeObj.update) {
				Extensions.processExtensions('com.foldingtext.editor.beforeUndoRedo', function (each) {
					each(outerThis, changeObj);
				});
			} else if (changeIsCommingFromCodeMirror) {
				//Extensions.processExtensions('com.foldingtext.editor.before', function (each) {
				//	each(outerThis, changeObj);
				//});				
			}

			// Make sure changes don't get combined if they happen at different locations.
			if (changeIsCommingFromFoldingText === 0 && cm.listSelections().length === 1) {
				var fromLineCh = changeObj.from,
					toLineCh = changeObj.to,
					lastChangeEnd = outerThis._lastChangeEnd,
					insertedText = changeObj.text.join('\n');

				if (lastChangeEnd && CodeMirror.cmpPos(lastChangeEnd, fromLineCh) !== 0 && CodeMirror.cmpPos(lastChangeEnd, toLineCh) !== 0) {
					doc.changeGeneration(true);
				}

				lastChangeEnd = CodeMirror.Pos(fromLineCh.line, fromLineCh.ch);
				lastChangeEnd.line += changeObj.text.length - 1;
				lastChangeEnd.ch += changeObj.text[changeObj.text.length - 1].length;
				outerThis._lastChangeEnd = lastChangeEnd;
			}			
		});

		cm.on('focus', function (cm) {
		});

		cm.on('blur', function (cm) {
		});

		cm.on('update', function (cm) {
		});

		cm.on('change', function (doc, changeObj) {
			if (Environment.debug) {
				assert.ok(tree.text() === cm.getValue(), 'Tree and CM must stay in sync');
			}
			Extensions.processExtensions('com.foldingtext.editor.change', function (each) {
				each(outerThis);
			});
		});

		cm.on('refresh', function (cm) {
			Extensions.processExtensions('com.foldingtext.editor.refresh', function (each) {
				each(outerThis);
			});
		});
		
		function scheduleCheckForViewportNodesChange() {
			outerThis._viewPortNodesChangedDelay.clear();
			outerThis._viewPortNodesChangedDelay.set(100, function () {
				var oldViewportNodes = outerThis._viewportNodes,
					oldViewportNodesSet = outerThis._viewportNodesSet,
					newViewportNodes = outerThis.visibleNodesInRange(outerThis.viewportRange());

				if (!oldViewportNodes.shallowEquals(newViewportNodes)) {
					var newViewportNodesSet = NodeSet.setFromArray(newViewportNodes),
						removedNodes = NodeSet.minusSets(oldViewportNodesSet, newViewportNodesSet),
						addedNodes = NodeSet.minusSets(newViewportNodesSet, oldViewportNodesSet);

					Extensions.processExtensions('com.foldingtext.editor.viewportNodesChanged', function (each) {
						each(outerThis, newViewportNodes, removedNodes, addedNodes);
					});

					outerThis._viewportNodes = newViewportNodes;
					outerThis._viewportNodesSet = newViewportNodesSet;
				}
			});			
		}


		cm.on('viewportChange', function (cm, from, to) {
			// willEndOperation can miss some cases because when it's called
			// viewport isn't updated yet. So need to call here too. For
			// normal typing this should'nt do anything since willEndOperation
			// has already ensureClassified but if lines are inserted/deleted
			// then this will clean things up
			var viewport = outerThis.viewport();
			tree.ensureClassified(
				tree.lineNumberToNode(viewport.from),
				tree.lineNumberToNode(viewport.to)
			);

			scheduleCheckForViewportNodesChange();
		});

		function applyCodeMirrorJustChangedToTree(changeObj) {
			if (window.nativeViewController) {
				/*jshint camelcase: false */
				window.nativeViewController.javascriptUpdateChangeCount_(0);
				/*jshint camelcase: true */
			}

			if (!changeIsCommingFromFoldingText) {
				changeIsCommingFromCodeMirror++;
				tree.beginUpdates();
				while (changeObj) {
					var from = changeObj.from,
						fromNode = tree.lineNumberToNode(from.line),
						to = changeObj.to,
						toNode = tree.lineNumberToNode(to.line),
						text = changeObj.text,
						range = tree.createRangeFromNodes(fromNode, from.ch, toNode, to.ch);

					tree.replaceTextInRange(range, text.join('\n'));
					changeObj = changeObj.next;
				}
				tree.endUpdates();
				changeIsCommingFromCodeMirror--;
			}
		}

		cm.on('justChanged', function (doc, changeObj) {
			applyCodeMirrorJustChangedToTree(changeObj);
		});

		cm.on('willEndOperation', function (cm) {
			// need to check that selection did actually change. Use this method instead of
			Extensions.processExtensions('com.foldingtext.editor.selectionDidChange', function (each) {
				each(outerThis);
			});

			var viewPort = outerThis.viewport();
			tree.ensureClassified(
				tree.lineNumberToNode(viewPort.from),
				tree.lineNumberToNode(viewPort.to)
			);

			scheduleCheckForViewportNodesChange();

			Extensions.processExtensions('com.foldingtext.editor.willEndOperation', function (each) {
				each(outerThis);
			});

			outerThis._needsRenderNodes.forEachNodeInSet(function (each) {
				if (each.isTracked) {
					// at this point cm and ft should be fully synced, so it
					// is safe to use node.lineNumber() to lookup a cm
					// lineHandle
					var lineNumber = each.lineNumber();
					nodeRenderer.renderNode(each, lineNumber, cm.getLineHandle(lineNumber));
				}
			});

			outerThis._needsRenderNodes.removeAllNodes();
		});

		tree.addEventListener('treeChanged', function (e) {
			changeIsCommingFromFoldingText++;

			outerThis.operation(function () {
				var deltas = e.deltas,
					length = deltas.length,
					eachDelta,
					i;

				for (i = 0; i < length; i++) {
					eachDelta = deltas[i];

					// Make sure CM text is in sync with changes from delta
					if (!changeIsCommingFromCodeMirror && eachDelta.textChanged()) {
						cm.replaceRange(eachDelta.insertedText, eachDelta.fromLineCh, eachDelta.toLineCh);
					}

					// Now that CM is in sync with latest delta render that
					// delta.
					nodeRenderer.renderDelta(eachDelta);
				}
			});

			changeIsCommingFromFoldingText--;

			outerThis.operation(function () {
				Extensions.processExtensions('com.foldingtext.editor.treeChange', function (each) {
					each(outerThis, e);
				});
			});
		});

		cm.setOption('externalMode', function (cm, line, stylesArray) {
			var lineNumber = cm.getLineNumber(line),
				node = tree.lineNumberToNode(lineNumber);			
			return node.lineAttributedString().toCodeMirrorStyleArray();
		});

		cm.on('renderLine', function (cm, line, element, builder, measure) {
			var lineNumber = cm.getLineNumber(line),
				node = tree.lineNumberToNode(lineNumber);

			//console.log('renderLine: ' + lineNumber);

			elementRenderer._node = node;
			elementRenderer._element = element;
			elementRenderer._lineHandle = line;
			elementRenderer._lineNumber = lineNumber;
			elementRenderer._cmBuilder = builder;
			elementRenderer._cmMeasure = measure;

			tree.taxonomy.renderNodeElementStyles(outerThis, node, elementRenderer);

			Extensions.processExtensions('com.foldingtext.editor.renderNodeElementStyles', function (each) {
				each(outerThis, node, elementRenderer);
			});
		});
	};

	Extensions.addCommand({
		name: 'copyAsHTML',
		performCommand: function (editor) {
			var taxonomy = editor.tree().taxonomy,
				range = editor.selectedRange();
			Pasteboard.writeString(taxonomy.rangeToHTML(range));
		}
	});

	Extensions.addCommand({
		name: 'focus',
		performCommand: function (editor) {
			editor.focus();
		}
	});

	exports.Editor = Editor;
});