//
// Copyright (c) 2014 Jesse Grosjean. All rights reserved.
//
// Permission is given for you to use this software ONLY while creating themes
// and plugins for the "FoldingText" software platform. Permission is NOT
// given for any other purpose.
//

/** @module ft/editor/elementrenderer */
define(function (require, exports, module) {
	'use strict';

	/**
	 * Set styling on the DOM elements used to display {@link
	 * module:ft/core/node.Node|Nodes}.
	 *
	 * An instance of this class is passed into the
	 * {@link module:ft/core/extensions.Extensions.addRenderNodeElementStyles|addRenderNodeElementStyles}
	 * extension point callback. Extensions then use this class to set
	 * styling on the DOM element that's being used to display the given Node.
	 *
	 * This class differs from the NodeRenderer because it's used right before
	 * the DOM element is displayed on screen. 
	 *
	 * @private
	 * @constructor
	 * @alias module:ft/editor/elementrenderer.ElementRenderer
	 */
	var ElementRenderer = function ElementRenderer(editor, cm) {
		this.editor = editor;
		this.cm = cm;
		this._cmBuilder = null;
		this._cmMeasure = null;
		this._node = null;
		this._element = null;
		this._lineHandle = null;
		this._lineNumber = null;
	};

	/**
	 * Set DOM Element style property.
	 *
	 * @param {String} - style name.
	 * @param {String} - style value.
	 */
	ElementRenderer.prototype.setStyle = function setStyle(name, value) {
		this._element.style[name] = value;
	};

	/**
	 * Set a DOM Element attribute.
	 *
	 * @param {String} - property name.
	 * @param {String} - property value.
	 */
	ElementRenderer.prototype.setAttribute = function setAttribute(name, value) {
		try {
			this._element.setAttribute(name, value);
		} catch (e) {
			console.log('Invalid DOM Node attribute name: ' + name + ' or value: ' + value);
		}
	};

	/**
	 * Class that will be used to render this "text" element of the node's line DOM.
	 *
	 * @returns {String}
	 * @methodgroup Classes
	 */
	ElementRenderer.prototype.textClass = function textClass() {
		return this._cmBuilder.textClass;
	};

	/**
	 * Add a new class to the "text" element of the node's line DOM.
	 *
	 * @param {String} - class name.
	 */
	ElementRenderer.prototype.addTextClass = function addTextClass(clss) {
		if (this._cmBuilder.textClass) {
			this._cmBuilder.textClass += ' ' + clss;
		} else {
			this._cmBuilder.textClass = clss;
		}
	};

	/**
	 * Class that will be used to render this "background" element of the
	 * node's line DOM.
	 *
	 * @returns {String}
	 */
	ElementRenderer.prototype.backgroundClass = function backgroundClass() {
		return this._cmBuilder.bgClass;
	};

	/**
	 * Add a new class to the "background" element of the node's line DOM.
	 *
	 * @param {String} - class name.
	 */
	ElementRenderer.prototype.addBackgroundClass = function addBackgroundClass(clss) {
		if (this._cmBuilder.bgClass) {
			this._cmBuilder.bgClass += ' ' + clss;
		} else {
			this._cmBuilder.bgClass = clss;
		}
	};

	/**
	 * Class that will be used to render this "wrap" element of the node's
	 * line DOM.
	 *
	 * @returns {String}
	 */
	ElementRenderer.prototype.wrapClass = function wrapClass() {
		return this._cmBuilder.wrapClass;
	};

	/**
	 * Add a new class to the "wrap" element of the node's line DOM.
	 *
	 * @param {String} - class name.
	 */
	ElementRenderer.prototype.addWrapClass = function addWrapClass(clss) {
		if (this._cmBuilder.wrapClass) {
			this._cmBuilder.wrapClass += ' ' + clss;
		} else {
			this._cmBuilder.wrapClass = clss;
		}
	};

	/**
	 * Measure a range of text in the rendered node. This is used most often
	 * when setting margins, making sure that everything lines up no matter
	 * what styles are applied, etc.
	 *
	 * @param {Number} - location in node's text
	 * @param {Number} - length to measure
	 * @returns {Rect}
	 * @methodgroup Geometry
	 */
	ElementRenderer.prototype.measureTextRange = function measureTextRange(location, length) {
		if (this._cmBuilder.textClass !== this._element.className) {
			this._element.className = this._cmBuilder.textClass;
		}
		return this.editor.measureNodeText(this._node, location, length, this._cmBuilder, this._cmMeasure);
	};

	exports.ElementRenderer = ElementRenderer;
});