//
// Copyright (c) 2014 Jesse Grosjean. All rights reserved.
//
// Permission is given for you to use this software ONLY while creating themes
// and plugins for the "FoldingText" software platform. Permission is NOT
// given for any other purpose.
//

/**
 * Options control how line widgets get displayed.
 *
 * The options are the same as the options describe in CodeMirror's
 * [addLineWidget](http://codemirror.net/doc/manual.html#addLineWidget)
 * feature with two additions:
 *
 * 1. `overlay` - A boolean, if given a value of true then the widget will
 *    draw over the node's line instead of getting inserted before or after
 *    it.
 *
 * 2. `positionWidget` - A function that takes widgets calculated width and
 *    heigh as parameters and returns and object with (all optional) top,
 *    bottom, left, and right style properties. This function gives the
 *    widget renderer the opportunity to precisely position the widget after
 *    the node's text has been inserted into the node's DOM element.
 *
 * @typedef RenderLineWidgetOptions
 * @type {Object}
 */

/** @module ft/editor/noderenderer */
define(function (require, exports, module) {
	'use strict';

	var CodeMirror = require('codemirror').CodeMirror,
		Extensions = require('ft/core/extensions').Extensions,
		Environment = require('ft/util/environment'),
		assert = require('ft/util/assert');

	/**
	 * Add DOM widgets to {@link module:ft/core/node.Node|Nodes}.
	 *
	 * An instance of this class is passed into the {@link
	 * module:ft/core/extensions.Extensions.addRenderNode|addRenderNode}
	 * extension point callback. Extensions can use it to add DOM widgets that
	 * will be displayed when the node displays.
	 *
	 * For example `todo` mode uses this class to add checkbox widgets to tasks.
	 * And `timer` mode uses this class to add widgets that show when things
	 * will be done.
	 *
	 * @private
	 * @constructor
	 * @alias module:ft/editor/noderenderer.NodeRenderer
	 */
	var NodeRenderer = function NodeRenderer(editor, cm) {
		this.editor = editor;
		this.cm = cm;
		this._node = null;
		this._cmLineNumber = null;
		this._cmLineHandle = null;
	};

	NodeRenderer.prototype._clearRenderMarks = function _clearRenderMarks(nodes) {
		var length = nodes.length,
			i;
		for (i = 0; i < length; i++) {
			var each = nodes[i],
				renderMarks = each._renderMarks,
				markersLength = renderMarks ? renderMarks.length : 0,
				j;
			for (j = 0; j < markersLength; j++) {
				renderMarks[j].clear();
			}
			each._renderMarks = [];
		}
	};

	NodeRenderer.prototype.renderDelta = function renderDelta(delta) {
		// Deltas are applied to codemirror one at a time. That means that in
		// this code it's not safe to assume that CodeMirror and FoldingText
		// are fully in sync. For example the two systems might have different
		// text content and things like node.lineNumber() can't translate into
		// a CM line number. You can assume that the delta lineCh numbers are
		// in sync when this method is called.

		var removedNodes = delta.removedNodes,
			updatedNode = delta.updatedNode,
			insertedNodes = delta.insertedNodes,
			renderNodes = insertedNodes.slice(0),
			cmStartLine = delta.fromLineCh.line,
			outerThis = this,
			cm = this.cm,
			i = 0;

		this._clearRenderMarks(removedNodes);

		if (updatedNode) {
			renderNodes.unshift(updatedNode);
		}

		// Wacky hacky. The problem is that delta.fromLineCh.line does not
		// always correspond to renderNodes[i]. If you insert a node using the
		// tree API then the inserted text is \ninserted node text. and that
		// means that the from.line is actually the line BEFORE the node
		// that's actually inserted. So correct for that here.
		if (delta.textChanged()) {
			if (!updatedNode && delta.insertedText[0] === '\n') {
				cmStartLine++;
			}
		}

		cm.eachLine(cmStartLine, cmStartLine + renderNodes.length, function (eachLineHandle) {
			outerThis.renderNode(renderNodes[i], cmStartLine + i, eachLineHandle);
			i++;
		});
	};

	NodeRenderer.prototype.renderNode = function renderNode(node, cmLineNumber, cmLineHandle) {
		this._node = node;
		this._cmLineNumber = cmLineNumber;
		this._cmLineHandle = cmLineHandle;

		// 1. clear existing render marks
		var cm = this.cm,
			renderMarks = node._renderMarks,
			length = renderMarks.length,
			i;

		if (length > 0) {
			for (i = 0; i < length; i++) {
				renderMarks[i].clear();
			}
			node._renderMarks = renderMarks = [];
		}

		// 2. Clear styles array
		cmLineHandle.styles = null;

		// 3. Make sure CM really knows line needs update
		cm.regLineChange(cmLineNumber, 'class');
		cm.regLineChange(cmLineNumber, 'text');

		node.tree.taxonomy.renderNode(this.editor, node, this);

		// 3. Allow extensions to add render marks
		var outerThis = this;
		Extensions.processExtensions('com.foldingtext.editor.renderNode', function (each) {
			each(outerThis.editor, node, outerThis);
		});

		this._node = null;
		this._cmLineNumber = null;
		this._cmLineHandle = null;
	};

	//
	// Render methods
	//

	/**
	 * Add an inline widget to the node. The widget will visually replace the
	 * given range of text with the provided widget. The given widget DOM node
	 * must be an inline element (as opposed to a block element)
	 *
	 * @param {Number} - Text start location.
	 * @param {Number} - Text length.
	 * @param {DOMElement} - HTML element.
	 */
	NodeRenderer.prototype.renderInlineWidget = function renderInlineWidget(location, length, widget) {
		var lineNumber = this._cmLineNumber;
		this._node._renderMarks.push(
			this.cm.markText({
				line: lineNumber,
				ch: location
			}, {
				line: lineNumber,
				ch: location + length
			}, {
				replacedWith: widget
			})
		);
	};

	/**
	 * Add a line widget to the node. Line widgets don't affect the node's
	 * text display like inline widgets do. Instead the widget is inserted
	 * above, below, or overlayed over the node's line according to the passed
	 * in options.
	 *
	 * @param {Object} - HTML element to add.
	 * @param {RenderLineWidgetOptions=}
	 */
	NodeRenderer.prototype.renderLineWidget = function renderLineWidget(widget, options) {
		this._node._renderMarks.push(this.cm.addLineWidget(this._cmLineHandle, widget, options));
	};

	exports.NodeRenderer = NodeRenderer;
});