define(function (require, exports, module) {
	'use strict';

	var CodeMirror = require('codemirror').CodeMirror,
		Extensions = require('ft/core/extensions').Extensions,
		dom = require('ft/util/dom');

	var CompletePopup = function CompletePopup(editor, completions, delegate) {
		var popupElement = document.createElement('ul');
		popupElement.className = 'completionPopup';

		this.isOpen = false;
		this._popupElement = popupElement;
		this._completions = completions;
		this._delegate = delegate;
		this._isUpdating = false;

		var length = completions.length;
		for (var i = 0; i < length; i++) {
			var eachItem = document.createElement('li');
			eachItem.textContent = completions[i];
			popupElement.appendChild(eachItem);
		}

		popupElement.style.top = '0px';
		popupElement.style.left = '-1000px';

		var outerThis = this;
		CodeMirror.on(popupElement, 'mousedown', function (e) {
			var index = Array.prototype.indexOf.call(popupElement.childNodes, e.target);
			if (index >= 0) {
				outerThis.updateSelectedIndex(index);
			}
			window.setTimeout(function () {
				editor.focus();
			}, 0);
			return true;
		});

		CodeMirror.on(popupElement, 'dblclick', function (e) {
			outerThis.close(false, false);
			window.setTimeout(function () {
				editor.focus();
			}, 0);
			return true;
		});

		document.body.appendChild(popupElement);
	};

	(function () {

		this.open = function show(editor, pos) {
			var localCursorCoords = editor.cursorCoords(pos, 'local'),
				element = this._popupElement,
				delegate = this._delegate,
				offsetHeight = element.offsetHeight,
				outerThis = this;

			editor.scrollRectToVisible({ top: localCursorCoords.top, left: 0, right: 0, bottom: localCursorCoords.bottom + offsetHeight });

			window.setTimeout(function () {
				var pageCursorCoords = editor.cursorCoords(pos, 'page'),
					top = pageCursorCoords.bottom,
					left = pageCursorCoords.left;
				element.style.top = top + 'px';
				element.style.left = left + 'px';
				delegate.willOpen(outerThis);
				outerThis.updateSelectedIndex(0);
				window.setTimeout(function () {
					outerThis.isOpen = true;
				}, 0);
			}, 0);
		};

		this.selectedText = function selectedText() {
			return this._completions[this._selectedIndex];
		};

		this.selectUp = function selectUp() {
			var selectedIndex = this._selectedIndex;
			if (selectedIndex > 0) {
				this.updateSelectedIndex(selectedIndex - 1);
			}
		};

		this.selectDown = function selectDown() {
			var selectedIndex = this._selectedIndex;
			if (selectedIndex < this._completions.length - 1) {
				this.updateSelectedIndex(selectedIndex + 1);
			}
		};

		this.updateSelectedIndex = function updateSelectedIndex(newSelectedIndex) {
			var oldSelectedIndex = this._selectedIndex,
				items = this._popupElement.children;

			if (oldSelectedIndex !== newSelectedIndex) {
				var oldItem = items[oldSelectedIndex],
					newItem = items[newSelectedIndex];

				if (oldItem) {
					dom.removeClass(oldItem, 'selected');
				}

				if (newItem) {
					dom.addClass(newItem, 'selected');
					newItem.scrollIntoViewIfNeeded();
				}

				this._selectedIndex = newSelectedIndex;
				this._isUpdating = true;
				this._delegate.updateChoice(this);
				this._isUpdating = false;
			}
		};

		this.isUpdating = function isUpdating() {
			return this._isUpdating;
		};

		this.close = function close(revert, goLeft) {
			if (this._closing) {
				return;
			}
			this._closing = true;
			var popupElement = this._popupElement,
				parentElement = popupElement.parentElement;

			if (parentElement) {
				parentElement.removeChild(popupElement);
			}

			this._isUpdating = true;
			this._delegate.didClose(this, revert, goLeft);
			this._isUpdating = false;
		};
	}).call(CompletePopup.prototype);

	function insertMatch(editor, match) {
		editor.replaceSelection(match.substring(editor._completePrefixLength), 'around');
	}

	Extensions.add('com.foldingtext.editor.methods', {
		complete: function () {
			var editor = this,
				prefixRange = editor.prefixWordRangeForComplete(),
				prefixLength = prefixRange.length();

			if (editor._completePopup) {
				editor._completePopup.close();
				return;
			}

			if (prefixLength === 0) {
				editor.beep();
				return;
			} else {
				editor._completePrefixLength = prefixLength;
			}

			var completions = editor.completionsForPrefixWordRange(prefixRange);

			if (!completions || completions.length === 0) {
				editor.beep();
				return;
			} else {
				var completePopup = new CompletePopup(editor, completions, {
						willOpen: function (popup) {
							editor._completePopup = popup;
							editor._completeSavedUndoHistory = editor.cm().getHistory();
							editor._completeOriginalSelection = editor.selectedRange().textInRange();
						},
						updateChoice: function (popup) {
							insertMatch(editor, popup.selectedText());
						},
						didClose: function (popup, revert, goLeft) {
							editor.replaceSelection(editor._completeOriginalSelection);
							editor.cm().setHistory(editor._completeSavedUndoHistory);
							if (!revert) {
								insertMatch(editor, popup.selectedText());
								if (goLeft) {
									editor.performCommand('moveLeft');
								} else {
									editor.performCommand('moveRight');
								}
							}
							editor.removeKeyMap('completePopup');
							editor._completeSavedUndoHistory = null;
							editor._completeOriginalSelection = null;
							editor._completePopup = null;
						}
					});

				completePopup.open(editor, prefixRange.startLineCh());

				editor.addKeyMap({
					name: 'completePopup',
					'Up' : function (editor) {
						editor._completePopup.selectUp();
					},
					'Down' : function (editor) {
						editor._completePopup.selectDown();
					},
					'Left' : function (editor) {
						editor._completePopup.close(false, true);
					},
					'Right' : function (editor) {
						editor._completePopup.close(false, false);
					},
					'Enter' : function (editor) {
						editor._completePopup.close(false, false);
					},
					'Esc' : function (editor) {
						editor._completePopup.close(true);
					},
					'Backspace' : function (editor) {
						editor._completePopup.close(true);
					},
					'Delete' : function (editor) {
						editor._completePopup.close(true);
					},
					'F5' : function (editor) {
						editor._completePopup.close(true);
					}
				});
			}
		},

		prefixWordRangeForComplete: function () {
			var editor = this,
				tree = editor.tree(),
				selectedRange = editor.selectedRange(),
				selectedLocation = selectedRange.location(),
				wordRange = selectedRange.rangeByExtendingToWord(),
				wordLocation = wordRange.location(),
				prefixRange = tree.createRangeFromLocation(wordLocation, selectedLocation - wordLocation);

			prefixRange = Extensions.processExtensions('com.foldingtext.editor.prefixWordRangeForComplete', function (each) {
				return each(prefixRange);
			}, true);

			return prefixRange;
		},

		completionsForPrefixWordRange: function (prefixWordRange) {
			var completions = Extensions.processExtensions('com.foldingtext.editor.completionsForPrefixWordRange', function (each) {
				return each(prefixWordRange);
			}, true);

			if (completions === null) {
				if (window.nativeViewController) {
					/*jshint camelcase: false */
					var string = prefixWordRange.startNode.line(),
						start = prefixWordRange.startOffset,
						length = prefixWordRange.endOffset - start;
					return window.nativeViewController.javascriptCompletionsForPartialWordStart_length_inString_(start, length, string);
				} else {
					return ['testing', 'one', 'two', 'three'];
				}
			} else {
				return completions;
			}
		}
	});

	Extensions.add('com.foldingtext.editor.beforeChange', function (editor, changeObj) {
		if (editor._completePopup && !editor._completePopup.isUpdating()) {
			editor._completePopup.close();
		}
	});

	Extensions.add('com.foldingtext.editor.mouse', {
		mouseDown: function (editor, e) {
			if (editor._completePopup) {
				editor._completePopup.close();
			}
		}
	});

	Extensions.add('com.foldingtext.editor.viewportDidScroll', function (editor) {
		if (editor._completePopup && editor._completePopup.isOpen) {
			editor._completePopup.close();
		}
	});

	Extensions.add('com.foldingtext.editor.willPerformCommand', function (editor) {
		if (editor._completePopup) {
			editor._completePopup.close();
		}
	});
});