define(function (require, exports, module) {
	'use strict';

	var Extensions = require('ft/core/extensions').Extensions,
		Selection = require('ft/extensions/selection'),
		ObjectUtils = require('ft/util/objectutils');

	function _syntaxBlockCandidateFunction(candidateRun) {
		var keys = ObjectUtils.arrayOfKeys(candidateRun.attributes);
		if ((keys.length === 1 && keys[0] === 'text') || keys.length === 0) {
			return false;
		}
		return true;
	}

	function replaceRangeMarkIfChanged(editor, varriableName, newRange, className, extensionPoint) {
		var currentMarkTextVarriableName = varriableName + 'Text',
			currentMarkOffsetVarriableName = varriableName + 'Offset',

			currentMark = editor[varriableName],
			currentMarkText = editor[currentMarkTextVarriableName],
			currentMarkOffset = editor[currentMarkOffsetVarriableName],

			hasNewRange = newRange && !newRange.isCollapsed(),

			newFrom = hasNewRange ? newRange.startLineCh() : null,
			newTo = hasNewRange ? newRange.endLineCh() : null,
			newText = hasNewRange ? newRange.textInRange() : null,
			newOffset = hasNewRange ? newRange.startOffset : null;

		if (!currentMark && !newRange) {
			return;
		}

		if (currentMark) {
			if (hasNewRange) {
				var currentRange = currentMark.find();
				if (currentRange) {
					var currentFrom = currentRange.from,
						currentTo = currentRange.to;

					if (currentFrom.line === newFrom.line &&
						currentFrom.ch === newFrom.ch &&
						currentTo.line === newTo.line &&
						currentTo.ch === newTo.ch) {

						if (currentMarkText === newText && currentMarkOffset === newOffset) {
							return;
						}
					}
				}
			}

			currentMark.clear();
			editor[varriableName] = null;
			editor[currentMarkTextVarriableName] = null;
			editor[currentMarkOffsetVarriableName] = null;
		}

		if (hasNewRange) {
			editor[varriableName] = editor.cm().markText(newFrom, newTo, {
				className: className
			});
			editor[currentMarkTextVarriableName] = newText;
			editor[currentMarkOffsetVarriableName] = newOffset;
		}

		if (extensionPoint) {
			Extensions.processExtensions(extensionPoint, function (each) {
				each(editor, newRange);
			});
		}
	}

	function updateCurrentMarks(editor) {
		var tree = editor.tree(),
			selectedRange = editor.selectedRange(),
			headNode = selectedRange.headNode(),
			headOffset = selectedRange.headOffset(),
			headRange = tree.createRangeFromNodes(headNode, headOffset),
			attributedString = headNode.lineAttributedString(),
			syntaxSpanRange = attributedString.attributeRangeCalculatedByFunction(headOffset, _syntaxBlockCandidateFunction),
			wordRange = null,
			sentenceRange = null;

		if ((!syntaxSpanRange || syntaxSpanRange.length === 0) && headOffset > 0) {
			syntaxSpanRange = attributedString.attributeRangeCalculatedByFunction(headOffset - 1, _syntaxBlockCandidateFunction);
		}

		var syntaxRange = null;
		if (syntaxSpanRange && syntaxSpanRange.length > 0) {
			syntaxRange = tree.createRangeFromNodes(headNode, syntaxSpanRange.location, headNode, syntaxSpanRange.location + syntaxSpanRange.length);
		}

		replaceRangeMarkIfChanged(editor, '_currentSyntaxMark', syntaxRange, 'cm-currentsyntax', null);

		wordRange = headRange.rangeByExtendingToWord();
		sentenceRange = headRange.rangeByExtendingToSentence();

		replaceRangeMarkIfChanged(editor, '_currentWordMark', wordRange, 'cm-currentword', 'com.foldingtext.editor.currentWordDidChange');
		replaceRangeMarkIfChanged(editor, '_currentSentenceMark', sentenceRange, 'cm-currentsentence', 'com.foldingtext.editor.currentSentenceDidChange');
	}

	Extensions.add('com.foldingtext.editor.willEndOperation', function (editor) {
		// need to update on willEndOperation instead of selection changed
		// because of syntax range which depends on syntax changes which can
		// happen without hcanging the selection
		editor.operation(function () {
			updateCurrentMarks(editor);
		});
	});

	// IN PROGESS... problem is that can't really use ":hover" in CSS to show
	// links because those links can be made of multiple spans, so the
	// underline only shows on part of the link and looks bad. So instead need
	// to add idea of "indicator" mark that will be added to full link range
	// on mouse over.

	function updateIndicatedMarks(editor, e) {
		var targets = editor.targetsAt(e.clientX, e.clientY),
			attributes = targets.attributes,
			lineCh = targets.lineCh,
			node = targets.node,
			target;

		//console.log(lineCh);

		//replaceRangeMarkIfChanged(editor, '_indicatedTextMark', indicatedTextRange, 'cm-indicatedText', 'com.foldingtext.editor.indicatedTextDidChange');
	}

	Extensions.add('com.foldingtext.editor.mouseover', function (editor, e) {
		//editor.operation(function () {
		//	updateIndicatedMarks(editor, e);
		//});
	});

	Extensions.add('com.foldingtext.editor.methods', /**@lends module:ft/editor/editor.Editor.prototype*/ {

		/**
		 * @returns {Range}
		 * @methodgroup Current
		 */
		currentWordRange: function () {
			var mark = this._currentWordMark;
			if (mark) {
				var range = mark.find();
				if (range) {
					return this.tree().createRangeFromLineChs(range.from, range.to);
				}
			}
			return null;
		},

		/**@returns {Range}*/
		currentSentenceRange: function () {
			var mark = this._currentSentenceMark;
			if (mark) {
				var range = mark.find();
				if (range) {
					return this.tree().createRangeFromLineChs(range.from, range.to);
				}
			}
			return null;
		}
	});
});