define(function (require, exports, module) {
	'use strict';

	require('searchcursor');
	require('ft/extensions/panel');

	var Extensions = require('ft/core/extensions').Extensions;

	var FindReplace = function FindReplace(editor) {
		this._editor = editor;
		this._searchCursor = null;
		this._eventsRegistered = false;
	};

	(function () {

		this.registerEvents = function registerEvents(editor, panel) {
			CodeMirror.on(panel, 'returnpressed', function (obj) {
				if (obj.name === 'find') {
					var command = obj.shiftKey ? 'findPrev' : 'findNext';
					editor.execCommand(command);
				} else {
					editor.execCommand('replace');
				}
			});
			this._eventsRegistered = true;
		};

		this.searchCursor = function searchCursor() {
			return this._searchCursor;
		};

		this.showFindDialog = function showFindDialog() {
			var editor = this._editor,
				panel = editor.panel();

			panel.show('bottom');
			if (!panel.hasField('find')) {
				panel.addFields(this, [
					{
						name: 'find',
						index: 0,
						value: 'Find: ',
						markerStart: 0,
						markerLength: 5,
						mode: null
					}
				]);
			}

			if (panel.hasField('replace')) {
				panel.removeField('replace');
			}

			if (!this._eventsRegistered) {
				this.registerEvents(editor, panel);
			}

			panel.focus();
		};

		this.showFindReplaceDialog = function showFindReplaceDialog() {
			var editor = this._editor,
				panel = editor.panel();

			panel.show('bottom');
			if (!panel.hasField('find')) {
				panel.addFields(this, [
					{
						name: 'find',
						index: 0,
						value: 'Find: ',
						markerLength: 5,
						mode: null
					}
				]);
			}
			if (!panel.hasField('replace')) {
				panel.addFields(this, [
					{
						name: 'replace',
						index: 1,
						value: '\nReplace: ',
						markerStart: 1,
						markerLength: 8,
						mode: null
					}
				]);
			}

			if (!this._eventsRegistered) {
				this.registerEvents(editor, panel);
			}

			panel.focus(0);
		};

		this.findImplementation = function findImplementation(direction, cm) {
			var forward = direction !== 'reverse',
				Pos = CodeMirror.Pos,
				selectionStart = cm.doc.getCursor('start'),
				selectionEnd = cm.doc.getCursor('end'),
				currentPosition = forward ? selectionEnd : selectionStart,
				nextFindPosition,
				query = this._editor.panel().fieldValue('find'),
				caseSensitive = query.toLowerCase() !== query,
				searchCursor = cm.getSearchCursor(query, currentPosition, !caseSensitive);

			if (!searchCursor.find(!forward)) {
				searchCursor = cm.getSearchCursor(query, forward ? new Pos(cm.firstLine(), 0) : new Pos(cm.lastLine()), !caseSensitive);
				if (!searchCursor.find(!forward)) {
					return null;
				}
			}

			this._searchCursor = searchCursor;
			return searchCursor;
		};

	}).call(FindReplace.prototype);

	function getSearchState(editor) {
		return editor._searchState || (editor._searchState = { 'query': 'e', 'replaceString': 'b' });
	}

	function getSearchCursor(editor, query, pos) {
		return editor.cm().getSearchCursor(query, pos, typeof query === 'string' && query === query.toLowerCase());
	}

	function showFindInterface(editor) {
		var panel = editor.panel();

		panel.show();

		if (!panel.hasField('find')) {
			panel.addField({
				id: 'find',
				label: 'Find:',
				value: '',
				onChange: function (oldValue, newValue) {
					setQuery(editor, newValue);
				}
			});
		}

		panel.focus();
	}

	function hideFindInterface(editor) {
		editor.panel().removeField('find');
	}

	function setQuery(editor, query) {
		getSearchState(editor).query = query;
	}

	function setReplaceString(editor, replaceString) {
		getSearchState(editor).replaceString = replaceString;
	}

	function findMatch(editor, previous) {
		var cm = editor.cm(),
			state = getSearchState(editor),
			cursor = getSearchCursor(editor, state.query, previous ? state.posFrom : state.posTo);

		if (!cursor.find(previous)) {
			cursor = getSearchCursor(editor, state.query, previous ? CodeMirror.Pos(cm.lastLine()) : CodeMirror.Pos(cm.firstLine(), 0));
			if (!cursor.find(previous)) {
				return false;
			}
		}

		cm.setSelection(cursor.from(), cursor.to());
		state.posFrom = cursor.from();
		state.posTo = cursor.to();

		return true;
	}

	function nextMatch(editor) {
		return findMatch(editor, false);
	}

	function previousMatch(editor) {
		return findMatch(editor, true);
	}

	function showReplaceInterface(editor) {
	}

	function hideReplaceInterface(editor) {
		editor.panel().removeField('replace');
	}

	function replace(editor, all) {
		var state = getSearchState(editor),
			replaceString = state.replaceString,
			replaced = false;

		editor.operation(function () {
			if (all) {
				while (nextMatch(editor)) {
					editor.replaceSelection(replaceString, 'end');
					replaced = true;
				}
			} else {
				editor.replaceSelection(replaceString, 'end');
				replaced = true;
			}
		});

		return replaced;
	}

	function replaceAndFind(editor) {
		replace(editor, false);
		nextMatch(editor);
	}

	function validateFind(editor) {
		return getSearchState(editor).query !== undefined;
	}

	function validateReplace(editor) {
		var searchState = getSearchState(editor);
		return searchState.query !== undefined && searchState.replaceString !== undefined;
	}

	Extensions.addCommand({
		name: 'showFindInterface',
		shortcut: 'Cmd-F',
		performCommand: showFindInterface
	});

	Extensions.addCommand({
		name: 'hideFindInterface',
		shortcut: 'Esc',
		performCommand: hideFindInterface
	});

	Extensions.addCommand({
		name: 'setQuery',
		performCommand: setQuery
	});

	Extensions.addCommand({
		name: 'nextMatch',
		shortcut: 'Cmd-G',
		validate: function (editor) {
			return getSearchState(editor).query !== undefined;
		},
		performCommand: nextMatch
	});

	Extensions.addCommand({
		name: 'previousMatch',
		shortcut: 'Shift-Cmd-G',
		validate: validateFind,
		performCommand: previousMatch
	});

	Extensions.addCommand({
		name: 'showReplaceInterface',
		shortcut: 'Cmd-Alt-F',
		performCommand: showReplaceInterface
	});

	Extensions.addCommand({
		name: 'hideReplaceInterface',
		shortcut: 'Esc',
		performCommand: hideReplaceInterface
	});

	Extensions.addCommand({
		name: 'replace',
		validate: validateReplace,
		performCommand: replace
	});

	Extensions.addCommand({
		name: 'replaceAndFind',
		validate: validateReplace,
		performCommand: replaceAndFind
	});

	Extensions.addCommand({
		name: 'replaceAll',
		validate: validateReplace,
		performCommand: function (editor) {
			replace(editor, true);
		}
	});

	exports.FindReplace = FindReplace;
});