define(function (require, exports, module) {
	'use strict';

	var CodeMirror = require('codemirror').CodeMirror,
		queryParameter = require('ft/util/queryparameter').QueryParameter,
		Extensions = require('ft/core/extensions').Extensions,
		Range = require('ft/core/range').Range,
		nextFoldID = 0;

	var Fold = function Fold(editor, textMarker, isManual, isBoundary) {
		this.id = '' + nextFoldID++;
		this._valid = true;
		this._editor = editor;
		this._isManual = isManual;
		this._isBoundary = isBoundary;
		this._textMarker = textMarker;

		textMarker.ftFold = this;

		var outerThis = this;

		CodeMirror.on(textMarker, 'clear', function () {
			outerThis._valid = false;
			outerThis._editor._invalidFolds = true;
		});

		CodeMirror.on(textMarker, 'hide', function () {
			outerThis._valid = false;
			outerThis._editor._invalidFolds = true;
		});

		CodeMirror.on(textMarker, 'unhide', function () {
			if (!outerThis._valid) {
				outerThis._valid = true;
				outerThis._editor._folds.push(outerThis);
			}
		});
	};

	(function () {

		this.valid = function valid() {
			return this._valid;
		};

		this.isManual = function isManual() {
			return this._isManual;
		};

		this.setIsManual = function setIsManual(isManual) {
			this._isManual = isManual;
		};

		this.isBoundary = function isBoundary() {
			return this._isBoundary;
		};

		this.setIsBoundary = function setIsBoundary(isBoundary) {
			this._isBoundary = isBoundary;
		};

		this.clear = function  clear() {
			this._textMarker.clear();
		};

		this.range = function range() {
			var pos = this._textMarker.find();
			if (pos) {
				return this._editor.tree().createRangeFromLineChs(pos.from, pos.to);
			}
			return undefined;
		};
	}).call(Fold.prototype);

	/*jshint validthis:true */

	function folds() {
		var f = this._folds;
		if (!f) {
			f = [];
			this._folds = f;
		} else if (this._invalidFolds) {
			var length = f.length,
				newFolds = [],
				each,
				i;
			for (i = 0; i < length; i++) {
				each = f[i];
				if (each._valid) {
					newFolds.push(each);
				}
			}
			f = newFolds;
			this._folds = f;
			this._invalidFolds = false;
		}
		return f;
	}

	function foldsInRange(range) {
		var cm = this.cm(),
			from = range.startLineCh(),
			to = range.endLineCh(),
			marks = cm.findMarksInRange(from, to),
			length = marks.length,
			folds = [],
			mark,
			each,
			i;

		for (i = 0; i < length; i++) {
			each = marks[i].ftFold;
			if (each) {
				folds.push(each);
			}
		}

		return folds;
	}

	function foldsToJSON(folds) {
		var length = folds.length,
			jsonFolds = [],
			eachFold,
			each,
			i;

		for (i = 0; i < length; i++) {
			eachFold = folds[i];
			each = eachFold._textMarker.find();
			if (each) {
				if (eachFold.isManual()) {
					each.isManual = eachFold.isManual();
				}
				if (eachFold.isBoundary()) {
					each.isBoundary = eachFold.isBoundary();
				}
				jsonFolds.push(each);
			}
		}

		return jsonFolds;
	}

	function normalizeFoldsJSON(foldsJSON, originLineCh) {
		foldsJSON.forEach(function (each) {
			each.from.line -= originLineCh.line;
			if (each.from.line === 0) {
				each.from.ch -= originLineCh.ch;
			}

			each.to.line -= originLineCh.line;
			if (each.to.line === 0) {
				each.to.ch -= originLineCh.ch;
			}
		});
		return foldsJSON;		
	}

	function perpareNormalizedFoldsJSONForDestination(foldsJSON, destinationLineCh) {
		foldsJSON.forEach(function (each) {
			if (each.from.line === 0) {
				each.from.ch += destinationLineCh.ch;
			}
			each.from.line += destinationLineCh.line;

			if (each.to.line === 0) {
				each.to.ch += destinationLineCh.ch;
			}				
			each.to.line += destinationLineCh.line;
		});
		return foldsJSON;
	}		

	function createFoldInRange(range, isManual, isBoundary) {
		var editor = this,
			cm = editor.cm(),
			tree = editor.tree(),
			folds = editor.foldsInRange(range),
			rangeStart = range.location(),
			rangeEnd = rangeStart + range.length(),
			adjustedFoldRanges = [],
			length = folds.length,
			usedExistingFold,
			eachRange,
			each,
			fold,
			i;

		// Sometimes range length can be -1 for tracked ranges. So don't just
		// test for === 0
		if (range.length() < 1) {
			return undefined;
		}

		editor.operation(function () {
			for (i = 0; i < length; i++) {
				each = folds[i];
				eachRange = each.range();

				if (eachRange.isEqualToRange(range)) {
					fold = each;
					usedExistingFold = true;
					return;
				} else {
					var eachStart = eachRange.location(),
						eachEnd = eachStart + eachRange.length(),
						containsEachStart = eachStart >= rangeStart && eachStart < rangeEnd,
						containsEachEnd = eachEnd > rangeStart && eachEnd <= rangeEnd,
						containedByEach = rangeStart > eachStart && rangeEnd < eachEnd;

					if (containsEachStart && !containsEachEnd) {
						// trim existing start
						each.clear();
						if ((eachEnd - rangeEnd) > 0) {
							adjustedFoldRanges.push(tree.createRangeFromLocation(rangeEnd, eachEnd - rangeEnd));
						}
					} else if (!containsEachStart && containsEachEnd) {
						// trim existing end
						each.clear();
						if ((rangeStart - eachStart) > 0) {
							adjustedFoldRanges.push(tree.createRangeFromLocation(eachStart, rangeStart - eachStart));
						}
					} else if (!containsEachStart && !containsEachEnd && containedByEach) {
						// split existing range in two ranges sorounding new
						each.clear();
						if ((eachEnd - rangeEnd) > 0) {
							adjustedFoldRanges.push(tree.createRangeFromLocation(rangeEnd, eachEnd - rangeEnd));
						}
						if ((rangeStart - eachStart) > 0) {
							adjustedFoldRanges.push(tree.createRangeFromLocation(eachStart, rangeStart - eachStart));
						}
					} else if (containsEachStart && containsEachEnd) {
						// leave existing as is, it will become a subrange of the new range.
					}
				}
			}

			var widget = document.createElement('span'),
				from = range.startLineCh(),
				to = range.endLineCh(),
				mark = cm.markText(from, to, {
					className: 'collapsed',
					replacedWith: widget,
					clearOnEnter: false
				});

			fold = new Fold(editor, mark, isManual, isBoundary);
			widget.className = 'cm-fold';

			CodeMirror.on(widget, 'mousedown', function (e) {
				fold.clear();
				e.preventDefault();
			});

			length = adjustedFoldRanges.length;
			for (i = 0; i < length; i++) {
				each = adjustedFoldRanges[i];
				editor.createFoldInRange(each);
			}
		});

		if (!usedExistingFold) {
			if (!this._folds) {
				this._folds = [];
			}
			this._folds.push(fold);
		}

		editor.invalidateRestorableState();

		return fold;
	}

	function splitAndRemoveFoldsInRange(range) {
		var folds = this.foldsInRange(range),
			outerThis = this;

		if (folds.length > 0) {
			outerThis.operation(function () {
				outerThis.createFoldInRange(range).clear();
				folds = outerThis.foldsInRange(range);
				for (var i = 0; i < folds.length; i++) {
					var each = folds[i],
						eachRange = each.range();

					if (range.containsRange(eachRange)) {
						each.clear();
					}
				}
			});
		}
	}

	function restoreJSONFolds(jsonFolds) {
		if (!jsonFolds) {
			return;
		}

		var length = jsonFolds.length,
			each,
			i;

		jsonFolds.sort(function (a, b) {
			var compare = b.from.line - a.from.line;
			if (compare === 0) {
				compare = b.from.ch - a.from.ch;
				if (compare === 0) {
					compare = b.to.line - a.to.line;
					if (compare === 0) {
						compare = b.to.ch - a.to.ch;
					}
				}
			}
			return compare;
		});

		var outerThis = this,
			tree = this.tree(),
			textLength = tree.textLength();

		this.operation(function () {
			for (i = 0; i < length; i++) {
				each = jsonFolds[i];
				if (tree.textOffsetForLineChar(each.to.line, each.to.ch) <= textLength) {
					outerThis.createFoldInRange(tree.createRangeFromLineChs(each.from, each.to), each.isManual, each.isBoundary);
				}
			}
		});
	}

	function removeFold(fold) {
		this.removeFolds([fold]);
	}

	function removeFolds(folds) {
		var length = folds.length,
			outerThis = this,
			each,
			i;

		this.operation(function () {
			for (i = 0; i < length; i++) {
				folds[i].clear();
			}
		});

		this.invalidateRestorableState();
	}

	function nodeIsHiddenInFold(node, dontCountLineFolds) {
		return this.lineIsHiddenInFold(this.tree().nodeToLineNumber(node), dontCountLineFolds);
	}

	function lineIsHiddenInFold(line, dontCountLineFolds) {
		if (line < 0) {
			return false;
		}

		var cm = this.cm(),
			lineHandle = cm.getLineHandle(line);

		if (cm.lineHandleIsHidden(lineHandle)) {
			return true;
		}

		if (!dontCountLineFolds) {
			return lineHandleStartsLineFold(this, lineHandle);
		}

		return false;
	}

	function nodeStartsLineFold(node) {
		return this.lineStartsLineFold(this.tree().nodeToLineNumber(node));
	}

	function lineStartsLineFold(line) {
		if (line < 0) {
			return false;
		}

		var cm = this.cm(),
			lineHandle = cm.getLineHandle(line);

		return lineHandleStartsLineFold(this, lineHandle);
	}

	function lineHandleStartsLineFold(editor, lineHandle) {
		var cm = editor.cm(),
			markedSpans = lineHandle.markedSpans;

		if (markedSpans && markedSpans.length > 0) {
			for (var i = 0; i < markedSpans.length; i++) {
				var mark = markedSpans[i];
				if (mark.marker && mark.marker.className === 'collapsed' && (mark.from === null || mark.from === 0) && (mark.to === null || mark.to === lineHandle.text.length)) {
					var to = mark.marker.find().to;
					if (cm.getLine(to.line).length === to.ch) {
						return mark.marker;
					}
				}
			}
		}

		return null;
	}

	function visibleNodesInRange(range) {
		//
		// This isn't really what I want. Should skip all nodes that are completely folded away.
		//
		var //cm = this.cm(),
			nodes = range.nodesInRange();
		//	i;

		//cm.eachLine(range.startLine(), range.endLine(), function (handle) {
		//	i++;
		//});

		return nodes;
	}

	function saveFolds(editor, state) {
		state.jsonFolds = editor.foldsToJSON(editor.folds());
	}

	function loadFolds(editor, state) {
		if (!state.fileChangedSinceLastStateSync) {
			editor.operation(function () {
				editor.removeFolds(editor.folds());
				editor.restoreJSONFolds(state.jsonFolds);
			});			
		}
	}

	Extensions.addState({
		readState: loadFolds,
		writeState: saveFolds
	});

	Extensions.addRenderNodeElementStyles(function (editor, node, elementRenderer) {
		var marker = lineHandleStartsLineFold(editor, elementRenderer._lineHandle);
		if (marker) {
			elementRenderer.addTextClass('cm-foldedLine');
			elementRenderer.addWrapClass('cm-foldedLineWrapper');

			var lineChRange = marker.find();
			if (lineChRange.from.line === 0) {
				elementRenderer.addWrapClass('cm-foldedLineWrapper-at-start');
			} else if (lineChRange.to.line === editor.tree().nodeCount() - 1) {
				elementRenderer.addWrapClass('cm-foldedLineWrapper-at-end');
			}

			elementRenderer.setStyle('marginLeft', '0');
			elementRenderer.setStyle('marginRight', '0');
			elementRenderer.setStyle('textIndent', '0');
		}
	});

	Extensions.add('com.foldingtext.editor.beforeUndoRedo', function (editor, changeObj) {
		var range = editor.tree().createRangeFromLineChs(changeObj.from, changeObj.to),
			folds = editor.foldsInRange(range);
		editor.removeFolds(folds);
	});

	Extensions.add('com.foldingtext.editor.methods', {
		folds: folds,
		foldsInRange: foldsInRange,
		foldsToJSON: foldsToJSON,
		normalizeFoldsJSON: normalizeFoldsJSON,
		perpareNormalizedFoldsJSONForDestination: perpareNormalizedFoldsJSONForDestination,
		restoreJSONFolds: restoreJSONFolds,
		createFoldInRange: createFoldInRange,
		splitAndRemoveFoldsInRange: splitAndRemoveFoldsInRange,
		removeFold: removeFold,
		removeFolds: removeFolds,
		nodeIsHiddenInFold: nodeIsHiddenInFold,
		lineIsHiddenInFold: lineIsHiddenInFold,
		nodeStartsLineFold: nodeStartsLineFold,
		lineStartsLineFold: lineStartsLineFold,
		visibleNodesInRange: visibleNodesInRange
	});
});