define(function (require, exports, module) {
	'use strict';

	require('ft/extensions/folding');

	var Extensions = require('ft/core/extensions').Extensions,
		Node = require('ft/core/node').Node;

	/*jshint validthis:true */

	function _getEndFold(editor, node) {
		var tree = editor.tree(),
			lineLength = node.line().length,
			folds = editor.foldsInRange(tree.createRangeFromNodes(node, lineLength)),
			length = folds.length,
			i;

		for (i = 0; i < length; i++) {
			var each = folds[i],
				eachRange = each.range();

			if (eachRange.startNode === node && eachRange.startOffset === lineLength) {
				return each;
			}
		}

		return null;
	}

	function isExpanded(node, testChildren) {
		if (_getEndFold(this, node) !== null) {
			return false;
		}

		if (testChildren) {
			var eachChild = node.firstChild;
			while (eachChild) {
				if (this.nodeIsHiddenInFold(eachChild)) {
					return false;
				}
				eachChild = eachChild.nextSibling;
			}
		}

		return true;
	}

	function isCollapsed(node, testChildren) {
		return !this.isExpanded(node, testChildren);
	}

	function expandNode(node, all) {
		this.expandNodes([node], all);
	}

	function expandNodes(nodes, all) {
		var outerThis = this,
			tree = this.tree();

		this.operation(function () {
			var allChildren = [],
				nodesLength = nodes.length,
				i;

			for (i = 0; i < nodesLength; i++) {
				var each = nodes[i],
					endFold = _getEndFold(outerThis, each);

				if (endFold) {
					endFold.clear();
				}

				var eachChild = each.firstChild,
					eachExpandRange;
				while (eachChild) {
					if (outerThis.nodeIsHiddenInFold(eachChild)) {
						outerThis.expandToRevealNode(eachChild);
					}
					allChildren.push(eachChild);
					eachChild = eachChild.nextSibling;
				}

				if (all) {
					outerThis.expandNodes(allChildren, true);
				}
			}
		});
	}

	function expandToRevealNode(node) {
		var startNode = node,
			startOffset = 0,
			endNode = node,
			endOffset = node.line().length,
			expandRange;

		if (node.previousLineNode()) {
			startNode = node.previousLineNode();
			startOffset = startNode.line().length;
		}

		if (node.nextLineNode() && !node.firstChild) {
			endNode = node.nextLineNode();
			endOffset = 0;
		}

		this.splitAndRemoveFoldsInRange(this.tree().createRangeFromNodes(startNode, startOffset, endNode, endOffset));
	}

	function collapseNode(node, all, folds) {
		return this.collapseNodes([node], all, folds);
	}

	function collapseNodes(nodes, all, folds) {
		if (!folds) {
			folds = [];
		}

		var tree = this.tree(),
			outerThis = this,
			length = nodes.length,
			eachNode,
			i;

		this.operation(function () {
			for (i = 0; i < length; i++) {
				eachNode = nodes[i];
				if (eachNode.firstChild) {
					var last = eachNode.lastDescendantOrSelf(),
						foldRange = tree.createRangeFromNodes(eachNode, eachNode.line().length, last, last.line().length);

					if (all) {
						var eachSibling = eachNode.firstChild;
						while (eachSibling) {
							outerThis.collapseNode(eachSibling, true, folds);
							eachSibling = eachSibling.nextSibling;
						}
					}
					folds.push(outerThis.createFoldInRange(foldRange, true));
				}
			}
		});

		return folds;
	}

	function hideNodes(nodes, trimEmpties) {
		var editor = this,
			tree = editor.tree(),
			length = nodes.length,
			hideStart,
			hideEnd,
			previous,
			each,
			i;


		function addFold(from, to) {
			if (trimEmpties) {
				while (from.type() === 'empty' && from !== to) {
					from = from.nextLineNode();
				}

				while (to.type() === 'empty' && to !== from) {
					to = to.previousLineNode();
				}				
			}

			if (from === to && from.type() === 'empty') {
				return;
			}
			editor.createFoldInRange(tree.createRangeFromNodes(from, 0, to, -1), false);
		}

		editor.operation(function () {
			for (i = 0; i < length; i++) {
				each = nodes[i];
				if (!previous || each.previousLineNode() === previous) {
					if (!hideStart) {
						hideStart = each;
					}
					hideEnd = each;
				} else {
					if (hideStart) {
						addFold(hideStart, hideEnd);
						hideStart = undefined;
						hideEnd = undefined;
					}
				}
				previous = each;
			}

			if (hideStart) {
				addFold(hideStart, hideEnd);
			}
		});
	}

	function _biggestFold(folds) {
		var length = folds.length,
			biggest = folds[0],
			biggestLength = biggest.range().length(),
			each,
			eachLength,
			i;

		for (i = 1; i < length; i++) {
			each = folds[i];
			eachLength = each.range().length();
			if (eachLength > biggestLength) {
				biggest = each;
				biggestLength = eachLength;
			}
		}

		return biggest;
	}

	function toggleNodeFold(editor, all) {
		var range = editor.selectedRange(),
			folds = editor.foldsInRange(range);

		if (folds.length > 0) {
			if (range.isCollapsed()) {
				editor.removeFold(_biggestFold(folds));
			} else {
				editor.removeFolds(folds);
			}
			return;
		}

		var branches = Node.commonAncestorsForNodes(range.rangeByAdjustingToEffectedNodesInEditor(editor).nodesInRange()),
			lineOrderedBranches = branches.nodesInLineOrder(),
			node = range.startNode;

		if (editor.isCollapsed(node)) {
			editor.expandNodes(lineOrderedBranches, all);
		} else {
			if (editor.nodeIsHiddenInFold(node)) {
				editor.expandToRevealNode(node);
			} else {
				if (node.firstChild) {
					editor.collapseNodes(lineOrderedBranches, all);
				} else {
					if (!node.parent.isRoot()) {
						editor.collapseNode(node.parent, all);
						editor.setSelectedRange(editor.tree().createRangeFromNodes(node.parent, node.parent.line().length));
					}
				}
			}
		}
	}

	function toggleAllFolds(editor) {
		var folds = editor.folds(),
			length = folds.length,
			i;

		if (length) {
			editor.removeFolds(folds);
		} else {
			editor.operation(function () {
				var each = editor.tree().root.firstChild;
				while (each) {
					editor.collapseNode(each, true);
					each = each.nextSibling;
				}
			});
		}
	}

	function sortedAndUniqueNodeStructureLevels(nodes) {
		var structureLevels = [];
		nodes.forEach(function (each) {
			structureLevels.push(each.structureLevel());
		});
		structureLevels.sort();
		structureLevels.sortedUniquify();
		return structureLevels;
	}

	function foldingLevel(nodes) {
		var tree = this.tree(),
			editor = this;

		if (!nodes) {
			nodes = tree.root.descendants();
		}

		if (nodes instanceof Node) {
			var node = nodes;
			nodes = node.descendants();
			nodes.unshift(node);
		}

		tree.ensureClassified();

		var nodesStructureLevels = sortedAndUniqueNodeStructureLevels(nodes),
			firstHiddenStructureLevel = Number.MAX_VALUE,
			firstHiddenNode;

		nodes.forEach(function(each) {
			if (editor.nodeIsHiddenInFold(each)) {
				if (each.structureLevel() < firstHiddenStructureLevel) {
					firstHiddenStructureLevel = each.structureLevel();
					firstHiddenNode = each;
				}
			}
		});

		if (firstHiddenStructureLevel === Number.MAX_VALUE) {
			return nodesStructureLevels.length;
		} else {
			return nodesStructureLevels.indexOf(firstHiddenStructureLevel);
		}
	}

	function setFoldingLevel(level, nodes) {		
		var tree = this.tree(),
			editor = this;
			
		if (!nodes) {
			nodes = tree.root.descendants();
		}

		if (nodes instanceof Node) {
			var node = nodes;
			nodes = node.descendants();
			nodes.unshift(node);
		}

		editor.operation(function () {
			var nodesRange = tree.createRangeFromNodes(nodes[0], 0, nodes.lastObject(), -1),
				nodesStructureLevels = sortedAndUniqueNodeStructureLevels(nodes),
				collapseStructureLevel = nodesStructureLevels[level - 1],
				hideStructureLevel = nodesStructureLevels[level];
				
			// remove existing folds
			editor.removeFolds(editor.foldsInRange(nodesRange));

			if (level < 0) {
				return;
			}

			// find nodes to collapse
			editor.collapseNodes(nodes.filter(function (each) {
				return each.structureLevel() === collapseStructureLevel;
			}), false);

			// find extra nodes to hide
			editor.hideNodes(nodes.filter(function (each) {
				return each.structureLevel() >= hideStructureLevel && !editor.nodeIsHiddenInFold(each);
			}), true);
		});
	}

	Extensions.addCommand({
		name: 'toggleNodeFold',
		performCommand: function (editor, options) {
			toggleNodeFold(editor, false);
		}
	});

	Extensions.addCommand({
		name: 'toggleNodeFoldNodeAndDescendants',
		performCommand: function (editor, options) {
			toggleNodeFold(editor, true);
		}
	});

	Extensions.addCommand({
		name: 'toggleAllFolds',
		performCommand: toggleAllFolds
	});

	Extensions.addCommand({
		name: 'foldingLevel',
		performCommand: function (editor, options) {
			var selectedRange = editor.selectedRange(),
				effectedRange = selectedRange.rangeByExtendingToBranch().rangeByAdjustingToEffectedNodesInEditor(editor),
				nodes = effectedRange.nodesInRange(),
				foldingLevel = editor.foldingLevel(nodes);
			editor.setFoldingLevel(options.interfaceItemTag, nodes);
		}
	});

	Extensions.addCommand({
		name: 'expandFoldingLevel',
		performCommand: function (editor, options) {
			var selectedRange = editor.selectedRange(),
				effectedRange = selectedRange.rangeByExtendingToBranch().rangeByAdjustingToEffectedNodesInEditor(editor),
				nodes = effectedRange.nodesInRange(),
				foldingLevel = editor.foldingLevel(nodes);
			editor.setFoldingLevel(foldingLevel + 1, nodes);
		}
	});

	Extensions.addCommand({
		name: 'contractFoldingLevel',
		performCommand: function (editor, options) {
			var selectedRange = editor.selectedRange(),
				effectedRange = selectedRange.rangeByExtendingToBranch().rangeByAdjustingToEffectedNodesInEditor(editor),
				nodes = effectedRange.nodesInRange(),
				foldingLevel = editor.foldingLevel(nodes);

			if (foldingLevel > 1) {
				editor.setFoldingLevel(foldingLevel - 1, nodes);
			}/* else if (foldingLevel === 1) {
				var firstParent = nodes[0].parent;
				if (!firstParent.isRoot()) {
					editor.setSelectedRange(editor.tree().createRangeFromNodes(firstParent, -1));
					editor.performCommand('contractFoldingLevel');
				}
			}*/
		}
	});

	Extensions.addCommand({
		name: 'hideSelection',
		performCommand: function (editor, options) {
			var selectedRange = editor.selectedRange(),
				effectedRange = selectedRange.rangeByAdjustingToEffectedNodesInEditor(editor),
				nodes = effectedRange.nodesInRange();
			editor.hideNodes(nodes, false);
		}
	});

	Extensions.add('com.foldingtext.editor.methods',  /**@lends module:ft/editor/editor.Editor.prototype*/ {
		foldingLevel: foldingLevel,
		setFoldingLevel: setFoldingLevel,

		/**
		 * Return true if the node doesn't have a fold at the end of it's
		 * line.
		 *
		 * @kind function
		 * @param {module:ft/core/node.Node} node
		 * @returns {Boolean}
		 * @methodgroup Folding
		 */		
		isExpanded: isExpanded,

		/**
		 * Return true if the node is "collapsed". A node is consided
		 * collapsed if there is a fold at the end of it's line.
		 *
		 * @kind function
		 * @param {module:ft/core/node.Node} node
		 * @returns {Boolean}
		 */
		isCollapsed: isCollapsed,

		/**
		 * Collapse the node by adding a fold to hide all of it's descendant
		 * nodes. Nothing will happen 
		 *
		 * @kind function
		 * @param {module:ft/core/node.Node} node
		 */
		collapseNode: collapseNode,
		collapseNodes: collapseNodes,

		/**
		 * Unlike `collapseNode` which hides the descendants of the provided
		 * node, this function hides the nodes in the passed in array in
		 * folds. This method doesn't care about the node tree structure... it
		 * will just hide the individual nodes passed in. So for instance you
		 * could hide one sibling node and leave other siblings still visible.
		 *
		 * @kind function
		 * @param {Array.<module:ft/core/node.Node>} nodes
		 */		
		hideNodes: hideNodes,

		/**
		 * Expand the node by removing the fold at the end of it's line.
		 *
		 * @kind function
		 * @param {module:ft/core/node.Node} node
		 */
		expandNode: expandNode,
		expandNodes: expandNodes,

		/**
		 * Remove all folds neccessary to reveal the given node in the editor.
		 *
		 * @kind function
		 * @param {module:ft/core/node.Node} node
		 */
		expandToRevealNode: expandToRevealNode
	});
});