define(function (require, exports, module) {
	'use strict';

	var Extensions = require('ft/core/extensions').Extensions,
		Selection = require('ft/extensions/selection');

	function insertLink(editor) {
		var selectedText = editor.selectedText(),
			selectedRange;

		editor.operation(function () {
			editor.replaceSelection('[' + selectedText + ']()', 'around');
			if (selectedText.length) {
				selectedRange = editor.selectedRange().rangeByOffsetting(0, -1).rangeByCollapsing(false);
			} else {
				selectedRange = editor.selectedRange().rangeByOffsetting(1, 0).rangeByCollapsing(true);
			}
			editor.setSelectedRange(selectedRange);
		});
	}

	function toggleSelectionFormatting(editor, before, after) {
        // set after to before if no argument given or empty string
        if (!after) {
            after = before;
        }

		var selectedRange = editor.selectedRange(),
			selectedText = selectedRange.textInRange(),
			beforeLength = before.length,
			afterLength = after.length,
			addedFormatting = true;

		editor.operation(function () {
			if (selectedRange.isCollapsed()) {
                // no selection
				editor.replaceSelection(before + after, 'around');
				selectedRange.startOffset += beforeLength;
				selectedRange.endOffset += afterLength;
				editor.setSelectedRange(selectedRange);
			} else if (selectedText.startsWith(before) && selectedText.endsWith(after)) {
                // selection contains wrapping strings
				editor.replaceSelection(selectedText.substr(beforeLength, selectedText.length - (beforeLength + afterLength)), 'around');
				addedFormatting = false;
			} else {
                // selection does not contain wrapping strings
				var extendedRange = selectedRange.rangeByOffsetting(-beforeLength, beforeLength + afterLength),
					extendedText = extendedRange.textInRange();

				if (extendedText.startsWith(before) && extendedText.endsWith(after)) {
                    // check text before and after selection
					editor.setSelectedRange(extendedRange);
					editor.replaceSelection(selectedText, 'around');
					addedFormatting = false;
				} else {
                    // add wrapping strings
					editor.replaceSelection(before + selectedText + after, 'around');
					editor.setSelectedRange(editor.selectedRange().rangeByOffsetting(beforeLength, -(beforeLength + afterLength)));
				}
			}
		});

		return addedFormatting;
	}

	Extensions.add('com.foldingtext.editor.methods', {
		toggleSelectionFormatting: toggleSelectionFormatting
	});

	Extensions.addCommand({
		name: 'toggleBold',
		performCommand: function (editor) {
			editor.toggleSelectionFormatting(editor, '**');
		}
	});

	Extensions.addCommand({
		name: 'toggleItalic',
		performCommand: function (editor) {
			editor.toggleSelectionFormatting(editor, '_');
		}
	});

	Extensions.addCommand({
		name: 'toggleCode',
		performCommand: function (editor) {
			editor.toggleSelectionFormatting(editor, '`');
		}
	});

	Extensions.addCommand({
		name: 'insertLink',
		performCommand: function (editor) {
			insertLink(editor);
		}
	});

	function formatSelectionType(editor, type, typeIndentLevel, postProcess) {
		var selectedRange = editor.selectedRange(),
			effectedRange = selectedRange.rangeByAdjustingToEffectedNodesInEditor(editor),
			nodes = effectedRange.nodesInRange(),
			tree = editor.tree();

		tree.ensureClassified();

		Selection.performNodeOperationWhileMaintainingEditorStateInRange(editor, selectedRange, effectedRange, function() {
			tree.beginUpdates();
			nodes.forEach(function(each) {
				if (each.type() !== 'empty' || selectedRange.isCollapsed()) {
					var level = typeIndentLevel === undefined ? each.typeIndentLevel() : typeIndentLevel;
					each.setType(type);
					each.setTypeIndentLevel(level);
				}
			});
			if (postProcess) {
				postProcess(nodes);
			}
			tree.endUpdates();
		});
	}

	Extensions.addCommand({
		name: 'formatBody',
		performCommand: function (editor, options) {
			formatSelectionType(editor, 'body', 0);
		}
	});

	Extensions.addCommand({
		name: 'formatBlockquote',
		performCommand: function (editor, options) {
			formatSelectionType(editor, 'blockquote');
		}
	});

	Extensions.addCommand({
		name: 'formatCodeblock',
		performCommand: function (editor, options) {
			formatSelectionType(editor, 'codeblock', 0);
		}
	});

	Extensions.addCommand({
		name: 'formatHeading',
		performCommand: function (editor, options) {
			formatSelectionType(editor, 'heading', options.interfaceItemTag ? options.interfaceItemTag : 1);
		}
	});

	Extensions.addCommand({
		name: 'formatUnorderedList',
		performCommand: function (editor, options) {
			formatSelectionType(editor, 'unordered', undefined);
		}
	});

	Extensions.addCommand({
		name: 'formatOrderedList',
		performCommand: function (editor, options) {
			formatSelectionType(editor, 'ordered', undefined, function(nodes) {
				var label = 1;

				nodes.forEach(function (each) {
					var match = each.line().match(/^(\t*)([0-9]+\.\s+)/);
					if (match) {
						each.replaceLineInRange('' + label, match.index + match[1].length, 1);
						label++;
					}					
				});
			});
		}
	});

	function adjustTypeLevel(editor, delta) {
		var selectedRange = editor.selectedRange(),
			effectedRange = selectedRange.rangeByAdjustingToEffectedNodesInEditor(editor),
			nodes = effectedRange.nodesInRange(),
			tree = editor.tree();

		tree.ensureClassified();

		Selection.performNodeOperationWhileMaintainingEditorStateInRange(editor, selectedRange, effectedRange, function() {
			tree.beginUpdates();
			nodes.forEach(function(each) {
				if (each.type() !== 'empty' || selectedRange.isCollapsed()) {
					each.setTypeIndentLevel(each.typeIndentLevel() + delta);
				}
			});
			tree.endUpdates();
		});
	}

	Extensions.addCommand({
		name: 'increaseTypeLevel',
		performCommand: function (editor, options) {
			adjustTypeLevel(editor, 1);
		}
	});

	Extensions.addCommand({
		name: 'decreaseTypeLevel',
		performCommand: function (editor, options) {
			adjustTypeLevel(editor, -1);
		}
	});

	Extensions.addCommand({
		name: 'clearFormatting',
		performCommand: function (editor, options) {
			var selectedRange = editor.selectedRange(),
				effectedRange = selectedRange.rangeByAdjustingToEffectedNodesInEditor(editor),
				tree = editor.tree(),
				clearedInline = false;

			tree.ensureClassified();

			var formattedText = {
				_format: null,
				_cleared: false,
				_node: null,
				_location: 0,
				_length: 0,
				format: function () {
					return this._format;
				},
				length: function () {
					return this._length;
				},
				string: function () {
					return this._node.lineInRange(this._location, this._length);
				},
				replaceInRange: function (inserted, location, length) {
					this._cleared = true;
					this._node.replaceLineInRange(inserted, this._location + location, length);
				}
			};

			function processExtension(eachExtension) {
				eachExtension(formattedText);
			}

			Selection.performNodeOperationWhileMaintainingEditorStateInRange(editor, selectedRange, effectedRange, function() {
				tree.beginUpdates();

				selectedRange.forEachNodeInRange(function (node, line, location, length, fullySelected) {
					var end = location + length,
						attributesRange = {},
						eachAttributeRange = {},
						index = location,
						attributes;

					while (index <= end) {
						attributes = node.lineAttributesAtIndex(index, attributesRange);
						if (attributes) {
							index = attributesRange.location + attributesRange.length;
							for (var eachAttribute in attributes) {
								node.lineAttributeAtIndex(eachAttribute, index, null, eachAttributeRange);

								formattedText._cleared = false;
								formattedText._format = eachAttribute;
								formattedText._node = node;
								formattedText._location = eachAttributeRange.location;
								formattedText._length = eachAttributeRange.length;

								Extensions.processExtensions('com.foldingtext.editor.clearInlineFormatting', processExtension);

								if (formattedText._cleared) {
									index = eachAttributeRange.location;
									tree.ensureClassified();
									clearedInline = true;
									break;
								}
							}
						} else {
							index++;
						}
					}
				});

				tree.endUpdates();				
			});

			if (!clearedInline) {
				formatSelectionType(editor, 'body', 0);
			}
		}
	});	
});