define(function (require, exports, module) {
	'use strict';

	require('ft/extensions/folding');
	require('ft/extensions/foldingnodes');

	var Extensions = require('ft/core/extensions').Extensions,
		Selection = require('ft/extensions/selection'),
		Node = require('ft/core/node').Node;

	function moveBranches(editor, direction) {
		editor.tree().ensureClassified();

		var selectedRange = editor.selectedRange(),
			branches = Node.commonAncestorsForNodes(selectedRange.rangeByAdjustingToEffectedNodesInEditor(editor).nodesInRange()).nodesInLineOrder(),
			firstBranch = branches[0],
			lastBranch = branches.lastObject(),
			lastEffected = lastBranch.lastDescendantOrSelf(),
			tree = editor.tree(),
			newNextSibling = null,
			newParent = null;

		if (direction === 'up') {
			newParent = firstBranch.parent;
			newNextSibling = firstBranch.previousSibling;
			if (!newNextSibling) {
				return;
			}
		} else if (direction === 'down') {
			newParent = lastBranch.parent;
			newNextSibling = lastBranch.nextSibling;
			if (!newNextSibling) {
				// If no next sibling see if next line after last effected is
				// an empty line. If so (assuming special empty line rules
				// that all taxonomies use so far) then we can move down.
				// Though implement that move by moving the empty line up.
				// Tricky. Probably tooo tricky. Will see.
				var nextLineNode = lastEffected.nextLineNode();
				if (nextLineNode && nextLineNode.type() === 'empty') {
					tree.beginUpdates();
					tree.insertNodeBefore(nextLineNode, firstBranch);
					tree.endUpdates();					
				}
				return;
			}
			newNextSibling = newNextSibling.nextSibling;
		} else if (direction === 'left') {
			newParent = firstBranch.parent.parent;
			newNextSibling = firstBranch.parent.nextSibling;
		} else if (direction === 'right') {
			// When moving right must bring along all previous empty lines. 
			// So this code collect them and adds them to branches. Assumes
			// special empty line taxonomy rules... that empty line takes on
			// type of next non empty line.
			var each = firstBranch.previousLineNode();
			while (each && each.type() === 'empty') {
				branches.unshift(each);
				each = each.previousLineNode();
			}

			firstBranch = branches[0];
			newParent = firstBranch.previousSibling;
		}

		if (!newParent) {
			return;
		}

		var validMove = true;

		branches.forEach(function (each) {
			if (!newParent.validateInsertChildBefore(each, newNextSibling)) {
				validMove = false;
			}
		});

		if (!validMove) {
			return;
		}

		var effectedRange = tree.createRangeFromNodes(firstBranch, 0, lastEffected, lastEffected.line().length);
		Selection.performNodeOperationWhileMaintainingEditorStateInRange(editor, selectedRange, effectedRange, function() {
			newParent.insertChildrenBefore(branches, newNextSibling);
		});
	}

	function moveBranchesLeft(editor) {
		moveBranches(editor, 'left');
	}

	function moveBranchesRight(editor) {
		moveBranches(editor, 'right');
	}

	function moveBranchesUp(editor) {
		moveBranches(editor, 'up');
	}

	function moveBranchesDown(editor) {
		moveBranches(editor, 'down');
	}

	function promoteChildBranches(editor) {
		var selectedRange = editor.selectedRange(),
			node = selectedRange.startNode;

		if (node.firstChild) {
			var tree = editor.tree(),
				children = node.children(),
				lastEffected = children.lastObject().lastDescendantOrSelf(),
				effectedRange = tree.createRangeFromNodes(children[0], 0, lastEffected, lastEffected.line().length);

			Selection.performNodeOperationWhileMaintainingEditorStateInRange(editor, selectedRange, effectedRange, function() {
				node.parent.insertChildrenBefore(children, node.nextSibling);
			});
		}
	}

	function demoteTrailingSiblingBranches(editor) {
		var selectedRange = editor.selectedRange(),
			node = selectedRange.startNode,
			each = node.nextSibling,
			trialingSiblings = [];

		while (each) {
			trialingSiblings.push(each);
			each = each.nextSibling;
		}

		if (trialingSiblings.length) {
			var tree = editor.tree(),
				lastEffected = trialingSiblings.lastObject().lastDescendantOrSelf(),
				effectedRange = tree.createRangeFromNodes(trialingSiblings[0], 0, lastEffected, lastEffected.line().length);

			Selection.performNodeOperationWhileMaintainingEditorStateInRange(editor, selectedRange, effectedRange, function() {
				node.insertChildrenBefore(trialingSiblings, null);
			});
		}
	}

	function deleteBranches(editor) {
		var tree = editor.tree(),
			selectedRange = editor.selectedRange(),
			branches = Node.commonAncestorsForNodes(selectedRange.rangeByAdjustingToEffectedNodesInEditor(editor).nodesInRange()).nodesInLineOrder();

		if (!selectedRange.isCollapsed()) {
			editor.replaceSelection('');
			return;
		}

		if (tree.firstLineNode() === branches[0] && tree.lastLineNode() === branches[branches.length - 1].lastDescendantOrSelf()) {
			tree.setText('');
		} else {
			tree.beginUpdates();
			branches.forEach(function (each) {
				each.removeFromParent();
			});
			tree.endUpdates();			
		}
	}

	Extensions.addCommand({
		name: 'moveBranchesUp',
		performCommand: moveBranchesUp
	});

	Extensions.addCommand({
		name: 'moveBranchesDown',
		performCommand: moveBranchesDown
	});

	Extensions.addCommand({
		name: 'moveBranchesLeft',
		performCommand: moveBranchesLeft
	});

	Extensions.addCommand({
		name: 'moveBranchesRight',
		performCommand: moveBranchesRight
	});

	Extensions.addCommand({
		name: 'promoteChildBranches',
		performCommand: promoteChildBranches
	});

	Extensions.addCommand({
		name: 'demoteTrailingSiblingBranches',
		performCommand: demoteTrailingSiblingBranches
	});

	Extensions.addCommand({
		name: 'deleteBranches',
		performCommand: deleteBranches
	});
});