define(function (require, exports, module) {
	'use strict';

	var Extensions = require('ft/core/extensions').Extensions,
		dom = require('ft/util/dom');

	var Panel = function (editor) {
		var panelElement = document.querySelector('.panel');
		if (!panelElement) {
			panelElement = document.createElement('div');
			panelElement.className = 'panel';
		}

		this._panelElement = panelElement;
		this._editor = editor;
		this._panelEditor = this.createPanelEditor();
		this._ignorePanelEditorChanges = 0;
		this._fields = [];

		var outerThis = this;
		this._panelEditor.addKeyMap({
			'Esc': function () {
				outerThis.hide();
				outerThis._editor.focus();
			},
			'Enter': function () { outerThis.returnPressed(false); },
			'Shift-Enter': function () { outerThis.returnPressed(true); },
			'Tab': function () { outerThis.tabPressed(false); },
			'Shift-Tab': function () { outerThis.tabPressed(true); },
			'Option-Tab': function () { outerThis._panelEditor.execCommand('indent'); }
		});

		this._panelEditor.on('change', this._panelEditorContentChanged.bind(this));
	};

	(function () {

		this.createPanelEditor = function createPanelEditor() {
			return new CodeMirror(this._panelElement, {
				value: '',
				lineWrapping: true,
				indentWithTabs: true,
				viewportMargin: Infinity,
				undoDepth: Infinity,
				indentUnit: 4,
				autoIndent: false,
				fieldsController: this,
				isFieldEditor: true
			});
		};

		//
		// Display
		//

		this.show = function show() {
			var panelElement = this._panelElement;
			if (!panelElement.parentElement) {
				document.body.appendChild(panelElement);
			}
		};

		this.hide = function hide() {
			var panelElement = this._panelElement;
			if (panelElement.parentElement) {
				panelElement.parentElement.removeChild(panelElement);
			}
			this._editor.focus();
		};

		this.remove = function remove() {
			this.hide();
			this._editor._panel = null;
		};

		//
		// Fields
		//

		this.hasField = function hasField(id) {
			for (var i = this._fields.length - 1; i >= 0; i--) {
				if (this._fields[i].id === id) {
					return true;
				}
			}
			return false;
		};

		this.addField = function addField(field) {
			var Pos = CodeMirror.Pos,
				id = field.id,
				label = field.label,
				value = field.value,
				index = (field.index !== undefined) ? field.index : -1,
				panelEditor = this._panelEditor,
				mode = field.mode || null,
				fields = this._fields,
				// get the field immediately preceding the field to be created
				prevField = index > 0,
				posToInsertField = new Pos(panelEditor.firstLine(), 0),
				marker = null;

			if (prevField) {
				posToInsertField = this.posOfFieldValue(index - 1, false);
			}

			this._ignorePanelEditorChanges++;

			panelEditor.replaceRange(label + ' ' + value, posToInsertField);

			var indexOfInsert = panelEditor.indexFromPos(posToInsertField),
				indexOfStart = indexOfInsert + 0,
				indexOfEnd = indexOfStart + label.length,
				posMarkStart = panelEditor.posFromIndex(indexOfStart),
				posMarkEnd = panelEditor.posFromIndex(indexOfEnd);

			marker = panelEditor.doc.markText(posMarkStart, posMarkEnd, {
				atomic: true,
				readOnly: true,
				inclusiveLeft: true,
				inclusiveRight: true,
				className: 'field'
			});

			if (marker) {
				var	newField = {
					id: id,
					marker: marker,
					onChange: field.onChange
				};

				if (index < 0) {
					this._fields.push(newField);
				} else {
					this._fields.splice(index, 0, newField);
				}

				newField.value = this._getFieldValueFromEditor(id);
			} else {
				console.error('Cannot add marker: ', field);
			}

			this._ignorePanelEditorChanges--;
		};

		this.removeField = function removeField(idOrIndex) {
			if (typeof idOrIndex === 'string') {
				idOrIndex = this.indexFromID(idOrIndex);
			}

			if (idOrIndex === -1) {
				return;
			}

			this._ignorePanelEditorChanges++;

			var panelEditor = this._panelEditor,
				startRange = this.posOfFieldMarker(idOrIndex, true),
				endRange = this.posOfFieldValue(idOrIndex, false),
				endRangeLine = panelEditor.getLine(endRange.line),
				field = this._fields.splice(idOrIndex, 1);

			// Clean up editor after removal of field
			// Since a lot of fields are probably going to be in their own line, we would like the field
			// to appear and dissapear without leaving any extra lines within the editor

			// If field was on it's own line, delete the entire line, not just the line's content
			if (startRange.ch === 0 && startRange.line > 0) {
				startRange = panelEditor.posFromIndex(panelEditor.indexFromPos(startRange) - 1);
			}

			// If field removes entire line, remove the trailing newline character too
			if (endRange.ch >= endRangeLine.length - 1) {
				endRange = panelEditor.posFromIndex(panelEditor.indexFromPos(endRange) + 1);
			}

			field[0].marker.clear();

			panelEditor.replaceRange('', startRange, endRange);

			this._ignorePanelEditorChanges--;
		};

		this.posOfFieldMarker = function posOfFieldMarker(index, start) {
			var field = this._fields[index],
				marker = field.marker,
				pos = marker.find();

			if (start) {
				return pos.from;
			} else {
				return pos.to;
			}
		};

		// Note: Currently this system relies on the fact that the order of fields
		// present within the _fields array is the same order present within the fields
		// editor. Please ensure this while adding fields. _field array index can be specified
		// by passing 'index': number within the construction options
		this.posOfFieldValue = function posOfFieldValue(index, start) {
			var panelEditor = this._panelEditor;

			if (start) {
				var markEnd = this.posOfFieldMarker(index, false);
				markEnd = panelEditor.posFromIndex(panelEditor.indexFromPos(markEnd) + 1);
				return markEnd;
			}

			var lastLine = panelEditor.getLine(panelEditor.lastLine()),
				nextFieldStart = new CodeMirror.Pos(panelEditor.lastLine(), lastLine.length),
				fields = this._fields,
				hasNextField = fields.length > index + 1;

			if (hasNextField) {
				nextFieldStart = this.posOfFieldMarker(index + 1, true);
				nextFieldStart = panelEditor.posFromIndex(panelEditor.indexFromPos(nextFieldStart) - 1);
			}

			return nextFieldStart;
		};

		this._getFieldValueFromEditor = function _getFieldValueFromEditor(idOrIndex) {
			if (typeof idOrIndex === 'string') {
				idOrIndex = this.indexFromID(idOrIndex);
			}

			if (idOrIndex < 0) {
				return null;
			}

			var start = this.posOfFieldValue(idOrIndex, true),
				end = this.posOfFieldValue(idOrIndex, false);

			var value = this._panelEditor.getRange(start, end);

			return value.replace(/[\n]+$/, '');	// only remove the trailing newlines
		};

		this.fieldValue = function fieldValue(idOrIndex) {
			if (typeof idOrIndex === 'string') {
				idOrIndex = this.indexFromID(idOrIndex);
			}

			if (idOrIndex < 0) {
				return null;
			}

			var field = this._fields[idOrIndex];
			if (field.value) {
				return field.value;
			} else {
				field.value = this._getFieldValueFromEditor(idOrIndex);
				return field.value;
			}
		};

		this.indexFromID = function indexFromID(id) {
			for (var i = this._fields.length - 1; i >= 0; i--) {
				var field = this._fields[i];
				if (field.id === id) {
					return i;
				}
			}
			return -1;
		};

		this._panelEditorContentChanged = function _panelEditorContentChanged(cm, changeObj) {
			if (this._ignorePanelEditorChanges > 0) {
				return;
			}

			// for now, I think i'll ignore the changeObj and query the value of each field. A bit more expensive, but I think
			// should be simpler to understand and maybe has less edge cases

			var fields = this._fields,
				fieldCount = fields.length;

			for (var i = 0; i < fieldCount; i++) {
				var field = fields[i],
					oldValue = field.value,
					newValue = this._getFieldValueFromEditor(i);

				if (oldValue !== newValue) {
					field.value = newValue;
					if (field.onChange) {
						field.onChange(oldValue, newValue);
					}
					CodeMirror.signal(this, 'fieldvaluechanged', {
						from: oldValue,
						to: newValue,
						id: field.id
					});
				}
			}
		};

		this.focus = function focus(nthField, selectAll) {
			this._panelEditor.focus();
			if (nthField !== undefined && nthField >= 0) {
				this._focusField(nthField, selectAll);
			}
		};

		this._focusField = function _focusField(index, selectAll) {
			if (index < 0) {
				index = 0;
			}

			if (index >= this._fields.length) {
				index = this._fields.length - 1;
			}

			if (selectAll) {
				this._panelEditor.setSelection(this.posOfFieldValue(index, true), this.posOfFieldValue(index, false));
			} else {
				this._panelEditor.setCursor(this.posOfFieldValue(index, true));
			}
		};

		this.currentlyActiveField = function currentlyActiveField() {
			var selStart = this._panelEditor.getCursor('start');

			for (var i = 0; i < this._fields.length; i++) {
				var start = this.posOfFieldValue(i, true),
					end = this.posOfFieldValue(i, false);

				if (selStart.line >= start.line && selStart.ch >= start.ch && selStart.line <= end.line && selStart.ch <= end.ch) {
					return i;
				}
			}

			return -1;
		};

		this.returnPressed = function returnPressed(withShiftKey) {
			var fieldIndex = this.currentlyActiveField();

			if (fieldIndex < 0) {
				return;
			}

			var field = this._fields[fieldIndex];

			CodeMirror.signal(this, 'returnpressed', {
				id: field.id,
				shiftKey: withShiftKey || false
			});
		};

		this.tabPressed = function tabPressed(withShiftKey) {
			var active = this.currentlyActiveField();

			if (withShiftKey) {
				// move back
				if (active > 0) {
					this._focusField(active - 1);
				} else if (active === 0) {
					// move to main editor
					this._editor.focus();
				}
			} else {
				var lastFieldIndex = this._fields.length - 1;
				if (active < lastFieldIndex) {
					this._focusField(active + 1);
				} else if (active === lastFieldIndex) {
					this._editor.focus();
				}
			}
		};

	}).call(Panel.prototype);

	Extensions.add('com.foldingtext.editor.methods', {
		panel: function () {
			var panel = this._panel;
			if (!this._panel) {
				panel = new Panel(this);
				this._panel = panel;
			}
			return panel;
		}
	});
});