define(function (require, exports, module) {
	'use strict';

	var Extensions = require('ft/core/extensions').Extensions;

	function performNodeOperationWhileMaintainingEditorStateInRange(editor, selectedRange, effectedRange, nodeOperation) {
		editor.operation(function () {
			var lastMoved = effectedRange.endNode,
				folds = editor.foldsInRange(effectedRange),
				foldRanges = [];

			selectedRange.beginTrackingOffsets();

			folds.forEach(function (each) {
				var foldRange = each.range();
				foldRange.isManualFoldRange = each.isManual();
				foldRanges.push(foldRange);
				foldRange.beginTrackingOffsets();
				each.clear();
			});

			nodeOperation();

			foldRanges.forEach(function (eachRange) {
				eachRange.endTrackingOffsets();
				editor.createFoldInRange(eachRange, eachRange.isManualFoldRange);
			});

			selectedRange.endTrackingOffsets();

			// Fix selection case where line is fully selected and nodes are
			// moved vertically. In that case the selection end node can get
			// left behind. This fixes that case.
			if (selectedRange.startNode !== selectedRange.endNode && selectedRange.endOffset === 0) {
				var afterLast = effectedRange.endNode.nextLineNode();
				if (afterLast) {
					selectedRange.endNode = afterLast;
				} else {
					selectedRange.endNode = lastMoved;
					selectedRange.endOffset = lastMoved.line().length;
				}
			}

			editor.setSelectedRange(selectedRange);
		});
	}

	function expandSelection(editor) {
		var expandSelectionStartRange = editor._expandSelectionStartRange,
			extendSelectionLevel = editor._extendSelectionLevel;

		if (!expandSelectionStartRange) {
			expandSelectionStartRange = editor.selectedRange().copy();
			extendSelectionLevel = 0;
		}

		var previousExpandToRange = expandSelectionStartRange.rangeByExtendingToLevel(extendSelectionLevel),
			expandToRange = previousExpandToRange.copy();

		while (!expandToRange.isSelectingAll() && previousExpandToRange.isEqualToRange(expandToRange)) {
			expandToRange = expandSelectionStartRange.rangeByExtendingToLevel(++extendSelectionLevel);
		}

		editor.setSelectedRange(expandToRange);
		editor._expandSelectionStartRange = expandSelectionStartRange;
		editor._extendSelectionLevel = extendSelectionLevel;
	}

	function contractSelection(editor) {
		var expandSelectionStartRange = editor._expandSelectionStartRange,
			extendSelectionLevel = editor._extendSelectionLevel;

		if (!extendSelectionLevel) {
			return;
		}

		var previousContractToRange = expandSelectionStartRange.rangeByExtendingToLevel(extendSelectionLevel),
			contractToRange = previousContractToRange.copy();

		while (extendSelectionLevel > 0 && previousContractToRange.isEqualToRange(contractToRange)) {
			contractToRange = expandSelectionStartRange.rangeByExtendingToLevel(--extendSelectionLevel);
		}

		editor.setSelectedRange(contractToRange);
		editor._extendSelectionLevel = extendSelectionLevel;
		editor._expandSelectionStartRange = expandSelectionStartRange;
	}

	Extensions.addCommand({
		name: 'expandSelectionToWord',
		performCommand: function (editor) {
			editor.setSelectedRange(editor.selectedRange().rangeByExtendingToWord());
		}
	});

	Extensions.addCommand({
		name: 'expandSelectionToSentence',
		performCommand: function (editor) {
			editor.setSelectedRange(editor.selectedRange().rangeByExtendingToSentence());
		}
	});

	Extensions.addCommand({
		name: 'expandSelectionToLine',
		performCommand: function (editor) {
			editor.setSelectedRange(editor.selectedRange().rangeByExtendingToLine());
		}
	});

	Extensions.addCommand({
		name: 'expandSelectionToParagraph',
		performCommand: function (editor) {
			editor.setSelectedRange(editor.selectedRange().rangeByExtendingToParagraph());
		}
	});

	Extensions.addCommand({
		name: 'expandSelectionToBranch',
		performCommand: function (editor) {
			editor.setSelectedRange(editor.selectedRange().rangeByExtendingToBranch());
		}
	});

	Extensions.addCommand({
		name: 'expandSelection',
		performCommand: expandSelection
	});

	Extensions.addCommand({
		name: 'contractSelection',
		performCommand: contractSelection
	});

	Extensions.add('com.foldingtext.editor.beforeSelectionChange', function (editor, candidateSelection) {
		editor._expandSelectionStartRange = null;
		editor._extendSelectionLevel = 0;
	});

	exports.performNodeOperationWhileMaintainingEditorStateInRange = performNodeOperationWhileMaintainingEditorStateInRange;
});