define(function (require, exports, module) {
	'use strict';

	/*jshint bitwise: false*/

	require('ft/extensions/current');

	var Extensions = require('ft/core/extensions').Extensions,
		Taxonomy = require('ft/core/taxonomy').Taxonomy,
		NodeSet = require('ft/core/nodeset').NodeSet,
		NSTextCheckingTypeOrthography = 1 << 0, // language identification
		NSTextCheckingTypeSpelling = 1 << 1, // spell checking
		NSTextCheckingTypeGrammar = 1 << 2, // grammar checking
		NSTextCheckingTypeDate = 1 << 3, // date/time detection
		NSTextCheckingTypeAddress = 1 << 4, // address detection
		NSTextCheckingTypeLink = 1 << 5, // link detection
		NSTextCheckingTypeQuote = 1 << 6, // smart quotes
		NSTextCheckingTypeDash = 1 << 7, // smart dashes
		NSTextCheckingTypeReplacement = 1 << 8, // fixed replacements, such as copyright symbol for (c)
		NSTextCheckingTypeCorrection = 1 << 9, // autocorrection
		NSTextCheckingTypeRegularExpression = 1 << 10, // regular expression matches
		NSTextCheckingTypePhoneNumber = 1 << 11, // phone number detection
		NSTextCheckingTypeTransitInformation = 1 << 12; // transit (e.g. flight) info detection

	function attributeNameForTextCheckingMatchType(type) {
		switch (type) {
		case NSTextCheckingTypeOrthography:
			return 'language';
		case NSTextCheckingTypeSpelling:
			return 'spelling';
		case NSTextCheckingTypeGrammar:
			return 'grammar';
		case NSTextCheckingTypeDate:
			return 'date';
		case NSTextCheckingTypeAddress:
			return 'address';
		case NSTextCheckingTypeLink:
			return 'smartlink';
		case NSTextCheckingTypeQuote:
			return 'quote';
		case NSTextCheckingTypeDash:
			return 'dash';
		case NSTextCheckingTypeReplacement:
			return 'replacement';
		case NSTextCheckingTypeCorrection:
			return 'correction';
		case NSTextCheckingTypeRegularExpression:
			return 'expression';
		case NSTextCheckingTypePhoneNumber:
			return 'phonenumber';
		case NSTextCheckingTypeTransitInformation:
			return 'transitinformation';
		}
	}

	function checkSpelling(editor, options) {
		var text = editor.textContent(),
			range = editor.cursorRange(),
			nextRangeString = checkSpellingOfString(text, range.location() + range.length(), true),
			nextRange = nextRangeString ? editor.tree().createRangeFromString(nextRangeString) : null;

		function updateSpellingPanelWithMisspelledWord() {
			if (window.nativeViewController) {
				/*jshint camelcase: false */
				window.nativeViewController.javascriptUpdateSpellingPanelWithMisspelledWord_(editor.selectedText());
				/*jshint camelcase: true */
			}
		}

		if (nextRange) {
			editor.removeFolds(editor.foldsInRange(nextRange));
			editor.setSelectedRange(nextRange);
			updateSpellingPanelWithMisspelledWord(editor.selectedText());
		} else {
			updateSpellingPanelWithMisspelledWord(editor.selectedText());
			editor.beep();
		}
	}

	function checkSpellingOfString(string, from, wrap) {
		if (window.nativeViewController) {
			/*jshint camelcase: false */
			var result = window.nativeViewController.javascriptCheckSpellingOfString_startingAt_wrap_(string, from, wrap);
			/*jshint camelcase: true */
			return result;
		}
		return null;
	}

	Extensions.addInit(function (editor) {
		editor._textCheckingNodes = NodeSet.setFromArray(editor.viewportRange().nodesInRange());
		editor._textCheckingPendingNodes = new NodeSet();
		window.handleTextCheckingResults = function (nodeID, nodeText, jsonResults) {
			editor.handleTextCheckingResults(nodeID, nodeText, jsonResults);
		};
		window.selectionHasSpellingError = function () {
			var range = editor.selectedRange(),
				length = range.length();

			if (length === 0) { return false; }
			if (range.startNode !== range.endNode) { return false; }

			var effectiveRange = {},
				longestEffectiveRange = {};
			if (range.startNode.lineAttributeAtIndex('spelling', range.startOffset, effectiveRange, longestEffectiveRange)) {
				if (longestEffectiveRange.location === range.startOffset && longestEffectiveRange.length === length) {
					return true;
				}
			}

			return false;
		};
	});

	Extensions.add('com.foldingtext.editor.viewportNodesChanged', function (editor, nodes, removedNodes, addedNodes) {
		removedNodes.forEachNodeInSet(function (node) {
			editor.endTextChecking(node);
		});
		addedNodes.forEachNodeInSet(function (node) {
			editor.beginTextChecking(node);
		});
	});

	Extensions.addTreeChanged(function (editor, e) {
		if (editor._isApplyingTextCheckingChanges) {
			return;
		}

		var deltas = e.deltas,
			deletasLength = deltas.length,
			textCheckingNodes = editor._textCheckingNodes,
			eachDelta,
			i;

		for (i = 0; i < deletasLength; i++) {
			eachDelta = deltas[i];

			if (!eachDelta.textChanged()) {
				// If only attributes changes skip.
				// Text checking only cares about text changes
				continue;
			}

			var updatedNode = eachDelta.updatedNode,
				each;

			if (updatedNode) {
				// If changed node is current then mark as changed allow currentWordDidChange to schedule checks.
				if (updatedNode === editor._textCheckingCurrentNode) {
					var effectiveRange = {},
						longestEffectiveRange = {},
						originalSpellingText = updatedNode.lineAttributeAtIndex('spelling', eachDelta.fromLineCh.ch, effectiveRange, longestEffectiveRange);

					if (originalSpellingText) {
						var longRangeLocation = longestEffectiveRange.location,
							longRangeLength = longestEffectiveRange.length,
							newSpellingText = updatedNode.lineInRange(longRangeLocation, longRangeLength);

						if (originalSpellingText !== newSpellingText) {
							var regex = /(\s*)(\S+)(\s*)/,
								matches = newSpellingText.match(regex),
								preWord = matches[1],
								word = matches[2],
								postWord =  matches[3];

							if (originalSpellingText !== word) {
								updatedNode.removeLineAttributeInRange('spelling', longRangeLocation, longRangeLength);
							} else {
								if (preWord) {
									updatedNode.removeLineAttributeInRange('spelling', longRangeLocation, preWord.length);
								}
								if (postWord) {
									updatedNode.removeLineAttributeInRange('spelling', longRangeLocation + preWord.length + word.length, postWord.length);
								}
							}
						}
					}
					editor._textCheckingPendingNodes.addNode(updatedNode);
				// Else schedule check if its in set of nodes that are activly being checked
				} else if (textCheckingNodes.containsNode(updatedNode)) {
					editor.scheduleCheckTextInNode(updatedNode);
				}
			}
		}
	});

	Extensions.add('com.foldingtext.editor.currentWordDidChange', function (editor, currentWordRange) {
		var currentNode = editor._textCheckingCurrentNode,
			newCurrentNode = currentWordRange ? currentWordRange.startNode : undefined,
			scheduled = false;

		if (newCurrentNode !== currentNode) {
			editor._textCheckingCurrentNode = newCurrentNode;
			editor.scheduleCheckPendingNodes();
		} else {
			var currentWordStart = editor._textCheckingCurrentWordStart,
				currentWordEditedText = editor._textCheckingCurrentWordEditedText,
				newCurrentWordStart = currentWordRange ? currentWordRange.startOffset : undefined,
				newCurrentWordText = currentWordRange ? currentWordRange.textInRange() : undefined;

			// Current word changed.
			if (newCurrentWordStart !== currentWordStart) {
				editor._textCheckingCurrentWordStart = newCurrentWordStart;
				editor._textCheckingCurrentWordEditedText = undefined;

				// Schedule pending check if needed
				editor.scheduleCheckPendingNodes();

			// Current word edited.
			} else if (newCurrentWordText !== currentWordEditedText) {
				editor._textCheckingCurrentWordEditedText = newCurrentWordText;

				// Remove checking attributes in word range.
				if (currentWordRange) {
					newCurrentNode.removeLineAttributesInRange(editor._textCheckingAttributes, currentWordRange.startOffset, newCurrentWordText.length);
				}
			}
		}
	});

	Extensions.add('com.foldingtext.editor.methods', {
		beginTextChecking: function (node) {
			this._textCheckingNodes.addNode(node);
			this.scheduleCheckTextInNode(node);
		},

		scheduleCheckPendingNodes: function () {
			var checkingNodes = this._textCheckingNodes,
				pendingNodes = this._textCheckingPendingNodes,
				outerThis = this;

			pendingNodes.forEachNodeInSet(function (each) {
				if (checkingNodes.containsNode(each)) {
					outerThis.scheduleCheckTextInNode(each);
				}
			});

			pendingNodes.removeAllNodes();
		},

		scheduleCheckTextInNode: function (node) {
			if (this._textCheckingTypes === 0) {
				return;
			}

			if (window.nativeViewController) {
				/*jshint camelcase: false */
				window.nativeViewController.javascriptScheduleCheckNodeString_nodeID_string_(this._textCheckingTypes, node.id, node.line());
				/*jshint camelcase: true */
			} else {
				// just testing...
				var outerThis = this,
					nodeID = node.id,
					line = node.line();
				window.setTimeout(function () {
					var index = line.indexOf('teh'),
						results = [];

					if (index !== -1) {
						results = [NSTextCheckingTypeSpelling, '', index, 3];
					} else {
						index = line.indexOf('Затеивать');
						if (index !== -1) {
							results = [NSTextCheckingTypeSpelling, '', index, 9];
						}
					}

					outerThis.handleTextCheckingResults(nodeID, line, JSON.stringify(results));
				}, 100);
			}
		},

		handleTextCheckingResults: function (nodeID, nodeText, jsonResults) {
			var node = this._textCheckingNodes.nodeForID(nodeID);
			if (node) {
				var results = JSON.parse(jsonResults),
					length = results.length,
					tree = this.tree(),
					i;

				if (node.line() !== nodeText) {
					// node text changed while background check was in progress.
					// Schedule new check and return
					this.scheduleCheckTextInNode(node);
					return;
				}

				this._isApplyingTextCheckingChanges = true;
				tree.beginUpdates();

				node.removeLineAttributesInRange(this._textCheckingAttributes, 0, node.line().length);
				for (i = 0; i < length; i += 4) {
					var matchType = results[i],
						matchValue = results[i + 1],
						matchLocation = results[i + 2],
						matchLength = results[i + 3];
					if (!matchValue) {
						matchValue = node.lineInRange(matchLocation, matchLength);
					}

					if (node === this._textCheckingCurrentNode) {
						var matchEnd = matchLocation + matchLength,
							currentWordStart = this._textCheckingCurrentWordStart,
							currentWordText = this._textCheckingCurrentWordEditedText;

						// Don't allow text checking marks on current edited word
						if (matchLocation === currentWordStart && !!currentWordText && currentWordText === this._textCheckingCurrentWordEditedText) {
							this._textCheckingPendingNodes.addNode(node);
							continue;
						}
					}

					node.addLineAttributeInRange(attributeNameForTextCheckingMatchType(matchType), matchValue, matchLocation, matchLength);
				}

				tree.endUpdates();
				this._isApplyingTextCheckingChanges = false;
			}
		},

		endTextChecking: function (node) {
			node.removeLineAttributesInRange(this._textCheckingAttributes, 0, node.line().length);
			this._textCheckingNodes.removeNode(node);
		},

		textCheckingTypes: function () {
			return this._textCheckingTypes;
		},

		redoTextChecking: function () {
			var tree = this.tree(),
				outerThis = this;

			tree.beginUpdates();

			this._textCheckingNodes.forEachNodeInSet(function (each) {
				outerThis.scheduleCheckTextInNode(each);
			});

			tree.endUpdates();
		},

		setTextCheckingTypes: function (textCheckingTypes) {
			if (this._textCheckingTypes !== textCheckingTypes) {
				var tree = this.tree(),
					oldTextCheckingAttributes = this._textCheckingAttributes,
					newTextCheckingAttributes = [],
					outerThis = this;

				this._textCheckingTypes = textCheckingTypes;
				this._textCheckingAttributes = newTextCheckingAttributes;

				if (textCheckingTypes & NSTextCheckingTypeOrthography) {
					newTextCheckingAttributes.push('language');
				}

				if (textCheckingTypes & NSTextCheckingTypeSpelling) {
					newTextCheckingAttributes.push('spelling');
				}

				if (textCheckingTypes & NSTextCheckingTypeGrammar) {
					newTextCheckingAttributes.push('grammar');
				}

				if (textCheckingTypes & NSTextCheckingTypeDate) {
					newTextCheckingAttributes.push('date');
				}

				if (textCheckingTypes & NSTextCheckingTypeAddress) {
					newTextCheckingAttributes.push('address');
				}

				if (textCheckingTypes & NSTextCheckingTypeLink) {
					newTextCheckingAttributes.push('smartlink');
				}

				if (textCheckingTypes & NSTextCheckingTypeQuote) {
					newTextCheckingAttributes.push('quote');
				}

				if (textCheckingTypes & NSTextCheckingTypeDash) {
					newTextCheckingAttributes.push('dash');
				}

				if (textCheckingTypes & NSTextCheckingTypeReplacement) {
					newTextCheckingAttributes.push('replacement');
				}

				if (textCheckingTypes & NSTextCheckingTypeCorrection) {
					newTextCheckingAttributes.push('correction');
				}

				if (textCheckingTypes & NSTextCheckingTypeRegularExpression) {
					newTextCheckingAttributes.push('expression');
				}

				if (textCheckingTypes & NSTextCheckingTypePhoneNumber) {
					newTextCheckingAttributes.push('phonenumber');
				}

				if (textCheckingTypes & NSTextCheckingTypeTransitInformation) {
					newTextCheckingAttributes.push('transitinformation');
				}

				if (this.initialized) {
					tree.beginUpdates();

					this._textCheckingNodes.forEachNodeInSet(function (each) {
						each.removeLineAttributesInRange(oldTextCheckingAttributes, 0, each.line().length);
					});

					this.redoTextChecking();

					tree.endUpdates();
				}
			}
		}
	});

	Extensions.add('com.foldingtext.editor.options', {
		name: 'continuousSpellChecking',
		defaultValue: true,
		onChange: function (editor, oldValue, newValue) {
			if (newValue) {
				editor.setTextCheckingTypes(editor.textCheckingTypes() | NSTextCheckingTypeSpelling);
			} else {
				editor.setTextCheckingTypes(editor.textCheckingTypes() & ~NSTextCheckingTypeSpelling);
			}
			editor.invalidateRestorableState();
		}
	});

	Extensions.addCommand({
		name: 'toggleContinuousSpellChecking',
		validate: function (editor, options) {
			options.interfaceItemState = editor.option('continuousSpellChecking');
			return true;
		},
		performCommand: function (editor) {
			editor.setOption('continuousSpellChecking', !editor.option('continuousSpellChecking'));
		}
	});

	Extensions.addCommand({
		name: 'checkSpelling',
		performCommand: checkSpelling
	});

	Extensions.addCommand({
		name: 'redoTextChecking',
		performCommand: function (editor) {
			editor.redoTextChecking();
		}
	});
});