define(function (require, exports, module) {
	'use strict';

	require('ft/util/parallel');

	var Extensions = require('ft/core/extensions').Extensions,
		Parallel = window.Parallel,
		textContentProcessQueue = [],
		textContentWordCount = 0,
		textContent = '',
		element;

	function formatNumber(number) {
		return number.toString().replace(/\B(?=(\d{3})+(?!\d))/g, ',');		
	}

	function updateDisplay(editor) {
		if (editor.option('wordCount')) {
			if (!element) {
				element = document.createElement('div');
				document.body.appendChild(element);
			}

			var selectedRange = editor.selectedRange(),
				newInnerHTML;

			if (selectedRange.isCollapsed()) {
				element.className = 'wordCount';
				newInnerHTML = formatNumber(textContentWordCount) + ' Words<br/>';
			} else {
				var selectedText = selectedRange.textInRange(),
					selectedWordCount = calculateWordCountDelta({
					insertedText : selectedText,
					replacedText : ''
				});
				element.className = 'wordCount wordCountSelected';
				newInnerHTML = formatNumber(selectedWordCount) + ' Words<br/>';
			}

			if (newInnerHTML !== element.innerHTML) {
				element.innerHTML = newInnerHTML;
			}
		} else if (element) {
			element.parentNode.removeChild(element);
			element = null;
		}
	}

	function wordTextForRange(start, length) {
		var end = start + length,
			lengthMinusOne = textContent.length - 1,
			ch;

		while (start > 0) {
			start--;
			ch = textContent[start];
			if (ch === ' ' || ch === '\n') {
				break;
			}
		}

		while (end < lengthMinusOne) {
			end++;
			ch = textContent[end];
			if (ch === ' ' || ch === '\n') {
				break;
			}
		}

		return textContent.substr(start, end - start);
	}

	function calculateWordCountDelta(data) {
		//var wordCountRegex = /[^~`!¡@#$%^&*()_\-+={}\[\]|\\:;"<,>.?¿\/\s]+/g;
		// http://stackoverflow.com/questions/2315488/using-javascript-how-can-i-count-a-mix-of-asian-characters-and-english-words
		var wordRegex = new RegExp(
			'[A-Za-z0-9_\'\u00C0-\u017F]+|'+ // ASCII letters (and accents)
			'[\u3040-\u309F]+|'+ // Hiragana
			'[\u30A0-\u30FF]+|'+ // Katakana
			'[\u4E00-\u9FFF\uF900-\uFAFF\u3400-\u4DBF]', // Single CJK ideographs
			'g');
		
		function wordCount(text) {
			var match = text.match(wordRegex);
			return match ? match.length : 0;
		}

		var replacedWordCount = wordCount(data.replacedText),
			insertedWordCount = wordCount(data.insertedText);
		return insertedWordCount - replacedWordCount;
	}

	function processNext(editor) {
		textContentProcessQueue[0].spawn(calculateWordCountDelta).then(function (wordCountDelta) {
			textContentProcessQueue.shift();
			textContentWordCount += wordCountDelta;
			if (textContentProcessQueue.length === 0) {
				updateDisplay(editor);
			} else {
				processNext(editor);
			}
		});
	}

	function scheduleProcessing(editor, location, replacedText, insertedText) {
		var replacedWordRangeText = wordTextForRange(location, replacedText.length);
		textContent = textContent.replaceRange(location, location + replacedText.length, insertedText);
		var insertedWordRangeText = wordTextForRange(location, insertedText.length);

		textContentProcessQueue.push(new Parallel({
			replacedText : replacedWordRangeText,
			insertedText : insertedWordRangeText
		}));

		if (textContentProcessQueue.length === 1) {
			processNext(editor);
		}
	}

	Extensions.addInit(function (editor) {
		scheduleProcessing(editor, 0, 0, editor.textContent());
	});

	Extensions.addTreeChanged(function (editor, e) {
		var deltas = e.deltas,
			deletasLength = deltas.length;
		for (var i = 0; i < deletasLength; i++) {
			var eachDelta = deltas[i];
			if (eachDelta.textChanged()) {
				scheduleProcessing(editor, eachDelta.location, eachDelta.replacedText, eachDelta.insertedText);
			}
		}
	});

	Extensions.addSelectionDidChange(function (editor) {
		updateDisplay(editor);
	});

	/*jshint multistr:true */
	Extensions.add('com.foldingtext.editor.styles',
		'.wordCount {\
			pointer-events: none;\
			position: fixed;\
			bottom: 0.5rem;\
			right: 1.5rem;\
			color: @secondaryTextColor;\
		}\
		.wordCountSelected {\
			color: @secondaryTextColor;\
		}'
	);
	/*jshint multistr:false */

	Extensions.add('com.foldingtext.editor.options', {
		name: 'wordCount',
		defaultValue: true,
		onChange: function (editor, oldValue, newValue) {
			updateDisplay(editor);
		}
	});

	Extensions.addCommand({
		name: 'toggleWordCount',
		validate: function (editor, options) {
			options.interfaceItemState = editor.option('wordCount');
			return true;
		},
		performCommand: function (editor) {
			editor.setOption('wordCount', !editor.option('wordCount'));
		}
	});

});