define(function (require, exports, module) {
	'use strict';

	var Extensions = require('ft/core/extensions').Extensions,
		CodeMirror = require('codemirror').CodeMirror,
		DateUtils = require('ft/util/date'),
		Node = require('ft/core/node').Node,
		replaceLocationRE = /[\-+*]/,
		shouldRenderNode = function shouldRenderNode(node) {
			return node.modeContext() === 'todo' && node.type() === 'unordered';
		},
		createArchiveNode = function createArchiveNode(tree) {
			return tree.createNode('# Archive');
		};

	Extensions.addMode({
		name: 'todo'
	});

	Extensions.addCommand({
		name: 'done',
		description: 'Toggle @done tag for selected lines.',
		performCommand: function (editor, options) {
			var node = options ? options.node : null,
				range = node ? node.tree.createRangeFromNodes(node, 0) : null;
			editor.toggleTag('done', new Date().format('isoDate'), range);
		}
	});

	Extensions.addCommand({
		name: 'archive @done',
		description: 'Move @done items to the archive section.',
		performCommand: function (editor, options) {
			var tree = editor.tree();

			tree.beginUpdates();
			tree.ensureClassified();

			var archiveNode = tree.evaluateNodePath("//@line:text = Archive")[0];
			if (!archiveNode) {
				archiveNode = createArchiveNode(tree);
				tree.appendNode(archiveNode);
			}

			Node.commonAncestorsForNodes(tree.evaluateNodePath("//@done except //@line:text = Archive//@done")).forEachNodeInSet(function(node) {
				archiveNode.insertChildBefore(node, archiveNode.firstChild);
			});

			tree.endUpdates();
		}
	});

	Extensions.addRenderNode(function (editor, node, nodeRenderer) {
		if (shouldRenderNode(node)) {
			var match = replaceLocationRE.exec(node.line());
			if (match) {
				var widget =  document.createElement('span'),
					input = document.createElement('input'),
					label = document.createElement('label'),
					checkboxID = node.id + '-todocheckbox';

				input.id = checkboxID;
				input.setAttribute('type', 'checkbox');
				input.checked = node.hasTag('done');

				label.setAttribute('for', checkboxID);

				CodeMirror.on(input, 'mousedown', function (e) {
					editor.performCommand('done', { node : node });
					e.preventDefault();
				});

				widget.appendChild(input);
				widget.appendChild(label);
				widget.className = 'cm-checkbox-renderwidget';
				nodeRenderer.renderInlineWidget(match.index, 1, widget);
			}
		}
	});

	function configure(shouldRenderNodeOverride, createArchiveNodeOverride) {
		if (shouldRenderNodeOverride) {
			shouldRenderNode = shouldRenderNodeOverride;			
		}
		if (createArchiveNodeOverride) {
			createArchiveNode = createArchiveNodeOverride;
		}
	}

	return configure;
});