//
// Copyright (c) 2014 Jesse Grosjean. All rights reserved.
//
// Permission is given for you to use this software ONLY while creating themes
// and plugins for the "FoldingText" software platform. Permission is NOT
// given for any other purpose.
//
define(function (require, exports, module) {
	'use strict';

	var CriticMarkupHelper = require('ft/taxonomy/helpers/criticmarkuphelper'),
		MagicLinkHelper = require('ft/taxonomy/helpers/magiclinkhelper'),
		HTMLHelper = require('ft/taxonomy/helpers/htmlhelper'),
		Extensions = require('ft/core/extensions').Extensions;


	/*jshint regexp: false */

	var CODE_SYNTAX_CHAR = '\uE000',
		CODE_CONTENT_SYNTAX_CHAR = '\uE001',
		ESCAPE_SYNTAX_CHAR = '\uE002',
		STRONG_SYNTAX_CHAR = '\uE003',
		EM_SYNTAX_CHAR = '\uE004',
		AUTOLINK_SYNTAX_CHAR = '\uE005',
		IGNORE_CHAR = '\uE006',
		doubleStrongSyntaxChar = STRONG_SYNTAX_CHAR + STRONG_SYNTAX_CHAR,
		escaptedSyntaxRegex = new RegExp(
			ESCAPE_SYNTAX_CHAR + '{2}|' +
			CODE_SYNTAX_CHAR + '+|' +
			STRONG_SYNTAX_CHAR + '+|' +
			EM_SYNTAX_CHAR + '+|' +
			AUTOLINK_SYNTAX_CHAR + '+', 'g'
		);

	var codeReplaceRegex = /(^|[^\\])(`+)([^\r]*?[^`])\2(?!`)/g,
		strongReplaceRegex = /(\*\*|__)(?=\S)([^\r]*?\S[*_]*)\1/g,
		emReplaceRegex = /(\*|_)(?=\S)([^\r]*?\S)\1/g,
		underscoreInWordRegex = /(\w)_(\w)/g,
		referenceLinkRegex = /(!?\[((?:\[[^\]]*\]|[^\[\]])*)\][ ]?(?:\n[ ]*)?\[(.*?)\])()()()()/g,
		inlineLineRegex = /(!?\[((?:\[[^\]]*\]|[^\[\]])*)\]\([ \t]*()<?((?:\([^)]*\)|[^()])*?)>?[ \t]*((['"])(.*?)\6[ \t]*)?\))/g,
		footnoteLinkRE = /\[\^([^\]]*)\]/g,
		webAutoLinkReplaceRegex = /<((https?|ftp|dict):[^'">\s]+)>/gi,
		emailAutoLinkReplaceRegex = /<(mailto:)?([\-.\w]+\@[\-a-z0-9]+(\.[\-a-z0-9]+)*\.[a-z]+)>/gi,
		linkLabelRE = /(?:(['"])[^'"]+\1)|(\(.*\))/,
		textRE = /^[^!\[\]*_\\<>` "'(.@]+/,
		attributesToClear = [
			'keyword',
			'code',
			'strong',
			'em',
			'linktarget',
			'linklabel',
			'linktitle',
			'footnotelink',
			'autolink',
			'referencelink',
			'inlinelink'
		].concat(CriticMarkupHelper.attributesToClear).concat(HTMLHelper.attributesToClear).concat(MagicLinkHelper.attributesToClear);

	function _pad(padChar, num) {
		var result = '';
		for (var i = 0; i < num; i++) {
			result += padChar;
		}
		return result;
	}

	function _escapeCodeSpanMatch(wholeMatch, m1, m2, m3, m4) {
		var codeSyntaxString = _pad(CODE_SYNTAX_CHAR, m2.length);
		var escapeCharString = _pad(CODE_CONTENT_SYNTAX_CHAR, m3.length);
		return m1 + codeSyntaxString + escapeCharString + codeSyntaxString;
	}

	function parseInlines(text, state, previousState, htmlMode, parseCriticMarkup, parseMagicLinks, inwordunderscores) {
		var escapedText = text.escapedText(),
			match;

		if (parseCriticMarkup) {
			CriticMarkupHelper.parseCriticMarkup(text, state, previousState);
		}

		var hasPotentialCode = escapedText.indexOf('`') !== -1;
		if (hasPotentialCode) {
			escapedText = escapedText.replace(codeReplaceRegex, _escapeCodeSpanMatch);
		}

		var hasPotentialStrong = escapedText.indexOf('**') !== -1 || escapedText.indexOf('__') !== -1;
		if (hasPotentialStrong) {
			escapedText = escapedText.replace(strongReplaceRegex, doubleStrongSyntaxChar + '$2' + doubleStrongSyntaxChar);
		}

		var hasUnderscore = escapedText.indexOf('_') !== -1,
			hasPotentialEmphasis = hasUnderscore || escapedText.indexOf('*') !== -1;
		if (hasPotentialEmphasis) {
			if (hasUnderscore && !inwordunderscores) {
				escapedText = escapedText.replace(underscoreInWordRegex, '$1' + IGNORE_CHAR + '$2');
			}
			escapedText = escapedText.replace(emReplaceRegex, EM_SYNTAX_CHAR + '$2' + EM_SYNTAX_CHAR);
		}

		var hasPotentialAutoLink = escapedText.indexOf('>') !== -1;
		if (hasPotentialAutoLink) {
			webAutoLinkReplaceRegex.lastIndex = 0;
			escapedText = escapedText.replace(webAutoLinkReplaceRegex, AUTOLINK_SYNTAX_CHAR + '$1' + AUTOLINK_SYNTAX_CHAR);
			hasPotentialAutoLink = escapedText.indexOf('>') !== -1;
			if (hasPotentialAutoLink) {
				emailAutoLinkReplaceRegex.lastIndex = 0;
				escapedText = escapedText.replace(emailAutoLinkReplaceRegex, AUTOLINK_SYNTAX_CHAR + '$1' + '$2' + AUTOLINK_SYNTAX_CHAR);
			}
		}

		if (text.text() !== escapedText) {
			var codeStart = -1;
			var strongStart = -1;
			var italicStart = -1;
			var autlinkStart = -1;
			var contentString;
			var keyword;

			while ((match = escaptedSyntaxRegex.exec(escapedText))) {
				var escaptedSyntax = match[0];
				var escaptedSyntaxChar = escaptedSyntax[0];
				var index = match.index;

				if (escaptedSyntaxChar === ESCAPE_SYNTAX_CHAR) {
					text.addAttributeInRange('keyword', '\\', index, 1);
				} else if (escaptedSyntaxChar === CODE_SYNTAX_CHAR) {
					if (codeStart >= 0) {
						contentString = text.textSubstring(codeStart, index + escaptedSyntax.length);
						keyword = contentString.substring(0, 2);
						text.addAttributeInRange('keyword', keyword, codeStart, escaptedSyntax.length);
						text.addAttributeInRange('code', contentString.substring(escaptedSyntax.length, contentString.length - escaptedSyntax.length), codeStart, contentString.length);
						text.addAttributeInRange('keyword', keyword, codeStart + contentString.length - escaptedSyntax.length, escaptedSyntax.length);
						codeStart = -1;
					} else {
						codeStart = match.index;
					}
				} else if (escaptedSyntaxChar === STRONG_SYNTAX_CHAR) {
					if (strongStart >= 0) {
						contentString = text.textSubstring(strongStart, index + 2);
						keyword = contentString.substring(0, 2);
						text.addAttributeInRange('keyword', keyword, strongStart, 2);
						text.addAttributeInRange('strong', contentString.substring(2, contentString.length - 2), strongStart, contentString.length);
						text.addAttributeInRange('keyword', keyword, strongStart + contentString.length - 2, 2);
						strongStart = -1;
					} else {
						strongStart = match.index;
					}
				} else if (escaptedSyntaxChar === EM_SYNTAX_CHAR) {
					if (italicStart >= 0) {
						contentString = text.textSubstring(italicStart, index + 1);
						keyword = contentString[0];
						text.addAttributeInRange('keyword', keyword, italicStart, 1);
						text.addAttributeInRange('em', contentString.substring(1, contentString.length - 1), italicStart, contentString.length);
						text.addAttributeInRange('keyword', keyword, italicStart + contentString.length - 1, 1);
						italicStart = -1;
					} else {
						italicStart = match.index;
					}
				} else if (escaptedSyntaxChar === AUTOLINK_SYNTAX_CHAR) {
					if (autlinkStart >= 0) {
						contentString = text.textSubstring(autlinkStart, index + 1);
						keyword = contentString[0];
						text.addAttributeInRange('autolink', contentString, autlinkStart, contentString.length);
						text.addAttributeInRange('keyword', keyword, autlinkStart, 1);
						if ((match = emailAutoLinkReplaceRegex.exec(contentString))) {
							if (match[1] === 'mailto:') {
								text.addAttributeInRange('linktarget', contentString.substring(1, contentString.length - 1), autlinkStart + 1, contentString.length - 2);
							} else {
								text.addAttributeInRange('linktarget', 'mailto:' + contentString.substring(1, contentString.length - 1), autlinkStart + 1, contentString.length - 2);
							}
						} else {
							text.addAttributeInRange('linktarget', contentString.substring(1, contentString.length - 1), autlinkStart + 1, contentString.length - 2);
						}

						text.addAttributeInRange('keyword', keyword, autlinkStart + contentString.length - 1, 1);
						autlinkStart = -1;
					} else {
						autlinkStart = index;
					}
				}
			}
		}

		//
		// HTML Tags
		//

		if (htmlMode) {
			HTMLHelper.parseHTML(text, null, false, htmlMode);
		}

		//
		// [] Links
		//

		var hasPotentialLinks = escapedText.indexOf(']') !== -1;
		if (hasPotentialLinks) {
			var linkTarget;
			var linkTitle;
			var linkLabel;
			var link;
			var linkStart;
			var linkLength;
			var isImage;

			//
			// Reference Links
			//

			while ((match = referenceLinkRegex.exec(escapedText))) {
				link = match[0];
				linkLabel = match[2];
				linkTarget = match[3];
				link = text.textSubstr(match.index, link.length);
				linkLabel = text.textSubstr(match.index + 1, linkLabel.length);
				linkTarget = text.textSubstr((match.index + link.length) - (1 + linkTarget.length), linkTarget.length);
				linkStart = match.index;
				linkLength = link.length;
				isImage = link[0] === '!';

				text.addAttributeInRange('referencelink', link, linkStart, link.length);

				if (isImage) {
					text.addAttributeInRange('keyword', '!', linkStart, 1);
					linkStart++;
					linkLength--;
				}

				text.addAttributeInRange('keyword', '[', linkStart, 1);
				text.addAttributeInRange('linklabel', linkLabel, linkStart + 1, linkLabel.length);
				text.addAttributeInRange('keyword', ']', linkStart + 1 + linkLabel.length, 1);

				text.addAttributeInRange('keyword', '[', linkStart + (linkLength - (linkTarget.length + 2)), 1);
				text.addAttributeInRange('linktarget', linkTarget, linkStart + (linkLength - (1 + linkTarget.length)), linkTarget.length);
				text.addAttributeInRange('keyword', ']', linkStart + linkLength - 1, 1);
			}

			//
			// Inline Links
			//

			while ((match = inlineLineRegex.exec(escapedText))) {
				link = match[0];
				linkLabel = match[2];
				linkTarget = match[4];
				link = text.textSubstr(match.index, link.length);
				linkLabel = text.textSubstr(match.index + 1, linkLabel.length);
				linkTarget = text.textSubstr(match.index + 1 + linkLabel.length + 1 + 1, linkTarget.length);
				linkStart = match.index;
				linkLength = link.length;
				isImage = link[0] === '!';

				text.addAttributeInRange('inlinelink', link, linkStart, link.length);

				if (isImage) {
					text.addAttributeInRange('keyword', '!', linkStart, 1);
					linkStart++;
					linkLength--;
				}

				text.addAttributeInRange('keyword', '[', linkStart, 1);
				text.addAttributeInRange('linklabel', linkLabel, linkStart + 1, linkLabel.length);
				text.addAttributeInRange('keyword', ']', linkStart + 1 + linkLabel.length, 1);

				text.addAttributeInRange('keyword', '(', linkStart + 1 + linkLabel.length + 1, 1);
				text.addAttributeInRange('linktarget', linkTarget, linkStart + 1 + linkLabel.length + 1 + 1, linkLength - (1 + linkLabel.length + 1 + 1 + 1));
				text.addAttributeInRange('keyword', ')', linkStart + linkLength - 1, 1);
			}

			//
			// Footnote links
			//

			while ((match = footnoteLinkRE.exec(escapedText))) {
				link = match[0];
				linkLabel = match[1];
				link = text.textSubstr(match.index, link.length);
				linkLabel = text.textSubstr(match.index + 2, linkLabel.length);
				linkStart = match.index;
				text.addAttributeInRange('footnotelink', link, linkStart, link.length);
				text.addAttributeInRange('keyword', '[^', linkStart, 2);
				text.addAttributeInRange('linklabel', linkLabel, linkStart + 2, linkLabel.length);
				text.addAttributeInRange('keyword', ']', linkStart + 2 + linkLabel.length, 1);
			}
		}

		//
		// Magic links, links without needing any extra syntax.
		//

		if (parseMagicLinks) {
			if (escapedText.indexOf('@') !== -1) {
				MagicLinkHelper.parseEmailLinks(text, state, previousState);
			}
			MagicLinkHelper.parseWebLinks(text, state, previousState);
		}
	}

	Extensions.add('com.foldingtext.editor.clearInlineFormatting', function (formattedText) {
		var format = formattedText.format();
		if (format === 'strong') {
			formattedText.replaceInRange('', formattedText.length() - 2, 2);
			formattedText.replaceInRange('', 0, 2);
		} else if (format === 'em') {
			formattedText.replaceInRange('', formattedText.length() - 1, 1);
			formattedText.replaceInRange('', 0, 1);
		} else if (format === 'code') {
			formattedText.replaceInRange('', formattedText.length() - 1, 1);
			formattedText.replaceInRange('', 0, 1);
		}
	});

	/*jshint regexp: true */

	return {
		parseInlines: parseInlines,
		webAutoLinkReplaceRegex: webAutoLinkReplaceRegex,
		emailAutoLinkReplaceRegex: emailAutoLinkReplaceRegex,
		attributesToClear: attributesToClear
	};
});