//
// Copyright (c) 2014 Jesse Grosjean. All rights reserved.
//
// Permission is given for you to use this software ONLY while creating themes
// and plugins for the "FoldingText" software platform. Permission is NOT
// given for any other purpose.
//
define(function (require, exports, module) {
	'use strict';

	var Extensions = require('ft/core/extensions').Extensions,
		modeRegExp = /\.(\w+)\s*$/;

	function supportsMode(modeName) {
		return Extensions.processExtensions('com.foldingtext.editor.mode', function (each) {
			if (each.name === modeName) {
				return true;
			}
		}, true);
	}

	function setMode(node, mode) {
		if (!mode || supportsMode(mode)) {
			var insertString = mode ? '.' + mode : '',
				line = node.line(),
				match;
			if ((match = modeRegExp.exec(line))) {
				node.replaceLineInRange(insertString, match.index, match[0].length);
			} else {
				node.replaceLineInRange(insertString, line.length, 0);
			}
			return mode;
		}
		return node.mode();
	}

	function modeStackEquals(a, b) {
		if (a === b) {
			return true;
		}

		if (a && b) {
			var aLen = a.length,
				bLen = b.length;

			if (aLen === bLen) {
				var i = aLen - 1;
				while (i >= 0) {
					var aRecord = a[i],
						bRecord = b[i];

					if (aRecord.mode !== bRecord.mode || aRecord.structureLevel !== bRecord.structureLevel) {
						return false;
					}

					i--;
				}

				return true;
			}
		}

		return false;
	}

	function modeStackClone(modeStack) {
		if (modeStack && modeStack.length) {
			return modeStack.slice(0);
		}
		return undefined;
	}

	//
	// Each lines state maintains a mode stack. This records all previous
	// nodes that are effecting the line. Generally each lines mode stack is
	// equal to a copy of the previous lines mode stack corrected based on
	// the state of the line. For example if the line itself declares a mode
	// then this lines mode stack will also include the mode that it declares.
	// Or if this line has a different structure level (say it's less indented)
	// then the previouse line, then some modes from the previous lines mode
	// stack might not be included in this lines mode stack.
	//
	// 1. parseMode, if a mode is found a new mode stack is created for this
	// and mode statck record pushed.
	//
	// 2. maintainValidModeStack, is called at the end of each phase of classification
	//

	function parseMode(text, state, previousState) {
		var match;
		if ((match = modeRegExp.exec(text.escapedText()))) {
			var mode = match[1].toLowerCase(),
				modeStack = state.modeStack;

			if (supportsMode(mode)) {
				if (!modeStack) {
					modeStack = [];
					state.modeStack = modeStack;
				}
				modeStack.push({
					node: text._node,
					mode: mode,
					structureLevel: state.typeLevel + state.typeIndentLevel
				});
				text.addAttributeInRange('mode', match[0], match.index + 1, match[0].length);
				text.addAttributeInRange('keyword', match[0], match.index + 1, match[0].length);
				text.sliceText(0, match.index);
			}
		}
	}

	function reclassifyDown(text, state, previousState) {
		if (previousState) {
			// structure levels of previous nodes (and mode stacks) could have
			// changed. Going to maintainValidModeStack again. So set things
			// up as they would be right after calling parseMode the first
			// time. That is if this line declared a mode then this line
			// should have single mode record on stack. Otherwise this line
			// should have undefined mode stack. maintainValidModeStack will
			// then fix things up again.
			var modeRecord = state.modeStack ? state.modeStack.lastObject() : null;
			if (modeRecord && modeRecord.node === text._node) {
				state.modeStack = [modeRecord];
			} else {
				state.modeStack = undefined;
			}
			maintainValidModeStack(text, state, previousState);
		}
	}

	function maintainValidModeStack(text, state, previousState) {
		// merge in all preceding mode records that have a structureLevel less
		// then the structure level of the current line.
		if (previousState && previousState.modeStack) {
			var previousModeStack = previousState.modeStack,
				modeStack = state.modeStack ? state.modeStack : state.modeStack = [],
				structureLevel = state.typeLevel + state.typeIndentLevel,
				i = previousModeStack.length - 1;

			while (i >= 0) {
				var precedingModeRecord = previousModeStack[i];
				if (precedingModeRecord.structureLevel < structureLevel) {
					modeStack.unshift(precedingModeRecord);
				}
				i--;
			}

			if (modeStack.length === 0) {
				state.modeStack = undefined;
			}
		}
	}

	return {
		setMode: setMode,
		parseMode: parseMode,
		maintainValidModeStack: maintainValidModeStack,
		reclassifyDown: reclassifyDown,
		modeStackEquals: modeStackEquals,
		modeStackClone: modeStackClone,
		attributesToClear: ['mode']
	};
});
