//
// Copyright (c) 2014 Jesse Grosjean. All rights reserved.
//
// Permission is given for you to use this software ONLY while creating themes
// and plugins for the "FoldingText" software platform. Permission is NOT
// given for any other purpose.
//
define(function (require, exports, module) {
	'use strict';

	// Attribute names are: http://www.w3.org/TR/REC-xml/#NT-Name
	// minus ':' since ':' is used to separate tag segments in node
	// path queries

	require('ft/util/string');
	var tagStartChars = '[A-Z_a-z\\u00C0-\\u00D6\\u00D8-\\u00F6\\u00F8-\\u02FF\\u0370-\\u037D\\u037F-\\u1FFF\\u200C-\\u200D\\u2070-\\u218F\\u2C00-\\u2FEF\\u3001-\\uD7FF\\uF900-\\uFDCF\\uFDF0-\\uFFFD]',
		tagWordChars =  '[\\-.0-9\\u00B7\\u0300-\\u036F\\u203F-\\u2040]',
		tagWordRegexString = tagStartChars + '(?:' + tagStartChars + '|' + tagWordChars + ')*',
		tagRegexTemplate = '(^|\\s+)@(TAGNAME)(?:\\(([^\\)]*)\\))?(?=\\s|$)',
		tagRegex = new RegExp(tagRegexTemplate.replace('TAGNAME', tagWordRegexString), 'g'),
		Extensions = require('ft/core/extensions').Extensions;

	function _tagRange(node, tagName) {
		var line = node.line(),
			matchArea = line.match(new RegExp('(^|\\s)@(' + tagName.escapeForRegex() + ')(\\([^\\)]*\\))?'));
		if (matchArea) {
			return { location: matchArea.index, length: matchArea[0].length };
		}
		return null;
	}

	function removeTag(node, tagName) {
		try {
			var range = _tagRange(node, tagName);
			if (range) {
				node.replaceLineInRange('', range.location, range.length);
			}
			return true;
		} catch (e) {
			return false;
		}
	}

	function addTag(node, tagName, tagValue) {
		try {
			var tagString = '@' + tagName;
			if (tagValue) {
				tagString += '(' + tagValue + ')';
			}

			var range = _tagRange(node, tagName);
			if (!range) {
				range = { location: node.line().length, length: 0 };
			}

			if (range.location > 0) {
				tagString = ' ' + tagString;
			}

			node.replaceLineInRange(tagString, range.location, range.length);
			return true;
		} catch (e) {
			return false;
		}
	}

	var reservedTagNamesLookupCache;
	function reservedTagNamesLookup() {
		if (!reservedTagNamesLookupCache) {
			reservedTagNamesLookupCache = {};
			Extensions.processExtensions('com.foldingtext.editor.reservedTagNames', function (tags) {
				var length = tags.length,
					i;
				for (i = 0; i < length; i++) {
					reservedTagNamesLookupCache[tags[i]] = '';
				}
			});
		}
		return reservedTagNamesLookupCache;
	}

	//var tagRegex = /(^|\s+)@(\w+)(?:\(([^\)]*)\))?(?=\s|$)/g;

	function parseTags(text, state, previousState) {
		var escapedText = text.escapedText();

		if (escapedText.indexOf('@') === -1) {
			return;
		}

		var tags = state.tags || {},
			reservedTags = reservedTagNamesLookup(),
			trailingTagsLeadingSpaceLength = 0,
			trailingTagsStart,
			trailingTagsEnd,
			match;

		while ((match = tagRegex.exec(escapedText))) {
			var leadingSpace = match[1],
				tag = match[2],
				value = match[3];

			if (tags[tag] === undefined && reservedTags[tag] === undefined) {
				if (value === undefined) {
					value = '';
				}

				var tagStart = match.index + leadingSpace.length;

				text.addAttributeInRange('keyword', '@', tagStart, 1);
				text.addAttributeInRange('tagname', tag, tagStart + 1, tag.length);

				var length = tag.length + 1;

				if (value) {
					text.addAttributeInRange('keyword', '(', tagStart + 1 + tag.length, 1);
					text.addAttributeInRange('tagvalue', value, tagStart + 1 + tag.length + 1, value.length);
					text.addAttributeInRange('keyword', ')', tagStart + 1 + tag.length + value.length + 1, 1);
					length += value.length + 2;
				}

				text.addAttributeInRange('tag', text.textSubstr(match.index + leadingSpace.length, length), tagStart, length);

				tags[tag] = value;

				if (trailingTagsEnd !== match.index) {
					// reset and look for new trailing tags range
					trailingTagsLeadingSpaceLength = leadingSpace.length;
					trailingTagsStart = tagStart;
					trailingTagsEnd = trailingTagsStart + length;
				} else {
					// extend trailing tags end
					trailingTagsEnd += match[0].length;
				}
			}
		}

		if (trailingTagsEnd !== undefined) {
			var trailingSpaceLength = escapedText.match(/\s*$/)[0].length;
			if (trailingTagsEnd === escapedText.length - trailingSpaceLength) {
				var leadingSpaceStart = trailingTagsStart - trailingTagsLeadingSpaceLength,
					trailingTagsString = text.textSubstring(trailingTagsStart);

				if (leadingSpaceStart > 0) {
					text.sliceText(0, leadingSpaceStart);
				} else {
					text.sliceText(0, trailingTagsStart);
				}
				text.addAttributeInRange('trailingtags', trailingTagsString, trailingTagsStart, (trailingTagsEnd + trailingSpaceLength) - trailingTagsStart);
			}
		}

		state.tags = tags;
	}

	function toggleTag(tagName, tagValue, inRange) {
		/*jshint validthis:true */
		var tree = this.tree(),
			range = inRange || this.selectedRange(),
			addTag;

		tree.beginUpdates();
		range.forEachNodeInRange(function (node) {
			if (addTag === undefined) {
				addTag = node.tag(tagName) === undefined ? true : false;
			}

			if (addTag) {
				node.addTag(tagName, tagValue);
			} else {
				node.removeTag(tagName);
			}
		});
		tree.endUpdates();
	}

	Extensions.add('com.foldingtext.editor.methods', {
		toggleTag: toggleTag
	});

	Extensions.add('com.foldingtext.editor.prefixWordRangeForComplete', function (prefixRange) {
		var startNode = prefixRange.startNode,
			startOffset = prefixRange.startOffset,
			line = startNode.line();

		if (line[startOffset - 1] === '@') {
			if (startOffset > 1 && line[startOffset - 2].trim() === '') {
				return prefixRange.rangeByOffsetting(-1, 1);
			}
		}

		return prefixRange;
	});

	Extensions.add('com.foldingtext.editor.completionsForPrefixWordRange', function (prefixWordRange) {
		var startNode = prefixWordRange.startNode,
			startOffset = prefixWordRange.startOffset,
			line = startNode.line();

		if (line[startOffset] === '@') {
			var completions = [],
				prefix = prefixWordRange.textInRange(),
				tags = startNode.tree.tags(true),
				length = tags.length,
				each, i;

			for (i = 0; i < length; i++) {
				each = '@' + tags[i];
				if (each !== prefix && each.startsWith(prefix)) {
					completions.push(each);
				}
			}

			return completions.sort();
		}
		return null;
	});

	Extensions.addInit(function (editor) {
		editor.addKeyMap({
			'\'@\'' : function (editor) {
				var selectedRange = editor.selectedRange(),
					ch = selectedRange.startNode.line()[selectedRange.startOffset - 1];

				if (ch === undefined || ch.match(/\s/)) {
					var tags = editor.tree().tags(true);
					editor.replaceSelection('@', 'end');
					if (tags.length > 0) {
						editor.performCommand('complete');
					}
				} else {
					return require('codemirror').CodeMirror.Pass;
				}
			}
		});
	});

	Extensions.add('com.foldingtext.editor.reservedTagNames', ['id', 'level', 'type', 'line', 'mode', 'modeContext', 'index', 'property', 'name', 'value']);

	return {
		addTag: addTag,
		removeTag: removeTag,
		parseTags: parseTags,
		reservedTagNamesLookup: reservedTagNamesLookup,
		attributesToClear: ['tag', 'tagname', 'tagvalue', 'trailingtags']
	};
});