//
// Copyright (c) 2014 Jesse Grosjean. All rights reserved.
//
// Permission is given for you to use this software ONLY while creating themes
// and plugins for the "FoldingText" software platform. Permission is NOT
// given for any other purpose.
//

define(function (require, exports, module) {
	'use strict';

	require('ft/util/string');

	var LinebreakCausingWhitespaceHelper = require('ft/taxonomy/helpers/linebreakcausingwhitespacehelper'),
		CodeMirrorModeHelper = require('ft/taxonomy/helpers/codemirrormodehelper'),
		markdownStyles = require('components/text/text!./markdown.less'),
		PropertyHelper = require('ft/taxonomy/helpers/propertyhelper'),
		InlinesHelper = require('ft/taxonomy/helpers/inlineshelper'),
		HTMLHelper = require('ft/taxonomy/helpers/htmlhelper'),
		ModeHelper = require('ft/taxonomy/helpers/modehelper'),
		TagsHelper = require('ft/taxonomy/helpers/tagshelper'),
		Classifier = require('ft/core/classifier').Classifier,
		Extensions = require('ft/core/extensions').Extensions,
		Taxonomy = require('ft/core/taxonomy').Taxonomy,
		ObjectUtils = require('ft/util/objectutils'),
		oop = require('ft/util/oop');

	// Structure Level Constants
	var ROOT_STRUCTURE_TYPE_LEVEL = -100,
		HEADER_STRUCTURE_TYPE_LEVEL = 300,
		BODY_STRUCTURE_TYPE_LEVEL = 400;

	var LINK_DEFINITION_CLOSED = 0,
		LINK_DEFINITION_SEEK_URL = 1,
		LINK_DEFINITION_SEEK_TITLE = 2;

	var MarkdownTaxonomy = function (spec) {
		Taxonomy.call(this, spec);

		/*jshint regexp: false */
		this.emptyLineMatchRE = /^\s*$/;
		this.headingMatchRE = /^(#{1,6})(\s*)(?:.*?)(?:(\s*?)(\[([^\]]*)\]))?(?:(\s*?)(#+))?(\s*)$/;
		this.headingUnderlineMatchRE = /^[=\-]+\s*$/;
		this.blockquoteMatchRE = /(>\s{0,3})+/;
		this.horizontalruleMatchRE = /^([\*|\-|\_])([ ]?[ ]?\1){2,}[ \t]*$/;
		this.unorderedMatchRE = /^([\-+*])(\s+)(\[(?: |x)\](?:$|\s))?/;
		this.orderedMatchRE = /^(\d+\.)(\s+)(\[(?: |x)\](?:$|\s))?/;
		this.linkDefinitionRE = /\[([^\]]+)\]:(?:(\s*)(?:(?:<([^>]+)>)|(?:(\S+)))?(?:(\s+)(["(](.+?)[")]))?)?/;
		this.linkDefinitionURLOnlyRE = /(?:(?:<([^>]+)>)|(?:(\S+)))/;
		this.linkDefinitionTitleOnlyRE = /^(?:["(](.+?)[")])/;
		this.fencedcodeRE = /^```(\S+)?/;
		this.definitionMatchRE = /^:(?:[ \t]*)/;
		this.whitespaceRE = /\s*/;
		this.renderLineHeadingRE = /^(#+ ?)/;
		this.renderLineUnorderedRE = /^(\t*)([\-+*]\s+)/;
		this.renderLineTaskRE = /^(\t*)([\-+*]|[0-9]+\.)(\s+)(\[(?: |x)\](?:$|\s))?/;
		this.renderLineOrderedRE = /^(\t*)([0-9]+\.\s+)/;
		this.renderLineLinkDefinitionRE = /\[([^\]]+)\]:(\s*)/;
		/*jshint regexp: true */

		this.foldingtext = !!spec.foldingtext;
		this.gitmarkdown = !!spec.gitmarkdown;
		this.criticMarkup = !!spec.criticMarkup;
		this.multimarkdown = !!spec.multimarkdown;
		this.magiclinks = this.foldingtext || this.gitmarkdown;
		this.inwordunderscores = !this.foldingtext && !this.gitmarkdown && !this.multimarkdown;

		this.htmlMode = CodeMirrorModeHelper.getMode({}, 'text/html');
	};

	oop.inherits(MarkdownTaxonomy, Taxonomy);

	(function () {

		//
		// Line content modification
		//

		this.setType = function setType(node, type) {
			if (node.type() === type) {
				return type;
			}

			var newLead,
				newLine;

			switch (type) {
			case 'heading':
				newLead = '# ';
				break;
			case 'definition':
				newLead = ':	';
				break;
			case 'blockquote':
				newLead = '> ';
				break;
			case 'codeblock':
				newLead = '    ';
				break;
			case 'empty':
				newLine = '';
				break;
			case 'fencedcodebottomboundary':
			case 'fencedcodetopboundary':
				newLine = '```';
				break;
			case 'footnotelinkdefinition':
				newLead = '[^]: ';
				break;
			case 'headingunderline':
				newLine = '--------';
				break;
			case 'horizontalrule':
				newLine = '* * *';
				break;
			case 'body':
				newLead = '';
				break;
			case 'linkdefinition':
				newLead = '[]: ';
				break;
			case 'ordered':
				newLead = '1. ';
				break;
			case 'unordered':
				newLead = '- ';
				break;
			case 'task':
				newLead = '- [ ] ';
				break;

			// Don't support setting these types
			case 'fencedcode':
			case 'linkdefinitiontitle':
			case 'linkdefinitionurl':
			case 'term':
				break;

			default:
				throw 'Invalid setType: ' + type;
			}

			if (newLine !== undefined) {
				node.setLine(newLine);
				return type;
			} else {
				if (newLead !== undefined) {
					node.replaceLineInRange(newLead, 0, node.textStart());
					return type;
				}
			}

			return node.type();
		};

		this.typeLevelForType = function typeLevelForType(type) {
			switch (type) {
			case 'heading':
				return HEADER_STRUCTURE_TYPE_LEVEL;
			case 'definition':
				return BODY_STRUCTURE_TYPE_LEVEL + 1;
			case 'blockquote':
			case 'codeblock':
			case 'empty':
			case 'fencedcode':
			case 'fencedcodebottomboundary':
			case 'fencedcodetopboundary':
			case 'footnotelinkdefinition':
			case 'headingunderline':
			case 'horizontalrule':
			case 'body':
			case 'linkdefinition':
			case 'linkdefinitiontitle':
			case 'linkdefinitionurl':
			case 'ordered':
			case 'term':
			case 'unordered':
			case 'task':
				return BODY_STRUCTURE_TYPE_LEVEL;
			default:
				throw 'Invalid typeLevelForType: ' + type;
			}
		};

		this.setTypeIndentLevel = function setTypeIndentLevel(node, level) {
			level = this.validatedTypeIndentLevel(node, level);

			var currentLevel = node.typeIndentLevel(),
				delta = level - currentLevel,
				indentationChars;

			if (delta !== 0) {
				switch (node.type()) {
				case 'empty':
					return level;
				case 'heading':
					indentationChars = '#';
					break;
				case 'blockquote':
					indentationChars = '> ';
					break;
				default:
					indentationChars = '\t';
					break;
				}

				if (delta < 0) {
					node.replaceLineInRange('', 0, Math.abs(delta) * indentationChars.length);
				} else {
					var indentation = [];
					for (var i = 0; i < delta; i++) {
						indentation.push(indentationChars);
					}
					node.replaceLineInRange(indentation.join(''), 0, 0);
				}
			}

			return level;
		};

		this.validatedTypeIndentLevel = function validatedTypeIndentLevel(node, level) {
			var type = node.type();
			switch (node.type()) {
			case 'definition':
			case 'horizontalrule':
				return node.typeIndentLevel() === level;
			case 'heading':
			case 'ordered':
			case 'unordered':
			case 'task':
				if (level < 1) {
					return 1;
				}
				break;
			default:
				if (level < 0) {
					return 0;
				}
				break;
			}

			return level;
		};

		this.setMode = function setMode(node, mode) {
			return ModeHelper.setMode(node, mode);
		};

		this.removeTag = function removeTag(node, tagName) {
			return TagsHelper.removeTag(node, tagName);
		};

		this.addTag = function addTag(node, tagName, tagValue) {
			return TagsHelper.addTag(node, tagName, tagValue);
		};

		this.setPropertyNodeValue = function setPropertyNodeValue(node, propertyValue) {
			return PropertyHelper.setPropertyNodeValue(node, propertyValue);
		};

		//
		// Classify Methods
		//

		this.types = function types() {
			return {
				fencedcode : true,
				fencedcodetopboundary : true,
				fencedcodebottomboundary : true,
				empty : true,
				codeblock : true,
				heading : true,
				headingunderline : true,
				blockquote : true,
				horizontalrule : true,
				linkdefinition : true,
				term : true,
				definition : true,
				unordered : true,
				task : true,
				ordered : true,
				body : true
			};
		};

		this.detectAndHighlightGFMCodeBlock = function detectAndHighlightGFMCodeBlock(text, state, previousState) {
			var prevType = previousState.type,
				continueFenced = prevType === 'fencedcodetopboundary' || prevType === 'fencedcode',
				fenceMatch = text.textCharAt(1) === '`' ? text.textMatch(this.fencedcodeRE) : null;

			if (continueFenced || fenceMatch) {
				state.type = 'fencedcode';
				state.tags.language = previousState.tags.language;

				if (fenceMatch) {
					text.addAttributeInRange('keyword', '```', 0, 3);
					if (continueFenced) {
						state.type = 'fencedcodebottomboundary';
					} else {
						state.type = 'fencedcodetopboundary';
						var language = fenceMatch[1];
						if (language) {
							state.tags.language = language;
							text.addAttributeInRange('language', language, 3, language.length);

							var mode = CodeMirrorModeHelper.getMode(null, language);
							if (mode) {
								state.codemirrorState = {
									mode : mode,
									state : mode.startState ? mode.startState() : null
								};
								if (mode.name !== 'null') {
									text.addAttributeInRange('keyword', language, 3, language.length);
								}
							}
						}
					}
				} else {
					var codemirrorState = CodeMirrorModeHelper.copyState(previousState.codemirrorState);
					if (codemirrorState) {
						CodeMirrorModeHelper.runMode(codemirrorState.mode, codemirrorState.state, text);
						state.codemirrorState = codemirrorState;
					}
				}
				return true;
			}

			return false;
		};

		this.detectAndHighlightEmpty = function detectAndHighlightEmpty(text, state, previousState) {
			if (text.text().match(this.emptyLineMatchRE)) {
				state.type = 'empty';
				state.listContextLevel = previousState.listContextLevel;
				text.sliceAllText();
				return true;
			}
			return false;
		};

		this.detectAndHighlightCodeBlock = function detectAndHighlightCodeBlock(text, state, previousState) {
			if (state.logicalIndent >= 1) {
				if (state.logicalIndent <= previousState.listContextLevel) {
					return false;
				}

				if (previousState.linkDefinitionState !== LINK_DEFINITION_CLOSED) {
					return false;
				}

				state.type = 'codeblock';
				state.listContextLevel = previousState.listContextLevel;
				if (previousState.listContextLevel) {
					state.typeIndentLevel = previousState.listContextLevel + 1;
				}

				return true;
			}
			return false;
		};

		this.detectAndHighlightAtxHeading = function detectAndHighlightAtxHeading(text, state, previousState) {
			var ch = text.textCharAt(0),
				i;

			if (ch === '#') {
				var match = text.textMatch(this.headingMatchRE),
					leadSyntax = match[1],
					spaceAfterLead = match[2],
					spaceBeforeLabel = match[3],
					labelWithSyntax = match[4],
					label = match[5],
					spaceBeforeTrail = match[6],
					trailSyntax = match[7],
					trailSpace = match[8],
					textStart = leadSyntax.length + spaceAfterLead.length,
					textEnd = text.textLength() - (trailSpace ? trailSpace.length : 0);

				state.type = 'heading';
				state.typeLevel = this.typeLevelForType(state.type);
				state.typeIndentLevel = leadSyntax.length;
				text.addAttributeInRange('keyword', leadSyntax, match.index, leadSyntax.length);
				text.addAttributeInRange('lead', leadSyntax + spaceAfterLead, match.index, leadSyntax.length + spaceAfterLead.length);

				if (trailSyntax) {
					var trailLength = spaceBeforeTrail.length + trailSyntax.length;
					text.addAttributeInRange('keyword', trailSyntax, textEnd - trailSyntax.length, trailSyntax.length);
					textEnd -= trailLength;
				}

				if (this.multimarkdown && labelWithSyntax) {
					text.addAttributeInRange('label', label, textEnd - (label.length + 2), label.length + 2);
					text.addAttributeInRange('keyword', '[', textEnd - (label.length + 2), 1);
					text.addAttributeInRange('keyword', ']', textEnd - 1, 1);
					textEnd -= spaceBeforeLabel.length + labelWithSyntax.length;
				}

				text.sliceText(textStart, textEnd);

				if (this.foldingtext) {
					TagsHelper.parseTags(text, state, previousState);
					ModeHelper.parseMode(text, state, previousState);
				}

				InlinesHelper.parseInlines(text, state, previousState, this.htmlMode, this.criticMarkup, this.magiclinks, this.inwordunderscores);

				return true;
			}

			return false;
		};

		this.detectAndHighlightSetextHeadingUnderline = function detectAndHighlightSetextHeadingUnderline(text, state, previousState) {
			var match,
				matchText,
				ch = text.textCharAt(0);

			if ((ch === '-' || ch === '=') && previousState.type === 'body' && (match = text.textMatch(this.headingUnderlineMatchRE))) {
				state.type = 'headingunderline';
				state.typeIndentLevel = (ch === '=' ? 1 : 2);
				state.nextAction = Classifier.RECLASSIFY_PREVIOUS;
				matchText = match[0];
				text.addAttributeInRange('keyword', matchText, match.index, matchText.length);
				text.sliceAllText();
				return true;
			}

			return false;
		};

		this.detectAndHighlightBlockquote = function detectAndHighlightBlockquote(text, state, previousState) {
			var match,
				matchText,
				ch = text.textCharAt(0);

			if ((ch === '>' || ch === ' ') && (match = text.textMatch(this.blockquoteMatchRE))) {
				matchText = match[0];
				var rtrimMatchText = matchText.rtrim();
				state.type = 'blockquote';
				state.typeIndentLevel = rtrimMatchText.match(/>/g).length - 1;

				var level = 0;
				for (var i = 0; i < rtrimMatchText.length; i++) {
					if (rtrimMatchText[i] === '>') {
						text.addAttributeInRange('keyword', '>' + level, i, 1);
						level++;
					}
				}

				text.addAttributeInRange('lead', rtrimMatchText, 0, rtrimMatchText.length);
				text.sliceText(matchText.length);

				if (this.foldingtext) {
					TagsHelper.parseTags(text, state, previousState);
					ModeHelper.parseMode(text, state, previousState);
				}

				InlinesHelper.parseInlines(text, state, previousState, this.htmlMode, this.criticMarkup, this.magiclinks, this.inwordunderscores);

				return true;
			}

			return false;
		};

		this.detectAndHighlightHorizontalRule = function detectAndHighlightHorizontalRule(text, state, previousState) {
			var match,
				matchText,
				ch = text.textCharAt(0);

			if (state.logicalIndent === 0 && (ch === '*' || ch === '-' || ch === '_') && (match = text.textMatch(this.horizontalruleMatchRE))) {
				state.type = 'horizontalrule';
				matchText = match[0];
				text.addAttributeInRange('keyword', matchText, 0, matchText.length);
				text.sliceAllText();
				return true;
			}

			return false;
		};

		function highlightURLAndReturnAdvance(i, autoLinkURL, anyLinkURL, text) {
			if (autoLinkURL) {
				text.addAttributeInRange('keyword', '<', i, 1);
				i += 1;
				text.addAttributeInRange('linktarget', autoLinkURL, i, autoLinkURL.length);
				i += autoLinkURL.length;
				text.addAttributeInRange('keyword', '>', i, 1);
				i += 1;
			} else if (anyLinkURL) {
				text.addAttributeInRange('linktarget', anyLinkURL, i, anyLinkURL.length);
				i += anyLinkURL.length;
			}
			return i;
		}

		function highlightTitleAndReturnAdvance(i, title, text, multimarkdown) {
			text.addAttributeInRange('keyword', '"', i, 1);
			i += 1;
			text.addAttributeInRange('linktitle', title, i, title.length);
			i += title.length;
			text.addAttributeInRange('keyword', '"', i, 1);
			i += 1;

			if (multimarkdown) {
				i = text.textLength();
			}

			return i;
		}

		function highlightTrailingErrorRange(i, text) {
			var trailLength = text.textLength() - i,
				hasError = false;

			if (trailLength > 0) {
				var trailText = text.textSubstr(i, trailLength);
				trailText = trailText.ltrim();
				var leadingSpace = trailLength - trailText.length;
				trailText = trailText.rtrim();
				var trailingSpace = (trailLength - leadingSpace) - trailText.length;
				if (trailText) {
					text.addAttributeInRange('error', trailText, i + leadingSpace, trailText.length);
				}
			}
		}

		this.detectAndHighlightLinkDefinition = function detectAndHighlightLinkDefinition(text, state, previousState) {
			var match,
				ch = text.textCharAt(0),
				previousLinkDefinitionState = previousState.linkDefinitionState,
				autoLinkURL,
				anyLinkURL,
				title,
				i = 0;

			if (previousLinkDefinitionState === LINK_DEFINITION_CLOSED && state.logicalIndent !== 0) {
				return;
			}

			if (ch === '[' && (match = text.textMatch(this.linkDefinitionRE))) {
				state.type = 'linkdefinition';
				state.linkDefinitionState = LINK_DEFINITION_SEEK_URL;

				var label = match[1],
					spaceAfterLabel = match[2],
					footnote = false;
				autoLinkURL = match[3];
				anyLinkURL = match[4];
				var spaceAfterURL = match[5],
					titleWithSyntax = match[6];
				title = match[7];

				if (this.multimarkdown && text.textCharAt(1) === '^') {
					footnote = true;
					state.type = 'footnotelinkdefinition';
					text.addAttributeInRange('keyword', '[^', i, 2);
					label = label.slice(1);
					i += 2;
				} else {
					text.addAttributeInRange('keyword', '[', i, 1);
					i += 1;
				}

				text.addAttributeInRange('linkid', label, i, label.length);
				i += label.length;
				text.addAttributeInRange('keyword', ']:', i, 2);

				if (footnote) {
					state.linkDefinitionState = LINK_DEFINITION_CLOSED;
					return true;
				}

				if (autoLinkURL || anyLinkURL) {
					i += (2 + spaceAfterLabel.length);
					i = highlightURLAndReturnAdvance(i, autoLinkURL, anyLinkURL, text);

					if (titleWithSyntax) {
						state.linkDefinitionState = LINK_DEFINITION_CLOSED;
						i += spaceAfterURL.length;
						i = highlightTitleAndReturnAdvance(i, title, text, this.multimarkdown);
					} else {
						state.linkDefinitionState = LINK_DEFINITION_SEEK_TITLE;
					}
				} else {
					state.linkDefinitionState = LINK_DEFINITION_SEEK_URL;
				}

				highlightTrailingErrorRange(i, text);

				return true;
			} else if (previousLinkDefinitionState === LINK_DEFINITION_SEEK_URL) {
				state.linkDefinitionState = LINK_DEFINITION_SEEK_TITLE;
				state.type = 'linkdefinitionurl';
				state.typeIndentLevel = state.logicalIndent;
				match = text.textMatch(this.linkDefinitionURLOnlyRE);
				i = highlightURLAndReturnAdvance(match.index, match[1], match[2], text);

				var fullRemainder = text.escapedText().slice(i);
				if (fullRemainder) {
					var ltrimRemainder = fullRemainder.ltrim();
					if (ltrimRemainder) {
						match = ltrimRemainder.match(this.linkDefinitionTitleOnlyRE);
						if (match) {
							i = highlightTitleAndReturnAdvance(i + (fullRemainder.length - ltrimRemainder.length), match[1], text, this.multimarkdown);
							state.linkDefinitionState = LINK_DEFINITION_CLOSED;
						}
					}
				}

				i = highlightTrailingErrorRange(i, text);

				return true;
			} else if (previousLinkDefinitionState === LINK_DEFINITION_SEEK_TITLE) {
				state.linkDefinitionState = LINK_DEFINITION_CLOSED;
				match = text.textMatch(this.linkDefinitionTitleOnlyRE);
				if (match) {
					state.type = 'linkdefinitiontitle';
					state.typeIndentLevel = state.logicalIndent;
					i = highlightTitleAndReturnAdvance(match.index, match[1], text, this.multimarkdown);
					i = highlightTrailingErrorRange(i, text);
					return true;
				}
			}
			return false;
		};

		this.detectAndHighlightDefinitionListDefinition = function detectAndHighlightDefinitionListDefinition(text, state, previousState) {
			var match,
				matchText,
				ch = text.textCharAt(0);

			if (ch === ':' && (previousState.type === 'body' || previousState.type === 'definition') && (match = text.textMatch(this.definitionMatchRE))) {
				state.type = 'definition';
				state.typeLevel = this.typeLevelForType(state.type);
				matchText = match[0];
				text.addAttributeInRange('keyword', ':', 0, 1);
				text.sliceText(matchText.length);
				state.nextAction = Classifier.RECLASSIFY_PREVIOUS;

				if (this.foldingtext) {
					TagsHelper.parseTags(text, state, previousState);
					ModeHelper.parseMode(text, state, previousState);
				}

				InlinesHelper.parseInlines(text, state, previousState, this.htmlMode, this.criticMarkup, this.magiclinks, this.inwordunderscores);

				return true;
			}
			return false;
		};

		function isListType(type) {
			return type === 'unordered' || type === 'ordered';
		}

		function canListStartAfterState(taxonomy, state) {
			if (taxonomy.foldingtext) {
				return true;
			} else {
				var type = state.type;
				return type === undefined || type === 'empty' || state.listContextLevel > 0 || isListType(type);
			}
		}

		this.detectAndHighlightList = function detectAndHighlightList(text, state, previousState) {
			if (canListStartAfterState(this, previousState)) {
				var match,
					matchText,
					textStart = 0,
					ch = text.textCharAt(0);

				if ((ch === '-' || ch === '*' || ch === '+') && (match = text.textMatch(this.unorderedMatchRE))) {
					state.type = 'unordered';
				} else if (!isNaN(ch) && (match = text.textMatch(this.orderedMatchRE))) {
					state.type = 'ordered';
				}

				if (match) {
					state.typeIndentLevel = state.logicalIndent + 1;
					state.listContextLevel = state.typeIndentLevel;

					matchText = match[1];
					var checkbox = match[3];

					textStart += matchText.length;
					text.addAttributeInRange('keyword', matchText, 0, matchText.length);
					if (!checkbox) {
						text.addAttributeInRange('lead', matchText, 0, matchText.length);
					}
					textStart += match[2].length;

					if (checkbox && this.gitmarkdown) {
						state.type = 'task';
						if (checkbox.substr(1, 1) !== ' ') {
							state.tags.checked = '';
						}
						text.addAttributeInRange('keyword', '[', textStart, 1);
						text.addAttributeInRange('keyword', ']', textStart + 2, 1);
						text.addAttributeInRange('checkbox', checkbox.substr(0, 3), textStart, 3);
						text.addAttributeInRange('lead', matchText + checkbox, 0, matchText.length + checkbox.length);
						textStart += checkbox.length;
					}
					text.sliceText(textStart);

					if (this.foldingtext) {
						TagsHelper.parseTags(text, state, previousState);
						ModeHelper.parseMode(text, state, previousState);
					}

					InlinesHelper.parseInlines(text, state, previousState, this.htmlMode, this.criticMarkup, this.magiclinks, this.inwordunderscores);

					return true;
				}
			}

			return false;
		};

		this.detectAndHighlightItem = function detectAndHighlightItem(text, state, previousState) {
			if (previousState.type === 'blockquote') {
				state.type = 'blockquote';
				state.typeLevel = previousState.typeLevel;
				state.typeIndentLevel = previousState.typeIndentLevel;
			} else {
				state.type = 'body';

				var foldingtext = this.foldingtext,
					listContextLevel = previousState.listContextLevel,
					continueItem = !foldingtext && isListType(previousState.type),
					continueList = listContextLevel && state.logicalIndent > 0;

				if (continueItem) {
					state.typeIndentLevel = listContextLevel;
				} else if (continueList) {
					state.typeIndentLevel = state.logicalIndent;
				} else if (foldingtext) {
					state.typeIndentLevel = state.logicalIndent;
				}

				if (continueList || continueItem) {
					state.listContextLevel = listContextLevel;
				}

				LinebreakCausingWhitespaceHelper.parseTrailingSpace(text, state, previousState);
			}

			if (this.foldingtext) {
				TagsHelper.parseTags(text, state, previousState);
				ModeHelper.parseMode(text, state, previousState);
			}

			InlinesHelper.parseInlines(text, state, previousState, this.htmlMode, this.criticMarkup, this.magiclinks, this.inwordunderscores);
		};

		this.classify = function classify(text, state, previousState) {
			var treeEndsWithEmpty = false;

			if (this.gitmarkdown && this.detectAndHighlightGFMCodeBlock(text, state, previousState)) {
			} else if (this.detectAndHighlightEmpty(text, state, previousState)) {
				treeEndsWithEmpty = text.isLastLine();
			} else if (HTMLHelper.detectAndHighlightHtmlBlock(text, state, previousState, this.htmlMode)) {
			} else if (this.detectAndHighlightAtxHeading(text, state, previousState)) {
			} else if (this.detectAndHighlightSetextHeadingUnderline(text, state, previousState)) {
			} else if (this.multimarkdown && this.detectAndHighlightDefinitionListDefinition(text, state, previousState)) {
			} else {
				var foldingtext = this.foldingtext,
					string = text.text(),
					length = string.length,
					logicalIndent = 0,
					spacesRun = 0,
					ch,
					i;

				for (i = 0; i < length; i++) {
					ch = string[i];
					if (ch === ' ') {
						logicalIndent++;
						spacesRun++;
						if (spacesRun === 4 && foldingtext) {
							i++;
							break;
						}
					} else if (ch === '\t') {
						logicalIndent += 4;
						spacesRun = 0;
					} else {
						break;
					}
				}

				if (logicalIndent) {
					logicalIndent = Math.floor(logicalIndent / 4);
					text.sliceText(i);
				}

				state.logicalIndent = logicalIndent;

				if (this.detectAndHighlightLinkDefinition(text, state, previousState)) {
				} else if ((!foldingtext || spacesRun === 4) && this.detectAndHighlightCodeBlock(text, state, previousState)) {
				} else if (this.detectAndHighlightBlockquote(text, state, previousState)) {
				} else if (this.detectAndHighlightHorizontalRule(text, state, previousState)) {
				} else if (foldingtext && PropertyHelper.parsePropertyType(text, state, previousState)) {
				} else if (this.detectAndHighlightList(text, state, previousState)) {
				} else {
					this.detectAndHighlightItem(text, state, previousState);
				}
			}

/*
In progress. Somethings not working because I was calling ModeHelper.parseMode here for all lines.
That was GOOD. Then I started only calling it for some types. That seemed good... then I relaized That
it broke mode stack maintinance. So a property node for instance would break the timer mode.
To sovle this i renamed classify to "parseMode". And took the popModeStackToMatchNodeStrucureLevel call out
of the implementation. Now im only calling parseMode in places that I want modes... and am always calling
popModeStackToMatchNodeStrucureLevel ... thought that would work. But it's not. Property nodes are still breaking timer
mode.


So next step is figuring out WHY!
*/
			ModeHelper.maintainValidModeStack(text, state, previousState);

			if (previousState.type === 'empty' && (treeEndsWithEmpty || state.type !== 'empty')) {
				state.nextAction = Classifier.RECLASSIFY_PREVIOUS;
			}

			Extensions.processExtensions('com.foldingtext.taxonomy.classifier', function (classifier) {
				classifier.classify(text, state, previousState);
			});

			state.text = text.text();
			state.textStart = text.textStart();
			text.addAttributeInRange('text', text.text(), 0, text.textLength());
		};

		this.reclassifyUp = function reclassifyUp(text, state, nextState) {
			if (state.type === 'empty') {
				if (nextState.type === 'codeblock' || nextState.type === 'potentialemptycodeblockmarker') {
					state.type = 'potentialemptycodeblockmarker';
				}
				state.typeLevel = nextState.typeLevel;
				state.typeIndentLevel = nextState.typeIndentLevel;
				state.nextAction = Classifier.RECLASSIFY_PREVIOUS;
			} else if (nextState.type === 'headingunderline' && (state.type === 'body')) {
				state.type = 'heading';
				state.typeLevel = this.typeLevelForType(state.type);
				state.typeIndentLevel = nextState.typeIndentLevel;
				state.nextAction = Classifier.CLASSIFY_NEXT;
			} else if ((nextState.type === 'definition' || nextState.type === 'term') && (state.type === 'body')) {
				state.type = 'term';
				state.typeLevel = this.typeLevelForType(state.type);
				state.typeIndentLevel = 0;
				state.nextAction = Classifier.RECLASSIFY_PREVIOUS;
			} else {
				state.nextAction = Classifier.CLASSIFY_NEXT;
			}
		};

		this.reclassifyDown = function reclassifyDown(text, state, previousState) {
			if (state.type === 'potentialemptycodeblockmarker') {
				if (previousState.type === 'codeblock') {
					state.type = 'codeblock';
				} else {
					state.type = 'empty';
				}
			}
			if (this.foldingtext) {
				ModeHelper.reclassifyDown(text, state, previousState);
				Extensions.processExtensions('com.foldingtext.taxonomy.classifier', function (classifier) {
					classifier.classify(text, state, previousState);
				});
			}
		};

		//
		// State Management
		//

		this.startState = function startState() {
			return {
				nextAction: Classifier.CLASSIFY_NEXT,
				type: undefined,
				typeLevel: BODY_STRUCTURE_TYPE_LEVEL,
				typeIndentLevel: 0,
				logicalIndent: 0,
				listContextLevel: 0,
				linkDefinitionState: LINK_DEFINITION_CLOSED,
				text: '',
				textStart: 0,
				tags: {},
				modeStack: undefined,
				codemirrorState: undefined
			};
		};

		this.copyState = function copyState(state) {
			var outerThis = this;
			return {
				nextAction: state.nextAction,
				type: state.type,
				typeLevel: state.typeLevel,
				typeIndentLevel: state.typeIndentLevel,
				logicalIndent: state.logicalIndent,
				listContextLevel: state.listContextLevel,
				linkDefinitionState: state.linkDefinitionState,
				text: state.text,
				textStart: state.textStart,
				tags: ObjectUtils.shallowClone(state.tags),
				modeStack: ModeHelper.modeStackClone(state.modeStack),
				codemirrorState: CodeMirrorModeHelper.copyState(state.codemirrorState)
			};
		};

		this.equalStates = function equalStates(a, b) {
			return a.nextAction === b.nextAction &&
				a.type === b.type &&
				a.typeLevel === b.typeLevel &&
				a.typeIndentLevel === b.typeIndentLevel &&
				a.logicalIndent === b.logicalIndent &&
				a.listContextLevel === b.listContextLevel &&
				a.linkDefinitionState === b.linkDefinitionState &&
				a.text === b.text &&
				a.textStart === b.textStart &&
				ObjectUtils.shallowEquals(a.tags, b.tags) &&
				ModeHelper.modeStackEquals(a.modeStack, b.modeStack) &&
				CodeMirrorModeHelper.stateEquals(a.codemirrorState, b.codemirrorState);
		};

		this.textAttributesToClearBeforeClassify = function textAttributesToClearBeforeClassify() {
			var attributesToClear = this.attributesToClear;
			if (!attributesToClear) {
				attributesToClear = [
					'text',
					'lead',
					'trail',
					'error',
					'language'
				];
				attributesToClear = LinebreakCausingWhitespaceHelper.attributesToClear.concat(attributesToClear);
				attributesToClear = InlinesHelper.attributesToClear.concat(attributesToClear);
				Extensions.processExtensions('com.foldingtext.taxonomy.classifier', function (classifier) {
					attributesToClear = attributesToClear.concat(classifier.attributesToClear);
				});

				if (this.foldingtext) {
					attributesToClear = TagsHelper.attributesToClear.concat(attributesToClear);
					attributesToClear = ModeHelper.attributesToClear.concat(attributesToClear);
					attributesToClear = PropertyHelper.attributesToClear.concat(attributesToClear);
				}

				attributesToClear = CodeMirrorModeHelper.attributesToClear.concat(attributesToClear);

				attributesToClear.sort();
				attributesToClear.sortedUniquify();
				this.attributesToClear = attributesToClear;
			}
			return attributesToClear;
		};


		//
		// Leads
		//

		var unordedLeadRegex = /(\s*)([\-+*])\s/,
			taskLeadRegex =    /(\s*)([\-+*]|[0-9]+\.)(\s+)(\[(?: |x)\]\s)/,
			orderedLeadRegex = /(\s*)([0-9]+)(\.\s)/,
			blockquoteLeadRegex = /(> ?)*/;

		this.matchLeadForNode = function matchLeadForNode(node) {
			switch (node.type()) {
			case 'unordered':
				return unordedLeadRegex.exec(node.line());
			case 'task':
				return taskLeadRegex.exec(node.line());
			case 'ordered':
				return orderedLeadRegex.exec(node.line());
			case 'blockquote':
				return blockquoteLeadRegex.exec(node.line());
			}

			/*jshint camelcase: false*/
			return MarkdownTaxonomy.super_.matchLeadForNode.call(this, node);
		};

		this.nextLeadFromNode = function nextLeadFromNode(node, after) {
			if (node.type() === 'ordered') {
				var nodeLeadMatch = this.matchLeadForNode(node),
					number = parseInt(nodeLeadMatch[2], 10);
				if (after) {
					number++;
				} else {
					number--;
				}
				if (number <= 0) {
					number = 1;
				}
				return nodeLeadMatch[1] + number + nodeLeadMatch[3];
			}
			/*jshint camelcase: false*/
			return MarkdownTaxonomy.super_.nextLeadFromNode.call(this, node, after);
		};

		//
		// Rendering
		//

		this.styles = function styles() {
			return markdownStyles;
		};

		/*function headingIndexPath(node) {
			var path = [];
			while (node && node.type() === 'heading') {
				var index = 0,
					each = node.previousSibling;
				while (each) {
					if (each.type() === 'heading') {
						index++;
					}
					each = each.previousSibling;
				}
				path.push(index + 1);
				node = node.parent;
			}
			return path.join('.');
		}

		this.renderNode = function renderNode(editor, node, nodeRenderer) {
			var type = node.type(),
				line = node.line(),
				outerThis = this,
				widget,
				match;

			if (type === 'heading') {
				widget = document.createElement('span');
				widget.textContent = headingIndexPath(node);
				nodeRenderer.renderInlineWidget(0, node.typeIndentLevel(), widget);
			} else if (node.type() === 'unordered') {
				widget = document.createElement('span');
				widget.textContent = '•';
				nodeRenderer.renderInlineWidget(node.typeIndentLevel() - 1, 1, widget);
			}
		};*/

		this.renderNodeElementStyles = function renderNodeElementStyles(editor, node, elementRenderer) {
			var type = node.type(),
				line = node.line(),
				defaultMarginLeft = editor.defaultPREStyling().marginLeft,
				charWidth = editor.defaultCharWidth(),
				marginLeftAdjust = 0,
				textIndent = 0,
				tabWidth,
				tabsWidth,
				match;

			/*jshint camelcase: false*/
			MarkdownTaxonomy.super_.renderNodeElementStyles.call(this, editor, node, elementRenderer);
			/*jshint camelcase: true*/

			if (type === 'heading' && (match = line.match(this.renderLineHeadingRE))) {
				textIndent = -elementRenderer.measureTextRange(match.index, match[0].length).width;
			} else if (type === 'blockquote') {
				match = line.match(/\s*(>\s*)+/);
				if (match) { // line right after > is also blockquite of not empty lines, but it won't match regex
					textIndent = -elementRenderer.measureTextRange(match.index, match[0].length).width;
				}
				marginLeftAdjust += -textIndent;
			} else if (type === 'definition') {
				match = line.match(this.definitionMatchRE);
				textIndent = -elementRenderer.measureTextRange(match.index, match[0].length).width;
				marginLeftAdjust += -textIndent;
			} else if (type === 'ordered' && (match = line.match(this.renderLineOrderedRE))) {
				tabWidth = editor.tabWidth();
				tabsWidth = match[1].length * tabWidth;
				marginLeftAdjust += tabWidth + tabsWidth;
				textIndent = -(tabsWidth + elementRenderer.measureTextRange(match.index + match[1].length, match[2].length).width);
			} else if (type === 'unordered' && (match = line.match(this.renderLineUnorderedRE))) {
				tabWidth = editor.tabWidth();
				tabsWidth = match[1].length * tabWidth;
				marginLeftAdjust += tabWidth + tabsWidth;
				textIndent = -(tabsWidth + elementRenderer.measureTextRange(match.index + match[1].length, match[2].length).width);
			} else if (type === 'task' && (match = line.match(this.renderLineTaskRE))) {
				tabWidth = editor.tabWidth();
				tabsWidth = match[1].length * tabWidth;
				marginLeftAdjust += tabWidth + tabsWidth;
				textIndent = -(tabsWidth + elementRenderer.measureTextRange(match.index + match[1].length, match[0].length - match[1].length).width);
			} else if (type === 'codeblock') {
				if (this.foldingtext) {
					match = line.match(/\t*[ ]{4}/);
				} else {
					match = line.match(/(?:[ ]{4})|\t/);
				}
				if (match) {
					textIndent = -elementRenderer.measureTextRange(match.index, match[0].length).width;
					marginLeftAdjust += -textIndent;
				}
			} else if (type === 'htmlblock' || type === 'fencedcode' || type === 'fencedcodetopboundary' || type === 'fencedcodebottomboundary') {
				marginLeftAdjust += charWidth * 4;
			} else {
				match = line.match(/\s*/);
				textIndent = -elementRenderer.measureTextRange(match.index, match[0].length).width;
				marginLeftAdjust += -textIndent;
			}

			var totalMarginLeft = defaultMarginLeft + marginLeftAdjust,
				totalNegativeOffset = totalMarginLeft + textIndent;

			if (totalNegativeOffset < charWidth) {
				totalMarginLeft += -totalNegativeOffset;
				totalMarginLeft += charWidth;
			}

			if (totalMarginLeft !== defaultMarginLeft) {
				elementRenderer.setStyle('marginLeft', Math.round(totalMarginLeft) + 'px');
			}

			if (textIndent !== 0) {
				elementRenderer.setStyle('textIndent', Math.round(textIndent) + 'px');
			}
		};

		//
		// Export
		//

		this.rangeToHTML = function rangeToHTML(range) {
			// do it this way so that core.js doens't need to include entire markdown lib.
			var markdown = 'ft/taxonomy/markdown';
			if (require.defined(markdown)) {
				return window.markdown.toHTML(range.textInRange());
			}
			return 'opps markdown lib wasnt loaded';
		};

	}).call(MarkdownTaxonomy.prototype);

	Extensions.add('com.foldingtext.editor.taxonomies', {
		name: 'markdown',
		configurations: [
			{ mime: 'text/x-markdown', spec: 'markdown' }
		],
		taxonomy: function (spec) {
			return new MarkdownTaxonomy(spec);
		}
	});

	exports.MarkdownTaxonomy = MarkdownTaxonomy;
});