//
// Copyright (c) 2014 Jesse Grosjean. All rights reserved.
//
// Permission is given for you to use this software ONLY while creating themes
// and plugins for the "FoldingText" software platform. Permission is NOT
// given for any other purpose.
//

define(function (require, exports, module) {
	'use strict';

	var Extensions = require('../core/extensions').Extensions,
		Classifier = require('../core/classifier').Classifier,
		Taxonomy = require('../core/taxonomy').Taxonomy,

		CodeMirrorModeHelper = require('./helpers/codemirrormodehelper'),
		InlinesHelper = require('./helpers/inlineshelper'),
		HTMLHelper = require('./helpers/htmlhelper'),
		TagsHelper = require('./helpers/tagshelper'),
		ModeHelper = require('./helpers/modehelper'),

		ObjectUtils = require('../util/objectutils'),
		oop = require('../util/oop');

	var OutlineTaxonomy = function (spec) {
		Taxonomy.call(this, spec);

		this.htmlMode = CodeMirrorModeHelper.getMode({}, 'text/html');
		this.emptyLineMatchRE = /^\s*$/;
		this.leadingTabsRE = /^\t*/;
	};

	oop.inherits(OutlineTaxonomy, Taxonomy);

	(function () {

		this.setTypeIndentLevel = function setTypeIndentLevel(node, typeIndentLevel) {
			if (typeIndentLevel < 0) {
				typeIndentLevel = 0;
			}
			if (node.type() !== 'empty') {
				var currentLevel = node.line().match(/^\t*/)[0].length,
					delta = typeIndentLevel - currentLevel;
				if (delta < 0) {
					node.replaceLineInRange('', 0, Math.abs(delta));
				} else if (delta > 0) {
					var indentation = [];
					for (var i = 0; i < delta; i++) {
						indentation.push('\t');
					}
					node.replaceLineInRange(indentation.join(''), 0, 0);
				}
			}
			return typeIndentLevel;
		};

		this.setMode = function setMode(node, mode) {
			return ModeHelper.setMode(node, mode);
		};

		this.removeTag = function removeTag(node, tagName) {
			return TagsHelper.removeTag(node, tagName);
		};

		this.addTag = function addTag(node, tagName, tagValue) {
			return TagsHelper.addTag(node, tagName, tagValue);
		};

		//
		// Classification
		//

		this.types = function types() {
			return {
				item : true,
				empty : true
			};
		};

		this.detectAndHighlightEmpty = function detectAndHighlightEmpty(text, state, previousState) {
			if (text.text().match(this.emptyLineMatchRE)) {
				state.type = 'empty';
				text.sliceAllText();
				return true;
			}
			return false;
		};

		this.detectAndHighlightItem = function detectAndHighlightItem(text, state, previousState) {
			state.type = 'item';
			var tabs = text.text().match(this.leadingTabsRE)[0].length;
			if (tabs) {
				state.typeIndentLevel = tabs;
				text.sliceText(tabs);
			}
			TagsHelper.parseTags(text, state, previousState);
			InlinesHelper.parseInlines(text, state, previousState, this.htmlMode, this.spec.criticMarkup, true);
		};

		this.classify = function classify(text, state, previousState) {
			var treeEndsWithEmpty = false;

			if (this.detectAndHighlightEmpty(text, state, previousState)) {
				treeEndsWithEmpty = text.isLastLine();
			} else if (HTMLHelper.detectAndHighlightHtmlBlock(text, state, previousState, this.htmlMode)) {
			} else {
				this.detectAndHighlightItem(text, state, previousState);
			}

			ModeHelper.parseMode(text, state, previousState);
			ModeHelper.maintainValidModeStack(text, state, previousState);

			if (previousState.type === 'empty' && (treeEndsWithEmpty || state.type !== 'empty')) {
				state.nextAction = Classifier.RECLASSIFY_PREVIOUS;
			}

			Extensions.processExtensions('com.foldingtext.taxonomy.classifier', function (classifier) {
				classifier.classify(text, state, previousState);
			});

			state.text = text.text();
			state.textStart = text.textStart();
			text.addAttributeInRange('text', text.text(), 0, text.textLength());
		};

		this.reclassifyUp = function reclassifyUp(text, state, nextState) {
			if (state.type === 'empty') {
				state.typeLevel = nextState.typeLevel;
				state.typeIndentLevel = nextState.typeIndentLevel;
				state.nextAction = Classifier.RECLASSIFY_PREVIOUS;
			} else {
				state.nextAction = Classifier.CLASSIFY_NEXT;
			}
		};

		this.reclassifyDown = function reclassifyDown(text, state, previousState) {
			ModeHelper.reclassifyDown(text, state, previousState);
			Extensions.processExtensions('com.foldingtext.taxonomy.classifier', function (classifier) {
				classifier.classify(text, state, previousState);
			});
		};

		this.textAttributesToClearBeforeClassify = function textAttributesToClearBeforeClassify() {
			var attributesToClear = this.attributesToClear;
			if (!attributesToClear) {
				attributesToClear = [
					'text'
				];
				attributesToClear = TagsHelper.attributesToClear.concat(attributesToClear);
				attributesToClear = InlinesHelper.attributesToClear.concat(attributesToClear);
				attributesToClear = ModeHelper.attributesToClear.concat(attributesToClear);
				Extensions.processExtensions('com.foldingtext.taxonomy.classifier', function (classifier) {
					attributesToClear = attributesToClear.concat(classifier.attributesToClear);
				});
				attributesToClear.sort();
				attributesToClear.sortedUniquify();
				this.attributesToClear = attributesToClear;
			}
			return attributesToClear;
		};

		this.startState = function startState() {
			return {
				type: undefined,
				typeLevel: 0,
				typeIndentLevel: 0,
				text: '',
				textStart: 0,
				modeStack: undefined,
				tags: {},
				codemirrorState: undefined,
				nextAction: Classifier.CLASSIFY_NEXT
			};
		};

		this.copyState = function copyState(state) {
			return {
				nextAction: state.nextAction,
				type: state.type,
				typeLevel: state.typeLevel,
				typeIndentLevel: state.typeIndentLevel,
				text: state.text,
				textStart: state.textStart,
				modeStack: ModeHelper.modeStackClone(state.modeStack),
				tags: ObjectUtils.shallowClone(state.tags),
				codemirrorState: CodeMirrorModeHelper.copyState(state.codemirrorState)
			};
		};

		this.equalStates = function equalStates(a, b) {
			return a.nextAction === b.nextAction &&
				a.type === b.type &&
				a.typeLevel === b.typeLevel &&
				a.typeIndentLevel === b.typeIndentLevel &&
				a.text === b.text &&
				a.textStart === b.textStart &&
				ModeHelper.modeStackEquals(a.modeStack, b.modeStack) &&
				ObjectUtils.shallowEquals(a.tags, b.tags) &&
				CodeMirrorModeHelper.stateEquals(a.codemirrorState, b.codemirrorState);
		};

		//
		// Rendering
		//

		this.renderNodeElementStyles = function renderNodeElementStyles(editor, node, elementRenderer) {
			var tabWidth = editor.tabWidth(),
				defaultMarginLeft = editor.defaultPREStyling().marginLeft,
				indentWidth = node.typeIndentLevel() * tabWidth;

			/*jshint camelcase: false*/
			OutlineTaxonomy.super_.renderNodeElementStyles.call(this, editor, node, elementRenderer);
			/*jshint camelcase: true*/

			elementRenderer.setStyle('marginLeft', Math.round(defaultMarginLeft + indentWidth) + 'px');
			elementRenderer.setStyle('textIndent', Math.round(-indentWidth) + 'px');
		};

	}).call(OutlineTaxonomy.prototype);

	Extensions.add('com.foldingtext.editor.taxonomies', {
		name: 'outline',
		configurations: [
			{ mime: 'text/x-outline', spec: 'outline' }
		],
		taxonomy: function (spec) {
			return new OutlineTaxonomy(spec);
		}
	});

	exports.OutlineTaxonomy = OutlineTaxonomy;
});