//
// Copyright (c) 2014 Jesse Grosjean. All rights reserved.
//
// Permission is given for you to use this software ONLY while creating themes
// and plugins for the "FoldingText" software platform. Permission is NOT
// given for any other purpose.
//
define(function (require, exports, module) {
	'use strict';

	var TagsHelper = require('ft/taxonomy/helpers/tagshelper'),
		ModeHelper = require('ft/taxonomy/helpers/modehelper'),
		Classifier = require('ft/core/classifier').Classifier,
		Extensions = require('ft/core/extensions').Extensions,
		Taxonomy = require('ft/core/taxonomy').Taxonomy,
		ObjectUtils = require('ft/util/objectutils'),
		oop = require('ft/util/oop');

	var TabsTaxonomy = function (spec) {
		Taxonomy.call(this, spec);

		this.emptyLineMatchRE = /^\s*$/;
		this.leadingTabsRE = /^\t*/;
	};

	oop.inherits(TabsTaxonomy, Taxonomy);

	(function () {

		//
		// Classification
		//

		this.detectAndHighlightEmpty = function detectAndHighlightEmpty(text, state, previousState) {
			if (text.text().match(this.emptyLineMatchRE)) {
				state.type = 'empty';
				text.sliceAllText();
				return true;
			}
			return false;
		};

		this.detectAndHighlightItem = function detectAndHighlightItem(text, state, previousState) {
			state.type = 'item';
			var tabs = text.text().match(this.leadingTabsRE)[0].length;
			if (tabs) {
				state.typeIndentLevel = tabs;
				text.sliceText(tabs);
			}
			TagsHelper.parseTags(text, state, previousState);
		};

		this.types = function types() {
			return {
				item : true,
				empty : true
			};
		};
		
		this.classify = function classify(text, state, previousState) {
			var treeEndsWithEmpty = false;

			if (this.detectAndHighlightEmpty(text, state, previousState)) {
				treeEndsWithEmpty = text.isLastLine();
			} else {
				this.detectAndHighlightItem(text, state, previousState);
			}

			ModeHelper.parseMode(text, state, previousState);
			ModeHelper.maintainValidModeStack(text, state, previousState);

			if (previousState.type === 'empty' && (treeEndsWithEmpty || state.type !== 'empty')) {
				state.nextAction = Classifier.RECLASSIFY_PREVIOUS;
			}

			Extensions.processExtensions('com.foldingtext.taxonomy.classifier', function (classifier) {
				classifier.classify(text, state, previousState);
			});

			state.text = text.text();
			state.textStart = text.textStart();
			text.addAttributeInRange('text', text.text(), 0, text.textLength());
		};

		this.reclassifyUp = function reclassifyUp(text, state, nextState) {
			if (state.type === 'empty') {
				state.typeLevel = nextState.typeLevel;
				state.typeIndentLevel = nextState.typeIndentLevel;
				state.nextAction = Classifier.RECLASSIFY_PREVIOUS;
			} else {
				state.nextAction = Classifier.CLASSIFY_NEXT;
			}
		};

		this.reclassifyDown = function reclassifyDown(text, state, previousState) {
			ModeHelper.reclassifyDown(text, state, previousState);
			Extensions.processExtensions('com.foldingtext.taxonomy.classifier', function (classifier) {
				classifier.classify(text, state, previousState);
			});
		};

		this.textAttributesToClearBeforeClassify = function textAttributesToClearBeforeClassify() {
			var attributesToClear = this.attributesToClear;
			if (!attributesToClear) {
				attributesToClear = [
					'text'
				];
				attributesToClear = TagsHelper.attributesToClear.concat(attributesToClear);
				attributesToClear = ModeHelper.attributesToClear.concat(attributesToClear);
				Extensions.processExtensions('com.foldingtext.taxonomy.classifier', function (classifier) {
					attributesToClear = attributesToClear.concat(classifier.attributesToClear);
				});
				attributesToClear.sort();
				attributesToClear.sortedUniquify();
				this.attributesToClear = attributesToClear;
			}
			return attributesToClear;
		};

		this.startState = function startState() {
			return {
				type: undefined,
				typeLevel: 0,
				typeIndentLevel: 0,
				text: '',
				textStart: 0,
				modeStack: undefined,
				tags: {},
				nextAction: Classifier.CLASSIFY_NEXT
			};
		};

		this.copyState = function copyState(state) {
			return {
				nextAction: state.nextAction,
				type: state.type,
				typeLevel: state.typeLevel,
				typeIndentLevel: state.typeIndentLevel,
				text: state.text,
				textStart: state.textStart,
				modeStack: ModeHelper.modeStackClone(state.modeStack),
				tags: ObjectUtils.shallowClone(state.tags)
			};
		};

		this.equalStates = function equalStates(a, b) {
			return a.nextAction === b.nextAction &&
				a.type === b.type &&
				a.typeLevel === b.typeLevel &&
				a.typeIndentLevel === b.typeIndentLevel &&
				a.text === b.text &&
				a.textStart === b.textStart &&
				ModeHelper.modeStackEquals(a.modeStack, b.modeStack) &&
				ObjectUtils.shallowEquals(a.tags, b.tags);
		};

	}).call(TabsTaxonomy.prototype);

	Extensions.add('com.foldingtext.editor.taxonomies', {
		name: 'tabs',
		configurations: [
			{ mime: 'text/x-tabs', spec: 'tabs' }
		],
		taxonomy: function (spec) {
			return new TabsTaxonomy(spec);
		}
	});

	exports.TabsTaxonomy = TabsTaxonomy;
});