//
// Copyright (c) 2014 Jesse Grosjean. All rights reserved.
//
// Permission is given for you to use this software ONLY while creating themes
// and plugins for the "FoldingText" software platform. Permission is NOT
// given for any other purpose.
//

define(function (require, exports, module) {
	'use strict';

	var InlinesHelper = require('ft/taxonomy/helpers/inlineshelper'),
		taskPaperStyles = require('components/text/text!./taskpaper.less'),
		TagsHelper = require('ft/taxonomy/helpers/tagshelper'),
		Classifier = require('ft/core/classifier').Classifier,
		Extensions = require('ft/core/extensions').Extensions,
		Taxonomy = require('ft/core/taxonomy').Taxonomy,
		ObjectUtils = require('ft/util/objectutils'),
		oop = require('ft/util/oop');

	var TaskPaperTaxonomy = function (spec) {
		Taxonomy.call(this, spec);
		this.emptyLineMatchRE = /^\s*$/;
		this.leadingTabsRE = /^\t*/;
	};

	oop.inherits(TaskPaperTaxonomy, Taxonomy);

	(function () {

		this.setTypeIndentLevel = function setTypeIndentLevel(node, typeIndentLevel) {
			if (typeIndentLevel < 0) {
				typeIndentLevel = 0;
			}
			if (node.type() !== 'empty') {
				var currentLevel = node.line().match(/^\t*/)[0].length,
					delta = typeIndentLevel - currentLevel;
				if (delta < 0) {
					node.replaceLineInRange('', 0, Math.abs(delta));
				} else if (delta > 0) {
					var indentation = [];
					for (var i = 0; i < delta; i++) {
						indentation.push('\t');
					}
					node.replaceLineInRange(indentation.join(''), 0, 0);
				}
			}
			return typeIndentLevel;
		};

		this.removeTag = function removeTag(node, tagName) {
			return TagsHelper.removeTag(node, tagName);
		};

		this.addTag = function addTag(node, tagName, tagValue) {
			return TagsHelper.addTag(node, tagName, tagValue);
		};

		//
		// Classification
		//

		this.types = function types() {
			return {
				project : true,
				task : true,
				note : true,
				empty : true
			};
		};

		this.detectAndHighlightEmpty = function detectAndHighlightEmpty(text, state, previousState) {
			if (text.text().match(this.emptyLineMatchRE)) {
				state.type = 'empty';
				text.sliceAllText();
				return true;
			}
			return false;
		};

		this.detectAndHighlightTask = function detectAndHighlightTask(text, state, previousState) {
			if (text.textCharAt(0) === '-' && text.textCharAt(1) === ' ') {
				text.addAttributeInRange('keyword', '-', 0, 1);
				text.addAttributeInRange('lead', '- ', 0, 2);
				state.type = 'task';
				text.sliceText(2);
				return true;
			}
			return false;
		};

		this.detectAndHighlightProject = function detectAndHighlightProject(text, state, previousState) {
			if (text.textCharAt(text.textLength() - 1) === ':') {
				state.type = 'project';
				text.addAttributeInRange('keyword', ':', text.textLength() - 1, 1);
				text.sliceText(0, text.textLength() - 1);
				return true;
			}
			return false;
		};

		this.classify = function classify(text, state, previousState) {
			var treeEndsWithEmpty = false,
				tabs = text.text().match(this.leadingTabsRE)[0].length;

			if (tabs) {
				state.typeIndentLevel = tabs;
				text.sliceText(tabs);
			}

			if (this.detectAndHighlightEmpty(text, state, previousState)) {
				treeEndsWithEmpty = text.isLastLine();
			} else {
				TagsHelper.parseTags(text, state, previousState);
				InlinesHelper.parseInlines(text, state, previousState, null, false, true, false);

				if (this.detectAndHighlightTask(text, state, previousState)) {
				} else if (this.detectAndHighlightProject(text, state, previousState)) {
				} else {
					state.type = 'note';
				}
			}

			if (previousState.type === 'empty' && (treeEndsWithEmpty || state.type !== 'empty')) {
				state.nextAction = Classifier.RECLASSIFY_PREVIOUS;
			}

			Extensions.processExtensions('com.foldingtext.taxonomy.classifier', function (classifier) {
				classifier.classify(text, state, previousState);
			});

			state.text = text.text();
			state.textStart = text.textStart();
			text.addAttributeInRange('text', text.text(), 0, text.textLength());
		};

		this.reclassifyUp = function reclassifyUp(text, state, nextState) {
			if (state.type === 'empty') {
				state.typeLevel = nextState.typeLevel;
				state.typeIndentLevel = nextState.typeIndentLevel;
				state.nextAction = Classifier.RECLASSIFY_PREVIOUS;
			} else {
				state.nextAction = Classifier.CLASSIFY_NEXT;
			}
		};

		this.reclassifyDown = function reclassifyDown(text, state, previousState) {
		};

		this.textAttributesToClearBeforeClassify = function textAttributesToClearBeforeClassify() {
			var attributesToClear = this.attributesToClear;
			if (!attributesToClear) {
				attributesToClear = [
					'text'
				];
				attributesToClear = TagsHelper.attributesToClear.concat(attributesToClear);
				attributesToClear = InlinesHelper.attributesToClear.concat(attributesToClear);
				Extensions.processExtensions('com.foldingtext.taxonomy.classifier', function (classifier) {
					attributesToClear = attributesToClear.concat(classifier.attributesToClear);
				});
				attributesToClear.sort();
				attributesToClear.sortedUniquify();
				this.attributesToClear = attributesToClear;
			}
			return attributesToClear;
		};

		this.startState = function startState() {
			return {
				type: undefined,
				typeLevel: 0,
				typeIndentLevel: 0,
				text: '',
				textStart: 0,
				tags: {},
				nextAction: Classifier.CLASSIFY_NEXT
			};
		};

		this.copyState = function copyState(state) {
			return {
				nextAction: state.nextAction,
				type: state.type,
				typeLevel: state.typeLevel,
				typeIndentLevel: state.typeIndentLevel,
				text: state.text,
				textStart: state.textStart,
				tags: ObjectUtils.shallowClone(state.tags)
			};
		};

		this.equalStates = function equalStates(a, b) {
			return a.nextAction === b.nextAction &&
				a.type === b.type &&
				a.typeLevel === b.typeLevel &&
				a.typeIndentLevel === b.typeIndentLevel &&
				a.text === b.text &&
				a.textStart === b.textStart &&
				ObjectUtils.shallowEquals(a.tags, b.tags);
		};

		//
		// Lead
		//

		var taskLeadRegex = /(\s*)([\-+*])\s/;

		this.matchLeadForNode = function matchLeadForNode(node) {
			if (node.type() === 'task') {
				return taskLeadRegex.exec(node.line());
			}
			/*jshint camelcase: false*/
			return TaskPaperTaxonomy.super_.matchLeadForNode.call(this, node);
		};

		//
		// Rendering
		//

		this.styles = function styles() {
			return taskPaperStyles;
		};

		this.renderNodeElementStyles = function renderNodeElementStyles(editor, node, elementRenderer) {
			var cm = editor.cm(),
				type = node.type(),
				line = node.line(),
				tabWidth = editor.tabWidth(),
				charWidth = editor.defaultCharWidth(),
				defaultMarginLeft = editor.defaultPREStyling().marginLeft,
				leadingTabs = line.match(/\t*/)[0].length,
				hangingWidth = leadingTabs * tabWidth,
				extraLeftMargin = hangingWidth,
				tabsWidth = 0;

			/*jshint camelcase: false*/
			TaskPaperTaxonomy.super_.renderNodeElementStyles.call(this, editor, node, elementRenderer);
			/*jshint camelcase: true*/

			if (type === 'task' && line[leadingTabs] === '-' && line[leadingTabs + 1] === ' ') {
				hangingWidth += elementRenderer.measureTextRange(leadingTabs, 2).width;
			}

			var totalMargin = defaultMarginLeft + extraLeftMargin + tabsWidth,
				totalTextIndent = -(hangingWidth + tabsWidth),
				totalNegativeOffset = totalMargin + totalTextIndent;

			// what was this for?
			//if (totalNegativeOffset < charWidth) {
			//	totalMargin += -totalNegativeOffset;
			//	totalMargin += charWidth;
			//}

			elementRenderer.setStyle('marginLeft', Math.round(totalMargin) + 'px');
			elementRenderer.setStyle('textIndent', Math.round(totalTextIndent) + 'px');
		};

	}).call(TaskPaperTaxonomy.prototype);

	Extensions.add('com.foldingtext.editor.taxonomies', {
		name: 'taskpaper',
		configurations: [
			{ mime: 'text/x-taskpaper', spec: 'taskpaper' }
		],
		taxonomy: function (spec) {
			return new TaskPaperTaxonomy(spec);
		}
	});

	exports.TaskPaperTaxonomy = TaskPaperTaxonomy;
});