/*
 * Copyright (c) 2011-2012 Hog Bay Software. All rights reserved.
 */

define(function (require) {
	'use strict';

	var Range = require('ft/core/range').Range,
		ObjectUtils = require('ft/util/objectutils'),
		AttributeRun = require('ft/core/attributerun').AttributeRun,
		AttributedString = require('ft/core/attributedstring').AttributedString;

	describe('AttributedString', function () {
		var attributedString;

		beforeEach(function () {
			attributedString = new AttributedString('Hello world!');
		});

		afterEach(function () {
		});

		it('should be able to add attribute run', function () {
			var effectiveRange = {};
			attributedString.addAttributeInRange('name', 'jesse', 0, 5);
			expect(attributedString.attributesAtIndex(0, effectiveRange).name).toEqual('jesse');
			expect(effectiveRange.location).toEqual(0);
			expect(effectiveRange.length).toEqual(5);
			expect(attributedString.attributesAtIndex(5, effectiveRange)).toEqual({});
			expect(effectiveRange.location).toEqual(5);
			expect(effectiveRange.length).toEqual(7);
		});

		it('should be able to add attribute run bordering start of string', function () {
			var effectiveRange = {};
			attributedString.addAttributeInRange('name', 'jesse', 0, 5);
			expect(attributedString.attributesAtIndex(0, effectiveRange).name).toEqual('jesse');
			expect(effectiveRange.location).toEqual(0);
			expect(effectiveRange.length).toEqual(5);
		});

		it('should be able to add attribute run bordering end of string', function () {
			var effectiveRange = {};
			attributedString.addAttributeInRange('name', 'jesse', 6, 6);
			expect(attributedString.attributesAtIndex(6, effectiveRange).name).toEqual('jesse');
			expect(effectiveRange.location).toEqual(6);
			expect(effectiveRange.length).toEqual(6);
		});

		it('should be able to find longest effective range for attribute', function () {
			var longestEffectiveRange = {};
			attributedString.addAttributeInRange('one', 'one', 0, 12);
			attributedString.addAttributeInRange('two', 'two', 6, 6);
			expect(attributedString.attributeAtIndex('one', 6, null, longestEffectiveRange)).toEqual('one');
			expect(longestEffectiveRange.location).toEqual(0);
			expect(longestEffectiveRange.length).toEqual(12);
		});

		it('should be able to find longest effective range for attributes', function () {
			var longestEffectiveRange = {};
			attributedString.addAttributeInRange('one', 'one', 0, 12);
			attributedString.addAttributeInRange('two', 'two', 6, 6);
			attributedString._indexOfAttributeRunForCharacterIndex(10); // artificial split
			attributedString.attributesAtIndex(6, null, longestEffectiveRange);
			expect(longestEffectiveRange.location).toEqual(6);
			expect(longestEffectiveRange.length).toEqual(6);
		});

		it('should be able to find attribute range delimited by attribute keys', function () {
			var shouldExtendRangeToIncludeRun = function (run) {
				return ObjectUtils.arrayOfKeys(run.attributes).length > 0;
			};

			attributedString.addAttributeInRange('two', 'two', 0, 5);

			var range = attributedString.attributeRangeCalculatedByFunction(0, shouldExtendRangeToIncludeRun);
			expect(range.location).toEqual(0);
			expect(range.length).toEqual(5);

			range = attributedString.attributeRangeCalculatedByFunction(5, shouldExtendRangeToIncludeRun);
			expect(range.location).toEqual(5);
			expect(range.length).toEqual(0);
		});

		it('should be able to add multiple attributes in same attribute run', function () {
			var effectiveRange = {};
			attributedString.addAttributeInRange('name', 'jesse', 0, 5);
			attributedString.addAttributeInRange('age', '35', 0, 5);
			expect(attributedString.attributesAtIndex(0, effectiveRange).name).toEqual('jesse');
			expect(attributedString.attributesAtIndex(0, effectiveRange).age).toEqual('35');
			expect(effectiveRange.location).toEqual(0);
			expect(effectiveRange.length).toEqual(5);
		});

		it('should be able to add attributes in overlapping ranges', function () {
			var effectiveRange = {};
			attributedString.addAttributeInRange('name', 'jesse', 0, 5);
			attributedString.addAttributeInRange('age', '35', 3, 5);
			expect(attributedString.attributesAtIndex(0, effectiveRange).name).toEqual('jesse');
			expect(attributedString.attributesAtIndex(0, effectiveRange).age).not.toBeDefined();
			expect(effectiveRange.location).toEqual(0);
			expect(effectiveRange.length).toEqual(3);
			expect(attributedString.attributesAtIndex(3, effectiveRange).name).toEqual('jesse');
			expect(attributedString.attributesAtIndex(3, effectiveRange).age).toEqual('35');
			expect(effectiveRange.location).toEqual(3);
			expect(effectiveRange.length).toEqual(2);
			expect(attributedString.attributesAtIndex(6, effectiveRange).name).not.toBeDefined();
			expect(attributedString.attributesAtIndex(6, effectiveRange).age).toEqual('35');
			expect(effectiveRange.location).toEqual(5);
			expect(effectiveRange.length).toEqual(3);
		});

		it('should not allow getting or setting attributes at end of string', function () {
			// don't quite get this... should be throwing exceptions here?
			expect(attributedString.attributesAtIndex(0, null)).not.toBeNull();
			expect(attributedString.attributesAtIndex(11, null)).not.toBeNull();
			expect(attributedString.attributesAtIndex(12, null)).toBeNull();
			attributedString.replaceCharactersInRange(0, 12, '');
			expect(attributedString.attributesAtIndex(0, null)).toBeNull();
		});

		it('should ignore add/remove attributes in ranges with length === 0', function () {
			attributedString.addAttributeInRange('name', 1000, 0); // no crash even though location is invalid
			attributedString.removeAttributeInRange('name', 1000, 0); // no crash even though location is invalid
		});

		it('should allow removing attributes in range', function () {
			attributedString.addAttributeInRange('name', 'jesse', 0, 12);
			attributedString.removeAttributeInRange('name', 0, 12);
			expect(attributedString.toString(true)).toEqual('(Hello world!/)');
			attributedString.addAttributeInRange('name', 'jesse', 0, 12);
			attributedString.removeAttributeInRange('name', 0, 3);
			expect(attributedString.toString(true)).toEqual('(Hel/)(lo world!/name="jesse")');
			attributedString.removeAttributeInRange('name', 9, 3);
			expect(attributedString.toString(true)).toEqual('(Hel/)(lo wor/name="jesse")(ld!/)');
		});

		it('should update attribute runs when attributed string is modified', function () {
			attributedString.addAttributeInRange('name', 'jesse', 0, 12);
			attributedString.replaceCharactersInRange(0, 12, 'Hello');
			expect(attributedString.toString(true)).toEqual('(Hello/name="jesse")');
			attributedString.replaceCharactersInRange(5, 0, ' World!');
			expect(attributedString.toString(true)).toEqual('(Hello World!/name="jesse")');
		});

		it('should update attribute runs when node text is paritially updated', function () {
			attributedString.addAttributeInRange('name', 'jesse', 0, 5);
			attributedString.addAttributeInRange('name', 'joe', 5, 7);
			expect(attributedString.toString(true)).toEqual('(Hello/name="jesse")( world!/name="joe")');
			attributedString.replaceCharactersInRange(3, 5, '');
			expect(attributedString.toString(true)).toEqual('(Hel/name="jesse")(rld!/name="joe")');
			attributedString.replaceCharactersInRange(3, 0, 'lo wo');
			expect(attributedString.toString(true)).toEqual('(Hello wo/name="jesse")(rld!/name="joe")');
		});

		it('should remove leading attribute run if text in run is fully replaced', function () {
			attributedString = new AttributedString('\ttwo');
			attributedString.addAttributeInRange('name', 'jesse', 0, 1);
			attributedString.replaceCharactersInRange(0, 1, '');
			expect(attributedString.toString()).toEqual('(two/)');
		});

		it('should be able to create copy', function () {
			attributedString.addAttributeInRange('name', 'jesse', 0, 12);
			var copy = attributedString.copy();
			expect(attributedString.toString()).toEqual(copy.toString());
		});

		/*xdescribe('Attribute based range extension / contraction', function () {

			var attributedString, text, angled, squared, parens, curlies;

			beforeEach(function () {
				text = '123 <123 [123 45 (123 45) {67} ] >';
				angled = ['<123 [123 45 (123 45) {67} ] >', 4];
				squared = ['[123 45 (123 45) {67} ]', 9, 5];
				parens = ['(123 45)', 17, 13, 8];
				curlies = ['{67}', 26, 22, 17];
				var angledLength = angled[0].length,
					squaredLength = squared[0].length,
					parensLength = parens[0].length,
					curliesLength = curlies[0].length;
				attributedString = new AttributedString(text);
				attributedString.addAttributeInRange('text', text, 0, text.length);

				attributedString.addAttributeInRange('angled', angled[0], angled[1], angledLength);
				attributedString.addAttributeInRange('syntax', angled[0][0], angled[1], 1);
				attributedString.addAttributeInRange('syntax', angled[0][angledLength - 1], angled[1] + angledLength - 1, 1);

				attributedString.addAttributeInRange('squared', squared[0], squared[1], squaredLength);
				attributedString.addAttributeInRange('syntax', squared[0][0], squared[1], 1);
				attributedString.addAttributeInRange('syntax', squared[0][squaredLength - 1], squared[1] + squaredLength - 1, 1);

				attributedString.addAttributeInRange('parens', parens[0], parens[1], parensLength);
				attributedString.addAttributeInRange('syntax', parens[0][0], parens[1], 1);
				attributedString.addAttributeInRange('syntax', parens[0][parensLength - 1], parens[1] + parensLength - 1, 1);

				attributedString.addAttributeInRange('curlies', curlies[0], curlies[1], curliesLength);
				attributedString.addAttributeInRange('syntax', curlies[0][0], curlies[1], 1);
				attributedString.addAttributeInRange('syntax', curlies[0][curliesLength - 1], curlies[1] + curliesLength - 1, 1);
			});

			it('should expand to cover inner text', function () {
				var origRange = {location: curlies[1] + 1, length: 0};
				var newRange = attributedString.extendedRangeFromRange(origRange);
				expect(newRange.location).toEqual(curlies[1] + 1);
				expect(newRange.length).toEqual(curlies[0].length - 2);
			});

			it('should expand from inner text to cover syntax too', function () {
				var origRange = {location: curlies[1] + 1, length: 0};
				var newRange = attributedString.extendedRangeFromRange(origRange);

				expect(newRange.location).toEqual(curlies[1] + 1);
				expect(newRange.length).toEqual(curlies[0].length - 2);

				newRange = attributedString.extendedRangeFromRange(newRange);

				expect(newRange).not.toBeNull();
				expect(newRange.location).toEqual(curlies[1]);
				expect(newRange.length).toEqual(curlies[0].length);
			});

			it('should expand from a syntax covered range to the interior of the second largest', function () {
				var origRange = {location: curlies[1] + 1, length: 0};
				var newRange = attributedString.extendedRangeFromRange(origRange);

				expect(newRange.location).toEqual(curlies[1] + 1);
				expect(newRange.length).toEqual(curlies[0].length - 2);

				newRange = attributedString.extendedRangeFromRange(newRange);

				expect(newRange).not.toBeNull();
				expect(newRange.location).toEqual(curlies[1]);
				expect(newRange.length).toEqual(curlies[0].length);

				newRange = attributedString.extendedRangeFromRange(newRange);

				expect(newRange.location).toEqual(squared[1]);
				expect(newRange.length).toEqual(squared[0].length);
			});

			it('should select the whole internal when dealing with a range comprised of multiple attributes', function () {
				// 123 <123 [12^3 45 (123 45) {6^7}]>
				var origRange = {location: squared[1] + 4, length: squared[0].length - 6};
				var newRange = attributedString.extendedRangeFromRange(origRange);

				expect(newRange.location).toEqual(squared[1] + 1);
				expect(newRange.length).toEqual(squared[0].length - 2);
			});

			it('should select a large span when starting with a multi-attributed range', function () {
				// 123 <123 [123 45 (12^3 45) {6^7}]>
				var origRange = {location: parens[1] + 3, length: (curlies[1] + 2) - (parens[1] + 3)};
				var newRange = attributedString.extendedRangeFromRange(origRange);
				expect(newRange.location).toEqual(squared[1] + 1);
				expect(newRange.length).toEqual(squared[0].length - 2);
			});

			it('should default to returning the entire string range if no suitable attribute is found', function () {
				// 12^3 <123 [123 45 (12^3 45) {6^7}]>^
				var origRange = {location: 2, length: text.length - 2};
				var newRange = attributedString.extendedRangeFromRange(origRange);
				expect(newRange.location).toEqual(0);
				expect(newRange.length).toEqual(text.length);
			});
		});*/
	});
});
