define(function (require) {
	'use strict';

	describe('Classifier', function () {
		var Outline = require('ft/taxonomy/outlinetaxonomy').Outline,
			Extensions = require('ft/core/extensions').Extensions,
			Tree = require('ft/core/tree').Tree,
			tree;

		beforeEach(function () {
			tree = new Tree('', 'outline');
		});

		afterEach(function () {
			tree.ensureClassified();
			validateTreeClassification(tree);
			tree.cleanupForCollection();
		});

		function generateRandomTree(maxLines) {
			var lines = [],
				lineCount = Math.round(Math.random() * maxLines);

			for (var i = 0; i < lineCount; i++) {
				switch (Math.round(Math.random() * 3)) {
				case 0:
					lines.push('');
					break;
				case 1:
					lines.push('\tone');
					break;
				case 2:
					lines.push('\t\ttwo');
					break;
				}
			}

			tree.cleanupForCollection();
			tree = new Tree('', 'outline');
			tree.setText(lines.join('\n'));
		}

		function consoleLogTestCase(originalTreeText, textAPIOperations, lineAPIOperations) {
			var test = 'tree.setText(\'' + originalTreeText + '\');';

			test += 'NEWLINE' + 'classifier.classify(Infinity);';
			test += 'NEWLINE' + 'validateTreeClassification(tree);';

			var i, op;
			for (i = 0; i < textAPIOperations.length; i++) {
				op = textAPIOperations[i];
				test += 'NEWLINE' + 'tree.replaceTextInRange(tree.createRangeFromLocation(' + op.location + ', ' + op.length + '), \'' + op.insertText + '\');';
			}
			for (i = 0; i < lineAPIOperations.length; i++) {
				op = lineAPIOperations[i];
				if (op.remove !== undefined) {
					test += 'NEWLINE' + 'tree.removeNode(tree.lineNumberToNode(' + op.remove + '));';
				} else {
					test += 'NEWLINE' + 'tree.insertNodeBefore(tree.createNode(\'this is a test line!\'), tree.lineNumberToNode(' + op.insert + '));';
				}
			}

			test += 'NEWLINE' + 'classifier.classify(Infinity);';
			test += 'NEWLINE' + 'validateTreeClassification(tree);';

			test = test.replace(/\n/g, '\\n');
			test = test.replace(/NEWLINE/g, '\n');

			console.log(test);
		}

		function validateTreeClassification(tree, originalTreeText, textAPIOperations, lineAPIOperations) {
			if (tree.classifier.invalidLineStarts.length !== 0) {
				consoleLogTestCase(originalTreeText, textAPIOperations, lineAPIOperations);
				expect(tree.classifier.invalidLineStarts.length).toEqual(0);
			}

			var each = tree.firstLineNode();
			while (each) {
				if (each.type() === 'empty') {
					var nextNotEmpty = each.nextLineNode();
					while (nextNotEmpty && nextNotEmpty.type() === 'empty') {
						nextNotEmpty = nextNotEmpty.nextLineNode();
					}

					if (nextNotEmpty) {
						if (each.typeIndentLevel() !== nextNotEmpty.typeIndentLevel()) {
							consoleLogTestCase(originalTreeText, textAPIOperations, lineAPIOperations);
						}
						expect(each.typeIndentLevel()).toEqual(nextNotEmpty.typeIndentLevel());
					} else {
						if (each.typeIndentLevel() !== 0) {
							consoleLogTestCase(originalTreeText, textAPIOperations, lineAPIOperations);
						}
						expect(each.typeIndentLevel()).toEqual(0);
					}
				} else {
					var eachLevel = each.line().match(/^\t*/)[0].length;
					if (each.typeIndentLevel() !== eachLevel) {
						consoleLogTestCase(originalTreeText, textAPIOperations, lineAPIOperations);
					}
					expect(each.typeIndentLevel()).toEqual(eachLevel);
				}
				each = each.nextLineNode();
			}
		}

		describe('Init', function () {
			it('should classify empty tree', function () {
				tree.ensureClassified();
				expect(tree.classifier.invalidLineStarts.length).toEqual(0);
				expect(tree.firstLineNode().type()).toEqual('empty');
				expect(tree.firstLineNode().typeIndentLevel()).toEqual(0);
			});

			it('should classify single line', function () {
				tree.setText('one');
				tree.ensureClassified();
				expect(tree.firstLineNode().type()).toEqual('item');
				expect(tree.firstLineNode().typeIndentLevel()).toEqual(0);
				tree.setText('\tone');
				tree.ensureClassified();
				expect(tree.firstLineNode().typeIndentLevel()).toEqual(1);
			});

			it('should classify single indented line', function () {
				tree.setText('\tone');
				tree.ensureClassified();
				expect(tree.firstLineNode().type()).toEqual('item');
				expect(tree.firstLineNode().typeIndentLevel()).toEqual(1);
			});

			it('should classify empty line followed by non empty line', function () {
				tree.setText('\n\ttwo');
				tree.ensureClassified();
				expect(tree.firstLineNode().type()).toEqual('empty');
				expect(tree.firstLineNode().typeIndentLevel()).toEqual(1);
				expect(tree.lastLineNode().type()).toEqual('item');
				expect(tree.lastLineNode().typeIndentLevel()).toEqual(1);
			});

			it('should indent empty lines to match next non empty or previous.', function () {
				tree.setText('one\n\n\n\tthree\n\t\tfour\nfive');
				tree.ensureClassified();
				var each = tree.firstLineNode();
				expect(each.typeIndentLevel()).toEqual(0);
				each = each.nextLineNode();
				expect(each.typeIndentLevel()).toEqual(1);
				each = each.nextLineNode();
				expect(each.typeIndentLevel()).toEqual(1);
				each = each.nextLineNode();
				expect(each.typeIndentLevel()).toEqual(1);
				each = each.nextLineNode();
				expect(each.typeIndentLevel()).toEqual(2);
				each = each.nextLineNode();
				expect(each.typeIndentLevel()).toEqual(0);
			});

			it('should never generate textChanged deltas', function () {
				var ignoreTest = false;
				tree.setText('one\n\n\n\tthree\n\t\tfour\nfive');
				tree.addEventListener('treeChanged', function (e) {
					if (ignoreTest) {
						return;
					}
					for (var i = 0; i < e.deltas.length; i++) {
						expect(e.deltas[i].textChanged()).toBeFalsy();
					}
				});
				tree.ensureClassified();
				ignoreTest = true;
				tree.lastLineNode().replaceLineInRange('\t\t', 0, 0);
				ignoreTest = false;
				tree.ensureClassified();
			});

			it('should not fire tree changed event if ensureClassified doesnt reclassifyUp any nodes', function () {
				tree.setText('one\n\n\n\tthree\n\t\tfour\nfive');
				tree.ensureClassified();
				tree.addEventListener('treeChanged', function (e) {
					expect(false).toBeTruthy();
				});
				tree.ensureClassified();
				tree.ensureClassified(tree.firstLineNode().nextLineNode(), null);
				tree.ensureClassified(null, tree.lastLineNode());
				tree.ensureClassified(tree.firstLineNode().nextLineNode(), tree.lastLineNode().previousLineNode());
			});

			it('should fire minimal tree changed event, only including nodes that really changed', function () {
				var ignoreTest = false;
				tree.setText('one\n\n\ttwo\n\n\t\tthree');
				tree.addEventListener('treeChanged', function (e) {
					if (ignoreTest) {
						return;
					}
					expect(e.deltas.length).toEqual(1);
					expect(e.deltas[0].textChanged()).toBeFalsy();
					expect(e.deltas[0].updatedNode).toEqual(tree.firstLineNode());
					expect(e.deltas[0].insertedNodes.length).toEqual(0);
					expect(e.deltas[0].removedNodes.length).toEqual(0);
				});
				ignoreTest = true;
				tree.ensureClassified();
				tree.firstLineNode().replaceLineInRange('new', 0, 0);
				ignoreTest = false;
				tree.ensureClassified();
			});			

			it('should mark all reclassified nodes with _reclassifiedRegion', function () {
				tree.setText('\n\nthree\nfour');
				tree.ensureClassified();
				var node = tree.firstLineNode();
				expect(node._reclassifiedRegion).toBeTruthy();
				node = node.nextLineNode();
				expect(node._reclassifiedRegion).toBeTruthy();
				node = node.nextLineNode();
				expect(node._reclassifiedRegion).toBeTruthy();
				node = node.nextLineNode();
				expect(node._reclassifiedRegion).toBeFalsy();
			});

			it('should move and classify without causing inconsistencies', function () {
				tree.setText('\none\ntwo\n\tthree\n\n	five\n\nseven');
				tree.ensureClassified();

				var root = tree.root,
					one = tree.lineNumberToNode(1),
					two = tree.lineNumberToNode(2),
					five = tree.lineNumberToNode(5),
					secondspace = tree.lineNumberToNode(6);

				root.insertChildBefore(one, secondspace);
				expect(tree.text()).toEqual('\ntwo\n\tthree\n\n	five\none\n\nseven');

				tree.ensureClassified();
				validateTreeClassification(tree);

				two.insertChildBefore(one, null);
				expect(tree.text()).toEqual('\ntwo\n\tthree\n\n	five\n\tone\n\nseven');

				tree.ensureClassified();
				expect(tree.text()).toEqual('\ntwo\n\tthree\n\n	five\n\tone\n\nseven');
			});

			it('should correctly classify randomly generated trees', function () {
				for (var testReps = 0; testReps < 100; testReps++) {
					generateRandomTree(10);
					tree.ensureClassified();
					validateTreeClassification(tree);
				}
			});
		});

		describe('Ensure', function () {
			it('should not crash when ensureClassified in region that will try to classify previous back over ensure start.', function () {
				tree.setText('\n\n\n\nhello');
				var last = tree.lastLineNode();
				var nextToLast = last.previousLineNode();
				tree.ensureClassified(nextToLast, last);
			});
		});

		describe('Update', function () {
			it('should update line type when it changes', function () {
				tree.ensureClassified();
				expect(tree.firstLineNode().type()).toEqual('empty');
				validateTreeClassification(tree);
				tree.firstLineNode().setLine('not empty');
				expect(tree.classifier.invalidLineStarts.length).toEqual(1);
				tree.ensureClassified();
				expect(tree.firstLineNode().type()).toEqual('item');
			});

			it('should not update more than necessary', function () {
				tree.setText('one\n# Two\nthree\n\n# Four\nfive');
				tree.ensureClassified();
				validateTreeClassification(tree);
				tree.lastLineNode().setLine('FIVE');
				expect(tree.classifier.invalidLineStarts.length).toEqual(1);
				var nodes = [];
				Extensions.add('com.foldingtext.taxonomy.classify', {
					classify: function (node, state) {
						nodes.push(node);
					}
				});
				tree.ensureClassified();

				//expect(nodes.length).toEqual(1);
			});

			it('should handle case where last line is updated to be empty and has previous empty lines.', function () {
				tree.setText('\n\n\n\thello world');
				tree.ensureClassified();
				validateTreeClassification(tree);
				tree.lastLineNode().setLine('');
				tree.ensureClassified();
				validateTreeClassification(tree);
			});

			it('should classify appended node.', function () {
				tree.insertNodeBefore(tree.createNode('seven'), null);
				tree.ensureClassified();
				validateTreeClassification(tree);
				tree.lastLineNode().text();
			});

			it('should handle random test case 1', function () {
				tree.setText('one\ntwo\nthree');
				tree.ensureClassified();
				tree.firstLineNode().nextLineNode().setLine('');
				expect(tree.classifier.invalidLineStarts.length).toEqual(1);
				tree.ensureClassified();
				validateTreeClassification(tree);
			});

			it('should handle random test case 2', function () {
				tree.setText('\n\nthree');
				tree.ensureClassified();
				tree.lastLineNode().setLine('\tthree');
				expect(tree.classifier.invalidLineStarts.length).toEqual(1);
				tree.ensureClassified();
				validateTreeClassification(tree);
			});

			it('should handle random test case 3', function () {
				tree.ensureClassified();
				validateTreeClassification(tree);
				tree.replaceTextInRange(tree.createRangeFromLocation(0, 0), '\n12');
				expect(tree.classifier.invalidLineStarts.length).toEqual(2);
				tree.ensureClassified();
				validateTreeClassification(tree);
			});

			it('should handle random test case 4', function () {
				tree.setText('\tone\n\tone\n\tone\n\t\ttwo\n\n\t\ttwo');
				tree.ensureClassified();
				validateTreeClassification(tree);
				tree.replaceTextInRange(tree.createRangeFromLocation(4, 0), '01\n');
				expect(tree.classifier.invalidLineStarts.length).toEqual(2);
				tree.ensureClassified();
				validateTreeClassification(tree);
			});

			it('should handle random test case 5', function () {
				tree.setText('\tone\n\t\tone\n');
				tree.ensureClassified();
				validateTreeClassification(tree);
				tree.removeNode(tree.lineNumberToNode(1));
				tree.ensureClassified();
				validateTreeClassification(tree);
			});

			it('should handle random test case 6', function () {
				tree.setText('\n\t\ttwo');
				tree.ensureClassified();
				validateTreeClassification(tree);
				tree.removeNode(tree.lineNumberToNode(1));
				tree.ensureClassified();
				validateTreeClassification(tree);
			});

			it('should handle random test case 7', function () {
				tree.setText('\tone\n\tone');
				tree.ensureClassified();
				validateTreeClassification(tree);
				tree.replaceTextInRange(tree.createRangeFromLocation(5, 4), '012');
				tree.ensureClassified();
				validateTreeClassification(tree);
			});

			it('should handle random test case 8', function () {
				tree.setText('\t\ttwo\n\tone\n\tone');
				tree.ensureClassified();
				validateTreeClassification(tree);
				tree.beginUpdates();
				tree.replaceTextInRange(tree.createRangeFromLocation(9, 2), '01\n');
				tree.removeNode(tree.lineNumberToNode(1));
				tree.endUpdates();
				tree.ensureClassified();
				validateTreeClassification(tree);
			});

			it('should handle random test case 9', function () {
				tree.setText('\tone\n\t\ttwo\n\tone\n\t\ttwo\n\tone\n\n\tone');
				tree.ensureClassified();
				validateTreeClassification(tree);
				tree.beginUpdates();
				tree.replaceTextInRange(tree.createRangeFromLocation(6, 11), '01\n');
				tree.removeNode(tree.lineNumberToNode(5));
				tree.endUpdates();
				tree.ensureClassified();
				validateTreeClassification(tree);
			});

			it('should handle random test case 10', function () {
				tree.setText('');
				tree.ensureClassified();
				validateTreeClassification(tree);
				tree.replaceTextInRange(tree.createRangeFromLocation(0, 0), '\n12');
				tree.insertNodeBefore(tree.createNode('this is a test line!'), tree.lineNumberToNode(0));
				tree.ensureClassified();
				validateTreeClassification(tree);
			});

			it('should handle random test case 11', function () {
				tree.cleanupForCollection();
				tree = new Tree('', 'outline');
				tree.setText('		two\n	one\n	one\n		two\n	one\n	one');
				tree.ensureClassified();
				validateTreeClassification(tree);
				tree.replaceTextInRange(tree.createRangeFromLocation(24, 7), '012');
				tree.replaceTextInRange(tree.createRangeFromLocation(12, 3), '012');
				tree.replaceTextInRange(tree.createRangeFromLocation(26, 1), '012');
				tree.replaceTextInRange(tree.createRangeFromLocation(5, 16), '\n12');
				tree.insertNodeBefore(tree.createNode('this is a test line!'), tree.lineNumberToNode(0));
				tree.insertNodeBefore(tree.createNode('this is a test line!'), tree.lineNumberToNode(1));
				tree.insertNodeBefore(tree.createNode('this is a test line!'), tree.lineNumberToNode(1));
				tree.insertNodeBefore(tree.createNode('this is a test line!'), tree.lineNumberToNode(2));
				tree.ensureClassified();
				validateTreeClassification(tree);
			});

			it('should handle random test case 12', function () {
				tree.setText('	one\n\n\n');
				tree.ensureClassified();
				validateTreeClassification(tree);
				tree.replaceTextInRange(tree.createRangeFromLocation(1, 4), '01\n');
				tree.replaceTextInRange(tree.createRangeFromLocation(6, 0), '012');
				tree.replaceTextInRange(tree.createRangeFromLocation(0, 8), '012');
				tree.insertNodeBefore(tree.createNode('this is a test line!'), tree.lineNumberToNode(1));
				tree.removeNode(tree.lineNumberToNode(1));
				tree.ensureClassified();
				validateTreeClassification(tree);
			});

			it('should handle random test case 13', function () {
				tree.setText('		two\n	one\n		two\n		two\n		two');
				tree.ensureClassified();
				validateTreeClassification(tree);
				tree.replaceTextInRange(tree.createRangeFromLocation(18, 3), '012');
				tree.replaceTextInRange(tree.createRangeFromLocation(19, 8), '0\n2');
				tree.removeNode(tree.lineNumberToNode(1));
				tree.removeNode(tree.lineNumberToNode(0));
				tree.ensureClassified();
				validateTreeClassification(tree);
			});

			it('should handle random test case 14', function () {
				tree.setText('\n\n		two\n		two');
				tree.ensureClassified();
				validateTreeClassification(tree);
				tree.replaceTextInRange(tree.createRangeFromLocation(6, 3), '012');
				tree.replaceTextInRange(tree.createRangeFromLocation(9, 0), '012');
				tree.insertNodeBefore(tree.createNode('this is a test line!'), tree.lineNumberToNode(1));
				tree.removeNode(tree.lineNumberToNode(1));
				tree.ensureClassified();
				validateTreeClassification(tree); 
			});

			it('should maintain correct classification under random tree mutations', function () {
				function stringWithRandomNewlines() {
					var n = Math.round(Math.random() * 1);
					var s = '012';
					var slength = s.length;
					for (var i = 0; i < n; i++) {
						s = s.replace(Math.round(Math.random() * slength), '\n');
					}
					return s;
				}

				for (var testReps = 0; testReps < 500; testReps++) {
					generateRandomTree(10);
					tree.ensureClassified();
					validateTreeClassification(tree);

					var operations = Math.round(Math.random() * 3),
						originalTreeText = tree.text(),
						textLength = tree.textLength(),
						n,
						i;

					tree.beginUpdates();

					// text api, random replacements.
					var textAPIOperations = [];
					for (i = 0; i < operations; i++) {
						var insertText = stringWithRandomNewlines(),
							insertLength = insertText.length,
							location = Math.round(Math.random() * textLength),
							replaceLength = Math.round(Math.random() * (textLength - location));
						textAPIOperations.push({ location : location, length : replaceLength, insertText : insertText });
						tree.replaceTextInRange(tree.createRangeFromLocation(location, replaceLength), insertText);
						textLength += insertLength - replaceLength;
					}

					// line api, random add / removes
					var lineAPIOperations = [];
					for (i = 0; i < operations; i++) {
						if (Math.random() > 0.5) {
							var insertBeforeLine = Math.round(Math.random() * tree.nodeCount()),
								each = tree.createNode('this is a test line!');
							tree.insertNodeBefore(each, tree.lineNumberToNode(insertBeforeLine));
							textLength += each.line().length;
							lineAPIOperations.push({ insert: insertBeforeLine });
						} else {
							var removeLine = Math.round(Math.random() * tree.nodeCount());
							n = tree.lineNumberToNode(removeLine);
							if (n) {
								lineAPIOperations.push({ remove: removeLine });
								tree.removeNode(n);
							}
						}
					}

					// node api, random add, remove, update
					/*var nodeAPIOperations = [];
					for (i = 0; i < operations; i++) {
						var r = Math.random();
						n = tree.lineNumberToNode(Math.round(Math.random() * tree.nodeCount()));
						if (n) {
							if (r > 0.66) {
								nodeAPIOperations.push({ setLine : tree.nodeToLineNumber(n), oldLine : n.line(), newLine : 'this node has been updated!' });
								n.setLine('this node has been updated!');
							} else if (r > 0.33) {
								nodeAPIOperations.push({ appendChild : tree.nodeToLineNumber(n) });
								n.appendChild(tree.createNode('this node has a new child!'));
							} else {
								nodeAPIOperations.push({ removefromparent : tree.nodeToLineNumber(n) });
								n.removeFromParent();
							}
						}
					}*/

					tree.endUpdates();
					try {
						tree.ensureClassified();
					} catch (e) {
						consoleLogTestCase(originalTreeText, textAPIOperations, lineAPIOperations);
					}
					validateTreeClassification(tree, originalTreeText, textAPIOperations, lineAPIOperations);


					// Repeat same case, useful for debugging.
					/*tree.cleanupForCollection();
					tree = new Tree('', 'outline');
					classifier = tree.classifier;
					tree.setText(originalTreeText);
					classifier.classify(Infinity);
					validateTreeClassification(tree);

					tree.beginUpdates();
					var op;
					for (i = 0; i < textAPIOperations.length; i++) {
						op = textAPIOperations[i];
						tree.replaceTextInRange(tree.createRangeFromLocation(op.location, op.length), op.insertText);
					}

					for (i = 0; i < lineAPIOperations.length; i++) {
						op = lineAPIOperations[i];
						if (op.remove !== undefined) {
							tree.removeNode(tree.lineNumberToNode(op.remove));
						} else {
							tree.insertNodeBefore(tree.createNode('this is a test line!'), tree.lineNumberToNode(op.insert));
						}
					}

					tree.endUpdates();
					classifier.classify(Infinity);
					validateTreeClassification(tree, originalTreeText, textAPIOperations, lineAPIOperations);*/
				}
			});
		});

		describe('Invalidated Line Starts', function () {

			var one, two, three, four, five, six;

			beforeEach(function () {
				tree.setText('one\n\ttwo\n\t\tthree\n\t\tfour\n\tfive\n\t\tsix');
				tree.ensureClassified();
				one = tree.root.firstChild;
				two = one.nextLineNode();
				three = two.nextLineNode();
				four = three.nextLineNode();
				five = four.nextLineNode();
				six = five.nextLineNode();
			});

			it('should calculated invalidated lines after updating node', function () {
				one.setLine('new');
				expect(tree.classifier.invalidLineStarts).toEqual([0]);
			});

			it('should calculated invalidated lines after removing first node', function () {
				tree.removeNode(one);
				expect(tree.classifier.invalidLineStarts).toEqual([0]);
			});

			it('should calculated invalidated lines after removing last node', function () {
				tree.removeNode(six);
				expect(tree.classifier.invalidLineStarts).toEqual([3, 4]);
			});

			it('should calculated invalidated lines from multiple changes', function () {
				tree.beginUpdates();
				tree.removeNode(five);
				tree.removeNode(six);
				tree.removeNode(one);
				tree.endUpdates();
				expect(tree.classifier.invalidLineStarts).toEqual([ 0, 1, 2 ]);
			});

			it('should calculated invalidated lines (and merge them as needed) from multiple at end', function () {
				tree.beginUpdates();
				six.removeFromParent();
				two.removeFromParent();
				tree.endUpdates();
				expect(tree.classifier.invalidLineStarts).toEqual([ 0, 1 ]);
			});

			it('should map from lines for each delta into current tree', function () {
				tree.replaceTextInRange(tree.createRangeFromLocation(3, 0), '\nnew');
				expect(tree.classifier.invalidLineStarts).toEqual([0, 1]);
			});
		});
	});
});