define(function (require) {
	'use strict';

	//	one
	//		two
	//			three @t
	//			four @t
	//		five
	//			six @t(23)

	describe('NodePath', function () {
		var treeText = 'one\n\ttwo\n\t\tthree @t\n\t\tfour @t\n\tfive\n\t\tsix @t(23)',
			MarkdownTaxonomy = require('ft/taxonomy/markdowntaxonomy').MarkdownTaxonomy,
			OutlineTaxonomy = require('ft/taxonomy/outlinetaxonomy').OutlineTaxonomy,
			AttributedString = require('ft/core/attributedstring').AttributedString,
			NodePath = require('ft/core/nodepath').NodePath,
			root, one, two, three, four, five, six,
			Tree = require('ft/core/tree').Tree,
			tree,
			cm;

		var NodeSet;

		beforeEach(function () {
			tree = new Tree(treeText, 'outline');
			tree.ensureClassified();
			root = tree.root;
			one = root.firstChild;
			two = one.nextLineNode();
			three = two.nextLineNode();
			four = three.nextLineNode();
			five = four.nextLineNode();
			six = five.nextLineNode();
		});

		afterEach(function () {
			tree.cleanupForCollection();
		});

		describe('Unions', function () {
			it('should union paths separated by "union"', function () {
				expect(root.evaluateNodePath('//one union //e')).toEqual([one, three, five]);
			});

			it('should union paths separated by "union" where first path ends in *', function () {
				expect(root.evaluateNodePath('/* union //two')).toEqual([one, two]);
			});
		});

		describe('Intersections', function () {
			it('should intersect paths separated by "intersect"', function () {
				expect(root.evaluateNodePath('//three intersect //@t')).toEqual([three]);
			});
		});

		describe('Exceptions', function () {
			it('should minus paths separated by "except"', function () {
				expect(root.evaluateNodePath('//* except //@t')).toEqual([one, two, five]);
			});

			it('should minus paths separated by "except" when expect results proceed orignal results', function () {
				expect(root.evaluateNodePath('(//three except //two)')).toEqual([three]);
			});
		});

		describe('Set Grouping', function () {
			it('should group path set operations with parentheses', function () {
				expect(root.evaluateNodePath('(//one union //two union //three except //two)')).toEqual([one, two, three]);
				expect(root.evaluateNodePath('((//one union //two union //three) except //two)')).toEqual([one, three]);
			});

			it('should allow slicing after set group', function () {
				expect(root.evaluateNodePath('(//@t)[0]')).toEqual([three]);
				expect(root.evaluateNodePath('(//@t union //e)[0]')).toEqual([one]);				
			});
		});

		describe('Locations', function () {
			it('should evaluate relative paths relative to currentNode', function () {
				expect(root.evaluateNodePath('one')).toEqual([one]);
				expect(four.evaluateNodePath('one')).toEqual([]);
			});

			it('should evaluate absoluote paths relative to currentNode root', function () {
				expect(root.evaluateNodePath('/one')).toEqual([one]);
				expect(four.evaluateNodePath('/one')).toEqual([one]);
			});

			it('should evaluate path starting with // as absoluote', function () {
				expect(four.evaluateNodePath('//one')).toEqual([one]);
			});
		});

		describe('Axes', function () {
			it('should parse error if axis doesnt have a predicate', function () {
				expect(NodePath.parse('/ancestor-or-self::').error.column).toEqual(20);
			});

			it('should evaluate ancestor-or-self axis for ancestor-or-self::', function () {
				expect(three.evaluateNodePath('ancestor-or-self::*')).toEqual([root, one, two, three]);
			});

			it('should evaluate ancestor axis for ancestor::', function () {
				expect(three.evaluateNodePath('ancestor::*')).toEqual([root, one, two]);
			});

			it('should evaluate ancestor axis from multiple contexts and merge results', function () {
				expect(three.evaluateNodePath('//one or two/ancestor::*')).toEqual([root, one]);
			});

			it('should evaluate descendant axis from multiple contexts and correctly merge results', function () {
				expect(three.evaluateNodePath('//one or two//*')).toEqual([two, three, four, five, six]);
			});

			it('should evaluate child axis as default or child::', function () {
				expect(two.evaluateNodePath('*')).toEqual([three, four]);
				expect(two.evaluateNodePath('child::*')).toEqual([three, four]);
			});

			it('should evaluate descendant-or-self axis for descendant-or-self::', function () {
				expect(two.evaluateNodePath('descendant-or-self::*')).toEqual([two, three, four]);
			});

			it('should evaluate decendent axis for // or descendant::', function () {
				expect(one.evaluateNodePath('/one//*')).toEqual([two, three, four, five, six]);
				expect(one.evaluateNodePath('/one/descendant::*')).toEqual([two, three, four, five, six]);
			});

			it('should evaluate filter descendants axis for /// or filter-descendants::', function () {
				tree.insertNodeBefore(tree.createNode('seven'), null);
				tree.ensureClassified();
				expect(one.evaluateNodePath('/one///four')).toEqual([one, two, four]);
			});

			it('should evaluate filter descendants and then be able to walk back up ancestors', function () {
				tree.setText('one\n\ttwo');
				tree.ensureClassified();
				one = tree.firstLineNode();
				two = one.nextLineNode();
				expect(one.evaluateNodePath('//two///*/ancestor-or-self::*')).toEqual([root, one, two]);
			});

			it('should evaluate filter ///@t or filter-descendants::', function () {
				five.removeFromParent();
				tree.ensureClassified();
				expect(one.evaluateNodePath('///@t')).toEqual([root, one, two, three, four]);

				four.removeFromParent();
				tree.ensureClassified();
				expect(one.evaluateNodePath('///@t')).toEqual([root, one, two, three]);
			});

			it('should evaluate following-sibling axis for following-sibling::', function () {
				expect(one.evaluateNodePath('/one/two/following-sibling::*')).toEqual([five]);
			});

			it('should evaluate following axis for following::', function () {
				expect(one.evaluateNodePath('/one/two/following::*')).toEqual([three, four, five, six]);
			});

			it('should evaluate parent axis for .. or parent::', function () {
				expect(two.evaluateNodePath('..*')).toEqual([one]);
				expect(two.evaluateNodePath('parent::*')).toEqual([one]);
			});

			it('should evaluate preceding-sibling axis for preceding-sibling::', function () {
				expect(one.evaluateNodePath('/one/five/preceding-sibling::*')).toEqual([two]);
			});

			it('should evaluate preceding axis for preceding::', function () {
				expect(one.evaluateNodePath('/one/five/preceding::*')).toEqual([one, two, three, four]);
			});

			it('should evaluate self axis for self', function () {
				expect(one.evaluateNodePath('/one/five/self::*')).toEqual([five]);
			});

			it('should support taxonomy declared type shortcut', function () {
				expect(one.evaluateNodePath('//item')).toEqual([one, two, three, four, five, six]);
				expect(one.evaluateNodePath('//item two')).toEqual([two]);
			});
		});

		describe('Predicate', function () {
			describe('Structure', function () {
				it('should accept a complete attribute, relation, value, slice predicate', function () {
					expect(root.evaluateNodePath('//@line = "one"[0]')).toEqual([one]);
				});

				it('should accept a attribute, relation, value predicate', function () {
					expect(root.evaluateNodePath('//@line = "one"')).toEqual([one]);
				});

				it('should default to case insensitive comparisons', function () {
					expect(root.evaluateNodePath('//@line = "oNe"')).toEqual([one]);
				});

				it('should accept attribute value predicates', function () {
					expect(root.evaluateNodePath('//@line one')).toEqual([one]);
				});

				it('should accept value predicates', function () {
					expect(root.evaluateNodePath('//one')).toEqual([one]);
				});

				it('should default to @line:text if no attribute is specified', function () {
					expect(root.evaluateNodePath('//= one')).toEqual([one]);
				});

				it('should default to contians if no relation is specified', function () {
					expect(root.evaluateNodePath('//@t 23')).toEqual([six]);
				});

				it('should test attribute for existance of no value is specified', function () {
					expect(root.evaluateNodePath('//@t')).toEqual([three, four, six]);
				});
			});

			describe('Boolean', function () {
				it('should evaluate predicates joined with "and"', function () {
					expect(root.evaluateNodePath('//w and o')).toEqual([two]);
				});

				it('should evaluate predicates joined with "or"', function () {
					expect(root.evaluateNodePath('//w or i')).toEqual([two, five, six]);
				});

				it('should evaluate predicates preceeded by "not"', function () {
					expect(root.evaluateNodePath('//not e')).toEqual([two, four, six]);
				});

				it('should accept any number of consecutive negates as a single not', function () {
					expect(root.evaluateNodePath('/one/two')).toEqual([two]);
					expect(root.evaluateNodePath('/one/not two')).toEqual([five]);
					expect(root.evaluateNodePath('/one/not not two')).toEqual([two]);
					expect(root.evaluateNodePath('/one/not not not two')).toEqual([five]);
				});

				it('should handle and or not in the proper order', function () {
					expect(root.evaluateNodePath('//(@line:text e and not @t) or @t = 23')).toEqual([one, five, six]);
					expect(root.evaluateNodePath('//e and @t or not @t')).toEqual([one, two, three, five]);
					expect(root.evaluateNodePath('//(e and @t) or not @t')).toEqual([one, two, three, five]);
					expect(root.evaluateNodePath('//e and (@t or not v)')).toEqual([one, three]);
				});

				it('should handle boolean grouping without leading axis', function () {
					expect(new NodePath('(a and b) and c').toString()).toEqual('(a and b) and c');
				});
			});

			describe('Nesting', function () {
				it('should accept parenthesis around predicates', function () {
					expect(root.evaluateNodePath('//(not @t)')).toEqual([one, two, five]);
					expect(root.evaluateNodePath('//((not @t))')).toEqual([one, two, five]);
					expect(root.evaluateNodePath('//(((not @t)) or @t = 23)')).toEqual([one, two, five, six]);
				});

				it('should accept negated groups', function () {
					expect(root.evaluateNodePath('//not (@t or e)')).toEqual([two]);
				});
			});

			describe('Attributes', function () {
				it('should consider built in attriubte @type, @line, @index, @id', function () {
					expect(root.evaluateNodePath('//@type = item')).toEqual([one, two, three, four, five, six]);
					expect(root.evaluateNodePath('//@index = 1')).toEqual([four, five]);
				});

				it('should work correctly for @index=0 case', function () {
					expect(root.evaluateNodePath('/@index = 0')).toEqual([one]);
				});

				it('should consider tag attributes defined in line @tag(value)', function () {
					expect(root.evaluateNodePath('//@t = 23')).toEqual([six]);
				});

				describe('Properties', function () {

					//I am **bold** and I am *italic*
					//	two
					//		my : value
					//			other : 2
					//	five
					//		six

					beforeEach(function () {
						// Need to load in FoldingText taxonomy for these cases
						var Taxonomies = require('ft/core/taxonomies'),
							taxonomy = Taxonomies.taxonomy({
								foldingtext: true,
								multimarkdown: true,
								gitmarkdown: true,
								criticMarkup: true
							}, 'markdown');

						tree = new Tree(treeText, taxonomy);
						root = tree.root;
						one = root.firstChild;
						two = one.nextLineNode();
						three = two.nextLineNode();
						four = three.nextLineNode();
						five = four.nextLineNode();
						six = five.nextLineNode();
						one.setLine('I am **bold** and I am *italic*');
						three.setLine('\t\tmy : 1');
						four.setLine('\t\t\tother : 2');
						tree.ensureClassified();
					});

					afterEach(function () {
						tree.cleanupForCollection();
					});

					it('should consider @property paths by examining nodes property child nodes', function () {
						expect(root.evaluateNodePath('//@property:my')).toEqual([two]);
						expect(root.evaluateNodePath('//@property:my 1')).toEqual([two]);
						expect(root.evaluateNodePath('//@property:my = 1')).toEqual([two]);
						expect(root.evaluateNodePath('//@property:my != 1')).toEqual([one, three, four, five, six]);
						expect(root.evaluateNodePath('//@property:my < 1')).toEqual([]);
						expect(root.evaluateNodePath('//@property:my > 1')).toEqual([]);
						expect(root.evaluateNodePath('//@property:my:other')).toEqual([two]);
					});

					it('should consider span attributes such as **bold** and *italic* in line', function () {
						expect(root.evaluateNodePath('//@line:strong')).toEqual([one]);
						expect(root.evaluateNodePath('//@line:strong = bold')).toEqual([one]);
						expect(root.evaluateNodePath('//@line:em')).toEqual([one]);
						expect(root.evaluateNodePath('//@line:em = italic')).toEqual([one]);
						expect(root.evaluateNodePath('//@line:em = "notfound"')).toEqual([]);
						expect(root.evaluateNodePath('//@line:notfound')).toEqual([]);
					});
				});
			});

			//	one
			//		two
			//			three @t(09)
			//			four @t
			//		five
			//			six @t(23)
			describe('Relations', function () {
				beforeEach(function () {
					three.addTag('t', '09');
					tree.ensureClassified();
				});

				it('should support =', function () {
					expect(root.evaluateNodePath('//= one')).toEqual([one]);
				});

				it('should support !=', function () {
					expect(root.evaluateNodePath('//@t != 23')).toEqual([one, two, three, four, five]);
				});

				it('should support <', function () {
					expect(root.evaluateNodePath('//@t < 23')).toEqual([three, four]);
				});

				it('should support <=', function () {
					expect(root.evaluateNodePath('//@t <= 23')).toEqual([three, four, six]);
				});

				it('should support >', function () {
					expect(root.evaluateNodePath('//@t > 09')).toEqual([six]);
				});

				it('should support >=', function () {
					expect(root.evaluateNodePath('//@t >= 09')).toEqual([three, six]);
				});

				it('should support beginswith', function () {
					expect(root.evaluateNodePath('//beginswith o')).toEqual([one]);
				});

				it('should support contains', function () {
					expect(root.evaluateNodePath('//contains ne')).toEqual([one]);
				});

				it('should support endswith', function () {
					expect(root.evaluateNodePath('//endswith ne')).toEqual([one]);
				});

				it('should support matches', function () {
					expect(root.evaluateNodePath('//matches ".*i.*"')).toEqual([five, six]);
				});

				it('should internally handle exception when matches is given a bad regex', function () {
					expect(root.evaluateNodePath('//@line matches " @\\\\w("')).toEqual([]);
				});				
			});

			describe('Optional Options', function () {
				it('should support options (AROV) formatted queries', function () {
					one.setLine('Being all INSENSITIVE');
					tree.ensureClassified();
					expect(root.evaluateNodePath('@line contains insensitive')).toEqual([one]);
					expect(root.evaluateNodePath('@line contains [s] insensitive')).toEqual([]);
					//expect(root.evaluateNodePath('@line contains [s] INSENSITIVE')).toEqual([one]);
				});

				it('should support AOV formatted queries', function () {
					one.setLine('INSENSITIVE');
					tree.ensureClassified();
					expect(root.evaluateNodePath('@line insensitive')).toEqual([one]);
					expect(root.evaluateNodePath('@line [s] insensitive')).toEqual([]);
					expect(root.evaluateNodePath('@line [s] INSENSITIVE')).toEqual([one]);
				});

				it('should support OV formatted queries', function () {
					one.setLine('INSENSITIVE');
					tree.ensureClassified();
					expect(root.evaluateNodePath('insensitive')).toEqual([one]);
					expect(root.evaluateNodePath('[s] insensitive')).toEqual([]);
					expect(root.evaluateNodePath('[s] INSENSITIVE')).toEqual([one]);
				});

				it('should support ROV formatted queries', function () {
					one.setLine('INSENSITIVE');
					tree.ensureClassified();
					expect(root.evaluateNodePath('= insensitive')).toEqual([one]);
					expect(root.evaluateNodePath('= [s] insensitive')).toEqual([]);
					expect(root.evaluateNodePath('= [s] INSENSITIVE')).toEqual([one]);
				});

				it('should support convert to number before compare option', function () {
					one.setLine('1.0');
					tree.ensureClassified();
					expect(root.evaluateNodePath('@line = 1.0')).toEqual([one]);
					expect(root.evaluateNodePath('@line = 1')).toEqual([]);
					expect(root.evaluateNodePath('@line = [n] 1')).toEqual([one]);
				});

				it('should support convert to date before compare option', function () {
					one.setLine('November 1, 2012');
					tree.ensureClassified();
					expect(root.evaluateNodePath('@line = November 1, 2012')).toEqual([one]);
					expect(root.evaluateNodePath('@line = November 01, 2012')).toEqual([]);
					expect(root.evaluateNodePath('@line = [d] November 01, 2012')).toEqual([one]);
				});
			});

			describe('Values', function () {
				it('should accept unquoted values', function () {
					one.setLine('find this string');
					tree.ensureClassified();
					expect(root.evaluateNodePath('find this string')).toEqual([one]);
				});

				it('should ignore reserved words in unquoted values unless followed by whitespace', function () {
					one.setLine('notandor');
					tree.ensureClassified();
					expect(root.evaluateNodePath('notandor')).toEqual([one]);
				});

				it('should accept ored unquoted strings', function () {
					one.setLine('one is here');
					two.setLine('two is here');
					tree.ensureClassified();
					expect(root.evaluateNodePath('//one is here or two is here')).toEqual([one, two]);
				});

				it('should identify double quoted strings', function () {
					expect(root.evaluateNodePath('"one"')).toEqual([one]);
				});

				it('should allow escaping of single and double quotes', function () {
					one.setLine('on\'e');
					tree.ensureClassified();
					expect(root.evaluateNodePath(NodePath.calculateMinNodePath(one))).toEqual([one]);
				});

				it('should not find operator inside quoted string', function () {
					one.setLine('one is here');
					two.setLine('two is here');
					three.setLine('one is here or two is here');
					tree.ensureClassified();
					expect(root.evaluateNodePath('//"one is here or two is here"')).toEqual([three]);
				});

				it('should allow attribute value to start with zero', function () {
					one.setLine('0one');
					tree.ensureClassified();
					expect(root.evaluateNodePath('0')).toEqual([one]);
					expect(root.evaluateNodePath('0one')).toEqual([one]);
				});

				it('should support tags and values ending with numbers', function () {
					one.addTag('t1', 'v1');
					tree.ensureClassified();
					expect(root.evaluateNodePath('//@t1 = v1')).toEqual([one]);
				});

				it('should accept quoted unicode strings', function () {
					var	umlat = tree.createNode('find a ü');
					tree.insertNodeBefore(umlat, null);
					expect(root.evaluateNodePath('//"ü"')).toEqual([umlat]);
				});

				it('should accept unquoted unicode strings', function () {
					// This test is breaking. Need to update nodepathparser.pegjs to use "letter" with unicode
					// sequences. also should update to latest pegjs while at it. Can't remember why I commented out
					// unicode sequences... mabye because of file size... or maybe was breaking something else.
					// so be on lookout for other breask when that's reenabled.
					var	umlat = tree.createNode('find a ü'),
						arabic = tree.createNode('find in arabic بِسْمِ ٱللّٰهِ ٱلرَّحْمـَبنِ ٱلرَّحِيمِ'),
						russian = tree.createNode('find this string in russian По оживлённым берегам'),
						francais = tree.createNode('find in français'),
						german = tree.createNode('find in german dürfen Anführungszeichen'),
						accents = tree.createNode('accept accents aàáâãäāăąȧXǎȁȃeèéêẽëēĕęėXěȅȇiìíîĩïīĭįiXǐȉȋoòóôõöŏǫȯőǒȍȏuùúûũüūŭųXűǔȕȗyỳýŷỹÿȳXXẏXXXX'),
						greek = tree.createNode('find in ancient greek τὰ πάντʼ ἂν ἐξήκοι σαφῆ'),
						sanskrit = tree.createNode('find in sanskrit पशुपतिरपि तान्यहानि कृच्छ्राद्'),
						chinese = tree.createNode('find in chinese 其為人也孝弟 而好犯上者 鮮矣'),
						tamil = tree.createNode('find in tamil ஸ்றீனிவாஸ ராமானுஜன் ஐயங்கார்');

					tree.insertNodeBefore(umlat, null);
					tree.insertNodeBefore(arabic, null);
					tree.insertNodeBefore(russian, null);
					tree.insertNodeBefore(francais, null);
					tree.insertNodeBefore(german, null);
					tree.insertNodeBefore(accents, null);
					tree.insertNodeBefore(greek, null);
					tree.insertNodeBefore(sanskrit, null);
					tree.insertNodeBefore(chinese, null);
					tree.insertNodeBefore(tamil, null);


					expect(root.evaluateNodePath('//find a ü')).toEqual([umlat]);
					expect(root.evaluateNodePath('//find in arabic بِسْمِ ٱللّٰهِ ٱلرَّحْمـَبنِ ٱلرَّحِيمِ')).toEqual([arabic]);
					expect(root.evaluateNodePath('//find this string in russian По оживлённым берегам')).toEqual([russian]);
					expect(root.evaluateNodePath('//find in français')).toEqual([francais]);
					expect(root.evaluateNodePath('//find in german dürfen Anführungszeichen')).toEqual([german]);
					expect(root.evaluateNodePath('//accept accents aàáâãäāăąȧXǎȁȃeèéêẽëēĕęėXěȅȇiìíîĩïīĭįiXǐȉȋoòóôõöŏǫȯőǒȍȏuùúûũüūŭųXűǔȕȗyỳýŷỹÿȳXXẏXXXX')).toEqual([accents]);
					expect(root.evaluateNodePath('//find in ancient greek τὰ πάντʼ ἂν ἐξήκοι σαφῆ')).toEqual([greek]);
					expect(root.evaluateNodePath('//find in sanskrit पशुपतिरपि तान्यहानि कृच्छ्राद्')).toEqual([sanskrit]);
					expect(root.evaluateNodePath('//find in chinese 其為人也孝弟 而好犯上者 鮮矣')).toEqual([chinese]);
					expect(root.evaluateNodePath('//find in tamil ஸ்றீனிவாஸ ராமானுஜன் ஐயங்கார்')).toEqual([tamil]);
				});
			});
		});

		describe('Slicing', function () {
			it('should slice from start', function () {
				expect(tree.evaluateNodePath('//@t[0]')).toEqual([three]);
				expect(tree.evaluateNodePath('//@t[1]')).toEqual([four]);
				expect(tree.evaluateNodePath('//@t[2]')).toEqual([six]);
			});

			it('should clamp slice from start', function () {
				expect(tree.evaluateNodePath('//@t[300]')).toEqual([six]);
			});

			it('should slice from end', function () {
				expect(tree.evaluateNodePath('//@t[-1]')).toEqual([six]);
				expect(tree.evaluateNodePath('//@t[-2]')).toEqual([four]);
				expect(tree.evaluateNodePath('//@t[-3]')).toEqual([three]);
			});

			it('should clamp slice from end', function () {
				expect(tree.evaluateNodePath('//@t[-300]')).toEqual([three]);
			});

			it('should slice range', function () {
				expect(tree.evaluateNodePath('//@t[0:1]')).toEqual([three]);
				expect(tree.evaluateNodePath('//@t[0:2]')).toEqual([three, four]);
				expect(tree.evaluateNodePath('//@t[0:3]')).toEqual([three, four, six]);
				expect(tree.evaluateNodePath('//@t[0:300]')).toEqual([three, four, six]);
				expect(tree.evaluateNodePath('//@t[0:-1]')).toEqual([three, four]);
				expect(tree.evaluateNodePath('//@t[0:-2]')).toEqual([three]);
				expect(tree.evaluateNodePath('//@t[0:-3]')).toEqual([]);
				expect(tree.evaluateNodePath('//@t[0:]')).toEqual([three, four, six]);
			});
		});

		describe('Reported Error Cases', function () {
			it('should return empty array when evaluating bad node path', function () {
				expect(root.evaluateNodePath('/////union')).toEqual([]);
			});
		});

		describe('To String', function () {
			it('should convert path object to path string', function () {
				expect(new NodePath('/*').toString()).toEqual('/*');

				expect(new NodePath('a union b').toString()).toEqual('a union b');
				expect(new NodePath('a intersect b').toString()).toEqual('a intersect b');
				expect(new NodePath('a except b').toString()).toEqual('a except b');

				expect(new NodePath('a union b union c').toString()).toEqual('a union (b union c)');
				expect(new NodePath('(a union b) union c').toString()).toEqual('(a union b) union c');

				expect(new NodePath('//a').toString()).toEqual('//a');
				expect(new NodePath('a//b').toString()).toEqual('a//b');
				expect(new NodePath('///a').toString()).toEqual('///a');
				expect(new NodePath('a///b').toString()).toEqual('a///b');
				expect(new NodePath('a/..b').toString()).toEqual('a/..b');
				expect(new NodePath('..b').toString()).toEqual('..b');
				expect(new NodePath('/a/ancestor::b').toString()).toEqual('/a/ancestor::b');
				expect(new NodePath('a/ancestor::b///c').toString()).toEqual('a/ancestor::b///c');

				expect(new NodePath('@line = one').toString()).toEqual('@line = one');
				expect(new NodePath('= one').toString()).toEqual('= one');
				expect(new NodePath('= oNe').toString()).toEqual('= oNe');
				expect(new NodePath('oNe').toString()).toEqual('oNe');

				expect(new NodePath('a and b').toString()).toEqual('a and b');
				expect(new NodePath('a or b').toString()).toEqual('a or b');
				expect(new NodePath('not a').toString()).toEqual('not a');

				expect(new NodePath('a and b and c').toString()).toEqual('a and (b and c)');
				expect(new NodePath('not (a and b)').toString()).toEqual('not (a and b)');

				expect(new NodePath('@done').toString()).toEqual('@done');
				expect(new NodePath('@line:strong').toString()).toEqual('@line:strong');

				expect(new NodePath('> 10').toString()).toEqual('> 10');
				expect(new NodePath('@line matches 10').toString()).toEqual('@line matches 10');

				expect(new NodePath('= [s] insensitive').toString()).toEqual('= [s] insensitive');

				expect(new NodePath('/or').toString()).toEqual('/or');
				expect(new NodePath('/"not a"').toString()).toEqual('/"not a"');
				expect(new NodePath('/other things').toString()).toEqual('/other things');
				expect(new NodePath('/"quoted strings"').toString()).toEqual('/quoted strings');
			});
		});

		describe('Min Paths', function () {
			it('should calculate the min nodepath that uniquly specifies a given node', function () {
				two.setLine('\tnot( needs quotes in nodepath');
				tree.ensureClassified();
				expect(NodePath.calculateMinNodePath(four)).toEqual('/one/"not("/four');
				expect(tree.evaluateNodePath(NodePath.calculateMinNodePath(four))).toEqual([four]);
			});

			it('should calculate the min nodepath using node id if node has no text', function () {
				two.setLine('\tnot( needs quotes in nodepath');
				four.setLine('\t\t@one @two');
				tree.ensureClassified();
				expect(NodePath.calculateMinNodePath(four)).toEqual('/one/"not("/@id = 4');
				expect(tree.evaluateNodePath(NodePath.calculateMinNodePath(four))).toEqual([four]);
			});

			it('should calculate the min nodepath using node id if two candidate nodes have same text', function () {
				five.setLine('\ttwo');
				tree.ensureClassified();
				var path = NodePath.calculateMinNodePath(five);
				expect(path.indexOf('id = ')).not.toEqual(-1);
				expect(tree.evaluateNodePath(path)).toEqual([five]);
			});

			it('should calculate the min nodepath using node id if two candidate nodes have same text, just different case', function () {
				five.setLine('\tTwo');
				tree.ensureClassified();
				var path = NodePath.calculateMinNodePath(five);
				expect(path.indexOf('id = ')).not.toEqual(-1);
				expect(tree.evaluateNodePath(path)).toEqual([five]);
			});
		});

		describe('Location & Filter', function () {
			it('should get location path string, part before ///', function () {
				expect(new NodePath('///filter').locationPathString()).toEqual('');
				expect(new NodePath('/location///filter').locationPathString()).toEqual('/location');
				expect(new NodePath('/location/location').locationPathString()).toEqual('/location/location');
			});

			it('should set location path string, part before ///', function () {
				expect(new NodePath('/location///filter').setLocationPathString('one/two').toString()).toEqual('one/two///filter');
				expect(new NodePath('/location').setLocationPathString('one/two').toString()).toEqual('one/two');
				expect(new NodePath('///filter').setLocationPathString('one/two').toString()).toEqual('one/two///filter');
				expect(new NodePath('/location///filter').setLocationPathString('').toString()).toEqual('///filter');
			});

			it('should get filter step string, part after ///', function () {
				expect(new NodePath('///filter').filterStepString()).toEqual('filter');
				expect(new NodePath('/location///filter').filterStepString()).toEqual('filter');
				expect(new NodePath('/location/location').filterStepString()).toEqual('');
			});

			it('should set filter step string, part after ///', function () {
				expect(new NodePath('///filter').setFilterStepString('new filter').toString()).toEqual('///new filter');
				expect(new NodePath('location///filter').setFilterStepString('new filter').toString()).toEqual('location///new filter');
				expect(new NodePath('location///filter').setFilterStepString(null).toString()).toEqual('location');
			});

			it('should be able to focus out based on location and filter', function () {
				expect(new NodePath('a union b').updateNodePathByFocusingOutOneLevel().toString()).toEqual('a///*');
				expect(new NodePath('/a').updateNodePathByFocusingOutOneLevel().toString()).toEqual('/a///*');
				expect(new NodePath('/a/b').updateNodePathByFocusingOutOneLevel().toString()).toEqual('/a/b///*');
				expect(new NodePath('/a/b///*').updateNodePathByFocusingOutOneLevel().toString()).toEqual('/a///*');
				expect(new NodePath('/a/b///mouse').updateNodePathByFocusingOutOneLevel().toString()).toEqual('/a/b///*');
				expect(new NodePath('/"abc/o"').updateNodePathByFocusingOutOneLevel().toString()).toEqual('/"abc/o"///*');
			});
		});

		describe('Syntax Highlighting', function () {
			it('should syntax highlight node paths', function () {
				var attributedString = new AttributedString('@a =[i] b');
				NodePath.syntaxHighlight(attributedString, null, null);
				expect(attributedString.toString()).toEqual('(@a/attributepath)( /)(=/relation)([i]/modifier)( /)(b/value)');
			});

			//it('should syntax highlight invalid node paths up to error point', function () {
			//	var attributedString = new AttributedString('Hello/a"/b');
			//	NodePath.syntaxHighlight(attributedString, null, null);
			//	expect(attributedString.toString()).toEqual('(Hello/chars)(//axis)(a/chars)("/b/invalid)');
			//});
		});
	});
});