/*
 * Copyright (c) 2011-2012 Hog Bay Software. All rights reserved.
 */

define(function (require) {
	'use strict';

/*
	one
		two
			three
			four
		five
			six
*/

	describe('Range', function () {
		var NodeSet = require('ft/core/nodeset').NodeSet,
			Range = require('ft/core/range').Range,
			Node = require('ft/core/node').Node,
			Tree = require('ft/core/tree').Tree;

		var treeText = 'one\n\ttwo\n\t\tthree\n\t\tfour\n\tfive\n\t\tsix',
			one, two, three, four, five, six,
			tree;

		beforeEach(function () {
			tree = new Tree(treeText, 'outline');
			one = tree.root.firstChild;
			two = one.nextLineNode();
			three = two.nextLineNode();
			four = three.nextLineNode();
			five = four.nextLineNode();
			six = five.nextLineNode();
		});

		afterEach(function () {
			tree.cleanupForCollection();
		});

		it('should know start node text before range', function () {
			var range = tree.createRangeFromLocation(0, 0);
			expect(range.startNodeTextBeforeRange()).toEqual('');
			range = tree.createRangeFromLocation(1, 0);
			expect(range.startNodeTextBeforeRange()).toEqual('o');
			range = tree.createRangeFromLocation(3, 0);
			expect(range.startNodeTextBeforeRange()).toEqual('one');
		});

		it('should know end node text after range', function () {
			var range = tree.createRangeFromLocation(0, 0);
			expect(range.endNodeTextAfterRange()).toEqual('one');
			range = tree.createRangeFromLocation(1, 0);
			expect(range.endNodeTextAfterRange()).toEqual('ne');
			range = tree.createRangeFromLocation(3, 0);
			expect(range.endNodeTextAfterRange()).toEqual('');
		});

		it('should know node after end node', function () {
			var range = tree.createRangeFromLocation(0, 0);
			expect(range.nodeAfterEndNode()).toEqual(two);
		});

		it('should be equivalent to create from locations or nodes', function () {
			var range1 = tree.createRangeFromLocation(3, 1);
			var range2 = tree.createRangeFromNodes(one, 3, two, 0);
			expect(range1.location()).toEqual(range2.location());
			expect(range1.length()).toEqual(range2.length());
			expect(range1.startNode).toEqual(range2.startNode);
			expect(range1.startOffset).toEqual(range2.startOffset);
			expect(range1.endNode).toEqual(range2.endNode);
			expect(range1.endOffset).toEqual(range2.endOffset);
		});

		it('should be able to mark intermediate nodes from the tree', function () {
			var range1 = tree.createRangeFromNodes(two, 1, five, 4);
			var effected = tree._nodesAndTextSegmentsInRange(range1);
			expect(effected.lines.join('\n')).toEqual('two\n\t\tthree\n\t\tfour\n\tfiv');
			expect(effected.nodes[0]).toEqual(two);
			expect(effected.nodes[1]).toEqual(three);
			expect(effected.nodes[2]).toEqual(four);
			expect(effected.nodes[3]).toEqual(five);
		});

		it('should call block once for each line in range', function () {
			var range = tree.createRangeFromNodes(one, 0, one, 0),
				count = 0,
				expectedLine = 0;
			range.forEachNodeInRange(function (node, line, location, length) {
				count++;
				expect(expectedLine).toEqual(line);
			});
			expect(count).toEqual(1);

			range = tree.createRangeFromNodes(one, 0, two, 0);
			count = 0;
			range.forEachNodeInRange(function (node, line, location, length) {
				count++;
				expect(expectedLine).toEqual(line);
				expectedLine++;
			});
			expect(count).toEqual(2);

			range = tree.createRangeFromNodes(one, 0, three, 0);
			count = 0;
			range.forEachNodeInRange(function (node, line, location, length) {
				count++;
			});
			expect(count).toEqual(3);
		});

		describe('Extend and Contract Selection', function () {

			it('should select to word boundaries', function () {
				one.setLine('Hello world. Look at this.');
				expect(tree.createRangeFromLocation(0, 0).rangeByExtendingToWord().textInRange()).toEqual('Hello');
				expect(tree.createRangeFromLocation(0, 0).rangeByExtendingToWord().rangeByExtendingToWord().textInRange()).toEqual('Hello');
			});

			it('should select to sentence boundaries', function () {
				one.setLine('Hello world. Look at this.');
				expect(tree.createRangeFromLocation(0, 0).rangeByExtendingToSentence().textInRange()).toEqual('Hello world. ');
				expect(tree.createRangeFromLocation(0, 0).rangeByExtendingToSentence().rangeByExtendingToSentence().textInRange()).toEqual('Hello world. ');
			});

			it('should select to paragraph boundaries', function () {
				one.setLine('Hello world. Look at this.');
				expect(tree.createRangeFromLocation(0, 0).rangeByExtendingToParagraph().textInRange()).toEqual('Hello world. Look at this.\n');
				expect(tree.createRangeFromLocation(0, 0).rangeByExtendingToParagraph().rangeByExtendingToParagraph().textInRange()).toEqual('Hello world. Look at this.\n');
			});

			it('should select to line boundaries', function () {
				one.setLine('Hello world. Look at this.');
				expect(tree.createRangeFromLocation(0, 0).rangeByExtendingToLine().textInRange()).toEqual('Hello world. Look at this.');
				expect(tree.createRangeFromLocation(0, 0).rangeByExtendingToLine().rangeByExtendingToLine().textInRange()).toEqual('Hello world. Look at this.');
			});

			it('should select to branch boundaries', function () {
				tree.ensureClassified();
				expect(tree.createRangeFromNodes(two, 0, two, 0).rangeByExtendingToBranch().textInRange()).toEqual('\ttwo\n\t\tthree\n\t\tfour\n');
				expect(tree.createRangeFromNodes(two, 0, two, 0).rangeByExtendingToBranch().rangeByExtendingToBranch().textInRange()).toEqual('\ttwo\n\t\tthree\n\t\tfour\n');
			});

			it('should select to sibling branches boundaries', function () {
				tree.ensureClassified();
				expect(tree.createRangeFromNodes(two, 0, two, 0).rangeByExtendingToSiblingBranches().textInRange()).toEqual('\ttwo\n\t\tthree\n\t\tfour\n\tfive\n\t\tsix');
				expect(tree.createRangeFromNodes(two, 0, two, 0).rangeByExtendingToSiblingBranches().rangeByExtendingToSiblingBranches().textInRange()).toEqual('\ttwo\n\t\tthree\n\t\tfour\n\tfive\n\t\tsix');
			});

			//it('should select to parent branches boundaries', function () {
			//	one.setLine('Hello world. Look at this.');
			//	expect(tree.createRangeFromLocation(0, 0).rangeByExtendingToLine().textInRange()).toEqual('Hello world. Look at this.');
			//	expect(tree.createRangeFromLocation(0, 0).rangeByExtendingToLine().rangeByExtendingToLine().textInRange()).toEqual('Hello world. Look at this.');
			//});

/*
			it('should expand selection to content of strong span before including syntax', function () {
				one.setLine('**one two**');
				treeView.setSelectedRange(tree.createRangeFromNodes(one, 2, one, 2));
				treeView.extendSelection();
				expect(treeView.selectedText()).toEqual('one');
				treeView.extendSelection();
				expect(treeView.selectedText()).toEqual('one two');
				treeView.extendSelection();
				expect(treeView.selectedText()).toEqual('**one two**');
			});

			it('should expand selection at end of line that ends with period without crashing', function () {
				one.setLine('a.');
				treeView.setSelectedRange(tree.createRangeFromNodes(one, 2, one, 2));
				treeView.extendSelection();
				expect(treeView.selectedText()).toEqual('a.');
			});

			it('should expand and contract selection based on syntax as well as word, sentence, line, and tree structure', function () {
				three.setLine('\t\tHello *world*. Look *at* this.');
				treeView.setSelectedRange(tree.createRangeFromNodes(three, 9, three, 9));
				treeView.extendSelection();
				expect(treeView.selectedText()).toEqual('world');
				treeView.extendSelection();
				expect(treeView.selectedText()).toEqual('*world*');
				treeView.extendSelection();
				expect(treeView.selectedText()).toEqual('\t\tHello *world*. ');
				treeView.extendSelection();
				expect(treeView.selectedText()).toEqual('\t\tHello *world*. Look *at* this.');
				treeView.contractSelection();
				expect(treeView.selectedText()).toEqual('\t\tHello *world*. ');
				treeView.contractSelection();
				expect(treeView.selectedText()).toEqual('*world*');
				treeView.contractSelection();
				expect(treeView.selectedText()).toEqual('world');
				treeView.contractSelection();
				expect(treeView.selectedText()).toEqual('');
			});

			it('should expand from collapsed to word', function () {
				three.setLine('\t\tHello *world*. **Look at this** now.');
				treeView.setSelectedRange(tree.createRangeFromNodes(three, 10, three, 10));
				treeView.extendSelection();

				expect(treeView.selectedText()).toEqual('world');

				treeView.setSelectedRange(tree.createRangeFromNodes(three, 25, three, 25));
				treeView.extendSelection();

				expect(treeView.selectedText()).toEqual('at');
				treeView.extendSelection();
				expect(treeView.selectedText()).toEqual('Look at this');
				treeView.extendSelection();
				expect(treeView.selectedText()).toEqual('**Look at this**');
			});

			it('should expand to sentence after extending through syntax', function () {
				three.setLine('\t\tHello *world*. **Look at this** now.');
				treeView.setSelectedRange(tree.createRangeFromNodes(three, 10, three, 10));

				treeView.extendSelection();
				expect(treeView.selectedText()).toEqual('world');

				treeView.extendSelection();
				expect(treeView.selectedText()).toEqual('*world*');

				treeView.extendSelection();
				expect(treeView.selectedText()).toEqual('\t\tHello *world*. ');

				treeView.extendSelection();
				expect(treeView.selectedText()).toEqual('\t\tHello *world*. **Look at this** now.');

				treeView.contractSelection();
				expect(treeView.selectedText()).toEqual('\t\tHello *world*. ');

				treeView.contractSelection();
				expect(treeView.selectedText()).toEqual('*world*');

				treeView.contractSelection();
				expect(treeView.selectedText()).toEqual('world');

				treeView.contractSelection();
				expect(treeView.selectedText()).toEqual('');
			});

			it('should work fine with numbered lists', function () {
				three.setLine('\t\t2. Look *world* look');
				treeView.setSelectedRange(tree.createRangeFromNodes(three, 13, three, 13));

				treeView.extendSelection();
				expect(treeView.selectedText()).toEqual('world');

				treeView.extendSelection();
				expect(treeView.selectedText()).toEqual('*world*');

				treeView.extendSelection();
				expect(treeView.selectedText()).toEqual('Look *world* look');

				treeView.extendSelection();
				expect(treeView.selectedText()).toEqual('\t\t2. Look *world* look');

				three.setLine('\t\t2. Look *world* look.');
				treeView.setSelectedRange(tree.createRangeFromNodes(three, 13, three, 13));

				treeView.extendSelection();
				expect(treeView.selectedText()).toEqual('world');

				treeView.extendSelection();
				expect(treeView.selectedText()).toEqual('*world*');

				treeView.extendSelection();
				expect(treeView.selectedText()).toEqual('Look *world* look.');

				treeView.extendSelection();
				expect(treeView.selectedText()).toEqual('\t\t2. Look *world* look.');
			});

			it('should expand to sentence after extending through syntax', function () {
				three.setLine('\t\tHello *world*. **Look at this** now.');
				treeView.setSelectedRange(tree.createRangeFromNodes(three, 25, three, 25));

				treeView.extendSelection();
				expect(treeView.selectedText()).toEqual('at');

				treeView.extendSelection();
				expect(treeView.selectedText()).toEqual('Look at this');

				treeView.extendSelection();
				expect(treeView.selectedText()).toEqual('**Look at this**');

				treeView.extendSelection();
				expect(treeView.selectedText()).toEqual('**Look at this** now.');

				treeView.extendSelection();
				expect(treeView.selectedText()).toEqual('Hello *world*. **Look at this** now.');

				treeView.setSelectedRange(tree.createRangeFromNodes(three, 28, three, 28));

				treeView.extendSelection();
				expect(treeView.selectedText()).toEqual('this');

				treeView.extendSelection();
				expect(treeView.selectedText()).toEqual('Look at this');

				treeView.extendSelection();
				expect(treeView.selectedText()).toEqual('**Look at this**');

				treeView.extendSelection();
				expect(treeView.selectedText()).toEqual('**Look at this** now.');

				treeView.extendSelection();
				expect(treeView.selectedText()).toEqual('Hello *world*. **Look at this** now.');
			});

			it('should expand to entire document from selection in last paragraph without crashing', function () {
				treeView.setSelectedRange(tree.createRangeFromNodes(six, 0));
				treeView.extendSelection();
				treeView.extendSelection();
				treeView.extendSelection();
				treeView.extendSelection();
				expect(treeView.selectedText()).toEqual(treeText);
			});

			it('should skip unselectable ranges when extending ranges in filtered view', function () {
				treeController.collapseNodes([two], false);
				treeView.setSelectedRange(tree.createRangeFromNodes(two));
				treeView.extendSelection();
				expect(treeView.selectedText()).toEqual('\ttwo\n\t\tthree\n\t\tfour\n');

				treeView.contractSelection();
				expect(treeView.selectedText()).toEqual('\ttwo');

				treeView.setSelectedRange(tree.createRangeFromNodes(two));
				treeView.extendSelectionToParagraph();
				expect(treeView.selectedText()).toEqual('\ttwo\n\t\tthree\n\t\tfour\n');
			});
			*/
		});
	});
});
