define(function (require) {
	'use strict';

	var Outline = require('ft/taxonomy/outlinetaxonomy').Outline,
		Node = require('ft/core/node').Node,
		Tree = require('ft/core/tree').Tree;

	describe('Tree', function () {

/*
	one
		two
			three
			four
		five
			six
*/

		var treeText = 'one\n\ttwo\n\t\tthree\n\t\tfour\n\tfive\n\t\tsix',
			one, two, three, four, five, six,
			Tree = require('ft/core/tree').Tree,
			tree;

		beforeEach(function () {
			tree = new Tree(treeText, 'outline');
			tree.ensureClassified();
			one = tree.root.firstChild;
			two = one.nextLineNode();
			three = two.nextLineNode();
			four = three.nextLineNode();
			five = four.nextLineNode();
			six = five.nextLineNode();
		});

		afterEach(function () {
			tree.cleanupForCollection();
		});

		describe('Red Black', function () {
			it('should reset red/black state of removed nodes', function () {
				two.removeFromParent();
				expect(two._RBParent).toBeNull();
				expect(two._RBLeft).toBeNull();
				expect(two._RBRight).toBeNull();
				expect(two._RBColor).toEqual(Node.BLACK);
			});

			it('should not infinite loop when reinserting removed node at new structure level', function () {
				tree.setText('one\ntwo\nthree');
				one = tree.root.firstChild;
				two = one.nextLineNode();
				three = two.nextLineNode();
				two.removeFromParent();
				two.setLine('\ttwo');
				two._structureLevel = 1;
				two._typeIndentLevel = 1;
				tree.insertNodeBefore(two, three);
			});
		});

		describe('Create Nodes', function () {
			it('should create nodes from lines', function () {
				var newNodes = tree.createNodesFromLines(['a', '\tone', 'b']);
				var roots = newNodes.roots;
				expect(roots[0].line()).toEqual('a');
				expect(roots[1].line()).toEqual('b');

				var all = newNodes.all;
				expect(all[0].line()).toEqual('a');
				expect(all[1].line()).toEqual('\tone');
				expect(all[2].line()).toEqual('b');
			});
		});

		describe('Node Content', function () {
			it('should maintain mapping from ids to nodes', function () {
				expect(tree.nodeForID(four.id)).toEqual(four);
				four.removeFromParent();
				expect(tree.nodeForID(four.id)).toBeUndefined();
			});

			it('should know how many nodes are in the treeModel', function () {
				expect(tree.nodeCount()).toEqual(6);
			});

			it('should post changed events when nodes are updated', function () {
				tree.addEventListener('treeChanged', function (e) {
					var d = e.deltas[0];
					expect(e.deltas.length).toEqual(1);
					expect(d.fromLineCh).toEqual({line: 0, ch: 0});
					expect(d.toLineCh).toEqual({line: 0, ch: 3});
					expect(d.insertedText).toEqual('new');
					expect(d.replacedText).toEqual('one');
					expect(d.updatedNode).toEqual(one);
				});
				one.setLine('new');
			});

			it('should post changed events when nodes are partially updated', function () {
				tree.addEventListener('treeChanged', function (e) {
					var d = e.deltas[0];
					expect(e.deltas.length).toEqual(1);
					expect(d.fromLineCh).toEqual({line: 0, ch: 1});
					expect(d.toLineCh).toEqual({line: 0, ch: 2});
					expect(d.insertedText).toEqual('new');
					expect(d.replacedText).toEqual('n');
					expect(d.updatedNode).toEqual(one);
				});
				one.replaceLineInRange('new', 1, 1);
			});

			it('should post changed events when nodes are inserted at start', function () {
				var newNode = tree.createNode('new');
				tree.addEventListener('treeChanged', function (e) {
					var d = e.deltas[0];
					expect(e.deltas.length).toEqual(1);
					expect(d.fromLineCh).toEqual({line: 0, ch: 0});
					expect(d.toLineCh).toEqual({line: 0, ch: 0});
					expect(d.insertedText).toEqual('new\n');
					expect(d.replacedText).toEqual('');
					expect(d.updatedNode).toBeNull();
					expect(d.insertedNodes).toEqual([newNode]);
				});
				tree.insertNodeBefore(newNode, one);
			});

			it('should post changed events when nodes are inserted at end', function () {
				var newNode = tree.createNode('new');
				tree.addEventListener('treeChanged', function (e) {
					var d = e.deltas[0];
					expect(e.deltas.length).toEqual(1);
					expect(d.fromLineCh).toEqual({line: 5, ch: 5});
					expect(d.toLineCh).toEqual({line: 5, ch: 5});
					expect(d.insertedText).toEqual('\nnew');
					expect(d.replacedText).toEqual('');
					expect(d.updatedNode).toBeNull();
					expect(d.insertedNodes).toEqual([newNode]);
				});
				tree.insertNodeBefore(newNode, null);
			});

			it('should post changed events when nodes are removed at start', function () {
				tree.addEventListener('treeChanged', function (e) {
					var d = e.deltas[0];
					expect(e.deltas.length).toEqual(1);
					expect(d.fromLineCh).toEqual({line: 0, ch: 0});
					expect(d.toLineCh).toEqual({line: 1, ch: 0});
					expect(d.insertedText).toEqual('');
					expect(d.replacedText).toEqual('one\n');
					expect(d.updatedNode).toBeNull();
					expect(d.removedNodes).toEqual([one]);
				});
				tree.removeNode(one);
			});

			it('should post changed events when nodes are removed in middle', function () {
				tree.addEventListener('treeChanged', function (e) {
					var d = e.deltas[0];
					expect(e.deltas.length).toEqual(1);
					expect(d.fromLineCh).toEqual({line: 3, ch: 0});
					expect(d.toLineCh).toEqual({line: 4, ch: 0});
					expect(d.insertedText).toEqual('');
					expect(d.replacedText).toEqual('\t\tfour\n');
					expect(d.updatedNode).toBeNull();
					expect(d.removedNodes).toEqual([four]);
				});
				tree.removeNode(four);
			});

			it('should post changed events when nodes are removed at end', function () {
				tree.addEventListener('treeChanged', function (e) {
					var d = e.deltas[0];
					expect(e.deltas.length).toEqual(1);
					expect(d.fromLineCh).toEqual({line: 4, ch: 5});
					expect(d.toLineCh).toEqual({line: 5, ch: 5});
					expect(d.insertedText).toEqual('');
					expect(d.replacedText).toEqual('\n\t\tsix');
					expect(d.updatedNode).toBeNull();
					expect(d.removedNodes).toEqual([six]);
				});
				tree.removeNode(six);
			});

			it('should post changed events when multiple nodes are removed', function () {
				tree.addEventListener('treeChanged', function (e) {
					var d = e.deltas[0];
					expect(e.deltas.length).toEqual(1);
					expect(d.fromLineCh).toEqual({line: 1, ch: 0});
					expect(d.toLineCh).toEqual({line: 4, ch: 0});
					expect(d.insertedText).toEqual('');
					expect(d.replacedText).toEqual('\ttwo\n\t\tthree\n\t\tfour\n');
					expect(d.updatedNode).toBeNull();
					expect(d.removedNodes).toEqual([two, three, four]);
				});
				two.removeFromParent();
			});

			it('should be able to ', function () {
				tree.addEventListener('treeChanged', function (e) {
					var d = e.deltas[0];
					expect(e.deltas.length).toEqual(1);
					expect(d.fromLineCh).toEqual({line: 1, ch: 0});
					expect(d.toLineCh).toEqual({line: 4, ch: 0});
					expect(d.insertedText).toEqual('');
					expect(d.replacedText).toEqual('\ttwo\n\t\tthree\n\t\tfour\n');
					expect(d.updatedNode).toBeNull();
					expect(d.removedNodes).toEqual([two, three, four]);
				});
				two.removeFromParent();
			});

		});

		describe('Tags Content', function () {
			it('should return empty list when no tags', function () {
				expect(tree.tags()).toEqual([]);
			});

			it('should merge tags returned by name', function () {
				one.addTag('tag1', '');
				one.addTag('tag1', '');
				four.addTag('tag1', '');
				five.addTag('tag1', '');
				tree.ensureClassified();
				expect(tree.tags()).toEqual(['tag1']);
			});
			it('should support option to only include parsed tags', function () {
				one._tags.mouse = 'mouse';
				expect(tree.tags()).toEqual(['mouse']);
				expect(tree.tags(true)).toEqual([]);
			});
		});

		describe('Text Content', function () {

			it('should normalize line endings', function () {
				expect(Tree.stringByNormalizingLineEndings('one\r\ntwo\nthree\rfour')).toEqual('one\ntwo\nthree\nfour');
			});

			it('should created with string value', function () {
				expect(tree.text()).toEqual(treeText);
				expect(tree.textLength()).toEqual(35);
				expect(tree.nodeCount()).toEqual(6);
			});

			it('should update with new single line string value', function () {
				tree.addEventListener('treeChanged', function (e) {
					var d = e.deltas[0];
					expect(e.deltas.length).toEqual(1);
					expect(d.fromLineCh).toEqual({line: 0, ch: 0});
					expect(d.toLineCh).toEqual({line: 5, ch: 5});
					expect(d.insertedText).toEqual('Hello world');
					expect(d.replacedText).toEqual('one\n\ttwo\n\t\tthree\n\t\tfour\n\tfive\n\t\tsix');
					expect(d.updatedNode).toEqual(one);
					expect(d.removedNodes).toEqual([two, three, four, five, six]);
				});
				tree.setText('Hello world');
				expect(tree.text()).toEqual('Hello world');
				expect(tree.textLength()).toEqual(11);
				expect(tree.nodeCount()).toEqual(1);
			});

			it('should update with new multiline line string value', function () {
				tree.setText('Hello world\nlook at\nthis');
				expect(tree.text()).toEqual('Hello world\nlook at\nthis');
				expect(tree.textLength()).toEqual(24);
				expect(tree.nodeCount()).toEqual(3);
				tree.setText('');
			});

			it('should update with empty string value', function () {
				tree.addEventListener('treeChanged', function (e) {
					var d = e.deltas[0];
					expect(e.deltas.length).toEqual(1);
					expect(d.fromLineCh).toEqual({line: 0, ch: 0});
					expect(d.toLineCh).toEqual({line: 5, ch: 5});
					expect(d.insertedText).toEqual('');
					expect(d.replacedText).toEqual('one\n\ttwo\n\t\tthree\n\t\tfour\n\tfive\n\t\tsix');
					expect(d.updatedNode).toEqual(one);
					expect(d.removedNodes).toEqual([two, three, four, five, six]);
				});
				tree.setText('');
				expect(tree.text()).toEqual('');
				expect(tree.textLength()).toEqual(0);
				expect(tree.nodeCount()).toEqual(1);
			});

			it('should update with null string value', function () {
				tree.addEventListener('treeChanged', function (e) {
					var d = e.deltas[0];
					expect(e.deltas.length).toEqual(1);
					expect(d.fromLineCh).toEqual({line: 0, ch: 0});
					expect(d.toLineCh).toEqual({line: 5, ch: 5});
					expect(d.insertedText).toEqual('');
					expect(d.replacedText).toEqual('one\n\ttwo\n\t\tthree\n\t\tfour\n\tfive\n\t\tsix');
					expect(d.updatedNode).toBeNull();
					expect(d.removedNodes).toEqual([one, two, three, four, five, six]);
				});
				tree.setText(null);
				expect(tree.text()).toEqual('');
				expect(tree.textLength()).toEqual(0);
				expect(tree.nodeCount()).toEqual(0);
			});

			it('should be able to add child to tree with no lines', function () {
				var newNode = tree.createNode('hello world');
				tree.setText(null);
				tree.addEventListener('treeChanged', function (e) {
					var d = e.deltas[0];
					expect(e.deltas.length).toEqual(1);
					expect(d.fromLineCh).toEqual({line: 0, ch: 0});
					expect(d.toLineCh).toEqual({line: 0, ch: 0});
					expect(d.insertedText).toEqual('hello world');
					expect(d.replacedText).toEqual('');
					expect(d.updatedNode).toBeNull();
					expect(d.insertedNodes).toEqual([newNode]);
				});
				tree.root.appendChild(newNode);
				expect(tree.text()).toEqual('hello world');
			});

			it('should be settable', function () {
				tree.setText('a\nb');
				expect(tree.root.firstChild.line()).toEqual('a');
				expect(tree.root.lastChild.line()).toEqual('b');
				expect(tree.nodeCount()).toEqual(2);
			});

			it('should be settable to empty', function () {
				tree.addEventListener('treeChanged', function (e) {
					expect(e.deltas[0].location).toEqual(0);
					expect(e.deltas[0].replacedText).toEqual(treeText);
					expect(e.deltas[0].insertedText).toEqual('');
				});

				tree.setText('');
				expect(tree.root.firstChild).toEqual(one);
				expect(one.line()).toEqual('');
				expect(one).toEqual(tree.root.lastChild);
				expect(tree.nodeCount()).toEqual(1);
			});

			it('should be settable from empty', function () {
				tree.setText('');
				tree.addEventListener('treeChanged', function (e) {
					expect(e.deltas[0].location).toEqual(0);
					expect(e.deltas[0].replacedText).toEqual('');
					expect(e.deltas[0].insertedText).toEqual('boo!');
				});

				tree.setText('boo!');
				expect(tree.root.firstChild.line()).toEqual('boo!');
				expect(tree.root.lastChild.line()).toEqual('boo!');
			});

			it('should be gettable in in range', function () {
				expect(tree.textInRange(tree.createRangeFromLocation(1, 2))).toEqual('ne');
			});

			it('should handle this case which is now breaking the RB tree', function () {
				six.removeFromParent();
				five.removeFromParent();
				four.removeFromParent();
				tree.removeNode(two);
			});

			it('should delete trailing line when entire range is replaced', function () {
				tree.addEventListener('treeChanged', function (e) {
					var d = e.deltas[0];
					expect(e.deltas.length).toEqual(1);
					expect(d.fromLineCh).toEqual({line: 0, ch: 0});
					expect(d.toLineCh).toEqual({line: 1, ch: 0});
					expect(d.insertedText).toEqual('');
					expect(d.replacedText).toEqual('one\n');
					expect(d.updatedNode).toEqual(one);
					expect(d.removedNodes).toEqual([two]);
				});
				tree.replaceTextInRange(tree.createRangeFromLocation(0, 4), '');
				expect(tree.text()).toEqual('\ttwo\n\t\tthree\n\t\tfour\n\tfive\n\t\tsix');
			});

			it('should map between node offsets and text offsets', function () {
				expect(tree.textOffsetToNodeOffset(3)).toEqual({ node : one, offset : 3 });
				expect(tree.textOffsetToNodeOffset(4)).toEqual({ node : two, offset : 0 });
				expect(tree.nodeOffsetToTextOffset(one, 3)).toEqual(3);
				expect(tree.nodeOffsetToTextOffset(two, 0)).toEqual(4);
				expect(tree.nodeOffsetToTextOffset(four, 2)).toEqual(19);
				expect(tree.textOffsetToNodeOffset(19)).toEqual({ node : four, offset : 2 });
			});

			it('should be replaceable in range', function () {
				tree.addEventListener('treeChanged', function (e) {
					var d = e.deltas[0];
					expect(e.deltas.length).toEqual(1);
					expect(d.fromLineCh).toEqual({line: 0, ch: 1});
					expect(d.toLineCh).toEqual({line: 1, ch: 4});
					expect(d.insertedText).toEqual('ops\ntested');
					expect(d.replacedText).toEqual('ne\n\ttwo');
					expect(d.updatedNode).toEqual(one);
					expect(d.removedNodes).toEqual([two]);
				});

				tree.replaceTextInRange(tree.createRangeFromLocation(1, 7), 'ops\ntested');
				expect(one.line()).toEqual('oops');
				expect(one.nextSibling.line()).toEqual('tested');
				expect(tree.nodeCount()).toEqual(6);
			});

			it('should preserve leading newline at start of document', function () {
				tree.setText('');
				tree.replaceTextInRange(tree.createRangeFromLocation(0, 0), '\n0');
				tree.replaceTextInRange(tree.createRangeFromLocation(1, 1), '');
				expect(tree.text()).toEqual('\n');
			});

			it('should preserve trailing newline at end of document', function () {
				tree.setText('');
				tree.replaceTextInRange(tree.createRangeFromLocation(0, 0), '0\n');
				tree.replaceTextInRange(tree.createRangeFromLocation(0, 1), '');
				expect(tree.text()).toEqual('\n');
			});

			it('should join currentline with nextline when deleting newline', function () {
				var originalText = tree.text();
				tree.beginUpdates();
				tree.replaceTextInRange(tree.createRangeFromLocation(3, 1), '');
				expect(one.line()).toEqual('one\ttwo');
			});

			it('should update node attribute ranges when replacing in range', function () {
				one.addLineAttributeInRange('name', 'Jesse', 1, 2);
				expect(one.lineAttributedString().toString(true)).toEqual('(o/text="one")(ne/name="Jesse", text="one")');
				tree.replaceTextInRange(tree.createRangeFromLocation(0, 1), '');
				expect(one.lineAttributedString().toString(true)).toEqual('(ne/name="Jesse", text="one")');
			});

			it('should pass all fail cases discovered by random tree mutation', function () {
				tree.setText('');
				tree.replaceTextInRange(tree.createRangeFromLocation(0, 0), '0\n2\n\n2');
				tree.replaceTextInRange(tree.createRangeFromLocation(0, 3), '');
				expect(tree.text()).toEqual('\n\n2');

				tree.setText('');
				tree.replaceTextInRange(tree.createRangeFromLocation(0, 0), '2\n1');
				tree.replaceTextInRange(tree.createRangeFromLocation(0, 1), '');
				expect(tree.text()).toEqual('\n1');

				tree.setText('');
				tree.replaceTextInRange(tree.createRangeFromLocation(0, 0), 'a\n2\n1');
				tree.replaceTextInRange(tree.createRangeFromLocation(0, 3), '');
				expect(tree.text()).toEqual('\n1');

				tree.setText('');
				tree.replaceTextInRange(tree.createRangeFromLocation(0, 0), '2\n1');
				tree.replaceTextInRange(tree.createRangeFromLocation(2, 1), '');
				expect(tree.text()).toEqual('2\n');

				tree.setText('');
				tree.replaceTextInRange(tree.createRangeFromLocation(0, 0), '\n1\n2');
				tree.replaceTextInRange(tree.createRangeFromLocation(0, 2), '');
				expect(tree.text()).toEqual('\n2');

				tree.setText('');
				tree.replaceTextInRange(tree.createRangeFromLocation(0, 0), 'a\n1\n2');
				tree.replaceTextInRange(tree.createRangeFromLocation(1, 2), '');
				expect(tree.text()).toEqual('a\n2');

				tree.setText('');
				tree.replaceTextInRange(tree.createRangeFromLocation(0, 0), '\n1\n\n');
				tree.replaceTextInRange(tree.createRangeFromLocation(0, 3), '');
				expect(tree.textLength()).toEqual(1);
				expect(tree.text()).toEqual('\n');

				tree.setText('');
				tree.replaceTextInRange(tree.createRangeFromLocation(0, 0), '\n\n1\n');
				tree.replaceTextInRange(tree.createRangeFromLocation(1, 3), '');
				expect(tree.text()).toEqual('\n');

				tree.setText('');
				tree.replaceTextInRange(tree.createRangeFromLocation(0, 0), '\n1\n\n12');
				tree.replaceTextInRange(tree.createRangeFromLocation(0, 3), '');
				expect(tree.text()).toEqual('\n12');

				tree.setText('');
				tree.replaceTextInRange(tree.createRangeFromLocation(0, 0), '0\n\n1\n\n');
				tree.replaceTextInRange(tree.createRangeFromLocation(2, 3), '');
				expect(tree.text()).toEqual('0\n\n');
			});
		});
	});
});