define(function (require) {
	'use strict';

	describe('Editor', function () {

		require('ft/taxonomy/outlinetaxonomy');
		require('ft/taxonomy/markdowntaxonomy');
		require('ft/taxonomy/testtaxonomy');

		var CodeMirror = require('codemirror').CodeMirror,
			Editor = require('ft/editor/editor').Editor,
			Taxonomies = require('ft/core/taxonomies'),
			Tree = require('ft/core/tree').Tree,
			lineCount = 118406,
			editor;

		beforeEach(function () {
			editor = new Editor('', '');
		});

		afterEach(function () {
			editor.removeAndCleanupForCollection();
		});

		var cachedShakeSpear;

		function shakespear() {
			if (!cachedShakeSpear) {
				var baseLines = [
					'# each line averaged a little under eight words.',
					'## each line averaged a little under eight words.',
					'### each line averaged a little under eight words.',
					'',
					'- each line @four(five) averaged a little under eight words.',
					'\t- each line <strong>averaged</strong> a little under [eight](www.apple.com) words.test @one @two(three)',
					'\t\t- **each *line* averaged** a ``little under`` eight words.',
					'\t\t- some lines have www.apple.com links and jesse@hogbaysoftware.com emails',
					''
				];

				var lines = [],
					count = lineCount / baseLines.length,
					chunk = baseLines.join('\n');
				for (var i = 0; i < count; i++) {
					lines.push(chunk);
				}

				cachedShakeSpear = lines.join('\n');
			}
			return cachedShakeSpear;
		}

		it('should load/parse/query/delete shakespear sized texts with null taxonomy', function () {
			var cm = editor.cm(),
				shakespearText = shakespear(),
				duration,
				start;

			console.profile('Load');
			start = new Date().getTime();
			editor.setTextContent(shakespearText);
			duration = + (new Date().getTime() - start) / 1000;
			console.log('Load Duration: ' + duration + ' seconds.');
			expect(duration).toBeLessThan(1.0);
			console.profileEnd();

			console.profile('Parse');
			start = new Date().getTime();
			editor.tree().ensureClassified();
			duration = + (new Date().getTime() - start) / 1000;
			console.log('Parse Duration: ' + duration + ' seconds.');
			expect(duration).toBeLessThan(1.0);
			console.profileEnd();

			console.profile('Query');
			start = new Date().getTime();
			expect(editor.tree().evaluateNodePath('//*').length).toEqual(editor.tree().nodeCount());
			duration = + (new Date().getTime() - start) / 1000;
			console.log('Query Duration: ' + duration + ' seconds.');
			expect(duration).toBeLessThan(0.05);
			console.profileEnd();

			console.profile('Delete');
			start = new Date().getTime();
			cm.replaceRange('', {line: 0, ch: 1}, {line: editor.tree().nodeCount(), ch: 1});
			duration = + (new Date().getTime() - start) / 1000;
			console.log('Delete Duration: ' + duration + ' seconds.');
			expect(editor.tree().nodeCount()).toEqual(1);
			expect(duration).toBeLessThan(1.0);
			console.profileEnd();

			console.profile('Insert');
			start = new Date().getTime();
			cm.replaceRange(shakespearText, {line: 0, ch: 1}, {line: 0, ch: 1});
			duration = + (new Date().getTime() - start) / 1000;
			console.log('Insert Duration: ' + duration + ' seconds.');
			expect(duration).toBeLessThan(1.0);
			console.profileEnd();
		});

		it('should classify text taxonomy that does nothing except assing structure levels', function () {
			var tree = new Tree(shakespear(), Taxonomies.taxonomy({
			}, 'test'));

			console.profile('Test Parse');
			var start = new Date().getTime();
			tree.ensureClassified();
			var duration = + (new Date().getTime() - start) / 1000;
			expect(duration).toBeLessThan(5.0);
			console.log('Test Parse Duration: ' + duration + ' seconds.');
			console.profileEnd();
		});

		it('should classify shakespear sized texts with outline taxonomy', function () {
			var tree = new Tree(shakespear(), Taxonomies.taxonomy({
			}, 'outline'));

			console.profile('Outline Parse');
			var start = new Date().getTime();
			tree.ensureClassified();
			var duration = + (new Date().getTime() - start) / 1000;
			expect(duration).toBeLessThan(5.0);
			console.log('Outline Parse Duration: ' + duration + ' seconds.');
			console.profileEnd();
		});

		it('should classify shakespear sized texts with markdown taxonomy', function () {
			var tree = new Tree(shakespear(), 'markdown');

			console.profile('Markdown Parse');
			var start = new Date().getTime();
			tree.ensureClassified();
			var duration = + (new Date().getTime() - start) / 1000;
			expect(duration).toBeLessThan(5.0);
			console.log('Markdown Parse Duration: ' + duration + ' seconds.');
			console.profileEnd();
		});

		it('should classify shakespear sized texts with foldingtext taxonomy', function () {
			var tree = new Tree(shakespear(), Taxonomies.taxonomy({
				gitmarkdown: false,
				multimarkdown: false,
				criticMarkup: false,
				foldingtext: true
			}, 'markdown'));

			console.profile('FoldingText Parse');
			var start = new Date().getTime();
			tree.ensureClassified();
			var duration = + (new Date().getTime() - start) / 1000;
			expect(duration).toBeLessThan(5.0);
			console.log('FoldingText Parse Duration: ' + duration + ' seconds.');
			console.profileEnd();
		});

			jasmine.DEFAULT_TIMEOUT_INTERVAL = 1000000;

		it('should scroll shakespear sized texts in foldingtext', function (done) {
			editor.removeAndCleanupForCollection();
/*			editor = new Editor(shakespear(), Taxonomies.taxonomy({
				gitmarkdown: false,
				multimarkdown: false,
				criticMarkup: false,
				foldingtext: true
			}, 'markdown'));
*/
			editor = new Editor(shakespear(), 'outline');

			editor.tree().ensureClassified();

			console.profile('FoldingText Scroll');

			var scrollLine = 0,
				start = new Date().getTime(),
				scrollIncrement = Math.round(lineCount / 100);

			function scrollToNextLine() {
				if (scrollLine < lineCount) {
					//if (scrollLine > 2000 && scrollLine < (2000 + scrollIncrement)) {
					//	debugger;
					//}

					//editor.scrollToLine(scrollLine);
					editor.cm().scrollIntoView(scrollLine);
					scrollLine += scrollIncrement;
					window.setTimeout(function () {
						scrollToNextLine();
					});
				} else {
					var duration =+ (new Date().getTime() - start) / 1000;
					expect(duration).toBeLessThan(5.0);
					console.log('FoldingText Scroll Duration: ' + duration + ' seconds.');
					console.profileEnd();	
					done();				
				}
			}

			scrollToNextLine();
		});

		it('should scroll shakespear sized texts in CodeMirror', function (done) {
			editor.removeAndCleanupForCollection();

			var myCodeMirror = new CodeMirror(document.body, {
				value: shakespear(),
				mode: 'markdown',
				lineWrapping: true
			});

			console.profile('CodeMirror Scroll');

			var scrollLine = 0,
				start = new Date().getTime(),
				scrollIncrement = Math.round(lineCount / 100);

			function scrollToNextLine() {
				if (scrollLine < lineCount) {
					//if (scrollLine > 2000 && scrollLine < (2000 + scrollIncrement)) {
					//	debugger;
					//}

					myCodeMirror.scrollIntoView(scrollLine);
					scrollLine += scrollIncrement;
					window.setTimeout(function () {
						scrollToNextLine();
					});
				} else {
					var duration =+ (new Date().getTime() - start) / 1000;
					expect(duration).toBeLessThan(5.0);
					console.log('CodeMirror Scroll Duration: ' + duration + ' seconds.');
					console.profileEnd();	
					done();				
				}
			}

			scrollToNextLine();
		});
	});
});