define(function (require) {
	'use strict';

/*
	one
		two
			three
			four
		five
			six
*/

	describe('Cut Copy Paste', function () {
		require('ft/extensions/cutcopypaste');
		require('ft/taxonomy/outlinetaxonomy');

		var treeText = 'one\n\ttwo\n\t\tthree\n\t\tfour\n\tfive\n\t\tsix',
			Editor = require('ft/editor/editor').Editor,
			Folding = require('ft/extensions/folding'),
			one, two, three, four, five, six,
			Tree = require('ft/core/tree').Tree,
			editor, tree,
			clipboard = {
				content: {},
				setData: function (type, data) {
					this.content[type] = data;
				},
				getData: function (type) {
					return this.content[type];
				},
				clear: function () {
					this.content = {};
				}
			};

		beforeEach(function () {
			editor = new Editor(treeText, 'outline', {indentWithTabs: true, indentUnit: 4});
			tree = editor.tree();
			one = tree.root.firstChild;
			two = one.nextLineNode();
			three = two.nextLineNode();
			four = three.nextLineNode();
			five = four.nextLineNode();
			six = five.nextLineNode();
			tree.ensureClassified();
		});

		afterEach(function () {
			editor.removeAndCleanupForCollection();
			clipboard.clear();
		});

		it('should copy', function () {
			editor.setSelectedLineCh({line: 2, ch: 1}, {line: 4, ch: 5});
			editor.copy(clipboard);
			expect(clipboard.getData('text/plain')).toEqual('\tthree\n\t\tfour\n\tfive');
		});

		it('should cut', function () {
			editor.setSelectedLineCh({line: 2, ch: 1}, {line: 4, ch: 5});
			editor.cut(clipboard);
			expect(clipboard.getData('text/plain')).toEqual('\tthree\n\t\tfour\n\tfive');
			expect(editor.textContent()).toEqual('one\n\ttwo\n\t\n\t\tsix');
		});

		it('should paste', function () {
			editor.setSelectedLineCh({line: 2, ch: 1}, {line: 4, ch: 5});
			clipboard.setData('text/plain', 'hello world');
			editor.paste(clipboard);
			expect(editor.textContent()).toEqual('one\n\ttwo\n\thello world\n\t\tsix');			
		});

		describe('Multiple Selections', function () {
			beforeEach(function () {				
				editor.setSelectedRanges([
					tree.createRangeFromLineChs({line: 0, ch: 0}, {line: 1, ch: 0}),
					tree.createRangeFromLineChs({line: 3, ch: 1}, {line: 4, ch: 3})
				]);
			});

			it('should copy', function () {
				editor.copy(clipboard);
				expect(clipboard.getData('text/plain')).toEqual('one\n\n\tfour\n\tfi');
			});

			it('should cut', function () {
				editor.cut(clipboard);
				expect(clipboard.getData('text/plain')).toEqual('one\n\n\tfour\n\tfi');
				expect(editor.textContent()).toEqual('\ttwo\n\t\tthree\n\tve\n\t\tsix');
			});

			it('should paste matching lines to selections', function () {
				clipboard.setData('text/plain', 'p1\np2');
				editor.paste(clipboard);
				expect(editor.textContent()).toEqual('p1\ttwo\n\t\tthree\n\tp2ve\n\t\tsix');
			});

			it('should paste not matching lines to selections', function () {
				clipboard.setData('text/plain', 'p1\np2\np3');
				editor.paste(clipboard);
				expect(editor.textContent()).toEqual('p1\np2\np3\ttwo\n\t\tthree\n\tp1\np2\np3ve\n\t\tsix');
			});
		});

		describe('Folded Ranges', function () {
			beforeEach(function () {
				editor.collapseNode(two);
				editor.collapseNode(five);
				editor.setSelectedRange(tree.createRangeFromLocation(0, editor.textContent().length));
			});

			it('should copy', function () {
				editor.copy(clipboard);
				expect(clipboard.getData('text/plain')).toEqual(treeText);
			});

			it('should cut', function () {
				editor.cut(clipboard);
				expect(clipboard.getData('text/plain')).toEqual(treeText);
				expect(editor.textContent()).toEqual('');
			});

			it('should paste restoring folds', function () {
				editor.cut(clipboard);
				editor.paste(clipboard);
				expect(editor.folds().length).toEqual(2);
				expect(editor.textContent()).toEqual(treeText);
			});
		});		
	});
});