define(function (require) {
	'use strict';

/*
	one
		two
			three
			four
		five
			six
*/

	describe('Folding', function () {

		require('ft/extensions/undo');
		require('ft/taxonomy/outlinetaxonomy');

		var Folding = require('ft/extensions/folding'),
			treeText = 'one\n\ttwo\n\t\tthree\n\t\tfour\n\tfive\n\t\tsix',
			Editor = require('ft/editor/editor').Editor,
			one, two, three, four, five, six,
			Tree = require('ft/core/tree').Tree,
			editor, tree;

		beforeEach(function () {
			jasmine.addMatchers(require('tests/custommatchers'));
			editor = new Editor(treeText, 'outline');
			tree = editor.tree();
			one = tree.root.firstChild;
			two = one.nextLineNode();
			three = two.nextLineNode();
			four = three.nextLineNode();
			five = four.nextLineNode();
			six = five.nextLineNode();
			tree.ensureClassified();
		});

		afterEach(function () {
			editor.removeAndCleanupForCollection();
		});

		it('should get folds', function () {
			expect(editor.folds()).toEqual([]);
		});

		it('should get folds in range', function () {
			editor.createFoldInRange(tree.createRangeFromLocation(0, 1));
			editor.createFoldInRange(tree.createRangeFromLocation(3, 1));
			expect(editor.foldsInRange(tree.createRangeFromLocation(10, 0)).length).toEqual(0);
			expect(editor.foldsInRange(tree.createRangeFromLocation(0, 0)).length).toEqual(1);
			expect(editor.foldsInRange(tree.createRangeFromLocation(0, 2)).length).toEqual(1);
			expect(editor.foldsInRange(tree.createRangeFromLocation(0, 3)).length).toEqual(2);
		});

		it('should create fold in range', function () {
			expect(editor.createFoldInRange(tree.createRangeFromLocation(0, 1))).not.toBeUndefined();
			expect(editor.folds().length).toEqual(1);
		});

		it('should not create fold in empty range', function () {
			expect(editor.createFoldInRange(tree.createRangeFromLocation(1, 0))).toBeUndefined();
			expect(editor.folds().length).toEqual(0);
		});

		it('should return existing fold if create a new fold over the same range', function () {
			var fold = editor.createFoldInRange(tree.createRangeFromLocation(0, 1));
			expect(editor.createFoldInRange(tree.createRangeFromLocation(0, 1))).toEqual(fold);
			expect(editor.folds().length).toEqual(1);
		});

		it('should adjust existing folds so that end doesnt overlap start of new fold range', function () {
			editor.createFoldInRange(tree.createRangeFromLocation(0, 3));
			editor.createFoldInRange(tree.createRangeFromLocation(2, 1));
			expect(editor.foldsInRange(tree.createRangeFromLocation(0, 0))[0].range().length()).toEqual(2);
			expect(editor.foldsInRange(tree.createRangeFromLocation(3, 0))[0].range().length()).toEqual(1);
			expect(editor.folds().length).toEqual(2);
		});

		it('should adjust existing folds so that start doesnt overlap end of new fold range', function () {
			editor.createFoldInRange(tree.createRangeFromLocation(0, 3));
			editor.createFoldInRange(tree.createRangeFromLocation(0, 1));
			expect(editor.foldsInRange(tree.createRangeFromLocation(0, 0))[0].range().length()).toEqual(1);
			expect(editor.foldsInRange(tree.createRangeFromLocation(2, 0))[0].range().length()).toEqual(2);
			expect(editor.folds().length).toEqual(2);
		});

		it('should create fold fully over another fold without any adjusting', function () {
			editor.createFoldInRange(tree.createRangeFromLocation(1, 1));
			editor.createFoldInRange(tree.createRangeFromLocation(0, 3));
			expect(editor.foldsInRange(tree.createRangeFromLocation(0, 0))[0].range().length()).toEqual(3);
			expect(editor.folds().length).toEqual(2);
		});

		it('should remove fold', function () {
			var fold = editor.createFoldInRange(tree.createRangeFromLocation(0, 1));
			editor.removeFold(fold);
			expect(editor.folds().length).toEqual(0);
		});

		it('should remove fold', function () {
			var fold = editor.createFoldInRange(tree.createRangeFromLocation(0, 1));
			editor.removeFold(fold);
			expect(editor.folds().length).toEqual(0);
		});

		it('should split and remove folds by trimming their start', function () {
			var fold = editor.createFoldInRange(tree.createRangeFromLocation(0, 3));
			editor.splitAndRemoveFoldsInRange(tree.createRangeFromLocation(0, 1));
			expect(editor.folds().length).toEqual(1);
			expect(editor.folds()[0].range().location()).toEqual(1);
			expect(editor.folds()[0].range().length()).toEqual(2);
		});

		it('should split and remove folds by trimming their end', function () {
			var fold = editor.createFoldInRange(tree.createRangeFromLocation(0, 3));
			editor.splitAndRemoveFoldsInRange(tree.createRangeFromLocation(2, 1));
			expect(editor.folds().length).toEqual(1);
			expect(editor.folds()[0].range().location()).toEqual(0);
			expect(editor.folds()[0].range().length()).toEqual(2);
		});

		it('should split and remove folds by splitting them into two', function () {
			var fold = editor.createFoldInRange(tree.createRangeFromLocation(0, 3));
			editor.splitAndRemoveFoldsInRange(tree.createRangeFromLocation(1, 1));
			expect(editor.folds().length).toEqual(2);
			expect(editor.folds()[0].range().location()).toEqual(2);
			expect(editor.folds()[0].range().length()).toEqual(1);
			expect(editor.folds()[1].range().location()).toEqual(0);
			expect(editor.folds()[1].range().length()).toEqual(1);
		});

		it('should split and remove folds that are fully sourounded', function () {
			var fold = editor.createFoldInRange(tree.createRangeFromLocation(1, 1));
			editor.splitAndRemoveFoldsInRange(tree.createRangeFromLocation(0, 3));
			expect(editor.folds().length).toEqual(0);
		});

		it('should convert folds to json', function () {
			editor.createFoldInRange(tree.createRangeFromLocation(0, 3));
			expect(editor.foldsToJSON(editor.folds())).toBeJsonEqual([ { from : { line : 0, ch : 0 }, to : { line : 0, ch : 3 } } ]);
		});

		it('should restore folds from json', function () {
			editor.restoreJSONFolds([ { from : { line : 0, ch : 0 }, to : { line : 0, ch : 3 } } ]);
			expect(editor.folds()[0].range().location()).toEqual(0);
			expect(editor.folds()[0].range().length()).toEqual(3);
			expect(editor.folds().length).toEqual(1);
		});

		describe('Undo', function () {
			it('should remove folds overlapping undo range', function () {
				editor.replaceTextInRange(tree.createRangeFromLocation(5, 0), 'new');
				editor.createFoldInRange(tree.createRangeFromLineChs({line: 1, ch: 2}, {line: 1, ch: 6}));
				editor.setSelectedLineCh({line: 0, ch: 0});
				editor.performCommand('toggleNodeFold');
				expect(editor.folds().length).toEqual(2);
				editor.performCommand('undo');
				expect(editor.folds().length).toEqual(0);
			});
		});
	});
});