define(function (require) {
	'use strict';

/*
	one
		two
			three
			four
		five
			six
*/

	describe('Folding Nodes', function () {

		require('ft/extensions/folding');
		require('ft/extensions/foldingnodes');
		require('ft/taxonomy/outlinetaxonomy');

		var treeText = 'one\n\ttwo\n\t\tthree\n\t\tfour\n\tfive\n\t\tsix',
			Editor = require('ft/editor/editor').Editor,
			one, two, three, four, five, six,
			Tree = require('ft/core/tree').Tree,
			editor, tree;

		beforeEach(function () {
			editor = new Editor(treeText, 'outline');
			tree = editor.tree();
			one = tree.root.firstChild;
			two = one.nextLineNode();
			three = two.nextLineNode();
			four = three.nextLineNode();
			five = four.nextLineNode();
			six = five.nextLineNode();
			tree.ensureClassified();
		});

		afterEach(function () {
			editor.removeAndCleanupForCollection();
		});

		it('should determine if node is expanded', function () {
			expect(editor.isExpanded(one)).toBeTruthy();
			editor.collapseNode(one);
			expect(editor.isExpanded(one)).toBeFalsy();
		});

		describe('Collapsing', function () {
			it('should collapse node', function () {
				editor.collapseNode(two);
				expect(editor.folds().length).toEqual(1);
				editor.collapseNode(two);
				expect(editor.folds().length).toEqual(1);
				expect(editor.folds()[0].isManual()).toBeTruthy();
			});

			it('should collapse all node descendents', function () {
				editor.collapseNode(one, true);
				expect(editor.folds().length).toEqual(3);
				expect(editor.folds()[0].isManual()).toBeTruthy();
				expect(editor.folds()[1].isManual()).toBeTruthy();
				expect(editor.folds()[2].isManual()).toBeTruthy();
			});
		});

		describe('Expanding', function () {
			it('should expand node', function () {
				editor.collapseNode(two);
				expect(editor.folds().length).toEqual(1);
				expect(editor.folds()[0].isManual()).toBeTruthy();
				editor.expandNode(two);
				expect(editor.folds().length).toEqual(0);
			});

			it('should expand all node descendents', function () {
				editor.collapseNode(one, true);
				editor.expandNode(one, true);
				expect(editor.folds().length).toEqual(0);
			});
		});

		describe('Toggling', function () {
			it('should toggle node collapsed', function () {
				expect(editor.isExpanded(one)).toBeTruthy();
				editor.performCommand('toggleNodeFold');
				expect(editor.isExpanded(one)).toBeFalsy();
				expect(editor.folds().length).toEqual(1);

				editor.performCommand('toggleNodeFold');
				expect(editor.isExpanded(one)).toBeTruthy();
				expect(editor.folds().length).toEqual(0);
			});

			it('should toggle only the biggest current fold', function () {
				editor.createFoldInRange(tree.createRangeFromLocation(2, 1));
				editor.createFoldInRange(tree.createRangeFromLocation(1, 2));
				editor.createFoldInRange(tree.createRangeFromLocation(0, 3));
				editor.setSelectedLineCh({line: 0, ch: 3});
				expect(editor.folds().length).toEqual(3);
				editor.performCommand('toggleNodeFold');
				expect(editor.folds().length).toEqual(2);
			});

			it('should toggle all selected roots collapsed', function () {
				editor.setSelectedLineCh({line: 1, ch: 0}, {line: 5, ch: 0});
				editor.performCommand('toggleNodeFold');
				expect(editor.folds().length).toEqual(2);
				editor.performCommand('toggleNodeFold');
				expect(editor.folds().length).toEqual(0);
			});

			it('should toggle fold of parent of current node has no children', function () {
				editor.setSelectedLineCh({line: 2, ch: 0}, {line: 2, ch: 0});
				editor.performCommand('toggleNodeFold');
				expect(editor.isExpanded(one)).toBeTruthy();
				expect(editor.isExpanded(two)).toBeFalsy();
				expect(editor.isExpanded(five)).toBeTruthy();
			});

			it('should not toggle fold of root', function () {
				editor.setTextContent('one');
				editor.setSelectedLineCh({line: 0, ch: 0});
				editor.performCommand('toggleNodeFold');
				expect(editor.folds().length).toEqual(0);
			});

			it('should toggle fold all on', function () {
				editor.performCommand('toggleAllFolds');
				expect(editor.isExpanded(one)).toBeFalsy();
				expect(editor.isExpanded(two)).toBeFalsy();
				expect(editor.isExpanded(five)).toBeFalsy();
				expect(editor.folds().length).toEqual(3);
			});

			it('should toggle fold all off', function () {
				editor.performCommand('toggleAllFolds');
				editor.performCommand('toggleAllFolds');
				expect(editor.folds().length).toEqual(0);
			});
		});

		describe('Folding Levels', function () {
			it('should get folding level', function () {
				expect(editor.foldingLevel()).toEqual(3);
				expect(editor.foldingLevel(one)).toEqual(3);
				expect(editor.foldingLevel(two)).toEqual(2);
				expect(editor.foldingLevel(three)).toEqual(1);

				editor.collapseNode(two);
				expect(editor.foldingLevel()).toEqual(2);
				expect(editor.foldingLevel(one)).toEqual(2);
				expect(editor.foldingLevel(two)).toEqual(1);
				expect(editor.foldingLevel(three)).toEqual(0);

				editor.collapseNode(one);
				expect(editor.foldingLevel()).toEqual(1);
				expect(editor.foldingLevel(one)).toEqual(1);
				expect(editor.foldingLevel(two)).toEqual(0);
				expect(editor.foldingLevel(three)).toEqual(0);
			});

			it('should set folding level', function () {
				editor.setFoldingLevel(2);
				expect(editor.isCollapsed(two)).toBeTruthy();
				expect(editor.isCollapsed(five)).toBeTruthy();

				editor.setFoldingLevel(1);
				expect(editor.isCollapsed(one)).toBeTruthy();

				editor.setFoldingLevel(3);
				expect(editor.isExpanded(one)).toBeTruthy();
				expect(editor.isExpanded(two)).toBeTruthy();
				expect(editor.isExpanded(five)).toBeTruthy();

				editor.setFoldingLevel(1, two);
				expect(editor.isCollapsed(two)).toBeTruthy();
			});

			it('should set folding error case', function () {
				four.removeFromParent();
				five.removeFromParent();
				tree.ensureClassified();

				expect(editor.foldingLevel()).toEqual(3);
				editor.setFoldingLevel(3);
				expect(editor.nodeIsHiddenInFold(one)).toBeFalsy();
				expect(editor.nodeIsHiddenInFold(two)).toBeFalsy();
				expect(editor.nodeIsHiddenInFold(three)).toBeFalsy();

				editor.setFoldingLevel(2);
				expect(editor.foldingLevel()).toEqual(2);
			});
		});
	});
});