define(function (require) {
	'use strict';

	describe('Formatting', function () {
		require('ft/taxonomy/markdowntaxonomy');
		require('ft/taxonomy/helpers/criticmarkuphelper');
		require('ft/extensions/formatting');

		var Taxonomies = require('ft/core/taxonomies'),
			Editor = require('ft/editor/editor').Editor,
			editor,
			tree;

		beforeEach(function () {
			editor = new Editor('', Taxonomies.taxonomy({
				criticMarkup: true
			}, 'markdown'));
			tree = editor.tree();
		});

		afterEach(function () {
			editor.removeAndCleanupForCollection();
		});

		describe('Link', function () {
			it('should add empty link no selection', function () {
				editor.performAction('insertLink');
				expect(editor.textContent()).toEqual('[]()');
				expect(editor.selectedRange().startOffset).toEqual(1);
				expect(editor.selectedRange().textInRange()).toEqual('');
			});

			it('should wrap text with link when text is selected', function () {
				editor.setTextContent('Hello world!');
				editor.setSelectedLineCh({line: 0, ch: 0}, {line: 0, ch: 5});
				editor.performAction('insertLink');
				expect(editor.textContent()).toEqual('[Hello]() world!');
				expect(editor.selectedRange().textInRange()).toEqual('');
				expect(editor.selectedRange().startOffset).toEqual(8);
			});
		});

		describe('Boundary', function () {
			it('should toggle selection boundary when no selection', function () {
				editor.performAction('toggleBold');
				expect(editor.textContent()).toEqual('****');
				expect(editor.selectedRange().startOffset).toEqual(2);
				expect(editor.selectedRange().textInRange()).toEqual('');
			});

			it('should add selection boundary when text is selected', function () {
				editor.setTextContent('Hello world!');
				editor.setSelectedLineCh({line: 0, ch: 0}, {line: 0, ch: 5});
				editor.performAction('toggleItalic');
				expect(editor.textContent()).toEqual('_Hello_ world!');
				expect(editor.selectedRange().textInRange()).toEqual('Hello');
				editor.performAction('toggleItalic');
				editor.setTextContent('Hello world!');
			});

			it('should remove boundary when selected text has boundary', function () {
				editor.setTextContent('`Hello` world!');
				editor.setSelectedLineCh({line: 0, ch: 0}, {line: 0, ch: 7});
				editor.performAction('toggleCode');
				expect(editor.textContent()).toEqual('Hello world!');
				expect(editor.selectedRange().textInRange()).toEqual('Hello');
			});
		});

		describe('Types', function () {
			it('should format heading from empty', function () {
				editor.setTextContent('');
				editor.setSelectedLineCh({line: 0, ch: 0}, {line: 0, ch: 0});
				editor.performAction('formatHeading');
				expect(editor.textContent()).toEqual('# ');
				expect(editor.selectedRange().startOffset).toEqual(2);
			});

			it('should format heading level', function () {
				editor.setTextContent('# heading');
				editor.setSelectedLineCh({line: 0, ch: 2}, {line: 0, ch: 2});

				tree.ensureClassified();
				editor.performAction('formatHeading');
				expect(editor.textContent()).toEqual('# heading');
				expect(editor.selectedRange().location()).toEqual(2);

				tree.ensureClassified();
				editor.performAction('formatHeading', 2);
				expect(editor.textContent()).toEqual('## heading');
				expect(editor.selectedRange().location()).toEqual(3);
			});

			it('should format unordered list', function () {
				editor.setTextContent('one\ntwo');
				editor.setSelectedLineCh({line: 0, ch: 0}, {line: 1, ch: 2});

				tree.ensureClassified();
				editor.performAction('formatUnorderedList');
				expect(editor.textContent()).toEqual('- one\n- two');
				expect(editor.selectedRange().textInRange()).toEqual('one\n- tw');
			});

			it('should format ordered list', function () {
				editor.setTextContent('one\ntwo');
				editor.setSelectedLineCh({line: 0, ch: 0}, {line: 1, ch: 2});

				tree.ensureClassified();
				editor.performAction('formatOrderedList');
				expect(editor.textContent()).toEqual('1. one\n2. two');
				expect(editor.selectedRange().textInRange()).toEqual('one\n2. tw');
			});

			it('should format blockquote', function () {
				editor.setTextContent('one\ntwo');
				editor.setSelectedLineCh({line: 0, ch: 0}, {line: 1, ch: 2});

				tree.ensureClassified();
				editor.performAction('formatBlockquote');
				expect(editor.textContent()).toEqual('> one\n> two');
				expect(editor.selectedRange().textInRange()).toEqual('one\n> tw');
			});

			it('should format codeblock', function () {
				editor.setTextContent('one\ntwo');
				editor.setSelectedLineCh({line: 0, ch: 0}, {line: 1, ch: 2});

				tree.ensureClassified();
				editor.performAction('formatCodeblock');
				expect(editor.textContent()).toEqual('    one\n    two');
				expect(editor.selectedRange().textInRange()).toEqual('one\n    tw');
			});

			it('should format body', function () {
				editor.setTextContent('- one\n- two');
				editor.setSelectedLineCh({line: 0, ch: 0}, {line: 1, ch: 2});

				tree.ensureClassified();
				editor.performAction('formatBody');
				expect(editor.textContent()).toEqual('one\ntwo');
				//expect(editor.selectedRange().textInRange()).toEqual('one\ntwo');
			});
		});

		describe('Clear Formatting', function () {
			it('should clear italic formatting', function () {
				editor.setTextContent('Hello world!');
				editor.setSelectedLineCh({line: 0, ch: 0}, {line: 0, ch: 5});
				editor.performAction('toggleItalic');
				editor.tree().ensureClassified();
				editor.performAction('clearFormatting');
				expect(editor.textContent()).toEqual('Hello world!');
			});

			it('should clear substitution formatting', function () {
				editor.setTextContent('Hello world!');
				editor.setSelectedLineCh({line: 0, ch: 0}, {line: 0, ch: 5});
				editor.performAction('toggleSubstitution');
				expect(editor.textContent()).toEqual('{~~Hello~>~~} world!');
				editor.tree().ensureClassified();
				editor.performAction('clearFormatting');
				expect(editor.textContent()).toEqual('Hello world!');
			});
		});
	});
});