define(function (require) {
	'use strict';

/*
	one
		two
			three
			four
		five
			six
*/

	describe('Move Branches', function () {
		require('ft/extensions/movebranch');
		require('ft/taxonomy/outlinetaxonomy');
		require('ft/taxonomy/markdowntaxonomy');

		var treeText = 'one\n\ttwo\n\t\tthree\n\t\tfour\n\tfive\n\t\tsix',
			Editor = require('ft/editor/editor').Editor,
			Folding = require('ft/extensions/folding'),
			one, two, three, four, five, six,
			Tree = require('ft/core/tree').Tree,
			editor, tree;

		beforeEach(function () {
			jasmine.addMatchers(require('tests/custommatchers'));
			editor = new Editor(treeText, 'outline');
			tree = editor.tree();
			one = tree.root.firstChild;
			two = one.nextLineNode();
			three = two.nextLineNode();
			four = three.nextLineNode();
			five = four.nextLineNode();
			six = five.nextLineNode();
			tree.ensureClassified();
		});

		afterEach(function () {
			editor.removeAndCleanupForCollection();
		});

		it('should move branch up', function () {
			editor.setSelectedLineCh({line: 4, ch: 0});
			editor.performCommand('moveBranchesUp');
			expect(editor.textContent()).toEqual('one\n\tfive\n\t\tsix\n\ttwo\n\t\tthree\n\t\tfour');
		});

		it('should not move branch up if it does not have a previous sibling', function () {
			editor.setSelectedLineCh({line: 1, ch: 0});
			editor.performCommand('moveBranchesUp');
			expect(editor.textContent()).toEqual('one\n\ttwo\n\t\tthree\n\t\tfour\n\tfive\n\t\tsix');
		});

		it('should move branch up in this bug case', function () {
			editor.setTextContent('one\ntwo\n\tthree\n\t\tfour');
			editor.setSelectedLineCh({line: 1, ch: 0}, {line: 1, ch: 0});
			editor.performCommand('moveBranchesUp');
			expect(editor.textContent()).toEqual('two\n\tthree\n\t\tfour\none');
		});

		it('should preserve folds when moving lines up', function () {
			editor.setSelectedLineCh({line: 4, ch: 0});
			expect(editor.folds().length).toEqual(0);
			editor.performCommand('toggleNodeFold');
			expect(editor.folds().length).toEqual(1);
			editor.performCommand('moveBranchesUp');
			expect(editor.folds().length).toEqual(1);
			editor.performCommand('moveBranchesUp');
			editor.performCommand('moveBranchesUp');
			editor.performCommand('moveBranchesUp');
			expect(editor.folds().length).toEqual(1);
		});

		it('should not include trailing newline when moving branch up', function () {
			editor.setSelectedLineCh({line: 3, ch: 0}, {line: 4, ch: 0});
			editor.performCommand('moveBranchesUp');
			expect(editor.textContent()).toEqual('one\n\ttwo\n\t\tfour\n\t\tthree\n\tfive\n\t\tsix');
			expect(editor.selectedText()).toEqual('\t\tfour\n');
		});

		it('should move branch down', function () {
			editor.setSelectedLineCh({line: 1, ch: 0});
			editor.performCommand('moveBranchesDown');
			expect(editor.textContent()).toEqual('one\n\tfive\n\t\tsix\n\ttwo\n\t\tthree\n\t\tfour');
		});

		it('should not move branch down if it does not have a next sibling', function () {
			editor.setSelectedLineCh({line: 4, ch: 0});
			editor.performCommand('moveBranchesDown');
			expect(editor.textContent()).toEqual('one\n\ttwo\n\t\tthree\n\t\tfour\n\tfive\n\t\tsix');
		});

		it('should always move down if next branch is empty, even though that empty line isnt next sibling', function () {
			editor.setTextContent('one\n\ttwo\n\t\tthree\n\n');
			tree.ensureClassified();
			editor.setSelectedLineCh({line: 1, ch: 0});

			editor.performCommand('moveBranchesDown');
			expect(editor.textContent()).toEqual('one\n\n\ttwo\n\t\tthree\n');
			expect(editor.selectedRange().startLineCh()).toBeJsonEqual({line: 2, ch: 0});

			editor.performCommand('moveBranchesDown');
			expect(editor.textContent()).toEqual('one\n\n\n\ttwo\n\t\tthree');
			expect(editor.selectedRange().startLineCh()).toBeJsonEqual({line: 3, ch: 0});
		});

		it('should handle this case that made me go insane for two days', function () {
			editor.setTextContent('\none\ntwo\n\tthree\n\n	five\n\nseven');
			tree.ensureClassified();
			editor.setSelectedLineCh({line: 1, ch: 0});
			editor.performCommand('moveBranchesDown');
			editor.performCommand('moveBranchesRight');
		});

		it('should preservefolds when moving branch down', function () {
			editor.setSelectedLineCh({line: 1, ch: 0});
			var baseMarks = editor.cm().getAllMarks().length;
			editor.performCommand('toggleNodeFold');
			editor.performCommand('moveBranchesDown');
			expect(editor.cm().getAllMarks().length - baseMarks).toEqual(1);
			editor.performCommand('moveBranchesDown');
			editor.performCommand('moveBranchesDown');
			editor.performCommand('moveBranchesDown');
			expect(editor.cm().getAllMarks().length - baseMarks).toEqual(1);
		});

		it('should preserve selection when moving right and trailing newline is selected', function () {
			editor.setSelectedLineCh({line: 3, ch: 0}, {line: 4, ch: 0});
			expect(editor.selectedText()).toEqual('\t\tfour\n');
			editor.performCommand('moveBranchesRight');
			expect(editor.selectedText()).toEqual('\t\tfour\n');
			expect(editor.textContent()).toEqual('one\n\ttwo\n\t\tthree\n\t\t\tfour\n\tfive\n\t\tsix');
		});

		it('should move branch down in this bug case', function () {
			editor.setTextContent('one\n\ttwo\n\t\tthree\nfour');
			editor.setSelectedLineCh({line: 1, ch: 0}, {line: 1, ch: 0});
			editor.performCommand('toggleNodeFold');
			editor.setSelectedLineCh({line: 0, ch: 0}, {line: 0, ch: 0});
			editor.performCommand('toggleNodeFold');
			editor.performCommand('moveBranchesDown');
			expect(editor.textContent()).toEqual('four\none\n\ttwo\n\t\tthree');
		});

		it('should move branch left', function () {
			editor.setSelectedLineCh({line: 4, ch: 0});
			editor.performCommand('moveBranchesLeft');
			expect(editor.textContent()).toEqual('one\n\ttwo\n\t\tthree\n\t\tfour\nfive\n\tsix');
		});

		it('should move branch to end of siblings before moving left if needed', function () {
			editor.setSelectedLineCh({line: 1, ch: 0});
			editor.performCommand('moveBranchesLeft');
			expect(editor.textContent()).toEqual('one\n\tfive\n\t\tsix\ntwo\n\tthree\n\tfour');
		});

		it('should move branch right', function () {
			editor.setSelectedLineCh({line: 4, ch: 0});
			editor.performCommand('moveBranchesRight');
			expect(editor.textContent()).toEqual('one\n\ttwo\n\t\tthree\n\t\tfour\n\t\tfive\n\t\t\tsix');
		});

		it('should move branch right based on proceeding non empty line', function () {
			editor.setTextContent('one\n\ntwo');
			tree.ensureClassified();
			editor.setSelectedLineCh({line: 2, ch: 0});
			editor.performCommand('moveBranchesRight');
			expect(editor.textContent()).toEqual('one\n\n\ttwo');
			editor.performCommand('moveBranchesRight'); // nothing should happen
			expect(editor.textContent()).toEqual('one\n\n\ttwo');
		});

		it('should move branch right that has empty lines in subtree', function () {
			editor.setTextContent('one\ntwo\n\n\tthree');
			tree.ensureClassified();
			editor.setSelectedLineCh({line: 1, ch: 0});
			editor.performCommand('moveBranchesRight');
			tree.validate();
			expect(editor.textContent()).toEqual('one\n\ttwo\n\n\t\tthree');
		});

		it('should delete branch', function () {
			editor.setSelectedLineCh({line: 1, ch: 0});
			editor.performCommand('deleteBranches');
			expect(editor.textContent()).toEqual('one\n\tfive\n\t\tsix');
		});

		it('should operate on visible line, not hidden line when cursor is right after fold', function () {
			editor.setSelectedLineCh({line: 1, ch: 0});
			editor.performCommand('toggleNodeFold');
			editor.setSelectedLineCh({line: 3, ch: 6});
			editor.performCommand('moveBranchesDown');
			expect(editor.textContent()).toEqual('one\n\tfive\n\t\tsix\n\ttwo\n\t\tthree\n\t\tfour');
		});

		it('should demote trailing sibling branches', function () {
			editor.setSelectedLineCh({line: 2, ch: 0}, {line: 2, ch: 0});
			editor.performCommand('demoteTrailingSiblingBranches');
			expect(editor.textContent()).toEqual('one\n\ttwo\n\t\tthree\n\t\t\tfour\n\tfive\n\t\tsix');
		});

		it('should not crash where their are no trailing siblings to demote', function () {
			editor.setSelectedLineCh({line: 4, ch: 0});
			editor.performCommand('demoteTrailingSiblingBranches');
			expect(editor.textContent()).toEqual(treeText);
		});

		it('should promote child branches', function () {
			editor.setSelectedLineCh({line: 0, ch: 0}, {line: 0, ch: 0});
			editor.performCommand('promoteChildBranches');
			expect(editor.textContent()).toEqual('one\ntwo\n\tthree\n\tfour\nfive\n\tsix');

			tree.ensureClassified();
			editor.setSelectedLineCh({line: 1, ch: 0}, {line: 1, ch: 0});
			editor.performCommand('promoteChildBranches');
			expect(editor.textContent()).toEqual('one\ntwo\nthree\nfour\nfive\n\tsix');
		});

		it('should not crash where their are no child branches to promote', function () {
			editor.setSelectedLineCh({line: 2, ch: 0});
			editor.performCommand('promoteChildBranches');
			expect(editor.textContent()).toEqual(treeText);
		});

		describe('Markdown', function () {
			beforeEach(function () {
				editor.removeAndCleanupForCollection();
				editor = new Editor('', 'markdown');
				tree = editor.tree();
			});			

			it('should move markdown list item down paste empty line', function () {
				editor.setTextContent('- one\n');
				tree.ensureClassified();
				editor.setSelectedLineCh({line: 0, ch: 0});
				editor.performCommand('moveBranchesDown');
				expect(editor.textContent()).toEqual('\n- one');
			});

			it('should move markdown list item down sibling with empty line after', function () {
				editor.setTextContent('- one\n- two\n');
				tree.ensureClassified();
				editor.setSelectedLineCh({line: 0, ch: 0});
				editor.performCommand('moveBranchesDown');
				expect(editor.textContent()).toEqual('- two\n- one\n');
			});

			it('should move markdown header down paste next header', function () {
				editor.setTextContent('\n# a\n- b\n# c');
				tree.ensureClassified();
				editor.setSelectedLineCh({line: 1, ch: 0});
				editor.performCommand('moveBranchesDown');
			});

			it('should not over-indent markdown header', function () {
				editor.setTextContent('# a\nb\n\n## c');
				tree.ensureClassified();
				editor.setSelectedLineCh({line: 3, ch: 0});
				editor.performCommand('moveBranchesRight');
				expect(editor.textContent()).toEqual('# a\nb\n\n## c');
			});
		});

		//it('should position cursor correctly after promote', function () {
		//	editor.setTextContent('one\n\ttwo');
		//	tree.ensureClassified();
		//	editor.setSelectedLineCh({line: 0, ch: 3});
		//	editor.performCommand('promoteChildBranches');
		//	expect(editor.cursorRange().startLineCh()).toEqual({ line : 0, ch : 3 });
		//});
	});
});