define(function (require) {
	'use strict';

/*
	one
		two
			three
			four
		five
			six
*/

	describe('Move Paragraph', function () {
		require('ft/extensions/moveparagraph');
		require('ft/taxonomy/outlinetaxonomy');

		var treeText = 'one\n\ttwo\n\t\tthree\n\t\tfour\n\tfive\n\t\tsix',
			Editor = require('ft/editor/editor').Editor,
			Folding = require('ft/extensions/folding'),
			one, two, three, four, five, six,
			Tree = require('ft/core/tree').Tree,
			editor, tree;

		beforeEach(function () {
			editor = new Editor(treeText, 'outline', {indentWithTabs: true, indentUnit: 4});
			tree = editor.tree();
			one = tree.root.firstChild;
			two = one.nextLineNode();
			three = two.nextLineNode();
			four = three.nextLineNode();
			five = four.nextLineNode();
			six = five.nextLineNode();
			tree.ensureClassified();
		});

		afterEach(function () {
			editor.removeAndCleanupForCollection();
		});

		it('should move up', function () {
			editor.setSelectedLineCh({line: 4, ch: 1}, {line: 4, ch: 5});
			editor.performCommand('moveParagraphsUp');
			expect(editor.textContent()).toEqual('one\n\ttwo\n\t\tthree\n\tfive\n\t\tfour\n\t\tsix');
			expect(editor.selectedText()).toEqual('five');
		});

		it('should move down', function () {
			editor.setSelectedLineCh({line: 4, ch: 1}, {line: 4, ch: 5});
			editor.performCommand('moveParagraphsDown');
			expect(editor.textContent()).toEqual('one\n\ttwo\n\t\tthree\n\t\tfour\n\t\tsix\n\tfive');
			expect(editor.selectedText()).toEqual('five');
		});

		it('should move left', function () {
			editor.setSelectedLineCh({line: 4, ch: 1}, {line: 4, ch: 5});
			editor.performCommand('moveParagraphsLeft');
			expect(editor.textContent()).toEqual('one\n\ttwo\n\t\tthree\n\t\tfour\nfive\n\t\tsix');
			expect(editor.selectedText()).toEqual('five');
		});

		it('should move right', function () {
			editor.setSelectedLineCh({line: 4, ch: 1}, {line: 4, ch: 5});
			editor.performCommand('moveParagraphsRight');
			expect(editor.textContent()).toEqual('one\n\ttwo\n\t\tthree\n\t\tfour\n\t\tfive\n\t\tsix');
			expect(editor.selectedText()).toEqual('five');
		});

		it('should move multiple lines right and undo in single operation', function () {
			editor.setSelectedLineCh({line: 0, ch: 1}, {line: 4, ch: 5});
			editor.performCommand('moveParagraphsRight');
			expect(editor.textContent()).toEqual('\tone\n\t\ttwo\n\t\t\tthree\n\t\t\tfour\n\t\tfive\n\t\tsix');
			editor.performCommand('undo');
			expect(editor.textContent()).toEqual(treeText);
		});

		describe('Skipping Folds', function () {
			it('should insert before collapsed node when moving up', function () {
				editor.setSelectedLineCh({line: 4, ch: 1}, {line: 4, ch: 5});
				editor.collapseNode(two);
				editor.performCommand('moveParagraphsUp');
				expect(editor.textContent()).toEqual('one\n\tfive\n\ttwo\n\t\tthree\n\t\tfour\n\t\tsix');
				expect(editor.selectedText()).toEqual('five');
			});

			it('should insert before top hidden line node when moving up', function () {
				editor.setSelectedLineCh({line: 4, ch: 1}, {line: 4, ch: 5});
				editor.createFoldInRange(tree.createRangeFromNodes(two, 0, four, -1));
				editor.performCommand('moveParagraphsUp');
				expect(editor.textContent()).toEqual('one\n\tfive\n\ttwo\n\t\tthree\n\t\tfour\n\t\tsix');
				expect(editor.selectedText()).toEqual('five');
			});

			it('should insert after last hidden node when moving down', function () {
				editor.setSelectedLineCh({line: 0, ch: 0}, {line: 0, ch: 3});
				editor.collapseNode(two);
				editor.performCommand('moveParagraphsDown');
				expect(editor.textContent()).toEqual('\ttwo\n\t\tthree\n\t\tfour\none\n\tfive\n\t\tsix');
				expect(editor.selectedText()).toEqual('one');
			});
		});

		describe('Carrying Folds', function () {
			it('should carry fold when moving collapsed node up', function () {
				editor.setSelectedLineCh({line: 1, ch: 1}, {line: 1, ch: 4});
				editor.collapseNode(two);
				editor.performCommand('moveParagraphsUp');
				expect(editor.textContent()).toEqual('\ttwo\n\t\tthree\n\t\tfour\none\n\tfive\n\t\tsix');
				expect(editor.selectedText()).toEqual('two');
				expect(editor.isCollapsed(two)).toBeTruthy();
			});

			it('should carry fold when moving collapsed node down', function () {
				editor.setSelectedLineCh({line: 1, ch: 1}, {line: 1, ch: 4});
				editor.collapseNode(two);
				editor.performCommand('moveParagraphsDown');
				expect(editor.textContent()).toEqual('one\n\tfive\n\ttwo\n\t\tthree\n\t\tfour\n\t\tsix');
				expect(editor.selectedText()).toEqual('two');
				expect(editor.isCollapsed(two)).toBeTruthy();
			});

			it('should not lose fold when moving collapsed node down when already at end', function () {
				editor.setSelectedLineCh({line: 4, ch: 1}, {line: 4, ch: 5});
				editor.collapseNode(five);
				editor.performCommand('moveParagraphsDown');
				expect(editor.textContent()).toEqual(treeText);
				expect(editor.selectedText()).toEqual('five');
				expect(editor.isCollapsed(five)).toBeTruthy();
			});

			it('should carry fold when moving collapsed node left', function () {
				editor.setSelectedLineCh({line: 1, ch: 1}, {line: 1, ch: 4});
				editor.collapseNode(two);
				editor.performCommand('moveParagraphsLeft');
				expect(editor.textContent()).toEqual('one\ntwo\n\tthree\n\tfour\n\tfive\n\t\tsix');
				expect(editor.selectedText()).toEqual('two');
				expect(editor.isCollapsed(two)).toBeTruthy();
			});

			it('should carry fold when moving collapsed node right', function () {
				editor.setSelectedLineCh({line: 1, ch: 1}, {line: 1, ch: 4});
				editor.collapseNode(two);
				editor.performCommand('moveParagraphsRight');
				expect(editor.textContent()).toEqual('one\n\t\ttwo\n\t\t\tthree\n\t\t\tfour\n\tfive\n\t\tsix');
				expect(editor.selectedText()).toEqual('two');
				expect(editor.isCollapsed(two)).toBeTruthy();
			});

			it('should one carry fold over another fold when moving down', function () {
				editor.setSelectedLineCh({line: 1, ch: 1}, {line: 1, ch: 4});
				editor.collapseNode(two);
				editor.collapseNode(five);
				editor.performCommand('moveParagraphsDown');
				expect(editor.textContent()).toEqual('one\n\tfive\n\t\tsix\n\ttwo\n\t\tthree\n\t\tfour');
				expect(editor.selectedText()).toEqual('two');
				expect(editor.isCollapsed(two)).toBeTruthy();
				expect(editor.isCollapsed(five)).toBeTruthy();
			});

			it('should one carry fold over another fold when moving up', function () {
				editor.setSelectedLineCh({line: 4, ch: 1}, {line: 4, ch: 5});
				editor.collapseNode(two);
				editor.collapseNode(five);
				editor.performCommand('moveParagraphsUp');
				expect(editor.textContent()).toEqual('one\n\tfive\n\t\tsix\n\ttwo\n\t\tthree\n\t\tfour');
				expect(editor.selectedText()).toEqual('five');
				expect(editor.isCollapsed(two)).toBeTruthy();
				expect(editor.isCollapsed(five)).toBeTruthy();
			});
		});
	});
});