define(function (require) {
	'use strict';

/*
	one
		two
			Hello world. Look at this.
			four
		five
			six
*/

	describe('Selection', function () {
		require('ft/extensions/selection');
		require('ft/taxonomy/outlinetaxonomy');

		var treeText = 'one\n\ttwo\n\t\tthree\n\t\tfour\n\tfive\n\t\tsix',
			Extensions = require('ft/core/extensions').Extensions,
			Editor = require('ft/editor/editor').Editor,
			one, two, three, four, five, six,
			Tree = require('ft/core/tree').Tree,
			editor, tree,
			selectedText,
			selectionChanged = function (editor) {
				selectedText = editor.selectedText();
			};

		beforeEach(function () {
			Extensions.addSelectionDidChange(selectionChanged);
			editor = new Editor(treeText, 'outline');
			tree = editor.tree();
			one = tree.root.firstChild;
			two = one.nextLineNode();
			three = two.nextLineNode();
			four = three.nextLineNode();
			five = four.nextLineNode();
			six = five.nextLineNode();
			three.setLine('\t\tHello world. Look at this.');
			tree.ensureClassified();
			editor.setSelectedLineCh({line: 2, ch: 9}, {line: 2, ch: 9});
		});

		afterEach(function () {
			Extensions.remove('com.foldingtext.editor.selectionDidChange', function (each) { return each === selectionChanged; });
			editor.removeAndCleanupForCollection();
			selectedText = null;
		});

		it('should get selection changed event on init', function () {
			expect(selectedText).toEqual('');
		});

		it('should get selection changed event on selection change', function () {
			editor.setSelectedRange(tree.createRangeFromLocation(0, 3));
			expect(selectedText).toEqual('one');
		});

		describe('Extend and Contract Selection', function () {
			it('should select to word boundaries', function () {
				editor.performAction('expandSelectionToWord');
				expect(editor.selectedRange().textInRange()).toEqual('world');
			});

			it('should select to sentence boundaries', function () {
				editor.performAction('expandSelectionToSentence');
				expect(editor.selectedRange().textInRange()).toEqual('\t\tHello world. ');
				editor.performAction('expandSelectionToSentence');
				expect(editor.selectedRange().textInRange()).toEqual('\t\tHello world. ');
			});

			it('should select to line boundaries', function () {
				editor.performAction('expandSelectionToLine');
				expect(editor.selectedRange().textInRange()).toEqual('\t\tHello world. Look at this.');
				editor.performAction('expandSelectionToLine');
				expect(editor.selectedRange().textInRange()).toEqual('\t\tHello world. Look at this.');
			});

			it('should select to paragraph boundaries', function () {
				editor.performAction('expandSelectionToParagraph');
				expect(editor.selectedRange().textInRange()).toEqual('\t\tHello world. Look at this.\n');
				editor.performAction('expandSelectionToParagraph');
				expect(editor.selectedRange().textInRange()).toEqual('\t\tHello world. Look at this.\n');
			});

			it('should expand and contract selection based on word, sentence, line, and tree structure', function () {
				editor.performAction('expandSelection');
				expect(editor.selectedText()).toEqual('world');
				editor.performAction('expandSelection');
				expect(editor.selectedText()).toEqual('\t\tHello world. ');
				editor.performAction('expandSelection');
				expect(editor.selectedText()).toEqual('\t\tHello world. Look at this.');
				editor.performAction('contractSelection');
				expect(editor.selectedText()).toEqual('\t\tHello world. ');
				editor.performAction('contractSelection');
				expect(editor.selectedText()).toEqual('world');
				editor.performAction('expandSelection');
				expect(editor.selectedText()).toEqual('\t\tHello world. ');
				editor.performAction('expandSelection');
				expect(editor.selectedText()).toEqual('\t\tHello world. Look at this.');
				editor.performAction('expandSelection');
				expect(editor.selectedText()).toEqual('\t\tHello world. Look at this.\n');
				editor.performAction('expandSelection');
				expect(editor.selectedText()).toEqual('\t\tHello world. Look at this.\n\t\tfour\n');
				editor.performAction('expandSelection');
				expect(editor.selectedText()).toEqual('\ttwo\n\t\tHello world. Look at this.\n\t\tfour\n');
				editor.performAction('expandSelection');
				expect(editor.selectedText()).toEqual('\ttwo\n\t\tHello world. Look at this.\n\t\tfour\n\tfive\n\t\tsix');
				editor.performAction('expandSelection');
				expect(editor.selectedText()).toEqual('one\n\ttwo\n\t\tHello world. Look at this.\n\t\tfour\n\tfive\n\t\tsix');
				editor.performAction('contractSelection');
				expect(editor.selectedText()).toEqual('\ttwo\n\t\tHello world. Look at this.\n\t\tfour\n\tfive\n\t\tsix');
				editor.performAction('contractSelection');
				expect(editor.selectedText()).toEqual('\ttwo\n\t\tHello world. Look at this.\n\t\tfour\n');
				editor.performAction('contractSelection');
				expect(editor.selectedText()).toEqual('\t\tHello world. Look at this.\n\t\tfour\n');
				editor.performAction('contractSelection');
				expect(editor.selectedText()).toEqual('\t\tHello world. Look at this.\n');
				editor.performAction('contractSelection');
				expect(editor.selectedText()).toEqual('\t\tHello world. Look at this.');
				editor.performAction('contractSelection');
				expect(editor.selectedText()).toEqual('\t\tHello world. ');
				editor.performAction('contractSelection');
				expect(editor.selectedText()).toEqual('world');
				editor.performAction('contractSelection');
				expect(editor.selectedText()).toEqual('');
			});
		});
	});
});