define(function (require) {
	'use strict';

	describe('FoldingText Taxonomy', function () {
		var MarkdownTaxonomy = require('ft/taxonomy/markdowntaxonomy').MarkdownTaxonomy,
			Extensions = require('ft/core/extensions').Extensions,
			Taxonomies = require('ft/core/taxonomies'),
			Tree = require('ft/core/tree').Tree,
			one, two, three, four, five, six,
			tree;

		Extensions.addMode({
			name: 'test1'
		});

		beforeEach(function () {
			tree = new Tree('', Taxonomies.taxonomy({
				foldingtext: true,
				criticMarkup: true
			}, 'markdown'));
		});

		afterEach(function () {
			tree.cleanupForCollection();
		});

		function setTextAndEnsureAndSetNodes(text) {
			tree.setText(text);
			tree.classifier.ensureClassified();
			one = two = three = four = five = six = null;
			one = tree.firstLineNode();
			two = one.nextLineNode();
			if (two) {
				three = two.nextLineNode();
				if (three) {
					four = three.nextLineNode();
					if (four) {
						five = four.nextLineNode();
						if (five) {
							six = five.nextLineNode();
						}
					}
				}
			}
		}

		it('should use tabs to create structure', function () {
			setTextAndEnsureAndSetNodes('one\n\ttwo');
			expect(tree.firstLineNode().typeIndentLevel()).toEqual(0);
			expect(tree.lastLineNode().typeIndentLevel()).toEqual(1);
		});

		it('should allow list block to start without being proceeded by newline', function () {
			setTextAndEnsureAndSetNodes('one\n\t- two');
			expect(tree.firstLineNode().typeIndentLevel()).toEqual(0);
			expect(tree.lastLineNode().typeIndentLevel()).toEqual(2);
			expect(tree.lastLineNode().type()).toEqual('unordered');
		});

		it('should detect code blocks with spaces only', function () {
			setTextAndEnsureAndSetNodes('    one');
			expect(one.type()).toEqual('codeblock');
			expect(one.lineAttributedString().toString()).toEqual('(    /)(one/text)');
		});

		describe('Property Base Type', function () {
			it('should detect valid property lines', function () {
				setTextAndEnsureAndSetNodes('my :');
				expect(one.type()).not.toEqual('property');

				setTextAndEnsureAndSetNodes('my : property');
				expect(one.type()).toEqual('property');

				setTextAndEnsureAndSetNodes('not my : property');
				expect(one.type()).not.toEqual('property');

				setTextAndEnsureAndSetNodes('notmy:property');
				expect(one.type()).not.toEqual('property');

				setTextAndEnsureAndSetNodes('my : property value');
				expect(one.type()).toEqual('property');

				setTextAndEnsureAndSetNodes('notmy: property value');
				expect(one.type()).not.toEqual('property');
			});

			it('should get correct \'name\' attribute', function () {
				setTextAndEnsureAndSetNodes('my : property value');
				expect(one.tags().name).toEqual('my');
			});

			it('should recognize extended UTF8 names and values', function () {
				setTextAndEnsureAndSetNodes('Prénom : Robert');
				expect(one.tags().name).toEqual('prénom');

				setTextAndEnsureAndSetNodes('Prénom : Robért');
				expect(one.tags().value).toEqual('Robért');
			});

			it('should get correct \'value\' attribute', function () {
				setTextAndEnsureAndSetNodes('my : property value');
				expect(one.tags().value).toEqual('property value');
			});

			it('should detect proper indent level', function () {
				setTextAndEnsureAndSetNodes('\t\tmy : property value');
				expect(one.type()).toEqual('property');
				expect(one.typeIndentLevel()).toEqual(2);
			});

			it('should highlight correct syntax attributes', function () {
				setTextAndEnsureAndSetNodes('\t\t\tmy : property');
				//expect(one.lineAttributedString().toString(true)).toEqual('(			/tabs="			")(my/propertyname="my", text="my : property")( /syntax=":", text="my : property")(:/text="my : property")( propert/propertyvalue="property", text="my : property")(y/text="my : property")');
			});
		});

		//
		// Tags and modes are more properly tested in outline.spec.js If something goes wrong
		// in foldingtext's tag/mode behavior, first try to reproduce the problem in outline mode.
		// then either fix there, or figure out what's different and fix in foldingtext mode.
		//

		it('should ignore trailing tags on headings', function () {
			setTextAndEnsureAndSetNodes('# simple @tag');
			expect(one.text()).toEqual('simple');
		});

		it('should work with both trailing tags and closing heading syntaxes', function () {
			setTextAndEnsureAndSetNodes('# simple @tag ##');
			expect(one.text()).toEqual('simple');
		});

		it('should find modes', function () {
			setTextAndEnsureAndSetNodes('# simple.test1');
			expect(one.text()).toEqual('simple');
			expect(one.mode()).toEqual('test1');
		});

		it('should find modes shadowed by trailing ## in heading', function () {
			setTextAndEnsureAndSetNodes('# simple.test1 ##');
			expect(one.text()).toEqual('simple');
			expect(one.mode()).toEqual('test1');
		});
	});
});