define(function (require) {
	'use strict';

	describe('GithubMarkdown Taxonomy', function () {
		var MarkdownTaxonomy = require('ft/taxonomy/markdowntaxonomy').MarkdownTaxonomy,
			Taxonomies = require('ft/core/taxonomies'),
			Tree = require('ft/core/tree').Tree,
			tree;

		beforeEach(function () {
			tree = new Tree('', Taxonomies.taxonomy({
				gitmarkdown: true
			}, 'markdown'));
		});

		afterEach(function () {
			tree.cleanupForCollection();
		});

		it('should highlight github style checkboxes', function () {
			tree.setText('- [x] one\n\ntwo\n\n1. [ ] three');
			tree.classifier.ensureClassified();

			var line = tree.firstLineNode();
			expect(line.type()).toEqual('task');
			expect(line.lineAttributedString().toString()).toEqual('(-/keyword, lead)( /lead)([/checkbox, keyword, lead)(x/checkbox, lead)(]/checkbox, keyword, lead)( /)(one/text)');

			line = line.nextLineNode();
			line = line.nextLineNode();
			line = line.nextLineNode();
			line = line.nextLineNode();
			expect(line.type()).toEqual('task');
			expect(line.lineAttributedString().toString()).toEqual('(1./keyword, lead)( /lead)([/checkbox, keyword, lead)( /checkbox, lead)(]/checkbox, keyword, lead)( /)(three/text)');
		});

		describe('Inline', function () {
			it('should accept email auto links with no angle syntax', function () {
				tree.setText('user@example.com foo');
				tree.classifier.ensureClassified();
				expect(tree.firstLineNode().lineAttributedString().toString()).toEqual('(user@example.com/linktarget, magiclink, text)( foo/text)');
			});

			it('should accept uri', function () {
				tree.setText('mailto:jesse@hogbaysoftware.com');
				tree.classifier.ensureClassified();
				expect(tree.firstLineNode().lineAttributedString().toString()).toEqual('(mailto:jesse@hogbaysoftware.com/linktarget, magiclink, text)');

				tree.setText('x-apple-reminder://DA8D6EFD-5FA9-4B26-BD04-2B5AFFA4E29E');
				tree.classifier.ensureClassified();
				expect(tree.firstLineNode().lineAttributedString().toString()).toEqual('(x-apple-reminder://DA8D6EFD-5FA9-4B26-BD04-2B5AFFA4E29E/linktarget, magiclink, text)');
			});

			it('should accept url auto links with no angle syntax', function () {
				tree.setText('www.foldingtext.com');
				tree.classifier.ensureClassified();
				expect(tree.firstLineNode().lineAttributedString().toString()).toEqual('(www.foldingtext.com/linktarget, magiclink, text)');
			});

			it('should accept url auto links with no angle syntax after inline line', function () {
				tree.setText('[a](www.a.com) www.b.com');
				tree.classifier.ensureClassified();
				expect(tree.firstLineNode().lineAttributedString().toString()).toEqual('([/inlinelink, keyword, text)(a/inlinelink, linklabel, text)(]/inlinelink, keyword, text)((/inlinelink, keyword, text)(www.a.com/inlinelink, linktarget, magiclink, text)()/inlinelink, keyword, text)( /text)(www.b.com/linktarget, magiclink, text)');
			});

			it('should detect em with dashes and stars', function () {
				tree.setText('_one_');
				tree.classifier.ensureClassified();
				expect(tree.firstLineNode().lineAttributedString().toString()).toEqual('(_/em, keyword, text)(one/em, text)(_/em, keyword, text)');

				tree.setText('a _one_');
				tree.classifier.ensureClassified();
				expect(tree.firstLineNode().lineAttributedString().toString()).toEqual('(a /text)(_/em, keyword, text)(one/em, text)(_/em, keyword, text)');

				tree.setText('*one*');
				tree.classifier.ensureClassified();
				expect(tree.firstLineNode().lineAttributedString().toString()).toEqual('(*/em, keyword, text)(one/em, text)(*/em, keyword, text)');

				tree.setText('a *one*');
				tree.classifier.ensureClassified();
				expect(tree.firstLineNode().lineAttributedString().toString()).toEqual('(a /text)(*/em, keyword, text)(one/em, text)(*/em, keyword, text)');
			});

			it('should not make italics when dashes are inside words', function () {
				tree.setText('_o_e_');
				tree.classifier.ensureClassified();
				expect(tree.firstLineNode().lineAttributedString().toString()).toEqual('(_/em, keyword, text)(o_e/em, text)(_/em, keyword, text)');

				tree.setText('_o_e');
				tree.classifier.ensureClassified();
				expect(tree.firstLineNode().lineAttributedString().toString()).toEqual('(_o_e/text)');
			});

			it('should make strong when stars are inside words', function () {
				tree.setText('**one**');
				tree.classifier.ensureClassified();
				expect(tree.firstLineNode().lineAttributedString().toString()).toEqual('(**/keyword, strong, text)(one/strong, text)(**/keyword, strong, text)');

				tree.setText('__one__');
				tree.classifier.ensureClassified();
				expect(tree.firstLineNode().lineAttributedString().toString()).toEqual('(__/keyword, strong, text)(one/strong, text)(__/keyword, strong, text)');

				tree.setText('__o__e__');
				tree.classifier.ensureClassified();
				expect(tree.firstLineNode().lineAttributedString().toString()).toEqual('(__/keyword, strong, text)(o/strong, text)(__/keyword, strong, text)(e__/text)');
			});

			it('should detect inline followed by punctuation', function () {
				// These are failing now because of parseInWordFormatting being set to false in multimarkdown mode.
				// That regex for em and strong needs to be updated to handle punctuation.
				tree.setText('*em*.');
				tree.classifier.ensureClassified();
				expect(tree.firstLineNode().lineAttributedString().toString()).toEqual('(*/em, keyword, text)(em/em, text)(*/em, keyword, text)(./text)');

				tree.setText('**strong**.');
				tree.classifier.ensureClassified();
				expect(tree.firstLineNode().lineAttributedString().toString()).toEqual('(**/keyword, strong, text)(strong/strong, text)(**/keyword, strong, text)(./text)');

				tree.setText('`code`.');
				tree.classifier.ensureClassified();
				expect(tree.firstLineNode().lineAttributedString().toString()).toEqual('(`/code, keyword, text)(code/code, text)(`/code, keyword, text)(./text)');
			});

		});

		describe('Fentced Code Block', function () {
			it('should detect fenced code blocks', function () {
				tree.setText('```html\n<p>hello</p>\n```');
				tree.classifier.ensureClassified();
				var line = tree.firstLineNode();
				expect(line.type()).toEqual('fencedcodetopboundary');
				expect(line.tags().language).toEqual('html');
				expect(line.lineAttributedString().toString()).toEqual('(```/keyword, text)(html/keyword, language, text)');

				line = line.nextLineNode();
				expect(line.type()).toEqual('fencedcode');
				expect(line.tags().language).toEqual('html');
				expect(line.lineAttributedString().toString()).toEqual('(</mode-bracket, mode-tag, text)(p/mode-tag, text)(>/mode-bracket, mode-tag, text)(hello/text)(<//mode-bracket, mode-tag, text)(p/mode-tag, text)(>/mode-bracket, mode-tag, text)');

				line = line.nextLineNode();
				expect(line.type()).toEqual('fencedcodebottomboundary');
				expect(line.tags().language).toEqual('html');
				expect(line.lineAttributedString().toString()).toEqual('(```/keyword, text)');
			});
		});

	});
});