define(function (require) {
	'use strict';

	describe('Markdown Taxonomy', function () {
		var MarkdownTaxonomy = require('ft/taxonomy/markdowntaxonomy').MarkdownTaxonomy,
			Taxonomies = require('ft/core/taxonomies'),
			Tree = require('ft/core/tree').Tree,
			tree;

		beforeEach(function () {
			tree = new Tree('', 'markdown');
		});

		afterEach(function () {
			tree.cleanupForCollection();
		});

		it('should mark simple blocks as item', function () {
			tree.setText('one');
			tree.classifier.ensureClassified();
			expect(tree.firstLineNode().type()).toEqual('body');
		});

		describe('Structure', function () {
			it('should set type', function () {
				tree.setText('# one\ntwo');
				var one = tree.firstLineNode();
				tree.classifier.ensureClassified();
				one.setType('body');
				expect(one.line()).toEqual('one');
			});

			it('should set from heading to unordered', function () {
				tree.setText('## one');
				var one = tree.firstLineNode();
				tree.classifier.ensureClassified();
				tree.beginUpdates();
				one.setType('unordered');
				one.setTypeIndentLevel('1');
				expect(one.line()).toEqual('- one');
				tree.endUpdates();
			});
		});

		describe('Empty Type', function () {
			it('should mark empty lines as empty', function () {
				tree.setText('\n\n');
				tree.classifier.ensureClassified();
				var line = tree.firstLineNode();
				expect(line.type()).toEqual('empty');
				expect(tree.firstLineNode().type()).toEqual('empty');
				line = line.nextLineNode();
				expect(tree.firstLineNode().type()).toEqual('empty');
				line = line.nextLineNode();
				expect(tree.firstLineNode().type()).toEqual('empty');
			});

			it('should structure empty lines to level of next line', function () {
				tree.setText('\n# two');
				tree.classifier.ensureClassified();
				var empty = tree.firstLineNode();
				expect(empty.type()).toEqual('empty');
				expect(empty._structureLevel).toEqual(empty.nextLineNode()._structureLevel);
			});
		});

		describe('Heading Type', function () {
			describe('ATX Style', function () {
				it('should detect ATX style headings', function () {
					tree.setText('# one');
					tree.classifier.ensureClassified();
					expect(tree.lastLineNode().type()).toEqual('heading');
				});

				it('should detect upto 6 headings', function () {
					tree.setText('# one\n## two\n### three\n#### four\n##### five\n###### six');
					var line = tree.firstLineNode();
					tree.classifier.ensureClassified();
					expect(line.type()).toEqual('heading');
					expect(line.typeIndentLevel()).toEqual(1);
					expect(line._structureLevel).toEqual(300 + 1);
					line = line.nextLineNode();
					expect(line.type()).toEqual('heading');
					expect(line.typeIndentLevel()).toEqual(2);
					expect(line._structureLevel).toEqual(300 + 2);
					line = line.nextLineNode();
					expect(line.type()).toEqual('heading');
					expect(line.typeIndentLevel()).toEqual(3);
					expect(line._structureLevel).toEqual(300 + 3);
					line = line.nextLineNode();
					expect(line.type()).toEqual('heading');
					expect(line.typeIndentLevel()).toEqual(4);
					expect(line._structureLevel).toEqual(300 + 4);
					line = line.nextLineNode();
					expect(line.type()).toEqual('heading');
					expect(line.typeIndentLevel()).toEqual(5);
					expect(line._structureLevel).toEqual(300 + 5);
					line = line.nextLineNode();
					expect(line.type()).toEqual('heading');
					expect(line.typeIndentLevel()).toEqual(6);
					expect(line._structureLevel).toEqual(300 + 6);
				});

				it('should not continue heading to next node', function () {
					tree.setText('# one\nhello');
					tree.classifier.ensureClassified();
					expect(tree.firstLineNode().type()).toEqual('heading');
					expect(tree.lastLineNode().type()).toEqual('body');
				});

				it('should detect starting # has keyword', function () {
					tree.setText('# one');
					tree.classifier.ensureClassified();
					var attr = tree.firstLineNode().lineAttributedString();
					expect(attr.toString()).toEqual('(#/keyword, lead)( /lead)(one/text)');
				});

				it('should detect ending # as keywords', function () {
					tree.setText('### one #\n# two #\n# three ###');
					tree.classifier.ensureClassified();

					var line = tree.firstLineNode();
					expect(line.lineAttributedString().toString()).toEqual('(###/keyword, lead)( /lead)(one/text)( /)(#/keyword)');

					line = line.nextLineNode();
					expect(line.lineAttributedString().toString()).toEqual('(#/keyword, lead)( /lead)(two/text)( /)(#/keyword)');

					line = line.nextLineNode();
					expect(line.lineAttributedString().toString()).toEqual('(#/keyword, lead)( /lead)(three/text)( /)(###/keyword)');
				});

				it('should not detect # with space before it', function () {
					tree.setText(' # one');
					tree.classifier.ensureClassified();
					expect(tree.firstLineNode().type()).not.toEqual('heading');
					tree.setText('# one');
					tree.classifier.ensureClassified();
					expect(tree.firstLineNode().type()).toEqual('heading');
				});

				it('should set heading type', function () {
					tree.setText('one');
					var one = tree.firstLineNode();
					tree.classifier.ensureClassified();
					one.setType('heading');
					expect(one.line()).toEqual('# one');
				});

				it('should set heading type indent level', function () {
					tree.setText('# one');
					var one = tree.firstLineNode();
					tree.classifier.ensureClassified();
					one.setTypeIndentLevel(2);
					expect(one.line()).toEqual('## one');
				});
			});

			describe('Setext Style', function () {
				it('should detect setext headings', function () {
					tree.setText('one\n==');
					tree.classifier.ensureClassified();
					var line = tree.firstLineNode(),
						attr = line.lineAttributedString();

					expect(line.type()).toEqual('heading');
					expect(line.typeIndentLevel()).toEqual(1);
					expect(line.nextLineNode().type()).toEqual('headingunderline');
				});

				it('should detect setext headings', function () {
					tree.setText('one\n--');
					tree.classifier.ensureClassified();
					var line = tree.firstLineNode(),
						attr = line.lineAttributedString();

					expect(line.type()).toEqual('heading');
					expect(line.typeIndentLevel()).toEqual(2);
					expect(line.nextLineNode().type()).toEqual('headingunderline');
				});


				it('should not detect setext headings with space before it', function () {
					tree.setText('one\n =');
					tree.classifier.ensureClassified();
					var line = tree.firstLineNode(),
						attr = line.lineAttributedString();

					expect(line.type()).not.toEqual('heading');
					expect(line.nextLineNode().type()).not.toEqual('headingunderline');
				});

				it('should only detect setext headings where previous type is "item"', function () {
					tree.setText('- one\n-');
					tree.classifier.ensureClassified();
					var line = tree.firstLineNode(),
						attr = line.lineAttributedString();

					expect(line.type()).toEqual('unordered');
					expect(line.nextLineNode().type()).toEqual('body');
				});

				it('should stop detecting setext headings when updated this way', function () {
					tree.setText('a\n\nb\n==');
					tree.classifier.ensureClassified();
					var underline = tree.lastLineNode(),
						heading = underline.previousLineNode();

					expect(heading.type()).toEqual('heading');
					expect(underline.type()).toEqual('headingunderline');

					tree.replaceTextInRange(tree.createRangeFromNodes(underline, 0), 'c');

					tree.classifier.ensureClassified();

					expect(heading.type()).toEqual('body');
					expect(underline.type()).toEqual('body');
				});

				it('should stop detecting setext headings when updated this other way', function () {
					tree.setText('b\n==');
					tree.classifier.ensureClassified();
					var heading = tree.firstLineNode(),
						underline = heading.nextLineNode();

					expect(heading.type()).toEqual('heading');
					expect(underline.type()).toEqual('headingunderline');

					tree.replaceTextInRange(tree.createRangeFromNodes(underline, 0), '\n');

					tree.classifier.ensureClassified();

					expect(heading.type()).toEqual('body');
					expect(tree.lastLineNode().type()).toEqual('body');
				});

				it('should stop detecting setext headings when updated this other other way', function () {
					tree.setText('b\n==');
					tree.classifier.ensureClassified();
					var heading = tree.firstLineNode(),
						underline = heading.nextLineNode();

					expect(heading.type()).toEqual('heading');
					expect(underline.type()).toEqual('headingunderline');

					tree.beginUpdates();
					tree.insertNodeBefore(tree.createNode('filler 1'), underline);
					tree.insertNodeBefore(tree.createNode('filler 2'), underline);
					tree.insertNodeBefore(tree.createNode('filler 3'), underline);
					tree.endUpdates();

					tree.classifier.ensureClassified();

					expect(heading.type()).toEqual('body');
					expect(underline.previousLineNode().type()).toEqual('heading');
					expect(underline.type()).toEqual('headingunderline');
				});
			});
		});

		describe('Codeblock Type', function () {
			it('should detect tab codeblocks', function () {
				tree.setText('\tone');
				tree.classifier.ensureClassified();
				expect(tree.firstLineNode().type()).toEqual('codeblock');
				expect(tree.firstLineNode().text()).toEqual('one');
				expect(tree.firstLineNode().lineAttributedString().toString()).toEqual('(	/)(one/text)');
			});

			it('should detect space codeblocks', function () {
				tree.setText('    one');
				tree.classifier.ensureClassified();
				expect(tree.firstLineNode().type()).toEqual('codeblock');
				expect(tree.firstLineNode().text()).toEqual('one');
				expect(tree.firstLineNode().lineAttributedString().toString()).toEqual('(    /)(one/text)');
			});

			it('should detect empty lines inside code block as code block', function () {
				tree.setText('    one\n\n    two');
				tree.classifier.ensureClassified();
				expect(tree.lineNumberToNode(0).type()).toEqual('codeblock');
				expect(tree.lineNumberToNode(1).type()).toEqual('codeblock');
				expect(tree.lineNumberToNode(2).type()).toEqual('codeblock');
			});

			it('should set codeblock type', function () {
				tree.setText('one');
				var one = tree.firstLineNode();
				tree.classifier.ensureClassified();
				one.setType('codeblock');
				expect(one.line()).toEqual('    one');
			});

			it('should set codeblock type indent level', function () {
				tree.setText('    one');
				var one = tree.firstLineNode();
				tree.classifier.ensureClassified();
				one.setTypeIndentLevel(2);
				expect(one.line()).toEqual('\t\t    one');
			});
		});

		describe('Lists Types', function () {
			it('should not start lists after non-empty lines', function () {
				tree.setText('one\n- welcome');
				tree.classifier.ensureClassified();
				expect(tree.lastLineNode().type()).not.toEqual('unordered');
			});

			it('should continue lists after newline', function () {
				tree.setText('- one\ntwo\n- three');
				tree.classifier.ensureClassified();
				var line = tree.firstLineNode();
				expect(line.typeIndentLevel()).toEqual(1);
				expect(line.type()).toEqual('unordered');

				line = line.nextLineNode();
				expect(line.typeIndentLevel()).toEqual(1);
				expect(line.type()).toEqual('body');

				line = line.nextLineNode();
				expect(line.typeIndentLevel()).toEqual(1);
				expect(line.type()).toEqual('unordered');
			});

			it('should stop lists after item with no indentation, proceeded by space', function () {
				tree.setText('- a\n\nb\n\n- c');
				tree.classifier.ensureClassified();

				var line = tree.firstLineNode();
				expect(line.typeIndentLevel()).toEqual(1);
				expect(line.type()).toEqual('unordered');

				line = line.nextLineNode().nextLineNode();
				expect(line.typeIndentLevel()).toEqual(0);
				expect(line.type()).toEqual('body');

				line = line.nextLineNode().nextLineNode();
				expect(line.typeIndentLevel()).toEqual(1);
				expect(line.type()).toEqual('unordered');
			});

			it('should set list type indent level', function () {
				tree.setText('- one');
				var one = tree.firstLineNode();
				tree.classifier.ensureClassified();
				one.setTypeIndentLevel(2);
				expect(one.line()).toEqual('\t- one');
			});

			describe('Unordered', function () {
				it('should detect unordered lists', function () {
					tree.setText('- hello');
					tree.classifier.ensureClassified();
					expect(tree.firstLineNode().type()).toEqual('unordered');
				});

				it('should detect unordered list indent levels', function () {
					tree.setText(' - a\n\t- b');
					tree.classifier.ensureClassified();
					expect(tree.firstLineNode().typeIndentLevel()).toEqual(1);
					expect(tree.lastLineNode().typeIndentLevel()).toEqual(2);
				});

				it('should detect starting -/+/* has keyword', function () {
					tree.setText('- hello\n+ welcome\n* back!');
					tree.classifier.ensureClassified();

					var line = tree.firstLineNode(),
						attr = line.lineAttributedString();

					expect(attr.toString()).toEqual('(-/keyword, lead)( /)(hello/text)');
					line = line.nextLineNode();
					attr = line.lineAttributedString();
					expect(attr.toString()).toEqual('(+/keyword, lead)( /)(welcome/text)');
					line = line.nextLineNode();
					attr = line.lineAttributedString();
					expect(attr.toString()).toEqual('(*/keyword, lead)( /)(back!/text)');
				});

				it('should set unorderd type', function () {
					tree.setText('one');
					var one = tree.firstLineNode();
					tree.classifier.ensureClassified();
					one.setType('unordered');
					expect(one.line()).toEqual('- one');
				});
			});

			describe('Ordered', function () {
				it('should detect ordered lists', function () {
					tree.setText('1. hello');
					tree.classifier.ensureClassified();
					expect(tree.firstLineNode().type()).toEqual('ordered');
				});

				it('should detect starting ordered number as keyword', function () {
					tree.setText('1. hello\n2. welcome\n3. back!');
					tree.classifier.ensureClassified();

					var line = tree.firstLineNode(),
						attr = line.lineAttributedString();

					expect(attr.toString()).toEqual('(1./keyword, lead)( /)(hello/text)');
					line = line.nextLineNode();
					attr = line.lineAttributedString();
					expect(attr.toString()).toEqual('(2./keyword, lead)( /)(welcome/text)');
					line = line.nextLineNode();
					attr = line.lineAttributedString();
					expect(attr.toString()).toEqual('(3./keyword, lead)( /)(back!/text)');
				});

				it('should set unorderd type', function () {
					tree.setText('- one @done');
					var one = tree.firstLineNode();
					tree.classifier.ensureClassified();
					one.setType('ordered');
					expect(one.line()).toEqual('1. one @done');
				});
			});
		});

		describe('Blockquotes', function () {
			it('should mark blockquotes', function () {
				tree.setText('> quote');
				tree.classifier.ensureClassified();
				expect(tree.firstLineNode().type()).toEqual('blockquote');
				expect(tree.firstLineNode().lineAttributedString().toString()).toEqual('(>/keyword, lead)( /)(quote/text)');
			});

			it('should mark multi-level blockquotes', function () {
				tree.setText('> quote\none\n\ntwo');
				tree.classifier.ensureClassified();
				var line = tree.firstLineNode();
				expect(line.type()).toEqual('blockquote');
				line = line.nextLineNode();
				expect(line.type()).toEqual('blockquote');
				expect(line.lineAttributedString().toString()).toEqual('(one/text)');
				expect(tree.lastLineNode().type()).toEqual('body');
			});

			it('should mark multiple blockquote levels', function () {
				tree.setText('> one\n> two\n> > three');
				tree.classifier.ensureClassified();
				var line = tree.firstLineNode();
				expect(line.type()).toEqual('blockquote');
				expect(line.typeIndentLevel()).toEqual(0);
				line = line.nextLineNode();
				expect(line.type()).toEqual('blockquote');
				expect(line.typeIndentLevel()).toEqual(0);
				line = line.nextLineNode();
				expect(line.type()).toEqual('blockquote');
				expect(line.typeIndentLevel()).toEqual(1);
			});

			it('should set blockquote indent level', function () {
				tree.setText('> one');
				var one = tree.firstLineNode();
				tree.classifier.ensureClassified();
				one.setTypeIndentLevel(2);
				expect(one.line()).toEqual('> > > one');
			});

			it('should set blockquote type', function () {
				tree.setText('- one @done');
				var one = tree.firstLineNode();
				tree.classifier.ensureClassified();
				one.setType('blockquote');
				expect(one.line()).toEqual('> one @done');
			});
		});

		describe('Horizontal Rules', function () {
			it('should mark horizontal rules', function () {
				tree.setText('* * *');
				tree.classifier.ensureClassified();
				var line = tree.firstLineNode();
				expect(line.type()).toEqual('horizontalrule');
			});

			it('should set horizontal rule type', function () {
				tree.setText('- one @done');
				var one = tree.firstLineNode();
				tree.classifier.ensureClassified();
				one.setType('horizontalrule');
				expect(one.line()).toEqual('* * *');
			});
		});

		describe('HTML Blocks', function () {
			it('should detect html', function () {
				tree.setText('\n<ul>\n\t<li>hello</li>\n</ul>\n\nhello');
				tree.classifier.ensureClassified();
				var line = tree.firstLineNode().nextLineNode(),
					attr = line.lineAttributedString();

				expect(attr.toString()).toEqual('(</mode-bracket, mode-tag, text)(ul/mode-tag, text)(>/mode-bracket, mode-tag, text)');
				expect(line.type()).toEqual('htmlblock');

				line = line.nextLineNode();
				attr = line.lineAttributedString();
				expect(attr.toString()).toEqual('(	/text)(</mode-bracket, mode-tag, text)(li/mode-tag, text)(>/mode-bracket, mode-tag, text)(hello/text)(<//mode-bracket, mode-tag, text)(li/mode-tag, text)(>/mode-bracket, mode-tag, text)');
				expect(line.type()).toEqual('htmlblock');

				line = line.nextLineNode();
				attr = line.lineAttributedString();
				expect(attr.toString()).toEqual('(<//mode-bracket, mode-tag, text)(ul/mode-tag, text)(>/mode-bracket, mode-tag, text)');
				expect(line.type()).toEqual('htmlblock');

				line = line.nextLineNode().nextLineNode();
				attr = line.lineAttributedString();
				expect(attr.toString()).toEqual('(hello/text)');
				expect(line.type()).toEqual('body');
			});

			it('should detect html comments', function () {
				tree.setText('<!-- this is a comment -->');
				tree.classifier.ensureClassified();
				expect(tree.firstLineNode().type().toString()).toEqual('htmlblock');
				expect(tree.firstLineNode().lineAttributedString().toString()).toEqual('(<!-- this is a comment -->/mode-comment, text)');
			});

			it('should not detect block if previous line is not empthy', function () {
				tree.setText('one\n<div>');
				tree.classifier.ensureClassified();
				expect(tree.firstLineNode().type().toString()).toEqual('body');
				expect(tree.firstLineNode().nextLineNode().type().toString()).toEqual('body');
			});
		});

		describe('Link Definition', function () {
			it('should detect link definition', function () {
				tree.setText('[foo]:');
				tree.classifier.ensureClassified();
				var one = tree.firstLineNode();
				expect(one.type()).toEqual('linkdefinition');
				expect(one.lineAttributedString().toString()).toEqual('([/keyword, text)(foo/linkid, text)(]:/error, keyword, text)');
			});

			it('should detect link definition with URL', function () {
				tree.setText('[foo]: http://www.example.com/');
				tree.classifier.ensureClassified();
				var one = tree.firstLineNode();
				expect(one.type()).toEqual('linkdefinition');
				expect(one.lineAttributedString().toString()).toEqual('([/keyword, text)(foo/linkid, text)(]:/keyword, text)( /text)(http://www.example.com//linktarget, text)');

				tree.setText('[foo]:http://www.example.com/');
				tree.classifier.ensureClassified();
				one = tree.firstLineNode();
				expect(one.type()).toEqual('linkdefinition');
				expect(one.lineAttributedString().toString()).toEqual('([/keyword, text)(foo/linkid, text)(]:/keyword, text)(http://www.example.com//linktarget, text)');
			});

			it('should detect link definition with angle URL', function () {
				tree.setText('[foo]: <http://www.example.com/>');
				tree.classifier.ensureClassified();
				var one = tree.firstLineNode();
				expect(one.type()).toEqual('linkdefinition');
				expect(one.lineAttributedString().toString()).toEqual('([/keyword, text)(foo/linkid, text)(]:/keyword, text)( /text)(</keyword, text)(http://www.example.com//linktarget, text)(>/keyword, text)');

				tree.setText('[foo]:<http://www.example.com/>');
				tree.classifier.ensureClassified();
				one = tree.firstLineNode();
				expect(one.type()).toEqual('linkdefinition');
				expect(one.lineAttributedString().toString()).toEqual('([/keyword, text)(foo/linkid, text)(]:/keyword, text)(</keyword, text)(http://www.example.com//linktarget, text)(>/keyword, text)');
			});

			it('should detect link definition with path', function () {
				tree.setText('[foo]: my.html');
				tree.classifier.ensureClassified();
				var one = tree.firstLineNode();
				expect(one.type()).toEqual('linkdefinition');
				expect(one.lineAttributedString().toString()).toEqual('([/keyword, text)(foo/linkid, text)(]:/keyword, text)( /text)(my.html/linktarget, text)');
			});

			it('should detect link definition with URL and title', function () {
				tree.setText('[foo]: http://www.example.com/ "look at this!"');
				tree.classifier.ensureClassified();
				var one = tree.firstLineNode();
				expect(one.type()).toEqual('linkdefinition');
				expect(one.lineAttributedString().toString()).toEqual('([/keyword, text)(foo/linkid, text)(]:/keyword, text)( /text)(http://www.example.com//linktarget, text)( /text)("/keyword, text)(look at this!/linktitle, text)("/keyword, text)');
			});

			it('should detect multiline links definitions (2 lines)', function () {
				tree.setText('[foo]:\n/url "title"');
				tree.classifier.ensureClassified();
				var node = tree.firstLineNode();
				expect(node.type()).toEqual('linkdefinition');
				node = node.nextLineNode();
				expect(node.type()).toEqual('linkdefinitionurl');
				expect(node.lineAttributedString().toString()).toEqual('(/url/linktarget, text)( /text)("/keyword, text)(title/linktitle, text)("/keyword, text)');
			});

			it('should detect multiline links definitions (3 lines)', function () {
				tree.setText('[foo]:\n/url\n"title"');
				tree.classifier.ensureClassified();
				var node = tree.firstLineNode();
				expect(node.type()).toEqual('linkdefinition');
				node = node.nextLineNode();
				expect(node.type()).toEqual('linkdefinitionurl');
				node = node.nextLineNode();
				expect(node.type()).toEqual('linkdefinitiontitle');
			});

			it('should detect multiline links definitions with tabs', function () {
				tree.setText('[foo]:\n\t/url\n\t"title"');
				tree.classifier.ensureClassified();
				var node = tree.firstLineNode();
				expect(node.type()).toEqual('linkdefinition');
				node = node.nextLineNode();
				expect(node.type()).toEqual('linkdefinitionurl');
				node = node.nextLineNode();
				expect(node.type()).toEqual('linkdefinitiontitle');
			});

			it('should detect link definition with errors', function () {
				tree.setText('[foo]: /url bad title');
				tree.classifier.ensureClassified();
				var one = tree.firstLineNode();
				expect(one.type()).toEqual('linkdefinition');
				expect(one.lineAttributedString().toString()).toEqual('([/keyword, text)(foo/linkid, text)(]:/keyword, text)( /text)(/url/linktarget, text)( /text)(bad title/error, text)');
			});

			it('should detect consecutive link definitions', function () {
				tree.setText('[a]: a\n[b]: b');
				tree.classifier.ensureClassified();

				var line = tree.firstLineNode();
				expect(line.type()).toEqual('linkdefinition');
				expect(line.lineAttributedString().toString()).toEqual('([/keyword, text)(a/linkid, text)(]:/keyword, text)( /text)(a/linktarget, text)');

				line = line.nextLineNode();
				expect(line.type()).toEqual('linkdefinition');
				expect(line.lineAttributedString().toString()).toEqual('([/keyword, text)(b/linkid, text)(]:/keyword, text)( /text)(b/linktarget, text)');
			});
		});

		describe('Inline', function () {
			it('should detect two or more trailing spaces on paragraphs', function () {
				tree.setText('hello world ');
				tree.classifier.ensureClassified();
				expect(tree.firstLineNode().lineAttributedString().toString()).toEqual('(hello world /text)');

				tree.setText('hello world  ');
				tree.classifier.ensureClassified();
				expect(tree.firstLineNode().lineAttributedString().toString()).toEqual('(hello world/text)( /)( /linebreak-causing-whitespace)');

				tree.setText('hello world   ');
				tree.classifier.ensureClassified();
				expect(tree.firstLineNode().lineAttributedString().toString()).toEqual('(hello world/text)(  /)( /linebreak-causing-whitespace)');
			});

			it('should detect em *', function () {
				tree.setText('one *em* two');
				tree.classifier.ensureClassified();
				expect(tree.firstLineNode().lineAttributedString().toString()).toEqual('(one /text)(*/em, keyword, text)(em/em, text)(*/em, keyword, text)( two/text)');
			});

			it('should detect multiple em within one line', function () {
				tree.setText('one *em* two *three*');
				tree.classifier.ensureClassified();
				expect(tree.firstLineNode().lineAttributedString().toString()).toEqual('(one /text)(*/em, keyword, text)(em/em, text)(*/em, keyword, text)( two /text)(*/em, keyword, text)(three/em, text)(*/em, keyword, text)');
			});

			it('should detect strongs with *', function () {
				tree.setText('one **strong** two');
				tree.classifier.ensureClassified();
				expect(tree.firstLineNode().lineAttributedString().toString()).toEqual('(one /text)(**/keyword, strong, text)(strong/strong, text)(**/keyword, strong, text)( two/text)');
			});

			it('should not trip on strong and em edgecases', function () {
				// Strong and Em edge cases
				var line;
				tree.setText(['there\'s something ** here',
					'not * * * strong'].join('\n'));
				tree.classifier.ensureClassified();
				line = tree.firstLineNode();
				expect(line.lineAttributedString().toString()).toEqual('(there\'s something ** here/text)');
				line = line.nextLineNode();
				expect(line.lineAttributedString().toString()).toEqual('(not * * * strong/text)');

				tree.setText('something * here');
				tree.classifier.ensureClassified();
				line = tree.firstLineNode();
				expect(line.lineAttributedString().toString()).toEqual('(something * here/text)');

				tree.setText('something * here');
				tree.classifier.ensureClassified();
				line = tree.firstLineNode();
				expect(line.lineAttributedString().toString()).toEqual('(something * here/text)');

				tree.setText('this ***** is strong');
				tree.classifier.ensureClassified();
				line = tree.firstLineNode();
				expect(line.lineAttributedString().toString()).toEqual('(this /text)(**/keyword, strong, text)(*/strong, text)(**/keyword, strong, text)( is strong/text)');

				tree.setText(['this ** nothing** is not strong',
					'this **nothing ** is not strong'].join('\n'));
				tree.classifier.ensureClassified();
				line = tree.firstLineNode();
				expect(line.lineAttributedString().toString()).toEqual('(this /text)(*/em, keyword, text)(* nothing/em, text)(*/em, keyword, text)(* is not strong/text)');
				line = line.nextLineNode();
				expect(line.lineAttributedString().toString()).toEqual('(this /text)(*/em, keyword, text)(*nothing */em, text)(*/em, keyword, text)( is not strong/text)');

				tree.setText('this *is maybe* required');
				tree.classifier.ensureClassified();
				line = tree.firstLineNode();
				expect(line.lineAttributedString().toString()).toEqual('(this /text)(*/em, keyword, text)(is maybe/em, text)(*/em, keyword, text)( required/text)');

				tree.setText('this *** is em');
				tree.classifier.ensureClassified();
				line = tree.firstLineNode();
				expect(line.lineAttributedString().toString()).toEqual('(this /text)(*/em, keyword, text)(*/em, text)(*/em, keyword, text)( is em/text)');

				tree.setText(['this *  nothing* is not em',
					'this *nothing * is not em'].join('\n'));
				tree.classifier.ensureClassified();
				line = tree.firstLineNode();
				expect(line.lineAttributedString().toString()).toEqual('(this *  nothing* is not em/text)');
				line = line.nextLineNode();
				expect(line.lineAttributedString().toString()).toEqual('(this *nothing * is not em/text)');

				tree.setText('this *should normally work ** is em');
				tree.classifier.ensureClassified();
				line = tree.firstLineNode();
				expect(line.lineAttributedString().toString()).toEqual('(this /text)(*/em, keyword, text)(should normally work */em, text)(*/em, keyword, text)( is em/text)');

				tree.setText('this ** should normally work* ha!');
				tree.classifier.ensureClassified();
				line = tree.firstLineNode();
				expect(line.lineAttributedString().toString()).toEqual('(this /text)(*/em, keyword, text)(* should normally work/em, text)(*/em, keyword, text)( ha!/text)');

				tree.setText('this *not really **strong**enough*');
				tree.classifier.ensureClassified();
				line = tree.firstLineNode();
				expect(line.lineAttributedString().toString()).toEqual('(this /text)(*/em, keyword, text)(not really /em, text)(**/em, keyword, strong, text)(strong/em, strong, text)(**/em, keyword, strong, text)(enough/em, text)(*/em, keyword, text)');

				tree.setText('this ***strong**enough*');
				tree.classifier.ensureClassified();
				line = tree.firstLineNode();
				expect(line.lineAttributedString().toString()).toEqual('(this /text)(**/keyword, strong, text)(*/em, keyword, strong, text)(strong/em, strong, text)(**/em, keyword, strong, text)(enough/em, text)(*/em, keyword, text)');

				// Using John Gruber's test case for strong-ems
				tree.setText(['***This is strong and em.***',
					'So is ***this*** word',
					'___This is strong and em.___',
					'So is ___this___ word'].join('\n'));
				tree.classifier.ensureClassified();
				line = tree.firstLineNode();
				expect(line.lineAttributedString().toString()).toEqual('(**/keyword, strong, text)(*/em, keyword, strong, text)(This is strong and em./em, strong, text)(*/em, keyword, strong, text)(**/keyword, strong, text)');
				line = line.nextLineNode();
				expect(line.lineAttributedString().toString()).toEqual('(So is /text)(**/keyword, strong, text)(*/em, keyword, strong, text)(this/em, strong, text)(*/em, keyword, strong, text)(**/keyword, strong, text)( word/text)');
				line = line.nextLineNode();
				expect(line.lineAttributedString().toString()).toEqual('(__/keyword, strong, text)(_/em, keyword, strong, text)(This is strong and em./em, strong, text)(_/em, keyword, strong, text)(__/keyword, strong, text)');
				line = line.nextLineNode();
				expect(line.lineAttributedString().toString()).toEqual('(So is /text)(__/keyword, strong, text)(_/em, keyword, strong, text)(this/em, strong, text)(_/em, keyword, strong, text)(__/keyword, strong, text)( word/text)');

				tree.setText('This case is un*believe*able! This d_oe_s not work!');
				tree.classifier.ensureClassified();
				line = tree.firstLineNode();
				expect(line.lineAttributedString().toString()).toEqual('(This case is un/text)(*/em, keyword, text)(believe/em, text)(*/em, keyword, text)(able! This d/text)(_/em, keyword, text)(oe/em, text)(_/em, keyword, text)(s not work!/text)');
			});

			it('should allow code spans inline', function () {
				tree.setText('foo `bar`');
				tree.classifier.ensureClassified();
				expect(tree.firstLineNode().lineAttributedString().toString()).toEqual('(foo /text)(`/code, keyword, text)(bar/code, text)(`/code, keyword, text)');
			});

			it('should ignore unclosed backticks', function () {
				tree.setText('foo `bar');
				tree.classifier.ensureClassified();
				expect(tree.firstLineNode().lineAttributedString().toString()).toEqual('(foo `bar/text)');
			});

			it('should allow double backticks', function () {
				tree.setText('``foo ` bar``');
				tree.classifier.ensureClassified();
				expect(tree.firstLineNode().lineAttributedString().toString()).toEqual('(``/code, keyword, text)(foo ` bar/code, text)(``/code, keyword, text)');
			});

			it('should allow multiple backticks', function () {
				tree.setText('`foo```bar`');
				tree.classifier.ensureClassified();
				expect(tree.firstLineNode().lineAttributedString().toString()).toEqual('(`/code, keyword, text)(foo```bar/code, text)(`/code, keyword, text)');
			});

			it('should allow simultaneous code segments', function () {
				tree.setText('`foo```bar` hello `world`');
				tree.classifier.ensureClassified();
				expect(tree.firstLineNode().lineAttributedString().toString()).toEqual('(`/code, keyword, text)(foo```bar/code, text)(`/code, keyword, text)( hello /text)(`/code, keyword, text)(world/code, text)(`/code, keyword, text)');
			});

			it('should balance code backticks', function () {
				tree.setText('``foo ``` bar` hello');
				tree.classifier.ensureClassified();
				expect(tree.firstLineNode().lineAttributedString().toString()).toEqual('(`/code, keyword, text)(`foo ``` bar/code, text)(`/code, keyword, text)( hello/text)');

				tree.setText('``foo ``` bar` hello`` world');
				tree.classifier.ensureClassified();
				expect(tree.firstLineNode().lineAttributedString().toString()).toEqual('(``/code, keyword, text)(foo ``` bar` hello/code, text)(``/code, keyword, text)( world/text)');
			});

			it('should detect inline followed by punctuation', function () {
				tree.setText('*em*.');
				tree.classifier.ensureClassified();
				expect(tree.firstLineNode().lineAttributedString().toString()).toEqual('(*/em, keyword, text)(em/em, text)(*/em, keyword, text)(./text)');

				tree.setText('**strong**.');
				tree.classifier.ensureClassified();
				expect(tree.firstLineNode().lineAttributedString().toString()).toEqual('(**/keyword, strong, text)(strong/strong, text)(**/keyword, strong, text)(./text)');

				tree.setText('`code`.');
				tree.classifier.ensureClassified();
				expect(tree.firstLineNode().lineAttributedString().toString()).toEqual('(`/code, keyword, text)(code/code, text)(`/code, keyword, text)(./text)');
			});

			it('should be able to escape syntax', function () {
				tree.setText('one \\*\\*two\\*\\* \\_three\\_ \\`four\\`');
				tree.classifier.ensureClassified();
				expect(tree.firstLineNode().lineAttributedString().toString()).toEqual('(one /text)(\\/keyword, text)(*/text)(\\/keyword, text)(*two/text)(\\/keyword, text)(*/text)(\\/keyword, text)(* /text)(\\/keyword, text)(_three/text)(\\/keyword, text)(_ /text)(\\/keyword, text)(`four/text)(\\/keyword, text)(`/text)');
			});

			it('should understand links', function () {
				tree.setText('[foo](http://example.com) bar');
				tree.classifier.ensureClassified();
				expect(tree.firstLineNode().lineAttributedString().toString()).toEqual('([/inlinelink, keyword, text)(foo/inlinelink, linklabel, text)(]/inlinelink, keyword, text)((/inlinelink, keyword, text)(http://example.com/inlinelink, linktarget, text)()/inlinelink, keyword, text)( bar/text)');
			});

			it('should detect links with title', function () {
				tree.setText('[foo](http://example.com/ \'bar\') hello');
				tree.classifier.ensureClassified();
				expect(tree.firstLineNode().lineAttributedString().toString()).toEqual('([/inlinelink, keyword, text)(foo/inlinelink, linklabel, text)(]/inlinelink, keyword, text)((/inlinelink, keyword, text)(http://example.com/ \'bar\'/inlinelink, linktarget, text)()/inlinelink, keyword, text)( hello/text)');
			});

			it('should accept em within links', function () {
				tree.setText('[*foo*](http://example.com/) bar');
				tree.classifier.ensureClassified();
				expect(tree.firstLineNode().lineAttributedString().toString()).toEqual('([/inlinelink, keyword, text)(*/em, inlinelink, keyword, linklabel, text)(foo/em, inlinelink, linklabel, text)(*/em, inlinelink, keyword, linklabel, text)(]/inlinelink, keyword, text)((/inlinelink, keyword, text)(http://example.com//inlinelink, linktarget, text)()/inlinelink, keyword, text)( bar/text)');
			});

			it('should accept strong within links', function () {
				tree.setText('[**foo**](http://example.com/) bar');
				tree.classifier.ensureClassified();
				expect(tree.firstLineNode().lineAttributedString().toString()).toEqual('([/inlinelink, keyword, text)(**/inlinelink, keyword, linklabel, strong, text)(foo/inlinelink, linklabel, strong, text)(**/inlinelink, keyword, linklabel, strong, text)(]/inlinelink, keyword, text)((/inlinelink, keyword, text)(http://example.com//inlinelink, linktarget, text)()/inlinelink, keyword, text)( bar/text)');
			});

			it('should accept strongem within links', function () {
				tree.setText('[***foo***](http://example.com/) bar');
				tree.classifier.ensureClassified();
				expect(tree.firstLineNode().lineAttributedString().toString()).toEqual('([/inlinelink, keyword, text)(**/inlinelink, keyword, linklabel, strong, text)(*/em, inlinelink, keyword, linklabel, strong, text)(foo/em, inlinelink, linklabel, strong, text)(*/em, inlinelink, keyword, linklabel, strong, text)(**/inlinelink, keyword, linklabel, strong, text)(]/inlinelink, keyword, text)((/inlinelink, keyword, text)(http://example.com//inlinelink, linktarget, text)()/inlinelink, keyword, text)( bar/text)');
			});

			it('should accept web auto links', function () {
				tree.setText('<http://example.com/> foo');
				tree.classifier.ensureClassified();
				expect(tree.firstLineNode().lineAttributedString().toString()).toEqual('(</autolink, keyword, text)(http://example.com//autolink, linktarget, text)(>/autolink, keyword, text)( foo/text)');
			});

			it('should accept email auto links', function () {
				tree.setText('<user@example.com> foo');
				tree.classifier.ensureClassified();
				expect(tree.firstLineNode().lineAttributedString().toString()).toEqual('(</autolink, keyword, text)(user@example.com/autolink, linktarget, text)(>/autolink, keyword, text)( foo/text)');
			});

			it('should detect inline html', function () {
				tree.setText('- hello <strong>world</strong>');
				tree.classifier.ensureClassified();
				expect(tree.firstLineNode().type().toString()).toEqual('unordered');
				expect(tree.firstLineNode().lineAttributedString().toString()).toEqual('(-/keyword, lead)( /)(hello /text)(</mode-bracket, mode-tag, text)(strong/mode-tag, text)(>/mode-bracket, mode-tag, text)(world/text)(<//mode-bracket, mode-tag, text)(strong/mode-tag, text)(>/mode-bracket, mode-tag, text)');
			});
		});

		describe('Text Extraction', function () {
			it('should handle simple items', function () {
				tree.setText('one two');
				tree.classifier.ensureClassified();
				expect(tree.firstLineNode().text()).toEqual('one two');
			});

			it('should handle leading #', function () {
				tree.setText('# one');
				tree.classifier.ensureClassified();
				expect(tree.firstLineNode().text()).toEqual('one');
			});

			it('should handle leading multiple #', function () {
				tree.setText('###### simple');
				tree.classifier.ensureClassified();
				expect(tree.firstLineNode().text()).toEqual('simple');
			});

			it('should handle leading -', function () {
				tree.setText('- simple');
				tree.classifier.ensureClassified();
				expect(tree.firstLineNode().text()).toEqual('simple');
			});

			it('should handle leading +', function () {
				tree.setText('+ simple');
				tree.classifier.ensureClassified();
				expect(tree.firstLineNode().text()).toEqual('simple');
			});

			it('should handle leading *', function () {
				tree.setText('* simple');
				tree.classifier.ensureClassified();
				expect(tree.firstLineNode().text()).toEqual('simple');
			});

			it('should handle leading blockquotes', function () {
				tree.setText('> simple');
				tree.classifier.ensureClassified();
				expect(tree.firstLineNode().text()).toEqual('simple');
			});

			it('should handle leading multilevel blockquotes', function () {
				tree.setText('> simple\ncontinuation\n> > indented\ncontinue2');
				tree.classifier.ensureClassified();

				var line = tree.firstLineNode();
				expect(line.text()).toEqual('simple');

				line = line.nextLineNode();
				expect(line.text()).toEqual('continuation');
				expect(line.type()).toEqual('blockquote');

				line = line.nextLineNode();
				expect(line.text()).toEqual('indented');
				expect(line.type()).toEqual('blockquote');

				line = line.nextLineNode();
				expect(line.text()).toEqual('continue2');
				expect(line.type()).toEqual('blockquote');
			});

			it('should ignore horizontal rules', function () {
				tree.setText('* * *');
				tree.classifier.ensureClassified();
				expect(tree.firstLineNode().text()).toEqual('');
			});

			it('should ignore leading tabs', function () {
				tree.setText('- simple\n\t- other\n\t\t- nope');
				tree.classifier.ensureClassified();

				var line = tree.firstLineNode();
				expect(line.text()).toEqual('simple');

				line = line.nextLineNode();
				expect(line.text()).toEqual('other');

				line = line.nextLineNode();
				expect(line.text()).toEqual('nope');
			});

			it('should ignore tailing # on headings', function () {
				tree.setText('# this works #');
				tree.classifier.ensureClassified();
				expect(tree.firstLineNode().text()).toEqual('this works');
			});

			it('should ignore mismatched # on headings', function () {
				tree.setText('## this works #\n# well ###');
				tree.classifier.ensureClassified();
				var line = tree.firstLineNode();
				expect(line.text()).toEqual('this works');
				line = line.nextLineNode();
				expect(line.text()).toEqual('well');
			});
		});

		describe('Undo Error Case', function () {
			it('should handle undo error case', function () {
				tree.setText('b\n# a\n');
				tree.classifier.ensureClassified();
				tree.replaceTextInRange(tree.createRangeFromLocation(0, 2), '');
				tree.replaceTextInRange(tree.createRangeFromLocation(4, 0), 'b\n');
				tree.classifier.ensureClassified();
			});

			it('should handle undo error case 2', function () {
				tree.setText('# b');
				tree.classifier.ensureClassified();
				var n = tree.createNode('a', false);
				tree.insertNodeBefore(n, tree.firstLineNode());
			});
		});
	});
});