define(function (require) {
	'use strict';

	describe('Outline Taxonomy', function () {
		var Extensions = require('ft/core/extensions').Extensions,
			Taxonomies = require('ft/core/taxonomies'),
			Tree = require('ft/core/tree').Tree;

		Extensions.addMode({
			name: 'test1'
		});

		Extensions.addMode({
			name: 'test2'
		});

		/*
			one
				two
					three
					four
				five
					six
		*/

		var treeText = 'one\n\ttwo\n\t\tthree\n\t\tfour\n\tfive\n\t\tsix',
			one, two, three, four, five, six,
			tree;

		beforeEach(function () {
			tree = new Tree(treeText, Taxonomies.taxonomy({
				tags: true,
				modes: true,
				criticMarkup: true
			}, 'outline'));
			tree.ensureClassified();
			one = tree.root.firstChild;
			two = one.nextLineNode();
			three = two.nextLineNode();
			four = three.nextLineNode();
			five = four.nextLineNode();
			six = five.nextLineNode();
		});

		afterEach(function () {
			tree.cleanupForCollection();
		});

		describe('Structure', function () {
			it('should set type indent level', function () {
				tree.setText('one\n\ttwo');
				one = tree.firstLineNode();
				tree.classifier.ensureClassified();
				one.setTypeIndentLevel(1);
				expect(one.line()).toEqual('	one');
			});

			it('should set mode', function () {
				tree.setText('one\n\ttwo');
				one = tree.firstLineNode();
				tree.classifier.ensureClassified();
				one.setMode('test1');
				expect(one.line()).toEqual('one.test1');
			});
		});

		describe('Critic', function () {
			it('should see addition syntax', function () {
				one.setLine('{++add++}');
				tree.classifier.ensureClassified();
				expect(one.lineAttributedString().toString()).toEqual('({/criticinsertion, criticstart, keyword, text)(++/criticinsertion, keyword, text)(add/criticinsertion, text)(++/criticinsertion, keyword, text)(}/criticend, criticinsertion, keyword, text)');
			});

			it('should not see overlapping addition syntax', function () {
				one.setLine('{++}');
				tree.classifier.ensureClassified();
				expect(one.lineAttributedString().toString()).toEqual('({++}/text)');
			});

			it('should not fail with just start syntax', function () {
				one.setLine('{++');
				tree.classifier.ensureClassified();
				expect(one.lineAttributedString().toString()).toEqual('({++/text)');

				one.setLine('{~~');
				tree.classifier.ensureClassified();
				expect(one.lineAttributedString().toString()).toEqual('({~~/text)');
			});

			it('should see deletion syntax', function () {
				one.setLine('{--delete--}');
				tree.classifier.ensureClassified();
				expect(one.lineAttributedString().toString()).toEqual('({/criticdeletion, criticstart, keyword, text)(--/criticdeletion, keyword, text)(delete/criticdeletion, text)(--/criticdeletion, keyword, text)(}/criticdeletion, criticend, keyword, text)');
			});

			it('should see subsitution syntax', function () {
				one.setLine('{~~delete~>add~~}');
				tree.classifier.ensureClassified();
				expect(one.lineAttributedString().toString()).toEqual('({/criticstart, criticsubstitution, keyword, text)(~~/criticsubstitution, keyword, text)(delete/criticdeletion, criticsubstitution, text)(~>/criticsubstitution, keyword, text)(add/criticinsertion, criticsubstitution, text)(~~/criticsubstitution, keyword, text)(}/criticend, criticsubstitution, keyword, text)');
			});

			it('should allow other syntax to follow subsitution syntax', function () {
				one.setLine('{~~delete~>add~~}{==wow!==}');
				tree.classifier.ensureClassified();
				expect(one.lineAttributedString().toString()).toEqual('({/criticstart, criticsubstitution, keyword, text)(~~/criticsubstitution, keyword, text)(delete/criticdeletion, criticsubstitution, text)(~>/criticsubstitution, keyword, text)(add/criticinsertion, criticsubstitution, text)(~~/criticsubstitution, keyword, text)(}/criticend, criticsubstitution, keyword, text)({/critichighlight, criticstart, keyword, text)(==/critichighlight, keyword, text)(wow!/critichighlight, text)(==/critichighlight, keyword, text)(}/criticend, critichighlight, keyword, text)');
			});

			it('should see comment syntax', function () {
				one.setLine('{>>boo!<<}');
				tree.classifier.ensureClassified();
				expect(one.lineAttributedString().toString()).toEqual('({/criticcomment, criticstart, keyword, text)(>>/criticcomment, keyword, text)(boo!/criticcomment, text)(<</criticcomment, keyword, text)(}/criticcomment, criticend, keyword, text)');
			});

			it('should see highlight syntax', function () {
				one.setLine('{==wow!==}');
				tree.classifier.ensureClassified();
				expect(one.lineAttributedString().toString()).toEqual('({/critichighlight, criticstart, keyword, text)(==/critichighlight, keyword, text)(wow!/critichighlight, text)(==/critichighlight, keyword, text)(}/criticend, critichighlight, keyword, text)');
			});

			it('should allow two blocks of critic syntax right next to each other', function () {
				one.setLine('{==a==}{>>b<<}');
				tree.classifier.ensureClassified();
				expect(one.lineAttributedString().toString()).toEqual('({/critichighlight, criticstart, keyword, text)(==/critichighlight, keyword, text)(a/critichighlight, text)(==/critichighlight, keyword, text)(}/criticend, critichighlight, keyword, text)({/criticcomment, criticstart, keyword, text)(>>/criticcomment, keyword, text)(b/criticcomment, text)(<</criticcomment, keyword, text)(}/criticcomment, criticend, keyword, text)');
			});

			it('should allow an extra { before syntax start', function () {
				one.setLine('{{==a==}');
				tree.classifier.ensureClassified();
				expect(one.lineAttributedString().toString()).toEqual('({/text)({/critichighlight, criticstart, keyword, text)(==/critichighlight, keyword, text)(a/critichighlight, text)(==/critichighlight, keyword, text)(}/criticend, critichighlight, keyword, text)');
			});
		});

		describe('Tags', function () {
			it('should add tags', function () {
				one.addTag('mytüg');
				expect(one.line()).toEqual('one @mytüg');
				expect(one.tags()).toEqual({ 'mytüg' : undefined });
			});

			it('should remove tags', function () {
				one.addTag('my_tüg');
				one.removeTag('my_tüg');
				expect(one.line()).toEqual('one');
				expect(one.tags()).toEqual({});
			});

			it('should add tags with values', function () {
				one.addTag('my_tag', 'my value');
				expect(one.line()).toEqual('one @my_tag(my value)');
				expect(one.tags()).toEqual({ 'my_tag' : 'my value' });
			});

			it('should not consider trailing tags are part of text', function () {
				one.addTag('mytüg');
				expect(one.text()).toEqual('one');
				tree.classifier.ensureClassified();
				expect(one.text()).toEqual('one');
			});
		});

		describe('Mode', function () {
			it('should set mode immediatly', function () {
				one.setMode('test1');
				expect(one.mode()).toEqual('test1');
				expect(one.modeContext()).toEqual('test1');
			});

			it('should replace mode immediatly', function () {
				one.setMode('test1');
				one.setMode('test2');
				expect(one.mode()).toEqual('test2');
				expect(one.modeContext()).toEqual('test2');
			});

			it('should remove mode immediatly', function () {
				one.setMode('test1');
				one.setMode(null);
				expect(one.mode()).toEqual(null);
				expect(one.modeContext()).toEqual(null);
			});

			it('should node include mode as part of text', function () {
				one.setMode('test1');
				tree.classifier.ensureClassified();
				expect(one.text()).toEqual('one');
			});

			it('should allow whitespace after mode', function () {
				one.setLine('one.test1 \t');
				tree.classifier.ensureClassified();
				expect(one.mode()).toEqual('test1');
				expect(one.text()).toEqual('one');
			});

			it('should apply mode context to decendents after ensureClassified', function () {
				one.setMode('test1');
				expect(two.modeContext()).toBeUndefined();
				expect(three.modeContext()).toBeUndefined();
				expect(four.modeContext()).toBeUndefined();
				expect(five.modeContext()).toBeUndefined();
				expect(six.modeContext()).toBeUndefined();
				tree.classifier.ensureClassified();
				expect(two.modeContext()).toEqual('test1');
				expect(three.modeContext()).toEqual('test1');
				expect(four.modeContext()).toEqual('test1');
				expect(five.modeContext()).toEqual('test1');
				expect(six.modeContext()).toEqual('test1');
			});

			it('should not allow empty lines to break mode context', function () {
				two.setLine('');
				one.setMode('test1');
				tree.classifier.ensureClassified();
				expect(two.modeContext()).toEqual('test1');
				expect(three.modeContext()).toEqual('test1');
				expect(four.modeContext()).toEqual('test1');
				expect(five.modeContext()).toEqual('test1');
				expect(six.modeContext()).toEqual('test1');
			});

			it('should not allow trailing empty lines to get mode context', function () {
				five.setLine('');
				six.setLine('');
				one.setMode('test1');
				tree.classifier.ensureClassified();
				expect(two.modeContext()).toEqual('test1');
				expect(three.modeContext()).toEqual('test1');
				expect(four.modeContext()).toEqual('test1');
				expect(five.modeContext()).toBeUndefined();
				expect(six.modeContext()).toBeUndefined();
			});

			it('should remove mode context from decendents after ensureClassified', function () {
				one.setMode('test1');
				tree.classifier.ensureClassified();
				one.setMode(null);
				tree.classifier.ensureClassified();
				expect(two.modeContext()).toBeUndefined();
				expect(three.modeContext()).toBeUndefined();
				expect(four.modeContext()).toBeUndefined();
				expect(five.modeContext()).toBeUndefined();
				expect(six.modeContext()).toBeUndefined();
			});

			it('should apply mode context to decendents in multiple levels after ensureClassified', function () {
				one.setMode('test1');
				two.setMode('test2');
				tree.classifier.ensureClassified();
				expect(two.modeContext()).toEqual('test2');
				expect(three.modeContext()).toEqual('test2');
				expect(four.modeContext()).toEqual('test2');
				expect(five.modeContext()).toEqual('test1');
				expect(six.modeContext()).toEqual('test1');

				two.setMode(null);
				tree.classifier.ensureClassified();
				expect(two.modeContext()).toEqual('test1');
				expect(three.modeContext()).toEqual('test1');
				expect(four.modeContext()).toEqual('test1');
				expect(five.modeContext()).toEqual('test1');
				expect(six.modeContext()).toEqual('test1');

				one.setMode(null);
				tree.classifier.ensureClassified();
				expect(two.modeContext()).toBeUndefined();
				expect(three.modeContext()).toBeUndefined();
				expect(four.modeContext()).toBeUndefined();
				expect(five.modeContext()).toBeUndefined();
				expect(six.modeContext()).toBeUndefined();
			});

			it('should apply mode context to decendents in multiple levels after ensureClassified (internal empty line shouldn\'t break things)', function () {
				one.setMode('test1');
				two.setMode('test2');
				three.setLine('');
				tree.classifier.ensureClassified();
				expect(two.modeContext()).toEqual('test2');
				expect(three.modeContext()).toEqual('test2');
				expect(four.modeContext()).toEqual('test2');
			});

			it('should pass this bug case', function () {
				tree.setText('one.test1\ntwo\nthree.test2');

				tree.classifier.ensureClassified();
				one = tree.firstLineNode();
				two = one.nextLineNode();
				three = two.nextLineNode();
				expect(one.mode()).toEqual('test1');
				expect(two.mode()).toEqual(undefined);
				expect(three.mode()).toEqual('test2');

				three.setLine('no more mode');
				tree.classifier.ensureClassified();
				expect(one.mode()).toEqual('test1');
				expect(two.mode()).toEqual(undefined);
				expect(three.mode()).toEqual(undefined);
			});
		});
	});
});